#include "osindependence.h"
#include "debug.h"
#include "procbase.h"

/****************************************************************************
Description:
    Initializes important flags needed for proper resource cleanup.
    Calls Processor's ctor
****************************************************************************/
void
ProcBaseConstructor(
    IN OUT PPROCBASE pThis,
    IN PCONFIG       pConfig
    )
{
    DPF("ProcBaseConstructor\n");

    pThis->Type = PROCBASE_SIGN;

    pThis->pConfig = pConfig;
    pThis->bISRInstalled = FALSE;
    pThis->bProcessorPlugged = FALSE;
    pThis->RangesCnt = 2;

    PCrptSpinLockInit( &pThis->InterruptSpinlock);

    DmaCtxConstructor( &pThis->SrcDmaCtx);
    DmaCtxConstructor( &pThis->DstDmaCtx);

    ProcessorConstructor( &pThis->Proc, pThis, pConfig);
}

/****************************************************************************
Description:
****************************************************************************/
void
ProcBaseDestructor(
    IN PPROCBASE pThis
    )
{
    DPF( "ProcBaseDestructor\n");

    ASSERT( pThis->Type == PROCBASE_SIGN);

    ProcessorDestructor( &pThis->Proc);

    DmaCtxDestructor( &pThis->SrcDmaCtx);
    DmaCtxDestructor( &pThis->DstDmaCtx);

    ASSERT( !pThis->bProcessorPlugged);
    ASSERT( !pThis->bISRInstalled);
}

/****************************************************************************
pDispatcher -
    needed merely to be passed to the PROCESSOR when it is plugged.
****************************************************************************/
BOOLEAN
ProcBaseAcquireHardware(
    IN PPROCBASE        pThis,
    IN PDISPATCHER      pDispatcher,
    IN device_t		dev
    )
{
    BOOLEAN bSuccess;             /* func.ret.value */

    DPF( "PBAcqHW\n");
    ASSERT( pThis->Type == PROCBASE_SIGN);

    bSuccess = FALSE;

    /*
     * Copy the processor board's hardware coordinates
     */
    pThis->dev = dev;

    if ( TranslateAndMapAddrs( pThis))
    {
        ASSERT( pThis->MemRanges[0].ResourceMapped);
        ASSERT( pThis->MemRanges[1].ResourceMapped);

#ifndef _NO_HW_
        /* Plug the processor into the processor base by letting
         * it know about DMA memory, hardware registers and the
         * DISPATCHER.
         */
        if ( BSD_SUCCESS(
                ProcessorPlug(
                    &pThis->Proc,
                    &pThis->SrcDmaCtx,
                    &pThis->DstDmaCtx,
                    pDispatcher,
                    (PVOID) pThis->MemRanges[1].MappedRangeStart, /* dma regs VA */
                    (PVOID) pThis->MemRanges[0].MappedRangeStart /* proc regs VA */
                )))
        {
            pThis->bProcessorPlugged = TRUE;

            /* Now that the PROCESSOR is operational, install ISR
             */
	    pThis->irid = 0;
	    pThis->irq = bus_alloc_resource(pThis->dev, SYS_RES_IRQ,
		&pThis->irid, 0, ~0, 1, RF_SHAREABLE | RF_ACTIVE);
	    if (pThis->irq != NULL) {
		bus_setup_intr(pThis->dev, pThis->irq, INTR_TYPE_BIO,
		    DeviceISR, pThis, &pThis->intrhand);
                pThis->bISRInstalled = TRUE;
                bSuccess = TRUE;
            }
            else
            {
                DPF( "pci_map_int failed\n");
            }
        }
        else
        {
            DPF( "ProcessorPlug failed\n");
        }
#endif
    }

    /*
     * Tell the dispatcher about this processor only if everything is Ok
     */
    if ( bSuccess)
    {
#ifndef _NO_HW_
        DispatcherRegisterProcessor( pDispatcher, &pThis->Proc);
#endif
    }

    return bSuccess;
}

/****************************************************************************

    This method is always called on driver unloading or
    if ProcBaseAcquireHardware() has failed, to release all the hardware
    resources.

****************************************************************************/
BOOLEAN
ProcBaseReleaseHardware(
    IN PPROCBASE       pThis
    )
{
    BOOLEAN             bSuccess;       /* func.ret.value */

    PADAPTER_MEM_RANGE  pCurRange;      /* walks mem.ranges array */

    DPF( "ProcBaseReleaseHardware\n");
    ASSERT( pThis->Type == PROCBASE_SIGN);
#ifdef foo
    /*
     * Check the resources acquisition flags and return all the resources
     * that must be returned.
     *
     * Some of the resources may not be acquired, if the earlier call
     * to ProcBaseAcquireHardware() has failed - this is why the flags
     * are needed
     */

    bSuccess = TRUE;

    if ( pThis->bISRInstalled)
    {
        if ( pci_unmap_int( pThis->CfgID) == 0)
        {
            DPF1( "pci_unmap_int(%08x) failed\n", pThis->CfgID);
        }
        pThis->bISRInstalled = FALSE;
    }

    if ( pThis->bProcessorPlugged)
    {
        ProcessorUnplug( &pThis->Proc);
        pThis->bProcessorPlugged = FALSE;
    }

#if 0
    /* */
    /* BUGBUG - pci_var.h doesn't expose an API for undoing the pci_map_mem()'s */
    /* work !!!! However, it doesn't provide an API for removing a PCI device */
    /* from the list where it was placed by pci_register_lkm() either, so there */
    /* is no reason for "undoing" - nobody can use this device anymore anyway - */
    /* it is associated with this driver forever. */
    /* */
    /*
     * Walk the mem.ranges array and un-map mapped ranges.
     */
    for ( nIndex = 0; nIndex < pThis->RangesCnt; nIndex++)
    {
        pCurRange = &pThis->MemRanges[ nIndex];
        if ( pCurRange->ResourceMapped)
        {
            DPF( "    Unmapping memory at 0x%08x", pCurRange->MappedRangeStart);
            ?????
        }
    }
#endif
#endif
    return bSuccess;
}

/****************************************************************************
Description:

    Translates a bus-specific address into the corresponding system logical
    address. Also, this will map a physical range to nonpaged system space if
    required.

    Most PCI devices respond to either an I/O range or a memory range, but
    not both.  For simplicity, this function simply maps all the resources
    for a given PCI device.  In reality, only one of the ranges might be
    used, so some cleanup may be needed here.

Arguments:

    pThis            - ptr to PROCBASE

****************************************************************************/
BOOLEAN
TranslateAndMapAddrs(
    IN PPROCBASE   pThis
    )
{
    BOOLEAN     bSuccess;       /* func.ret.value */
    PADAPTER_MEM_RANGE  pCurRange;      /* walks mem.ranges array */

    DPF( "    Xlate&mapAddrs(%08x)", pThis);
    ASSERT( pThis->Type == PROCBASE_SIGN);

    bSuccess = TRUE;

    /*
     * Map the first memory range
     */
    pCurRange = &pThis->MemRanges[ 0];
    pThis->rid0 = PCI_BASE_ADDRESS_0;
    pThis->res0 = bus_alloc_resource(pThis->dev, SYS_RES_MEMORY,
	&pThis->rid0, 0, ~0, 1, RF_ACTIVE);
    if (pThis->res0 != NULL) {
	pThis->bhandle0 = rman_get_bushandle(pThis->res0);
	pThis->btag0 = rman_get_bustag(pThis->res0);
	pThis->vhandle0 = (vm_offset_t)rman_get_virtual(pThis->res0);
	pCurRange->MappedRangeStart = pThis->vhandle0;
	pCurRange->RangeStart = vtophys(pThis->vhandle0);
        DPF1( "    PCI memory at PA=%08x mapped to VA=%08x\n", pCurRange->RangeStart, pCurRange->MappedRangeStart);
        pCurRange->ResourceMapped = TRUE;

        /*
         * Map the second memory range
         */
        pCurRange = &pThis->MemRanges[ 1];
	pThis->rid1 = PCI_BASE_ADDRESS_1;
	pThis->res1 = bus_alloc_resource(pThis->dev, SYS_RES_MEMORY,
	    &pThis->rid1, 0, ~0, 1, RF_ACTIVE);
	if (pThis->res1 != NULL) {
	    pThis->bhandle1 = rman_get_bushandle(pThis->res1);
	    pThis->btag1 = rman_get_bustag(pThis->res1);
	    pThis->vhandle1 = (vm_offset_t)rman_get_virtual(pThis->res1);
	    pCurRange->MappedRangeStart = pThis->vhandle1;
	    pCurRange->RangeStart = vtophys(pThis->vhandle1);
            DPF1( "    PCI memory at PA=%08x mapped to VA=%08x\n", pCurRange->RangeStart, pCurRange->MappedRangeStart);
            pCurRange->ResourceMapped = TRUE;
        }
        else
        {
            DPF1( "pci_map_mem failed for PA=%08x\n", pCurRange->RangeStart);
            bSuccess = FALSE;
        }
    }
    else
    {
        DPF1( "pci_map_mem failed for PA=%08x ", pCurRange->RangeStart);
        bSuccess = FALSE;
    }

    return bSuccess;
}

/****************************************************************************
    The standard BSD PCI interrupt handler
****************************************************************************/
void
DeviceISR
    (
        void    *arg
    )
{
    PPROCBASE   pThis;

    /*
     * We arranged this in the call to the pci_map_int() func.
     */
    pThis = (PPROCBASE) arg;
    ASSERT( pThis->Type == PROCBASE_SIGN);

    /*
     * Call PROCESSOR's interrupt method which return FALSE if it sees
     * that its 7751 board hasn't generated this interrupt. Then we don't need
     * to queue the DPC and should quickly return FALSE too.
     */
#ifndef _NO_HW_
    if ( ProcessorConfirmAndRegisterInterrupt( &pThis->Proc))
    {
        ProcessorServiceRegisteredInterrupts( &pThis->Proc);
    }
#endif
}

/****************************************************************************
    Acquire interrupt spinlock of our interrupt and call the specified routine
    with the specified parm, so that that routine could be sure the interrupt
    won't happen during its execution.
****************************************************************************/
BOOLEAN
ProcBaseSynchronizeExecution
    (
        IN PPROCBASE       pThis,
        IN PPROCBASE_SYNCHRONIZE_ROUTINE pSynchronizeRoutine,
        IN PVOID  SynchronizeContext
    )
{
    BOOLEAN bRetVal;

    DPF( "ProcBaseSynchronizeExecution\n");
    ASSERT( pThis->Type == PROCBASE_SIGN);

    PCrptSpinLockAcquire( &pThis->InterruptSpinlock);
    bRetVal = (pSynchronizeRoutine)( SynchronizeContext);
    PCrptSpinLockRelease( &pThis->InterruptSpinlock);
    return bRetVal;
}

