/**********************************************************************

    Module: testio.c

        Product: kernel mode driver for "PowerCrypt" encryption
        acceleration board based on Hi/fn 7751 chip
        by Ashley Laurent, Inc.

    ---------------------------------------------------------------

    Copyright:

        Ashley Laurent, Inc., 1999

        All Rights Reserved.

    ---------------------------------------------------------------

    Functional Description:

        Test program for the FreeBSD  and Linux PowerCrypt drivers.
        Opens the driver device and sends IOCTL requests.

    ---------------------------------------------------------------

    Environment:

        FreeBSD user mode

    ---------------------------------------------------------------

    Revision history:

        *   05-Mar-99   initial revision in NT
        *   30-Apr-99   changed system calls to their Unix variants

**********************************************************************/
/*
 * This program tests the driver's ability to perform DES encryption
 *
 * Then it tries to open SLOTS_TO_OPEN slots and then closes all the opened
 * slots.
 */

#include <stdio.h>
#include <sys/errno.h>
#include <sys/fcntl.h>
#include <sys/ioctl.h>

/*
 * IOCTL interface to the driver
 */
#include "pwrcryptioctl.h"

/*
 * The encode command that we're going to send
 */
#include "descomm.h"

/*
 * Make C more human-readable
 */
#define bool_t  int

#ifndef FALSE
#   define FALSE    (0)
#endif
#ifndef TRUE
#   define TRUE (1)
#endif
#ifndef NULL
#   define NULL 0
#endif

#define LOWORD(l)   ((unsigned short)(l))

#ifdef TRIPLE_DES
#define ALGO CRYPCOM_ALGO_3DES
#else
#define ALGO CRYPCOM_ALGO_DES
#endif

#ifdef USE_ECB
#define DESMODE CRYPCOM_NEW_IV | CRYPCOM_DES_MODE_ECB
#else /* use CBC by default */
#define DESMODE CRYPCOM_NEW_IV | CRYPCOM_DES_MODE_CBC
#endif

/*
 * Define how many slots we will try to open, the data buffer size and
 * keys for the encode/decode operations 
 */
#define SLOTS_TO_OPEN          5 
#define BUFFER_SIZE_TO_ALLOCATE 32*1024

#ifndef 3DES
static  unsigned char   DESKey[8]   = { 0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77, 0x88 };

#else
static  unsigned char   DESKey[24]   = { 0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77, 0x88, 0x99, 0xAA, 0xBB, 0xCC, 0xDD, 0xEE, 0xFF, 0, 1, 0x23, 0x45, 0x67, 0x89,0, 0, 0 };
#endif

static  unsigned char   DESIV[8]   = { 01, 0x23, 0x45, 0x67, 0x89, 0xab, 0xcd, 0xef };


int main (int argc, char *argv[])
{
    /* We will issue 3 different IOCTLs - here are their data packets
     */
    IOCTL_PWRCRYPT_RESERVE_SLOT_INPUT_OUTPUT    ReserveSlotInOutPack;
    IOCTL_PWRCRYPT_FREE_SLOT_INPUT_OUTPUT       FreeSlotInOutPack;
    IOCTL_PWRCRYPT_CALCULATE_INPUT_OUTPUT       CalcInOutPack;

    /* Slot number that we reserve for our single calculation request
     */
    u_long   CalcSlotNum;

    /* Statistics counter
     */
    u_long   nReqsSent;

    /* Ptrs to 2 parts of the command that we use
     */
    PDES_COMMAND    pCom;
    PDES_RESULT     pRlt;

    /* We will try to open as many slots as we can - here are the
     * storage for returned slot numbers
     */
    u_long      SlotNumber[ SLOTS_TO_OPEN];
    u_long      nSlotsOpened;
    bool_t      bFailure;

    int     fd;                 /* file descriptor for PcrptUMI device */

    unsigned char    *pBuf, *pBuf2;
    int     i;
    int     chKey;

#ifdef TRIPLE_DES
    printf("This version of testio compiled for Triple DES testing\n");
#endif
    

    /*
     * Allocate a single data buffer
     */
    pBuf = (unsigned char*) malloc( BUFFER_SIZE_TO_ALLOCATE);
    if ( pBuf == NULL)
    {
        printf( "malloc failed\n");
        return 1;
    }

    pBuf2 = (unsigned char*) malloc( BUFFER_SIZE_TO_ALLOCATE);
    if ( pBuf2 == NULL)
    {
        printf( "malloc failed\n");
        return 1;
    
    }

    memset(pBuf2, NULL, BUFFER_SIZE_TO_ALLOCATE);

    printf("testio:buffer %08x:%08x\n", pBuf, BUFFER_SIZE_TO_ALLOCATE);
    printf("testio:buffer %08x:%08x\n", pBuf2, BUFFER_SIZE_TO_ALLOCATE);

    /*
     * Open a device created by the driver
     */
    fd = open( "/dev/"PWRCRYPT_UM_INTERFACE_DEVICE_NAME"",
               O_RDWR,
               0);

    if ( fd == -1)
    {
        perror( "CreateFile 1 failed");
        free( pBuf);
	free( pBuf2);
        return 1;
    }

    /*
     * Reserve a slot to be able to make a calculation request
     */
    bFailure = ioctl( fd,
                      IOCTL_PWRCRYPT_RESERVE_SLOT,
                      &ReserveSlotInOutPack) == -1;

    if ( !bFailure)
    {
        /* Remember the reserved slot number
         */
        CalcSlotNum = ReserveSlotInOutPack.nReservedSlotNumber;
        printf( "Reserved calc slot number 0x%04x\n", CalcSlotNum);
    }
    else
    {
        perror( "Calc slot reservation has failed");
        free( pBuf);
	free( pBuf2);
        return 1;
    }

    /*
     * Fill the allocated buffer with some eye-catching data
     */
    for ( i = 0; i < BUFFER_SIZE_TO_ALLOCATE; i++)
    {
        pBuf[ i] = (char) i;
    }

    nReqsSent = 0;
    while( nReqsSent < 2) /*nReqsSent < 1) */
    {
        printf( "Requests sent %d\n", nReqsSent);
        nReqsSent += 2;

        /*
         * Output some data from the buffer
         */
        printf("Source data\n");
        for ( i = 0; i < 16; i++)
        {
            printf( "%02x ", pBuf[ i]);
        }
        printf("\n");

        /************************************************************************
         * Assemble the calc.request
         */
        CalcInOutPack.nSlotNumber = CalcSlotNum;

        CalcInOutPack.pSrcData = pBuf;
        CalcInOutPack.nTotalSrcCnt = BUFFER_SIZE_TO_ALLOCATE;

        CalcInOutPack.pDstBuffer = pBuf2;
        CalcInOutPack.nTotalDstCnt = CalcInOutPack.nTotalSrcCnt;

        /*-----------------------------------------------------------------------
         * Assemble the 7751 instruction in the allocated memory
         */

        /* These two ptrs are needed because we don't use the two areas'
         * structures as they are defined in the IOCTL buffer. This is
         * the proper use, however, since the definition contains mere
         * placeholders which are _intended_ to be filled differently
         * depending on required instruction
         */
        pCom = ( PDES_COMMAND) &CalcInOutPack.Command;
        pRlt = ( PDES_RESULT) &CalcInOutPack.Result;

        /*
         * Fill in the Command part of the 7751 instruction.
         */
        BASECOM_FILL(
            &pCom->base,
            BASECOM_SCRAMBLE | BASECOM_CRYP,
            CalcInOutPack.nTotalSrcCnt,
            CalcInOutPack.nTotalDstCnt);

        /*
         * (0 is the size of data to skip at the head of buffer. We skip nothing)
         */
        CRYPCOM_FILL(
            &pCom->cryp,
            CRYPCOM_NEW_KEY | CRYPCOM_DONT_STORE_CTX | ALGO | DESMODE,
            0,
            CalcInOutPack.nTotalSrcCnt);

        /*
         * Finally, copy enc.key and IV into DES_COMMAND
         */
        memcpy( &pCom->EncKey[0], &DESKey[0], sizeof( pCom->EncKey));
        memcpy( &pCom->EncIV[0],  &DESIV[0],  sizeof( pCom->EncIV));

        /*
         * Send the calc.req.
         */
        if ( ioctl( fd,
                    IOCTL_PWRCRYPT_CALCULATE,
                    &CalcInOutPack) == -1)
        {
            perror( "Encode failed");
            printf( "Press a key...");
            chKey = getchar();
        }

        /*
         * Output some data from the buffer
         */
        printf("Encoded data rlt sz=%d\n", CalcInOutPack.nResultSize);
        for ( i = 0; i < 16; i++)
        {
            printf( "%02x ", pBuf2[ i]);
        }
        printf("\n");

        /************************************************************************

         * Now decode the same data
         *
         * Fill in the Command part of the 7751 instruction.
         */
        BASECOM_FILL(
            &pCom->base,
            BASECOM_UNSCRAMBLE | BASECOM_CRYP,
            CalcInOutPack.nTotalSrcCnt,
            CalcInOutPack.nTotalDstCnt);

        /*
         * (0 is the size of data to skip at the head of buffer. We skip nothing)
         */
        CRYPCOM_FILL(
            &pCom->cryp,
            CRYPCOM_NEW_KEY | CRYPCOM_DONT_STORE_CTX | ALGO | DESMODE, 
            0,
            CalcInOutPack.nTotalSrcCnt);

	/* Switch the input and output */
	
	CalcInOutPack.pSrcData = pBuf2;
	CalcInOutPack.pDstBuffer = pBuf;

	/* Zero out the buffer so we can see if it came back ok */

	memset(pBuf, 0, CalcInOutPack.nTotalDstCnt);
 
        /*
         * Send the calc.req.
         */
        if ( ioctl( fd,
                    IOCTL_PWRCRYPT_CALCULATE,
                    &CalcInOutPack) == -1)
        {
            perror( "Decode failed");
            printf( "Press a key...");
            chKey = getchar();
        }

        /*
         * Output some data from the buffer
         */
        printf("Decoded data rlt sz=%d\n", CalcInOutPack.nResultSize);
        for ( i = 0; i < 16; i++)
        {
            printf( "%02x ", pBuf[ i]);
        }
        printf("\n");
    }
/*    return 0; */
    /************************************************************************
     * Free the testing data buffer
     */
    free( pBuf);
    free( pBuf2);

    close(fd);
    return 0;

    /* NOTREACHED */

    /*
     * The next test - try to open the defined number of slots
     */
    nSlotsOpened = 0;
    bFailure = FALSE;

    while ( nSlotsOpened < SLOTS_TO_OPEN && !bFailure)
    {
        /* Send RESERVE_SLOT request
         */
        bFailure = ioctl( fd,
                          IOCTL_PWRCRYPT_RESERVE_SLOT,
                          &ReserveSlotInOutPack) == -1;
        if ( !bFailure)
        {
            /* Remember the reserved slot number in the array
             */
            SlotNumber[ nSlotsOpened] = ReserveSlotInOutPack.nReservedSlotNumber;
            printf( "Reserved slot number 0x%04x\n", SlotNumber[ nSlotsOpened]);
            nSlotsOpened++;
        }
        else
        {
            printf( "Slot reservation #%d has failed: %s\n", nSlotsOpened, sys_errlist[errno]);
        }
    }

    printf( "Press a key...");
    chKey = getchar();

    /*
     * The next test - close the slots that we've opened
     */
    printf("nSlotsOpened = %d\n", nSlotsOpened);
    while ( nSlotsOpened > 0)
    {
        /* We always move the index to the next array entry, even if the slot
         * freeing has failed. Notice, that that failure should never normally
         * happen, since we try to free only our slots that we've reserved
         */
        nSlotsOpened--;

        /* Take a stored slot number from the array and put it into the IOCTL packet
         */
        FreeSlotInOutPack.nSlotNumberToFree = SlotNumber[ nSlotsOpened];

        /* Send FREE_SLOT request
         */
        bFailure = ioctl( fd,
                          IOCTL_PWRCRYPT_FREE_SLOT,
                          &FreeSlotInOutPack) == -1;
        if ( !bFailure)
        {
            printf( "Freed slot number 0x%04x\n", SlotNumber[ nSlotsOpened]);
        }
        else
        {
            printf( "Slot freeing for slot 0x%04x has failed: %s\n", SlotNumber[ nSlotsOpened], sys_errlist[errno]);
        }
    }

    printf( "Press a key...");
    chKey = getchar();

    /*
     * Close the device handle
     */
    close( fd);

    return 0;
}

