/**********************************************************************

    Module: userbuf.h

        Product: kernel mode driver for "PowerCrypt" encryption
        acceleration board based on Hi/fn 7751 chip
        by Ashley Laurent, Inc.

    ---------------------------------------------------------------

    Copyright:

        Ashley Laurent, Inc., 1999
        All Rights Reserved.

    ---------------------------------------------------------------

    Functional Description:

        FreeBSD implementation of functions for locking pageable buffer
        with MDL creation and for unlocking it with MDL desctruction.

    ---------------------------------------------------------------

    Environment:

        FreeBSD kernel mode

    ---------------------------------------------------------------

    Author:

        Stan Shkolnyy

    ---------------------------------------------------------------

    Revision history:

        *   9-May-99   initial revision

**********************************************************************/

#include "osindependence.h"
#include "debug.h"
#include "userbuf.h"

ULONG   BytesLocked = 0;

/****************************************************************************
 * Allocate/free MDL
 */
static
PMDL
AllocateMdl(
    IN PVOID  VirtualAddress,
    IN ULONG  Length)
{
    PMDL            pMdl;
    CSHORT          MdlSize;
    ULONG           PagesInMdl;

    DebugPrint(( 3, "AllocateMdl %08x:%08x\n", VirtualAddress, Length));

    /*
     * MDL consists of struct MDL + array of physical addresses of pages
     */
    PagesInMdl = ADDRESS_AND_SIZE_TO_SPAN_PAGES( VirtualAddress, Length);
    MdlSize = sizeof( *pMdl) + PagesInMdl * sizeof( ULONG);

    pMdl = malloc( MdlSize, M_TEMP, M_WAITOK);
    {
        PULONG  pPages;
        ULONG   iPage;

        pMdl->Size          = MdlSize;
        pMdl->StartVa       = PAGE_ALIGN( VirtualAddress);
        pMdl->ByteCount     = Length;
        pMdl->ByteOffset    = BYTE_OFFSET( VirtualAddress);

        /*
         * Fill the page array with invalid values of -1
         */
        pPages = (PULONG)(pMdl + 1);
        for ( iPage = 0; iPage < PagesInMdl; iPage++)
        {
            pPages[ iPage] = (ULONG) -1;
        }
    }
    return pMdl;
}

#define FreeMdl( pMdl)  free( pMdl, M_TEMP)

/****************************************************************************
 * Verify the caller's right to access the buffer in specified manner,
 * bring it in memory, lock and fill the MDL with the pages that back
 * the locked buffer.
 *
 * The function's ideas are borrowed from the physio() (kern_physio.c)
 * and vmapbuf() (machdep.c) implementations.
 */

static
PWRCRYPT_STATUS 
ProbeAndLockPages(
    IN OUT  PMDL                    pMdl,
    IN      PWRCRYPT_PROCESSOR_MODE BufAccessMode,
    IN      BOOLEAN                 bWriteAccessRequired
    )
{
    caddr_t     pBuf;
    ULONG       BufSize;
    caddr_t     CurPageVa;
    vm_offset_t CurPagePa;

    ULONG       PagesInBuffer;
    CSHORT      NeededMdlSize;
    PULONG      pPages;
    ULONG       iPage;

    pBuf = (caddr_t) MmGetMdlVirtualAddress( pMdl);
    BufSize = MmGetMdlByteCount( pMdl);

    DebugPrint(( 3, "ProbeAndLockPages %08x:%08x\n", pBuf, BufSize));

    /*
     * MDL consists of struct MDL + array of physical addresses of pages.
     * Is the supplied MDL big enough to describe the buffer?
     */
    PagesInBuffer = ADDRESS_AND_SIZE_TO_SPAN_PAGES( pBuf, BufSize);
    NeededMdlSize = sizeof( *pMdl) + PagesInBuffer * sizeof( ULONG);
    if ( NeededMdlSize > pMdl->Size)
    {
        panic("PwrCrypt.ProbeAndLockPages: MDL too small");
    }

    /*
     * The following is from the physio(), only physio() works with user
     * memory only and therefore doesn't use kernacc()
     */
    if ( BufAccessMode == UserMode)
    {
        DebugPrint(( 3, "    useracc\n"));
#if __FreeBSD__ > 4
        if ( !useracc( pBuf, BufSize, bWriteAccessRequired ? VM_PROT_WRITE : VM_PROT_READ))
#else
        if ( !useracc( pBuf, BufSize, bWriteAccessRequired ? B_WRITE : B_READ))
#endif
        {
            DebugPrint(( 3, "-fail\n"));
            return PWRCRYPT_STATUS_INVALID_USER_BUFFER;
        }
    }
    else
    {
        DebugPrint(( 3, "    kernacc\n"));
#if __FreeBSD__ > 4
        if ( !kernacc( pBuf, BufSize, bWriteAccessRequired ? VM_PROT_WRITE : VM_PROT_READ))
#else
        if ( !kernacc( pBuf, BufSize, bWriteAccessRequired ? B_WRITE : B_READ))
#endif
        {
            DebugPrint(( 3, "-fail\n"));
            return PWRCRYPT_STATUS_INVALID_USER_BUFFER;
        }
    }
    DebugPrint(( 3, "\n"));

    /*
     * Then physio() calls vmapbuf().
     * The following is from the vmapbuf(), only vmapbuf() also maps the memory
     * into the kernel VA space and this func doesn't. 
     *
     * Fill the MDL's virtual address fields
     * Fill the MDL's page array with physical page addresses
     */
    pMdl->StartVa       = PAGE_ALIGN( pBuf);
    pMdl->ByteCount     = BufSize;
    pMdl->ByteOffset    = BYTE_OFFSET( pBuf);

    DebugPrint((3, "StartVa=0x%x, ByteOffset=0x%x\n", pMdl->StartVa,
		pMdl->ByteOffset)); 
    
    /*
     * Point to the MDL's page address array
     */
    pPages = (PULONG)(pMdl + 1);

    for( CurPageVa = (caddr_t) trunc_page((vm_offset_t) pBuf),
            iPage = 0;
         CurPageVa < pBuf + BufSize &&
            iPage < PagesInBuffer;
         CurPageVa += PAGE_SIZE,
            iPage++)
    {
        vm_fault_quick(
            CurPageVa,
            bWriteAccessRequired ? (VM_PROT_READ | VM_PROT_WRITE) : VM_PROT_READ);
	
        CurPagePa = trunc_page( pmap_kextract( (vm_offset_t) CurPageVa));
        DebugPrint(( 3, "page %ld, va=0x%x, pa=0x%08x\n", iPage, CurPageVa, CurPagePa));
        if ( CurPagePa == 0)
        {
            panic("PwrCrypt.ProbeAndLockPages: page not present");
        }
        vm_page_hold( PHYS_TO_VM_PAGE( CurPagePa));
        pPages[ iPage] = CurPagePa;
    }

    return PWRCRYPT_STATUS_SUCCESS;
}

/****************************************************************************
 * Perform the operation reverse to ProbeAndLockPages()
 * Note, that the only purpose this function can use the buffer VA and length
 * for is the calculation of the number of pages backing the buffer. This is
 * because it is invoked in the arbitrary context
 */
static
void
UnlockPages(
    IN OUT  PMDL    pMdl
    )
{
    caddr_t     pBuf;
    ULONG       BufSize;
    caddr_t     CurPageVa;
    vm_offset_t CurPagePa;

    ULONG       PagesInBuffer;
    CSHORT      NeededMdlSize;
    PULONG      pPages;
    ULONG       iPage;

    pBuf = (caddr_t) MmGetMdlVirtualAddress( pMdl);
    BufSize = MmGetMdlByteCount( pMdl);

    DebugPrint(( 3, "UnlockPages %08x:%08x\n", pBuf, BufSize));

    /*
     * MDL consists of struct MDL + array of physical addresses of pages.
     * Is the supplied MDL big enough to describe the buffer?
     */
    PagesInBuffer = ADDRESS_AND_SIZE_TO_SPAN_PAGES( pBuf, BufSize);
    NeededMdlSize = sizeof( *pMdl) + PagesInBuffer * sizeof( ULONG);
    if ( NeededMdlSize > pMdl->Size)
    {
        panic("PwrCrypt.UnlockPages: MDL corrupted");
    }

    /*
     * Unlock the memory described by the MDL
     */
    pPages = (PULONG)(pMdl + 1);
    for( iPage = 0; iPage < PagesInBuffer; iPage++)
    {
        vm_page_unhold( PHYS_TO_VM_PAGE( pPages[ iPage]));
    }
}

/****************************************************************************
 * Lock user buffer and create new MDL for it
 */
PWRCRYPT_STATUS
LockBufferCreateMdl (
    IN  PVOID            pBuf,
    IN  ULONG            BufSize,
    IN  BOOLEAN          bChargeQuota,
    IN  PWRCRYPT_PROCESSOR_MODE  AccessMode,
    IN  BOOLEAN          bWriteAccessRequired,
    OUT PMDL             *ppMdl
    )
{
    PWRCRYPT_STATUS status;        /* func.ret.value */
    PMDL            pMdl;
    CSHORT          MdlSize;
    ULONG           PagesInMdl;

    DebugPrint((3, "LockUserBuffer\n"));

    *ppMdl = NULL;

    status = PWRCRYPT_STATUS_NO_MEMORY;

    pMdl = AllocateMdl( pBuf, BufSize);
    if ( pMdl != NULL)
    {
        /*
         * Now probe and lock the buffer described by the Mdl. On error,
         * deallocate the Mdl and return the appropriate status.
         */
        status = ProbeAndLockPages(
                    pMdl,
                    AccessMode,
                    bWriteAccessRequired
                    );
        if ( BSD_SUCCESS( status))
        {
            /* If we are here, the call has worked - return MDL* to the caller
             *
             * Now, as it can be seen
             * pMdl->MdlFlags = MDL_PAGES_LOCKED | MDL_WRITE_OPERATION
             */
            *ppMdl = pMdl;

            BytesLocked += MmGetMdlByteCount(pMdl);

            /* here I verify my understanding of things */
            ASSERT( pBuf == MmGetMdlVirtualAddress( pMdl));
            ASSERT( BufSize == MmGetMdlByteCount( pMdl));
/*            DebugPrint((3, "    pbuf=%08x size=%08x mdl=%08x\n", pBuf, BufSize, pMdl)); */
        }
        else
        {
            ASSERT(0);
            DPF("    ProbeAndLockPages error 0x%08x\n", status);
            FreeMdl( pMdl);
        }
    }
    else
    {
        DebugPrint((2, "    Mdl alloc.failed\n"));
        ASSERT(0);
    }

    return status;
}


/****************************************************************************

UnlockBufferDelMdl()

Description:
    Unlocks the buffer specified by the MDL and deletes MDL

Arguments:

Return:
    None

****************************************************************************/
void
UnlockBufferDelMdl (
    IN PMDL    pMdl
    )
{
    DebugPrint((3, "UnlockUserBuffer pbuf=%08x size=%08x mdl=%08x\n", MmGetMdlVirtualAddress( pMdl), MmGetMdlByteCount(pMdl), pMdl));

    BytesLocked -= MmGetMdlByteCount(pMdl);

    UnlockPages( pMdl);
    FreeMdl( pMdl);
}
