/**********************************************************************

    Module: dma_ctx.c

        Product: kernel mode driver for "PowerCrypt" encryption
        acceleration board based on Hi/fn 7751 chip
        by Ashley Laurent, Inc.

    ---------------------------------------------------------------

    Copyright:

        Ashley Laurent, Inc., 1999
        All Rights Reserved.

    ---------------------------------------------------------------

    Functional Description:

        Implementation of the DMA_CTX object that presents the interface
        through which PROCESSOR uses computer DMA facility

    ---------------------------------------------------------------

    Environment:

        NT or BSD kernel mode

    ---------------------------------------------------------------

    Author:

        Stan Shkolnyy

    ---------------------------------------------------------------

    Revision history:

        *   05-Mar-99   initial revision

**********************************************************************/

#include "osindependence.h"
#include "driverdef.h"
#include "xfer.h"
#include "debug.h"
#include "dma_ctx.h"

/****************************************************************************
****************************************************************************/
void
DmaCtxConstructor(
    PDMA_CTX            pThis
    )
{
    DPF("DmaCtxConstructor\n");

    pThis->Type = DMA_CTX_SIGN;

    pThis->pCommBuf = NULL;

#ifdef DBG
    pThis->MapRegsClaimed = 0;
    pThis->OverlapCnt = 0;
    pThis->MapsInProgressCnt = 0;
#endif

}

/****************************************************************************
****************************************************************************/
void
DmaCtxDestructor(
    PDMA_CTX            pThis
    )
{
    DPF("DmaCtxDestructor\n");
    ASSERT( pThis->Type == DMA_CTX_SIGN);

    /*
     * Free common buffer.
     */
    if ( pThis->pCommBuf != NULL)
    {
        DPF("    Freeing common buffer at 0x%08x\n", pThis->pCommBuf);
        free( pThis->pCommBuf, M_DEVBUF);
    }
}

/****************************************************************************
****************************************************************************/
int
DmaCtxInit(
    IN OUT PDMA_CTX     pThis,
    IN ULONG            MapRegsDesired,
    IN ULONG            MaxXferSize
    )
{
    DPF("DmaCtxGetAdapter\n");
    ASSERT( pThis->Type == DMA_CTX_SIGN);
    ASSERT( pThis->pCommBuf == NULL);

    pThis->MaxMapRegs = MapRegsDesired;

    return STATUS_SUCCESS;
}

/****************************************************************************
****************************************************************************/
PVOID
DmaCtxAllocateDmaMemory(
    IN OUT  PDMA_CTX    pThis,
    IN      ULONG       SgCtlAreaSize,
    OUT     ULONG       *pDmaAddr
    )
{
    BOOLEAN     bSuccess;

    DPF("DmaCtxAllocateDmaMemory size=0x%08x\n", SgCtlAreaSize);
    ASSERT( pThis->Type == DMA_CTX_SIGN);

    ASSERT( pThis->pCommBuf == NULL);

    //
    // BUGBUG
    //
    pThis->CommBufSize = 4096; //SgCtlAreaSize;

    pThis->pCommBuf = malloc( pThis->CommBufSize, M_DEVBUF, M_WAITOK);
    ASSERT( ( (ULONG)pThis->pCommBuf & 3) == 0);

    pThis->CommBufPA = vtophys( pThis->pCommBuf);
    *pDmaAddr = pThis->CommBufPA;

    DPF("    Allocated common buffer at VA=0x%08x PA=0x%08x Size=0x%04x\n", pThis->pCommBuf, pThis->CommBufPA, pThis->CommBufSize);

    return pThis->pCommBuf;
}

/****************************************************************************
****************************************************************************/
void 
DmaCtxAllocateAdapterChannel(
    IN OUT PDMA_CTX     pThis,
    IN ULONG            NumberOfMapRegisters,
    IN PMDL             pMdl,
    IN PMAP_CONTROL     MapControlRoutine,
    IN PVOID            MapControlRoutineContext
    )
{
    DebugPrint((2, "    mdl=0x%08x #mapregs= 0x%02x\n", pMdl, NumberOfMapRegisters));
    ASSERT( pThis->Type == DMA_CTX_SIGN);

    ASSERT( NumberOfMapRegisters == 0x41);

#ifdef DBG
    if ( pThis->MapRegsClaimed != 0)
    {
        pThis->OverlapCnt++;
    }
    pThis->MapRegsClaimed += NumberOfMapRegisters;

    /*
     * The driver algorithms are not supposed to claim more map regs than
     * it is available. In practice, this would probably lead
     * to pDevice to be queued despite the fact that nobody is executing
     * its AdapterControl, due to map regs shortage. Again, I don't know
     * for sure. BUGBUG - however, I don't rely on this in the driver, so
     * it doesn't matter.
     */
    ASSERT( pThis->MapRegsClaimed <= pThis->MaxMapRegs);

    pThis->MapsInProgressCnt++;
#endif

    /* Call the routine registered by our client. In that routine, it will
     * map its phys.RAM range to log.DMA range by calling our DmaCtxMapTransfer
     * one or more times
     */
    (MapControlRoutine)( MapControlRoutineContext, (PVOID) 0x12345678);

}

/****************************************************************************
#define MmGetMdlVirtualAddress(Mdl)  ((PVOID) ((PCHAR) (Mdl)->StartVa + (Mdl)->ByteOffset))
#define MmGetMdlByteCount(Mdl)  ((Mdl)->ByteCount)
#define MmGetMdlByteOffset(Mdl)  ((Mdl)->ByteOffset)
****************************************************************************/
ULONG
DmaCtxMapTransfer(
    IN OUT PDMA_CTX     pThis,
    IN PVOID            MapRegisterBase,
    IN PMDL             pMdl,
    IN PVOID            CurrentVa,
    IN OUT PULONG       pLength
    )
{
    ULONG   RetLogAddr;
    ULONG   MaxLength;

    PULONG  pPages;
    ULONG   iPage;

    DPF( "DmaCtxMapTransfer mdl=0x%08x\n", pMdl);
    ASSERT( pThis->Type == DMA_CTX_SIGN);
    ASSERT( (ULONG) MapRegisterBase == 0x12345678);

    #ifdef DBG

        /*
         * Is the caller-specified area out of the lower MDL bound?
         */
        if ( CurrentVa < MmGetMdlVirtualAddress( pMdl))
        {
            DPF( "    invalid CurrentVa\n");
        }

        /*
         * Is the caller-specified area out of the upper MDL bound?
         */
        if ( (PUCHAR) CurrentVa + *pLength >
             (PUCHAR) (MmGetMdlVirtualAddress( pMdl)) + MmGetMdlByteCount( pMdl))
        {
            DPF( "    invalid *pLength or CurrentVa\n");
        }
    #endif

    /*
     * Calculate DMA logical address which is equal to RAM physical address
     * on Intel machines
     */
    pPages = (PULONG)(pMdl + 1);
    iPage = ( (ULONG) ( PAGE_ALIGN( CurrentVa)) - (ULONG) pMdl->StartVa) >> PAGE_SHIFT;
    RetLogAddr = pPages[ iPage] + BYTE_OFFSET( CurrentVa);

    /*
     * Calculate the length of the mapped buffer chunk to return to the caller
     * (it is always <= PAGE_SIZE). *pLength is both IN and OUT. If it specifies
     * less then we can map, we don't touch it, else we assign the length
     * we can map to it.
     */
    MaxLength = PAGE_SIZE - BYTE_OFFSET( CurrentVa);
    if ( *pLength > MaxLength)
    {
        *pLength = MaxLength;
    }
    return RetLogAddr;
}

/****************************************************************************
****************************************************************************/
VOID
DmaCtxCompleteTransfer(
    IN OUT PDMA_CTX     pThis,
    IN PVOID            MapRegisterBase,
    IN ULONG            NumberOfMapRegisters,
    IN PMDL             pMdl
//    IN PVOID            TransferVa,
//    IN ULONG            TransferLength
    )
{
    DPF("DmaCtxCompleteTransfer    mdl=0x%08x #mapregs= 0x%02x\n", pMdl, NumberOfMapRegisters);
    ASSERT( pThis->Type == DMA_CTX_SIGN);
    ASSERT( (ULONG) MapRegisterBase == 0x12345678);
    ASSERT( NumberOfMapRegisters == 0x41);

    #ifdef DBG
        ASSERT( pThis->MapsInProgressCnt <= DMA_CTX_MAX_PARALLEL_MAPS &&
                pThis->MapsInProgressCnt > 0);
        pThis->MapsInProgressCnt--;
        pThis->MapRegsClaimed -= NumberOfMapRegisters;
    #endif
}
