/**********************************************************************

    Module: instrset.h

        Product: kernel mode driver for "PowerCrypt" encryption
        acceleration board based on Hi/fn 7751 chip
        by Ashley Laurent, Inc.

    ---------------------------------------------------------------

    Copyright:

        Ashley Laurent, Inc., 1999
        All Rights Reserved.

    ---------------------------------------------------------------

    Functional Description:

        Defines structure of the 7751 commands and results (instruction set).

        A client assembles 7751 commands and puts them into the request packets
        sent to the driver. The structure of the request packet, which includes
        space for a command and a result, is defined in pwrcryptioctl.h. The
        driver returns results by copying them into fields of the same request
        packet.

        Bear in mind that data structures defined in this file are
        HARDWARE-DEFINED and hence can't be changed.


    ---------------------------------------------------------------

    Environment:

        NT or BSD kernel and user mode

    ---------------------------------------------------------------

    Author:

        Stan Shkolnyy

    ---------------------------------------------------------------

    Revision history:

        *   17-Feb-99   initial revision

**********************************************************************/

#ifndef __INSTRSET_H__
#define __INSTRSET_H__

#ifdef _WIN32
    #pragma message("_WIN32 defined")
    #pragma pack(push, _INSTRSET_STRUCT_DEF_, 1)
#else
/*    #pragma message("_WIN32 not defined") */
/*    #warning No pragma pack() ??? */
#endif

/*
 * 7751 has 18 bits for each data "total counter", hence max.possible
 * DMA transfer = 256KB - 4 (with respect to the 4-byte counter alignment)
 */
#define PROCESSOR_MAX_XFER_SIZE   (256*1024 - 4)

/* Counters inside of commands are broken into two fragments - high 2 bits
 * and low 16 bits. These macros help to break ULONGs.
 */
#define LOWORD16(l)   ((unsigned short)(l))
#define HIWORD02(l)   ((unsigned short)(((unsigned long)(l) >> 16) & 3))

/****************************************************************************
 ****************************************************************************
 * Base command structure - the first data that comes to the processor
 */
typedef struct _BASE_COMMAND
{
    unsigned short  Command;
    unsigned short  SlotNo        :12;
    unsigned short  DstCntHi      :2;
    unsigned short  SrcCntHi      :2;
    unsigned short  SrcCntLo;
    unsigned short  DstCntLo;

} BASE_COMMAND, *PBASE_COMMAND;

/****************************************************************************
 * The smallest command that may be issued - includes base command and 8 bytes
 * of zeros that do not have other effects than the command DMA engine's happiness
 */
typedef struct _SMALLEST_COMMAND
{
    BASE_COMMAND    base;
    unsigned char   zerobytes[ 8];

} SMALLEST_COMMAND, *PSMALLEST_COMMAND;

/*
 * Bits for the Command member.
 */

/*
 * Direction of the operation - either encrypt, comp or decrypt, decomp
 */
#define BASECOM_SCRAMBLE        0x0000
#define BASECOM_UNSCRAMBLE      0x2000
#define BASECOM_READ_RAM        0x4000
#define BASECOM_WRITE_RAM       0x6000

/*
 * Which operations to perform - any combination or none
 */
#define BASECOM_CRYP            0x0800
#define BASECOM_MAC             0x0400
#define BASECOM_PAD             0x0200
#define BASECOM_COMP            0x0100

/*
 * Misc
 */
#define BASECOM_DISABLE_DEST_FIFO   0x1000

/*
 * Macros and functions to easy the work with the BaseCommand
 */
#define BASECOM_IS_COMP( pBaseCom)      ( (pBaseCom)->Command & BASECOM_COMP)
#define BASECOM_IS_PAD( pBaseCom)       ( (pBaseCom)->Command & BASECOM_PAD)
#define BASECOM_IS_MAC( pBaseCom)       ( (pBaseCom)->Command & BASECOM_MAC)
#define BASECOM_IS_CRYP( pBaseCom)      ( (pBaseCom)->Command & BASECOM_CRYP)

#define BASECOM_FILL( pBaseCom, CommBits, SrcCnt, DstCnt) \
                                                                        \
    (pBaseCom)->Command = CommBits;                                     \
    (pBaseCom)->DstCntHi = HIWORD02( DstCnt);                           \
    (pBaseCom)->SrcCntHi = HIWORD02( SrcCnt);                           \
    (pBaseCom)->SrcCntLo = LOWORD16( SrcCnt);                           \
    (pBaseCom)->DstCntLo = LOWORD16( DstCnt)

#define BASECOM_GET_COMCODE_BITS( pBaseCom)  ( (pBaseCom)->Command & 0xe000)

/****************************************************************************
 ****************************************************************************
 * Crypto command structure - follows the base command on the way to the
 * processor
 */
typedef struct _CRYP_COMMAND
{
    unsigned short  Command       :14;
    unsigned short  SrcCntHi      :2;
    unsigned short  SrcHdrCnt;
    unsigned short  SrcCntLo;
    unsigned short  Reserved;

} CRYP_COMMAND, *PCRYP_COMMAND;

/* Bits for SetCommand()'s parm.
 * Some of the following flags must be OR-ed and the result passed to the
 * method. The method assigns 0 to any other bits.
 */

/* Indicate that additional context info follows the command
 */
#define CRYPCOM_NEW_IV          0x1000
#define CRYPCOM_NEW_KEY         0x0800

/* Indicates that the context is not to be written to on-board RAM -
 * the RAM remains UNCHANGED ( hi/fn p.62)
 */
#define CRYPCOM_DONT_STORE_CTX  0x0040

/* If enc.algorithm is not DES or 3DES, set "non-DES" constant
 * otherwise, set one of "DES" constants
 */
#define CRYPCOM_MODE_BITMASK    0x0018
#define CRYPCOM_NON_DES_MODE    0x0000

#define CRYPCOM_DES_MODE_ECB    0x0000
#define CRYPCOM_DES_MODE_CBC    0x0008
#define CRYPCOM_DES_MODE_CFB    0x0010
#define CRYPCOM_DES_MODE_OFB    0x0018

#define CRYPCOM_ENC_ALG_DES     0x0000

/* Crypto-algorithm
 */
#define CRYPCOM_ALGO_BITMASK    0x0003
#define CRYPCOM_ALGO_DES        0x0000
#define CRYPCOM_ALGO_3DES       0x0001
#define CRYPCOM_ALGO_RC4        0x0002

/*
 * Macros and functions to easy the work with the CryptoCommand
 */
#define CRYPCOM_FILL( pComm, CommBits, SrcHeaderCnt, SrcCnt)     \
                                                                    \
    (pComm)->Command = CommBits;                                    \
    (pComm)->SrcCntHi = HIWORD02( SrcCnt);                          \
    (pComm)->SrcCntLo = LOWORD16( SrcCnt);                          \
    (pComm)->SrcHdrCnt = SrcHeaderCnt;                              \
    (pComm)->Reserved = 0

#define CRYPCOM_IS_NEW_CTX( pCrypCom)      ( (pCrypCom)->Command & CRYPCOM_NEW_KEY)
#define CRYPCOM_IS_DONT_STORE_CTX( pCrypCom)      ( (pCrypCom)->Command & CRYPCOM_DONT_STORE_CTX)

/****************************************************************************
 ****************************************************************************
 * Compression command structure - follows the base command on the way to the
 * processor
 */
typedef struct _COMP_COMMAND
{
    unsigned short  Command       :14;
    unsigned short  SrcCntHi      :2;
    unsigned short  SrcHdrCnt;
    unsigned short  SrcCntLo;
    unsigned short  Reserved;

} COMP_COMMAND, *PCOMP_COMMAND;

/* Performance mode - numerical value 0..15
 */
#define COMPCOM_PERF_BITMASK    0x0f00

/* Indicates that the 7751 must clear compression history before compression
 * (hifn p.53)
 */
#define COMPCOM_CLEAR_HISTORY   0x0010

/* Indicates that the 7751 must pass the data unaltered but update comp.history
 * as so the data were actually processed (hifn p.54)
 */
#define COMPCOM_UPDATE_HISTORY  0x0008

/* The same bit, but performs different actions depending on comp.method
 * (specified by bit 0) (hifn p.54)
 */ 
#define COMPCOM_LZS_STRIP0      0x0004
#define COMPCOM_MPPC_RESTART    0x0004

/* Compression algorithm (hifn p.55)
 */
#define COMPCOM_ALGO_BITMASK    0x0001
#define COMPCOM_ALGO_LZS        0x0000
#define COMPCOM_ALGO_MPPC       0x0001

/*
 * Macros and functions to easy the work with the CompCommand
 */
#define COMPCOM_FILL( pComm, CommBits, Performance, SrcHeaderCnt, SrcCnt)     \
                                                                    \
    (pComm)->Command = CommBits | ( Performance << 8);              \
    (pComm)->SrcCntHi = HIWORD02( SrcCnt);                          \
    (pComm)->SrcCntLo = LOWORD16( SrcCnt);                          \
    (pComm)->SrcHdrCnt = SrcHeaderCnt;                              \
    (pComm)->Reserved = 0

#define COMPCOM_IS_NEW_CTX( pCompCom)      ( (pCompCom)->Command & COMPCOM_CLEAR_HISTORY)

/****************************************************************************
 ****************************************************************************
 * Padding command structure - follows the base command on the way to the
 * processor. If the padding engine is the only engine enabled, 4 bytes of zero
 * must be appended to the end of the pad command
 */
typedef struct _PAD_COMMAND
{
    unsigned short  Command       :14;
    unsigned short  SrcCntHi      :2;
    unsigned short  SrcCntLo;

} PAD_COMMAND, *PPAD_COMMAND;

/*
 * Macros and functions to easy the work with the PadCommand
 */
#define PADCOM_FILL( pComm, CommBits, SrcCnt)     \
                                                                    \
    (pComm)->Command = CommBits;                                    \
    (pComm)->SrcCntHi = HIWORD02( SrcCnt);                          \
    (pComm)->SrcCntLo = LOWORD16( SrcCnt)

/****************************************************************************
 ****************************************************************************
 * Authentication command structure - follows the base command on the way to
 * the processor
 */
typedef struct _MAC_COMMAND
{
    unsigned short  Command       :14;
    unsigned short  SrcCntHi      :2;
    unsigned short  SrcHdrCnt;
    unsigned short  SrcCntLo;
    unsigned short  Reserved;

} MAC_COMMAND, *PMAC_COMMAND;

/* Indicate that additional context info follows the command
 */
#define MACCOM_NEW_KEY          0x0800

#define MACCOM_ORDER_IPSEC	0x0200

/* Indicates that the MAC key will be inserted into the data stream on encode
 * operation or stripped from the data stream on encode (hi/fn p.60)
 */
#define MACCOM_INSERT_MAC       0x0040

/* Specifies that the 7751 must add the calculated MAC value to the MAC result
 * structure on encode operation (hi/fn p.60)
 */
#define MACCOM_MAC_RESULT       0x0020

/* Specifies that the 7751 must truncate the calculated MAC value to 12 bytes
 * (hi/fn p.60). Otherwise, the value is 20b for SHA and 16b for MD5.
 */
#define MACCOM_MAC_TRUNCATE     0x0010

/* MAC-algorithm
 */

#define MACCOM_MODE_HMAC	0x0000
#define MACCOM_MODE_SSL		0x0004
#define MACCOM_MODE_HASH	0x0008

#define MACCOM_ALGO_BITMASK     0x0003
#define MACCOM_ALGO_SHA         0x0000
#define MACCOM_ALGO_MD5         0x0001

/*
 * Macros and functions to easy the work with the MACCommand
 */
#define MACCOM_FILL( pComm, CommBits, SrcHeaderCnt, SrcCnt)     \
                                                                    \
    (pComm)->Command = CommBits;                                    \
    (pComm)->SrcCntHi = HIWORD02( SrcCnt);                          \
    (pComm)->SrcCntLo = LOWORD16( SrcCnt);                          \
    (pComm)->SrcHdrCnt = SrcHeaderCnt;                              \
    (pComm)->Reserved = 0

#define MACCOM_IS_NEW_CTX( pMACCom)      ( (pMACCom)->Command & MACCOM_NEW_KEY)

/****************************************************************************
   The biggest command that may be issued - includes all the subcommands.
   It is defined for the only purpose of illustration and using in other
   structures as a placeholder for the biggest command
*/
typedef struct _BIGGEST_COMMAND
{
    BASE_COMMAND    base;
    COMP_COMMAND    comp;
    PAD_COMMAND     pad;
    MAC_COMMAND     mac;
    CRYP_COMMAND    cryp;
    unsigned char           MACKey  [  64];
    union
    {
        struct
        {
            unsigned char           EncKey  [  8];
            unsigned char           EncIV   [  8];
        } DESKeys;

        struct
        {
            unsigned char           EncKey  [ 24];
            unsigned char           EncIV   [  8];
        } TripleDESKeys;

        struct
        {
            unsigned char           EncKey  [ 260];
        } RC4Keys;

    } EncryptionContext;

} BIGGEST_COMMAND, *PBIGGEST_COMMAND;

/****************************************************************************
 ****************************************************************************
 * Base result structure - the last but one data that comes out of the
 * processor.
 * (Its more than close relation to the base command structure is obvious)
 */
typedef struct _BASE_RESULT
{
    unsigned short  Reserved;
    unsigned short  SessNo        :12;
    unsigned short  DstCntHi      :2;
    unsigned short  SrcCntHi      :2;
    unsigned short  SrcCntLo;
    unsigned short  DstCntLo;

} BASE_RESULT, *PBASE_RESULT;

/****************************************************************************
 ****************************************************************************
 * Compression result structure - follows the base result on the way out of the
 * processor
 */
typedef struct _COMP_RESULT
{
    unsigned short PrematureStop :1;     /*CompSrcCnt went to 0, but TotalSrcCnt still != 0 */
    unsigned short Reserved1     :15;
    unsigned short Reserved2;

} COMP_RESULT, *PCOMP_RESULT;

/****************************************************************************
 ****************************************************************************
 * Authentication result structure - follows the base result on the way out of the
 * processor
 */
typedef struct _MAC_RESULT
{
    unsigned short  PrematureStop :1;     /*MacSrcCnt went to 0, but TotalSrcCnt still != 0 */
    unsigned short  Miscompare    :1;
    unsigned short  Reserved1     :14;
    unsigned short  Reserved2;
    unsigned char  MAC[ 20];             /*20 bytes maximum */

} MAC_RESULT, *PMAC_RESULT;

/****************************************************************************
 ****************************************************************************
 * Crypto result structure - follows the base result on the way out of the
 * processor
 */
typedef struct _CRYP_RESULT
{
    unsigned short  PrematureStop :1;     /*EncSrcCnt went to 0, but TotalSrcCnt still != 0 */
    unsigned short  Reserved1     :15;
    unsigned short  Reserved2;

} CRYP_RESULT, *PCRYP_RESULT;

/****************************************************************************
Note, that PrematureStop may be normal, if prepending compression has decreased
the amount of source data before feeding them to the unit which indicated the
condition.
****************************************************************************/

/****************************************************************************
 * The biggest result that may be generated - includes all the subresults.
   It is defined for the only purpose of illustration and using in other
   structures as a placeholder for the biggest result
 */
typedef struct _BIGGEST_RESULT
{
    BASE_RESULT     base;
    COMP_RESULT     comp;
    MAC_RESULT      mac;
    CRYP_RESULT     cryp;

} BIGGEST_RESULT, *PBIGGEST_RESULT;

/****************************************************************************
 * The smallest result that may be generated - includes only base result.
 */
typedef struct _SMALLEST_RESULT
{
    BASE_RESULT     base;

} SMALLEST_RESULT, *PSMALLEST_RESULT;


#define _7751_MAC_KEY_SIZE      64
#define _7751_MAC_CONTEXT_SIZE  40
#define _7751_DES_KEY_SIZE       8
#define _7751_3DES_KEY_SIZE     24
#define _7751_DES_IV_SIZE        8
#define _7751_3DES_IV_SIZE       8
#define _7751_RC4_KEY_SIZE     260
#define _7751_RC4_CONTEXT_SIZE 260

#ifdef _WIN32
    #pragma pack(pop, _INSTRSET_STRUCT_DEF_)
#endif

#endif
