/**********************************************************************

    Module: km_interface.h

        Product: kernel mode driver for "PowerCrypt" encryption
        acceleration board based on Hi/fn 7751 chip
        by Ashley Laurent, Inc.

    ---------------------------------------------------------------

    Copyright:

        Ashley Laurent, Inc., 1999
        All Rights Reserved.

    ---------------------------------------------------------------

    Functional Description:

        Service entry points exported by the driver for kernel-mode
        clients. NT drivers dynamically and BSD drivers statically
        are linked to these entry points.

    ---------------------------------------------------------------

    Environment:

        NT or BSD kernel mode

    ---------------------------------------------------------------

    Author:

        Stan Shkolnyy

    ---------------------------------------------------------------

    Revision history:

        *   15-Feb-99   initial revision

**********************************************************************/

#ifndef __KM_INTERFACE_H__
#define __KM_INTERFACE_H__

/****************************************************************************
 * Ensure inclusion of necessary headers
 */

/*---------------------------------------------------------------------------
 * osindependence.h must be included before this file since it contains defines
 * for our symbols, such as PWRCRYPT_STATUS.
 */
#ifndef __OSINDEPENDENCE_H__
/*    #error include 'osindependence.h' before including this file */
    #include "osindependence.h"
#endif

/*---------------------------------------------------------------------------
 * pwrcryptioctl.h must be included before this file since it contains defines
 * for entry point parameters - PIOCTL_PWRCRYPT_XXX_INPUT_OUTPUT
 */
#ifndef __PWRCRYPTIOCTL_H__
/*    #error include 'pwrcryptioctl.h' before including this file */
    #include "pwrcryptioctl.h"
#endif

#include "briefcomm.h"

#ifdef _WIN32
    #pragma pack(push, _KM_INTERFACE_STRUCT_DEF_, 1)
#else
/*    #warning No pragma pack() ??? */
#endif

/****************************************************************************
 * Define device name for NT kernel mode clients. They must open a handle
 * to this device before they can call the PwrCrypt driver services
 */
#ifdef  _NT_DRIVER_BUILD_
    #define PWRCRYPT_LINK_DEVICE_UNICODE_NAME   L"\\Device\\PwrCryptLink"
#endif

/****************************************************************************

The entry points to the external KM client interface are prototyped below.
PIOCTL_PWRCRYPT_XXXX_INPUT_OUTPUT structures are defined in pwrcryptioctl.h

IMPORTANT common remarks and requirements.
------------------------------------------
In distinction of the user mode programs, the kernel-mode clients make calls
to separate entry points exported by the driver and therefore don't use IOCTL
codes defined in pwrcryptioctl.h

However, the structure of input and output packets is the same for both types
of clients.

The kernel-mode callers must allocate request packets in NON-PAGED SYSTEM MEMORY.

It must be non-paged, because the driver will touch it at the time when page
faults are prohibited. It must be system, because the driver will touch it in
an arbitrary context.

*/

/****************************************************************************

PwrcryptCalculate() -

    kernel mode equivalent of IOCTL_PWRCRYPT_CALCULATE. See pwrcryptioctl.h
    for request packet description.

    However, kernel mode client must also fill in the KMWorkArea member of the
    request packet, before calling PwrcryptCalculate().

*/

/*---------------------------------------------------------------------------
 * Prototype of the routine called on the completion of a calculation request
 * started by PwrcryptCalculate()
 */
typedef
void
(*PPWRCRYPT_CALCULATE_COMPLETION_ROUTINE)
    (
        unsigned long   nComplRtnParm1,
        unsigned long   nComplRtnParm2
    );

/*---------------------------------------------------------------------------
 * Internal structure of KMWorkArea member of the
 * IOCTL_PWRCRYPT_CALCULATE_INPUT_OUTPUT request packet which must be filled by
 * KM callers before calling PwrcryptCalculate().
 */

typedef struct _PWRCRYPT_CALCULATE_KM_WORKAREA
{
    IN  unsigned long       flBufsDescType;
    IN  PPWRCRYPT_CALCULATE_COMPLETION_ROUTINE
                            pComplRtn;
    IN  unsigned long       nComplRtnParm1;
    IN  unsigned long       nComplRtnParm2;
    IN  unsigned char       bChargeQuota;
    IN  unsigned short      nMACKeySize;
    IN  unsigned char       *pMACKey;
    IN  unsigned short      nEncKeySize;
    IN  unsigned char       *pEncKey;
    IN  unsigned short      nEncIVSize;
    IN  unsigned char       *pEncIV;

    IN  unsigned short      nCommandSize;

        BRIEF_COMMAND       BriefComm;

} PWRCRYPT_CALCULATE_KM_WORKAREA, *PPWRCRYPT_CALCULATE_KM_WORKAREA;

/*
    flBufsDescType -
        flags indicating the input form of the data buffers pointed to
        by the pointers in the request. The buffers can be specified by:

            - virtual address of a pageable memory
            - virtual address of a non-pageable memory
            - memory descriptor list, specifying physical memory.

    *_USER must be 'or'ed, if the corresponding buffer address is passed by
        virt.addr., not MDL and it is known that the addr. value is originated
        in a user-mode program. This flag lets the PwrCrypt know that the
        virt.addr. must be verified on to be accessible from the user mode.
        This prevents a user mode program from passing a virt.addr. pointing
        to the system space. If the buffers are allocated by a kernel mode
        driver, the flag must not be set.
*/
#define PWRCRYPT_CALCULATE_IN_PAGE          0x00000001L
#define PWRCRYPT_CALCULATE_IN_NONPAGE       0x00000002L
#define PWRCRYPT_CALCULATE_IN_MDL           0x00000004L
#define PWRCRYPT_CALCULATE_IN_USER          0x00000008L

#define PWRCRYPT_CALCULATE_OUT_PAGE         0x00010000L
#define PWRCRYPT_CALCULATE_OUT_NONPAGE      0x00020000L
#define PWRCRYPT_CALCULATE_OUT_MDL          0x00040000L
#define PWRCRYPT_CALCULATE_OUT_USER         0x00080000L

/*
    pComplRtn -
        ptr to a caller-supplied completion routine that will be called on the
        request completion.

    nComplRtnParm1, nComplRtnParm2 -
        any caller-defined values that will be passed to the completion routine

    bChargeQuota -
        Indicates whether the current process quotas will be charged by the
        certain system calls made by the driver. Should be != 0, if it's
        known that the driver is called in the context of the process that has
        requested the instruction execution. 

    pMACKey, pEncKey, pEncIV, nMACKeySize, nEncKeySize, nEncIVSize -
        coordinates of auth.key, enc.key and enc.IV which are parts of the
        "context" part of the instruction. The kernel mode caller can place
        them anywhere in the NON-PAGED SYSTEM MEMORY. PwrCrypt doesn't alter
        these data, therefore the client can use the same values in several
        requests simultaneously.

    nCommandSize -
        size of the Command in the IOCTL_PWRCRYPT_CALCULATE_INPUT_OUTPUT
        request. The size is the sum of sizes of the Base Command and all
        attached subcommands (but not including context information (keys
        and IVs) - their sizes are indicated in different fields (see above)!!)
        This size is provided for the sole purpose of optimization. The client
        already knows this size, so why not save the PwrCrypt some work?

    nResultSize -
        size of the Result generated by the 7751 and copied to the
        IOCTL_PWRCRYPT_CALCULATE_INPUT_OUTPUT "Result" member. This is the
        only output parameter in the KM workarea (another output parameter
        is the Result in the IOCTL_PWRCRYPT_CALCULATE_INPUT_OUTPUT).

    BriefComm -
        the only field that the caller doesn't need to fill. It is used by
        the PwrCrypt driver.

Remarks:

    Buffer passing methods.

    The driver doesn't need virtual addresses for two data buffers specified
    in the request, since it never interprets the contents of these buffers.
    Rather, it needs them described in terms of physical memory pages, to
    program the 7751 DMA engines.

    Kernel mode users of this function can pass pointers to memory descriptor
    lists instead of virtual addresses. If this is not the case, the driver
    creates needed MDLs by itself. These driver-created MDLs are deleted before
    the driver calls the kernel-mode client's completion routine.

    Hence, if a kernel-mode caller chooses to pass data buffers by their
    virtual addresses, it must call this service in the context where these
    addresses are valid (the buffers allocated in the system memory space are
    valid in an arbitrary context). The service DOESN'T and CAN'T detect invalid
    context.

    However, the caller CAN pass user-mode virtual addresses in the context of
    the requesting process. The service DOES detect this and returns with an
    error, if the user-mode client has passed invalid addresses.
*/

#ifdef _WIN32
    #pragma pack(pop, _KM_INTERFACE_STRUCT_DEF_)
#endif

/*---------------------------------------------------------------------------
 * Prototype of the entry point itself
 */
PWRCRYPT_STATUS
PwrcryptCalculate
    (
        PIOCTL_PWRCRYPT_CALCULATE_INPUT_OUTPUT  pCalcReq
    );

/****************************************************************************

PwrcryptReserveSlot() -

    kernel mode equivalent of IOCTL_PWRCRYPT_RESERVE_SLOT. See pwrcryptioctl.h
    for request packet description.
*/

PWRCRYPT_STATUS
PwrcryptReserveSlot
    (
        PIOCTL_PWRCRYPT_RESERVE_SLOT_INPUT_OUTPUT  pReserveSlotReq
    );

/****************************************************************************

PwrcryptFreeSlot() -

    kernel mode equivalent of IOCTL_PWRCRYPT_FREE_SLOT. See pwrcryptioctl.h
    for request packet description.
*/

PWRCRYPT_STATUS
PwrcryptFreeSlot
    (
        PIOCTL_PWRCRYPT_FREE_SLOT_INPUT_OUTPUT  pFreeSlotReq
    );

/****************************************************************************

PwrcryptGetCaps() -

    kernel mode equivalent of IOCTL_PWRCRYPT_GET_CAPS. See pwrcryptioctl.h
    for request packet description.
*/

PWRCRYPT_STATUS
PwrcryptGetCaps
    (
        PIOCTL_PWRCRYPT_GET_CAPS_INPUT_OUTPUT  pGetCapsReq
    );

#endif
