/****************************************************************************

    Module: pwrcryptioctl.h

        Product: kernel mode driver for "PowerCrypt" encryption
        acceleration board based on Hi/fn 7751 chip
        by Ashley Laurent, Inc.

    ---------------------------------------------------------------

    Copyright:

        Ashley Laurent, Inc., 1999
        All Rights Reserved.

    ---------------------------------------------------------------

    Functional Description:

        Defines IOCTL codes and parameters for services provided
        by the PwrCrypt drivers. Used by both user and kernel mode
        clients, though the kernel mode ones don't need IOCTL
        codes defined here.

    ---------------------------------------------------------------

    Environment:

        NT or BSD user and kernel mode

    ---------------------------------------------------------------

    Author:

        Stan Shkolnyy

    ---------------------------------------------------------------

    Revision history:

        *   17-Feb-99   initial revision

****************************************************************************/

/****************************************************************************

IMPORTANT common remarks and requirements.

The IOCTL codes defined below are needed for user mode clients, since
operating systems provide a single IOCTL path from the user mode to the
kernel mode. The codes are needed to distinct desired functions.

All defined NT IOCTL codes specify BUFFERED method. This means that
the IOCTL input buffer is copied to the NON-PAGED SYSTEM MEMORY before
forwarding it to the driver. Upon request completion, the driver decides
and defines whether the contents of the copy will be copied back to th
original user buffer.

All BSD IOCTL packets are always copied by the system into the non-paged
kernel memory before calling drivers. The BSD IOCTL code provide no indication
about that. However, instead of that, they provide indication whether the
copy of the user packet needs to be copied back to the original user buffer.
This is done by using _IOWR macro. Also, BSD IOCTL code indicates the size of
the IOCTL packet.

The copy of the user IOCTL request must be non-paged, because the driver will
touch it at the time when page faults are prohibited. It must be in the system
memory, because the driver will touch it in an arbitrary context.

****************************************************************************/

#ifndef __PWRCRYPTIOCTL_H__
#define __PWRCRYPTIOCTL_H__

#ifndef IN
#define IN
#endif

#ifndef OUT
#define OUT
#endif

#include "instrset.h"

#ifdef _WIN32
    #pragma pack( push, _PWRCRYPTIOCTL_STRUCT_DEF_, 1)
#endif

/****************************************************************************

Define IOCTL codes known to the driver and the name of the device that user
mode programs need to open to send requests to the driver

IOCTL codes are defined by different rules for NT and FreeBSD.

The name of the device and IOCTL packet structures are the same for both OSes

****************************************************************************/

/* Autodetect stuff to make lee's life easier */

#if !defined(_NT_DRIVER_BUILD_) && !defined(_BSD_DRIVER_BUILD_)

#ifdef _WIN32
#define	_NT_DRIVER_BUILD_ 1
#elif defined(__FreeBSD__)
#define _BSD_DRIVER_BUILD_ 1
#elif defined(linux)
#define _BSD_DRIVER_BUILD_ 1
#else
#error Cannot seem to recognise your OS
#endif

#endif

/******************
 * NT
 */
#ifdef  _NT_DRIVER_BUILD_

#define PWRCRYPT_UM_INTERFACE_DEVICE_NAME   "PwrCryptUMInterface"

#define IOCTL_PWRCRYPT_GET_CAPS \
    CTL_CODE( FILE_DEVICE_UNKNOWN, 0x801, METHOD_BUFFERED, FILE_ANY_ACCESS)

#define IOCTL_PWRCRYPT_RESERVE_SLOT \
    CTL_CODE( FILE_DEVICE_UNKNOWN, 0x802, METHOD_BUFFERED, FILE_ANY_ACCESS)

#define IOCTL_PWRCRYPT_FREE_SLOT \
    CTL_CODE( FILE_DEVICE_UNKNOWN, 0x803, METHOD_BUFFERED, FILE_ANY_ACCESS)

#define IOCTL_PWRCRYPT_CALCULATE \
    CTL_CODE( FILE_DEVICE_UNKNOWN, 0x804, METHOD_BUFFERED, FILE_ANY_ACCESS)

#endif  /* _NT_DRIVER_BUILD_ */

/******************
 * BSD
 */
#ifdef  _BSD_DRIVER_BUILD_

#define PWRCRYPT_UM_INTERFACE_DEVICE_NAME   "pwrcrypt"

#define IOCTL_PWRCRYPT_GET_CAPS \
    _IOWR( 'J', 11, struct _IOCTL_PWRCRYPT_GET_CAPS_INPUT_OUTPUT)

#define IOCTL_PWRCRYPT_RESERVE_SLOT \
    _IOWR( 'J', 12, struct _IOCTL_PWRCRYPT_RESERVE_SLOT_INPUT_OUTPUT)

#define IOCTL_PWRCRYPT_FREE_SLOT \
    _IOWR( 'J', 13, struct _IOCTL_PWRCRYPT_FREE_SLOT_INPUT_OUTPUT)

#define IOCTL_PWRCRYPT_CALCULATE \
    _IOWR( 'J', 14, struct _IOCTL_PWRCRYPT_CALCULATE_INPUT_OUTPUT)

#endif  /* _BSD_DRIVER_BUILD_ */

/****************************************************************************
*****************************************************************************

Parameters to the IOCTL codes

****************************************************************************/

/****************************************************************************

IOCTL_PWRCRYPT_GET_CAPS

Hardware capabilities checking service

****************************************************************************/

typedef struct _IOCTL_PWRCRYPT_GET_CAPS_INPUT_OUTPUT
{
    OUT unsigned long   nNumOfDESSlots;
    OUT unsigned long   nNumOfRC4Slots;
    OUT unsigned long   nNumOfCompHistSlots;

} IOCTL_PWRCRYPT_GET_CAPS_INPUT_OUTPUT, *PIOCTL_PWRCRYPT_GET_CAPS_INPUT_OUTPUT;

/*
    NumOfSlots - 
        contains number of slots that PwrCrypt maintains in its current 
        configuration (which is programmed by the driver according to the
        driver configuration). This is used by the PcrptUMI driver to properly
        scale its internal slot tracking data

    fCapabilities -
        contains flags identifying which encryption methods can be handled by 
        the hardware in its current configuration (which is programmed by the 
        driver according to the driver configuration)

Flags:
*/

#define SLOT_CAP_FLAG_STORE_RC4_KEY         0x00000001L
#define SLOT_CAP_FLAG_STORE_COMP_HISTORY    0x00000002L

/****************************************************************************

IOCTL_PWRCRYPT_RESERVE_SLOT

Slot reservation service - reserves a slot in the 7751 context RAM, so that
the client can safely use it to store keys between calls for data transformation
without conflicts with other clients. Flags specified with this call
affect internal slot reservation. The slot can't be used with MPPC or RC4, if
so was not indicated during reservation.

****************************************************************************/

typedef struct _IOCTL_PWRCRYPT_RESERVE_SLOT_INPUT_OUTPUT
{
   IN   unsigned long   fCapabilities;
   OUT  unsigned long   nReservedSlotNumber;

} IOCTL_PWRCRYPT_RESERVE_SLOT_INPUT_OUTPUT, *PIOCTL_PWRCRYPT_RESERVE_SLOT_INPUT_OUTPUT;

/*
    fCapabilities -
        flags that identify in which kinds of data transformation
        the client plans to use the slot.

    Flags:
*/
#define SLOT_CAP_FLAG_STORE_RC4_KEY         0x00000001L
#define SLOT_CAP_FLAG_STORE_COMP_HISTORY    0x00000002L

/*
    nReservedSlotNumber -
        the number of the reserved slot or PWRCRYPT_INVALID_SLOT_NUMBER, if
        there are no free slots.
*/

#define PWRCRYPT_INVALID_SLOT_NUMBER    -1L

/****************************************************************************

IOCTL_PWRCRYPT_FREE_SLOT

Slot freeing service - as the client doesn't need the allocated slot
anymore, it must dispose it by calling this service.

****************************************************************************/

typedef struct _IOCTL_PWRCRYPT_FREE_SLOT_INPUT_OUTPUT
{
    IN  unsigned long   nSlotNumberToFree;

} IOCTL_PWRCRYPT_FREE_SLOT_INPUT_OUTPUT, *PIOCTL_PWRCRYPT_FREE_SLOT_INPUT_OUTPUT;

/*
    nSlotNumberToFree -
        slot number that was previously returned by the slot reservation service.
*/

/****************************************************************************

IOCTL_PWRCRYPT_CALCULATE

Calculating service - to perform a calculation, the client specifies a buffer
filled with source data, a buffer for output data, a type of desired
calculation and a pre-allocated context number. The keys, specified for the
context, are used to process the data, if encryption and/or authentication
are to be performed. If the compression is used, its history may be stored
within specified context, depending on the processors' internal RAM configuration.


IMPORTANT. In NT, the user-mode clients must create a single instance of
IOCTL_PWRCRYPT_CALCULATE_INPUT_OUTPUT but put its coordinates to BOTH input and
output IOCTL parameters of DeviceIoControl(). It's necessary to pass it as the
_input_ buffer, for the driver to receive a copy of it. It's necessary to pass
it as the _output_ buffer, for the client to receive the result of the call.

The IOCTL_PWRCRYPT_CALCULATE_INPUT_OUTPUT is usually only a header of the IOCTL
buffer that can be significantly large. The SMALLEST_COMMAND member at the bottom
is intended to represent a space for the _smallest_ possible command. However,
the command that is actually placed at the SMALLEST_COMMAND location can be
larger, up to BIGGEST_COMMAND in length. The actual length is defined by the
contents of a BASE_COMMAND, which is the first member of both SMALLEST and
BIGGEST COMMANDs, and by the contents of other sub-commands that follow the
BASE_COMMAND, since those sub-commands may require some encryption or
authentication keys to be added to the command.

The above defined actual length must be specified as a IOCTL buffer parameter
to the IOCTL sending API function. The PcrptUMI driver verifies the specified
length against the contents of the command sent and refuses to accept the
request if any inconsistency detected.

Execution of the request produces the 7751 Result which is returned in the
Result member of the calc.request packet. The size of that Result is returned
through the IOCTL API return value which gives the size of returned data
(for Win32, this is the "BytesReturned" value produced by the DeviceIoControl())
The Result is defined as BIGGEST_RESULT, but only "BytesReturned" bytes at the
top of it are actually used for result. Again, like in the Command case, the
structure of the returned result is not the BIGGEST_RESULT structure. Its
actual structure corresponds to the Command sent.

****************************************************************************/

typedef struct _IOCTL_PWRCRYPT_CALCULATE_INPUT_OUTPUT
{
    /*
     * Don't even think about moving these two fields. KM drivers rely on
     * their order and location at the top of the IOCTL packet
     */
    OUT unsigned short      nResultSize;
    OUT BIGGEST_RESULT      Result;

        unsigned long       KMWorkArea[16];

    IN  unsigned long       nSlotNumber;
    IN  void                *pSrcData;
    IN  unsigned long       nTotalSrcCnt;
    IN  void                *pDstBuffer;
    IN  unsigned long       nTotalDstCnt;

/*    IN  SMALLEST_COMMAND    Command; */
    IN  BIGGEST_COMMAND     Command;

} IOCTL_PWRCRYPT_CALCULATE_INPUT_OUTPUT, *PIOCTL_PWRCRYPT_CALCULATE_INPUT_OUTPUT;

/*
    Result -
        this member will, upon the command completion, contain the command
        result. It is not always the BIGGEST_RESULT, it can be smaller, up
        to the SMALLEST_RESULT, depending on the command that has been executed.

    KMWorkArea -
        reserved space for use by the kernel mode driver. The user mode program
        doesn't touch this space.

    pSrcData, nTotalSrcCnt -
        pointer to and size of the client's source data buffer.

    pDstBuffer, nTotalDstCnt -
        pointer to and size of the client's output buffer where the transformed
        source data are to be copied.

    (the size of a buffer is limited to PROCESSOR_MAX_XFER_SIZE (256K-4)) 

    Command -
        the space for the caller to place a valid 7751 command that is to
        transform its source data in needed way.
*/

#ifdef _WIN32
    #pragma pack(pop, _PWRCRYPTIOCTL_STRUCT_DEF_)
#endif

#endif
