/**********************************************************************

    Module: globobj.c

        Product: kernel mode driver for "PowerCrypt" encryption
        acceleration board based on Hi/fn 7751 chip
        by Ashley Laurent, Inc.

    ---------------------------------------------------------------

    Copyright:

        Ashley Laurent, Inc., 1999
        All Rights Reserved.

    ---------------------------------------------------------------

    Functional Description:

        Construction/init/destruction of the driver's global objects and data

    ---------------------------------------------------------------

    Environment:

        BSD kernel mode

    ---------------------------------------------------------------

    Author:

        Stan Shkolnyy

    ---------------------------------------------------------------

    Revision history:

        *   05-Mar-99   initial revision

**********************************************************************/

#include "osindependence.h"
//#include "debug.h"
#include "globobj.h"

/****************************************************************************
 * Driver's global objects and data
 */
GLOBAL_OBJECTS glob;

/****************************************************************************
****************************************************************************/
void
GlobalObjectsConstructor( PGLOBAL_OBJECTS pThis)
{
    u_long  nIndex;

//    WatchConstructor( &pThis->Watch);

    ConfigConstructor( &pThis->Config);
    SlotmanConstructor( &pThis->SlotMan);
    SlotTrackerConstructor( &pThis->SlotTracker);
    DispatcherConstructor( &pThis->Dispatcher);

    /*
     * No PROCBASEs exist yet
     */
    pThis->NumOfValidProcBases = 0;
    for ( nIndex = 0; nIndex < MAXIMUM_DEVICES_SUPPORTED; nIndex++)
    {
        pThis->pProcBase[ nIndex] = NULL;
    }

    /*
     * No link device exists yet
     */
//    pThis->pLnkDev = NULL;
}

/****************************************************************************
****************************************************************************/
BSDSTATUS
GlobalObjectsInit( PGLOBAL_OBJECTS pThis)
{
//    WatchInit( &pThis->Watch);

    /*
     * First, try to retrieve driver configuration, because it is needed
     * for use by other objects
     */
    ConfigInit( &pThis->Config);
    if ( !ConfigIsValid( &pThis->Config))
    {
	
        return ENXIO;
    }

    /*
     * Init. slot manager - it allocates memory, so can fail
     */
    SlotmanInit( &pThis->SlotMan, &pThis->Config);
    if ( !SlotmanIsValid( &pThis->SlotMan))
    {
        return ENOMEM;
    }

    return STATUS_SUCCESS;
}

/****************************************************************************
****************************************************************************/
void
GlobalObjectsDestructor( PGLOBAL_OBJECTS pThis)
{
    int         status;
    u_long      i;

    /*
     * Destruct all existing PROC_BASEs and free their memory
     */
    for ( i = 0; i < MAXIMUM_DEVICES_SUPPORTED; i++)
    {
        if ( pThis->pProcBase[ i] != NULL)
        {
            status = ProcBaseReleaseHardware( pThis->pProcBase[ i]);
            printk("pwrcrypt: ProcBaseReleaseHardware returns %d\n", status);
            ASSERT( BSD_SUCCESS( status));
            ProcBaseDestructor( pThis->pProcBase[ i]);
            kfree( pThis->pProcBase[ i]);
            pThis->pProcBase[ i] = NULL;
        }
    }

    /*
     * Delete link device
     */
//    if ( pThis->pLnkDev != NULL)
//    {
//        IoDeleteDevice( pThis->pLnkDev);
//        pThis->pLnkDev = NULL;
//    }


    SlotmanDestructor( &pThis->SlotMan);
    SlotTrackerDestructor( &pThis->SlotTracker);
    DispatcherDestructor( &pThis->Dispatcher);
    ConfigDestructor( &pThis->Config);
//    WatchDestructor( &pThis->Watch);
}
