#include "osindependence.h"
#include "debug.h"
#include "procbase.h"

/****************************************************************************
Description:
    Initializes important flags needed for proper resource cleanup.
    Calls Processor's ctor
****************************************************************************/
void
ProcBaseConstructor(
    IN OUT PPROCBASE pThis,
    IN PCONFIG       pConfig
    )
{
    DPF("ProcBaseConstructor\n");

    pThis->Type = PROCBASE_SIGN;

    pThis->pConfig = pConfig;
    pThis->bISRInstalled = FALSE;
    pThis->bProcessorPlugged = FALSE;
    pThis->RangesCnt = 2;

    DmaCtxConstructor( &pThis->SrcDmaCtx);
    DmaCtxConstructor( &pThis->DstDmaCtx);

    ProcessorConstructor( &pThis->Proc, pThis, pConfig);
}

/****************************************************************************
Description:
****************************************************************************/
void
ProcBaseDestructor(
    IN PPROCBASE pThis
    )
{
    DPF( "ProcBaseDestructor\n");

    ASSERT( pThis->Type == PROCBASE_SIGN);

    ProcessorDestructor( &pThis->Proc);

    DmaCtxDestructor( &pThis->SrcDmaCtx);
    DmaCtxDestructor( &pThis->DstDmaCtx);

    ASSERT( !pThis->bProcessorPlugged);
    ASSERT( !pThis->bISRInstalled);
}

/****************************************************************************
pDispatcher -
    needed merely to be passed to the PROCESSOR when it is plugged.
****************************************************************************/
BOOLEAN
ProcBaseAcquireHardware(
    IN PPROCBASE        pThis,
    IN PDISPATCHER      pDispatcher,
    IN linux_pcitag	CfgID
    )
{
    BOOLEAN bSuccess;             // func.ret.value
    unsigned char irq;			  // Our IRQ line
    int status;

    DPF( "PBAcqHW\n");
    ASSERT( pThis->Type == PROCBASE_SIGN);

    bSuccess = FALSE;

    /*
     * Copy the processor board's hardware coordinates
     */
    pThis->CfgID = CfgID;

    if ( TranslateAndMapAddrs( pThis))
    {
        ASSERT( pThis->MemRanges[0].ResourceMapped);
        ASSERT( pThis->MemRanges[1].ResourceMapped);

#ifndef _NO_HW_
        /* Plug the processor into the processor base by letting
         * it know about DMA memory, hardware registers and the
         * DISPATCHER.
         */
        if ( BSD_SUCCESS(
                ProcessorPlug(
                    &pThis->Proc,
                    &pThis->SrcDmaCtx,
                    &pThis->DstDmaCtx,
                    pDispatcher,
                    (PVOID) pThis->MemRanges[1].MappedRangeStart, // dma regs VA
                    (PVOID) pThis->MemRanges[0].MappedRangeStart // proc regs VA
                )))
        {
            pThis->bProcessorPlugged = TRUE;

            /* Now that the PROCESSOR is operational, install ISR
             */
            
	    pcibios_read_config_byte((pThis->CfgID).bus,
	       			     (pThis->CfgID).function,
				      PCI_INTERRUPT_LINE,
				      &irq);
            status = request_irq(
                    irq,
                    DeviceISR,
                    SA_SHIRQ,
                    "pwrcrypt",
                    pThis);
					    
           printk("pwrcrypt: procbase.c: irq=%d, request_irq=%d\n", irq, status); 	
 	   if(!status) 
		    
            {
                pThis->bISRInstalled = TRUE;
                bSuccess = TRUE;
            }
            else
            {
                DPF( "request_irq(irq=%d) failed: errno %d\n", irq, status);
            }
        }
        else
        {
            DPF( "ProcessorPlug failed\n");
        }
#endif
    }

    /*
     * Tell the dispatcher about this processor only if everything is Ok
     */
    if ( bSuccess)
    {
#ifndef _NO_HW_
        DispatcherRegisterProcessor( pDispatcher, &pThis->Proc);
#endif
    }

    return bSuccess;
}

/****************************************************************************

    This method is always called on driver unloading or
    if ProcBaseAcquireHardware() has failed, to release all the hardware
    resources.

****************************************************************************/
BOOLEAN
ProcBaseReleaseHardware(
    IN PPROCBASE       pThis
    )
{
    BOOLEAN             bSuccess;       // func.ret.value

    PADAPTER_MEM_RANGE  pCurRange;      // walks mem.ranges array
    ULONG nIndex;
    UCHAR irq;

    DPF( "ProcBaseReleaseHardware\n");
    ASSERT( pThis->Type == PROCBASE_SIGN);
    

    /*
     * Check the resources acquisition flags and return all the resources
     * that must be returned.
     *
     * Some of the resources may not be acquired, if the earlier call
     * to ProcBaseAcquireHardware() has failed - this is why the flags
     * are needed
     */

    bSuccess = 0;

    if ( pThis->bISRInstalled)
    {
        pcibios_read_config_byte((pThis->CfgID).bus,
				 (pThis->CfgID).function,
				 PCI_INTERRUPT_LINE,
				 &irq);
	
	free_irq(irq, pThis);

        pThis->bISRInstalled = FALSE;
    }

    if ( pThis->bProcessorPlugged)
    {
        ProcessorUnplug( &pThis->Proc);
        pThis->bProcessorPlugged = FALSE;
    }

    /* Walk the mem.ranges array and un-map mapped ranges.
     */
    
    for ( nIndex = 0; nIndex < pThis->RangesCnt; nIndex++)
    {
        pCurRange = &pThis->MemRanges[ nIndex];
        if ( pCurRange->ResourceMapped)
        {
            DPF( "    Unmapping memory at 0x%08x\n", pCurRange->MappedRangeStart);
            vfree((void *)pCurRange->MappedRangeStart);
        }
    }

    return bSuccess;
}

/****************************************************************************
Description:

    Translates a bus-specific address into the corresponding system logical
    address. Also, this will map a physical range to nonpaged system space if
    required.

    Most PCI devices respond to either an I/O range or a memory range, but
    not both.  For simplicity, this function simply maps all the resources
    for a given PCI device.  In reality, only one of the ranges might be
    used, so some cleanup may be needed here.

Arguments:

    pThis            - ptr to PROCBASE

****************************************************************************/
BOOLEAN
TranslateAndMapAddrs(
    IN PPROCBASE   pThis
    )
{
    BOOLEAN     bSuccess;       // func.ret.value
    PADAPTER_MEM_RANGE  pCurRange;      // walks mem.ranges array
    unsigned int pciaddr;	// temporary storage for pcibios_read_*

    DPF( "    Xlate&mapAddrs(%08x)", pThis);
    ASSERT( pThis->Type == PROCBASE_SIGN);
    

    bSuccess = TRUE;

    /*
     * Map the first memory range
     */
    pCurRange = &pThis->MemRanges[ 0];
    
    pcibios_read_config_dword((pThis->CfgID).bus,
			      (pThis->CfgID).function,
			       PCI_BASE_ADDRESS_0,
			       &pciaddr);
    
    pCurRange->RangeStart = pciaddr;
    pCurRange->MappedRangeStart = (ULONG)ioremap(pciaddr, 4096); /* remap the memory into virtual space */
    pCurRange->ResourceMapped = TRUE;

    /* Now do the second one... */
    
    pCurRange = &pThis->MemRanges[1];

    pcibios_read_config_dword((pThis->CfgID).bus,
			      (pThis->CfgID).function,
			       PCI_BASE_ADDRESS_1,
			       &pciaddr);
    
    pCurRange->RangeStart = pciaddr;
    pCurRange->MappedRangeStart = (ULONG)ioremap(pciaddr, 4096); 
    pCurRange->ResourceMapped = TRUE;

    
    return bSuccess;
}

/****************************************************************************
    The Linux PCI interrupt handler
****************************************************************************/
void
DeviceISR
    (
        int irq,
	void    *arg,
	struct pt_regs *regs
    )
{
    PPROCBASE   pThis;

    /*
     * We arranged this in the call to the register_irq() func.
     */
    pThis = (PPROCBASE) arg;
    ASSERT( pThis->Type == PROCBASE_SIGN);

    /*
     * Call PROCESSOR's interrupt method which return FALSE if it sees
     * that its 7751 board hasn't generated this interrupt. Then we don't need
     * to queue the DPC and should quickly return FALSE too.
     */
#ifndef _NO_HW_
    if ( ProcessorConfirmAndRegisterInterrupt( &pThis->Proc))
    {
        ProcessorServiceRegisteredInterrupts( &pThis->Proc);
    }
#endif
}

/****************************************************************************
    Raise spl to spl of our interrupt and call the specified routine with
    the specified parm, so that that routine could be sure the interrupt
    won't happen during its execution.
****************************************************************************/
BOOLEAN
ProcBaseSynchronizeExecution
    (
        IN PPROCBASE       pThis,
        IN PPROCBASE_SYNCHRONIZE_ROUTINE pSynchronizeRoutine,
        IN PVOID  SynchronizeContext
    )
{
    BOOLEAN bRetVal;

    DPF( "ProcBaseSynchronizeExecution\n");
    ASSERT( pThis->Type == PROCBASE_SIGN);

   
    PCrptSpinLockAcquire( &pThis->InterruptSpinlock);
    bRetVal = (pSynchronizeRoutine)( SynchronizeContext);
    PCrptSpinLockRelease( &pThis->InterruptSpinlock); 
    return bRetVal;
}
