/**********************************************************************

    Module: userbuf.h

        Product: kernel mode driver for "PowerCrypt" encryption
        acceleration board based on Hi/fn 7751 chip
        by Ashley Laurent, Inc.

    ---------------------------------------------------------------

    Copyright:

        Ashley Laurent, Inc., 1999
        All Rights Reserved.

    ---------------------------------------------------------------

    Functional Description:

        FreeBSD implementation of functions for locking pageable buffer
        with MDL creation and for unlocking it with MDL desctruction.

    ---------------------------------------------------------------

    Environment:

        FreeBSD kernel mode

    ---------------------------------------------------------------

    Author:

        Stan Shkolnyy

    ---------------------------------------------------------------

    Revision history:

        *   9-May-99   initial revision

**********************************************************************/

#include <asm/bitops.h>
#include <asm/io.h>
#include "osindependence.h"
#include "debug.h"
#include "userbuf.h"
#include <asm/pgtable.h>

ULONG   BytesLocked = 0;

unsigned int uvirt_to_phys(unsigned long);

/****************************************************************************
 * Allocate/free MDL
 */
static
PMDL
AllocateMdl(
    IN PVOID  VirtualAddress,
    IN ULONG  Length)
{
    PMDL            pMdl;
    CSHORT          MdlSize;
    ULONG           PagesInMdl;

    DebugPrint(( 3, "AllocateMdl %08x:%08x\n", VirtualAddress, Length));

    /*
     * MDL consists of struct MDL + array of physical addresses of pages
     */
    PagesInMdl = ADDRESS_AND_SIZE_TO_SPAN_PAGES( VirtualAddress, Length);
    MdlSize = sizeof( *pMdl) + PagesInMdl * sizeof( ULONG);

    pMdl = kmalloc( MdlSize, GFP_KERNEL);
    {
        PULONG  pPages;
        ULONG   iPage;

        pMdl->Size          = MdlSize;
        pMdl->StartVa       = PAGE_ALIGN( VirtualAddress);
        pMdl->ByteCount     = Length;
        pMdl->ByteOffset    = BYTE_OFFSET( VirtualAddress);

        /*
         * Fill the page array with invalid values of -1
         */
        pPages = (PULONG)(pMdl + 1);
        for ( iPage = 0; iPage < PagesInMdl; iPage++)
        {
            pPages[ iPage] = (ULONG) -1;
        }
    }
    return pMdl;
}

#define FreeMdl( pMdl)  kfree( pMdl)

/****************************************************************************
 * Verify the caller's right to access the buffer in specified manner,
 * bring it in memory, lock and fill the MDL with the pages that back
 * the locked buffer.
 *
 * The function's ideas are borrowed from the physio() (kern_physio.c)
 * and vmapbuf() (machdep.c) implementations.
 */

static
PWRCRYPT_STATUS 
ProbeAndLockPages(
    IN OUT  PMDL                    pMdl,
    IN      PWRCRYPT_PROCESSOR_MODE BufAccessMode,
    IN      BOOLEAN                 bWriteAccessRequired
    )
{
    caddr_t     pBuf;
    ULONG       BufSize;
    caddr_t     CurPageVa;
    vm_offset_t CurPagePa;

    ULONG       PagesInBuffer;
    CSHORT      NeededMdlSize;
    PULONG      pPages;
    ULONG       iPage;
    int		status = STATUS_SUCCESS;

    pBuf = (caddr_t) MmGetMdlVirtualAddress( pMdl);
    BufSize = MmGetMdlByteCount( pMdl);

    DebugPrint(( 3, "ProbeAndLockPages %08x:%08x\n", pBuf, BufSize));

    /*
     * MDL consists of struct MDL + array of physical addresses of pages.
     * Is the supplied MDL big enough to describe the buffer?
     */
    PagesInBuffer = ADDRESS_AND_SIZE_TO_SPAN_PAGES( pBuf, BufSize);
    NeededMdlSize = sizeof( *pMdl) + PagesInBuffer * sizeof( ULONG);
    if ( NeededMdlSize > pMdl->Size)
    {
        panic("PwrCrypt.ProbeAndLockPages: MDL too small");
    }

    /*
     * Then physio() calls vmapbuf().
     * The following is from the vmapbuf(), only vmapbuf() also maps the memory
     * into the kernel VA space and this func doesn't. 
     *
     * Fill the MDL's virtual address fields
     * Fill the MDL's page array with physical page addresses
     */
    pMdl->StartVa       = PAGE_ALIGN( pBuf);
    pMdl->ByteCount     = BufSize;
    pMdl->ByteOffset    = BYTE_OFFSET( pBuf);

    DebugPrint((3, "    StartVa=0x%x, ByteCount=%ld, ByteOffset=0x%lx\n",
		(unsigned int)pMdl->StartVa, pMdl->ByteCount, pMdl->ByteOffset));
    /*
     * Point to the MDL's page address array
     */
    pPages = (PULONG)(pMdl + 1);

    for( CurPageVa = (caddr_t) PAGE_ALIGN(pBuf),
            iPage = 0;
         CurPageVa < pBuf + BufSize &&
            iPage < PagesInBuffer;
         CurPageVa += PAGE_SIZE,
            iPage++)
    {
         
 	/* In Linux, we do this backwards, it seems. Lock the page down, then
           fault it in by reading/writing to it (as appropriate) with get_user
           and put_user. */

	
	CurPagePa = uvirt_to_phys((unsigned long)CurPageVa);
 	DebugPrint((3, "   page %ld: va=0x%x, pa=0x%x\n", iPage, (unsigned int) CurPageVa, (unsigned int) CurPagePa));

	if ( CurPagePa == 0)
        {
            panic("PwrCrypt.ProbeAndLockPages: page not present");
        }
	
	
 	/* the following is a reasonable clone of vm_fault_quick() from BSD */

    	set_bit(PG_locked, &(mem_map[MAP_NR(CurPagePa)].flags));

	get_user(CurPageVa);
    
        pPages[ iPage] = CurPagePa;
    }

    return PWRCRYPT_STATUS_SUCCESS;
}

/****************************************************************************
 * Perform the operation reverse to ProbeAndLockPages()
 * Note, that the only purpose this function can use the buffer VA and length
 * for is the calculation of the number of pages backing the buffer. This is
 * because it is invoked in the arbitrary context
 */
static
void
UnlockPages(
    IN OUT  PMDL    pMdl
    )
{
    caddr_t     pBuf;
    ULONG       BufSize;

    ULONG       PagesInBuffer;
    CSHORT      NeededMdlSize;
    PULONG      pPages;
    ULONG       iPage;

    pBuf = (caddr_t) MmGetMdlVirtualAddress( pMdl);
    BufSize = MmGetMdlByteCount( pMdl);

    DebugPrint(( 3, "UnlockPages %08x:%08x\n", pBuf, BufSize));

    /*
     * MDL consists of struct MDL + array of physical addresses of pages.
     * Is the supplied MDL big enough to describe the buffer?
     */
    PagesInBuffer = ADDRESS_AND_SIZE_TO_SPAN_PAGES( pBuf, BufSize);
    NeededMdlSize = sizeof( *pMdl) + PagesInBuffer * sizeof( ULONG);
    if ( NeededMdlSize > pMdl->Size)
    {
        panic("PwrCrypt.UnlockPages: MDL corrupted");
    }

    /*
     * Unlock the memory described by the MDL
     */
    pPages = (PULONG)(pMdl + 1);
    for( iPage = 0; iPage < PagesInBuffer; iPage++)
    {
	if(test_bit(PG_locked, &(mem_map[MAP_NR(pPages[iPage])].flags)))
   		clear_bit(PG_locked, &(mem_map[MAP_NR(pPages[iPage])].flags)); 
    }
}

/****************************************************************************
 * Lock user buffer and create new MDL for it
 */
PWRCRYPT_STATUS
LockBufferCreateMdl (
    IN  PVOID            pBuf,
    IN  ULONG            BufSize,
    IN  BOOLEAN          bChargeQuota,
    IN  PWRCRYPT_PROCESSOR_MODE  AccessMode,
    IN  BOOLEAN          bWriteAccessRequired,
    OUT PMDL             *ppMdl
    )
{
    PWRCRYPT_STATUS status;        // func.ret.value
    PMDL            pMdl;
    CSHORT          MdlSize;
    ULONG           PagesInMdl;

    DebugPrint((3, "LockUserBuffer\n"));

    *ppMdl = NULL;

    status = PWRCRYPT_STATUS_NO_MEMORY;

    pMdl = AllocateMdl( pBuf, BufSize);
    if ( pMdl != NULL)
    {
        /*
         * Now probe and lock the buffer described by the Mdl. On error,
         * deallocate the Mdl and return the appropriate status.
         */
        status = ProbeAndLockPages(
                    pMdl,
                    AccessMode,
                    bWriteAccessRequired
                    );
        if ( BSD_SUCCESS( status))
        {
            /* If we are here, the call has worked - return MDL* to the caller
             *
             * Now, as it can be seen
             * pMdl->MdlFlags = MDL_PAGES_LOCKED | MDL_WRITE_OPERATION
             */
            *ppMdl = pMdl;

            BytesLocked += MmGetMdlByteCount(pMdl);

            // here I verify my understanding of things
            ASSERT( pBuf == MmGetMdlVirtualAddress( pMdl));
            ASSERT( BufSize == MmGetMdlByteCount( pMdl));
//            DebugPrint((3, "    pbuf=%08x size=%08x mdl=%08x\n", pBuf, BufSize, pMdl));
        }
        else
        {
            ASSERT(0);
            FreeMdl( pMdl);
        }
    }
    else
    {
        DebugPrint((2, "    Mdl alloc.failed\n"));
        ASSERT(0);
    }

    return status;
}


/****************************************************************************

UnlockBufferDelMdl()

Description:
    Unlocks the buffer specified by the MDL and deletes MDL

Arguments:

Return:
    None

****************************************************************************/
void
UnlockBufferDelMdl (
    IN PMDL    pMdl
    )
{
    DebugPrint((3, "UnlockUserBuffer pbuf=%08x size=%08x mdl=%08x\n", MmGetMdlVirtualAddress( pMdl), MmGetMdlByteCount(pMdl), pMdl));

    BytesLocked -= MmGetMdlByteCount(pMdl);

    UnlockPages( pMdl);
    FreeMdl( pMdl);
}

/*

	uvirt_to_phys -- utility function for converting user virtual addrs
	to physical addresses. 

*/

unsigned int uvirt_to_phys(unsigned long adr) {

	pgd_t *pgd;
	pmd_t *pmd;
	pte_t *ptep, pte;

	pgd = pgd_offset(current->mm, adr);
	if (pgd_none(*pgd))
	   return 0;
	pmd = pmd_offset(pgd, adr);
	if (pmd_none(*pmd))
		return 0;
	ptep = pte_offset(pmd, adr);
	pte = *ptep;
	if(pte_present(pte))
		return virt_to_phys((void *) (pte_page(pte)|(adr&(PAGE_SIZE - 1))));
	else return 0;
}
