/**********************************************************************

    Module: dma_ctx_bsd.h

        Product: kernel mode driver for "PowerCrypt" encryption
        acceleration board based on Hi/fn 7751 chip
        by Ashley Laurent, Inc.

    ---------------------------------------------------------------

    Copyright:

        Ashley Laurent, Inc., 1999
        All Rights Reserved.

    ---------------------------------------------------------------

    Functional Description:

        Declaration of the FreeBSD-specific part of the DMA_CTX object
        that presents the interface through which PROCESSOR uses
        computer DMA facility

    ---------------------------------------------------------------

    Environment:

        FreeBSD kernel mode

    ---------------------------------------------------------------

    Author:

        Stan Shkolnyy

    ---------------------------------------------------------------

    Revision history:

        *   05-Mar-99   initial revision

**********************************************************************/

#ifndef __DMA_CTX_BSD_H__
#define __DMA_CTX_BSD_H__

/****************************************************************************
 * This many transfers can be simultaneously mapped in a single DMA context
 */
#define DMA_CTX_MAX_PARALLEL_MAPS   2

/****************************************************************************
 * DMA context
 *
 * The purpose of DMA_CTX is to hide as much NT specific details as possible
 * from the PROCESSOR. Each PROCESSOR uses 2 DMA_CTX instances, for two
 * transfer directions.
 *
 * The context is half-duplex - the direction of transfers is specified in a
 * constructor parm and can't be changed.
 *
 * Though it was stated above that _two_ transfers can be in progress in a
 * single DMA_CTX simultaneously, this parallelism has a limitation.
 *
 * Two transfers can be simultaneously in progress, but not two MAP_CONTROL
 * routines. A new MAP_CONTROL routine can't be scheduled by the
 * DmaCtxAllocateAdapterChannel until the previuosly scheduled routine is called.
 *
 * "Two transfers in progress" should be understood as two MAP_CONTROL routines
 * scheduled and called without overlapping and two map register sets used for
 * DMA mapping, each mapping made in its corresponding MAP_CONTROL routine.
 * This way, two map reg sets are used simultaneously.
 */

typedef struct _DMA_CTX
{
    /************************************************************************
     * CONST members - filled once and never change
     */

    ULONG           Type;

    /*-----------------------------------------------------------------------
     * Returned by HalGetCommonBuffer - valid if ( pSgCtlArea != NULL)
     */
    PVOID           pCommBuf;   // VA of the scatter/gather area
    ULONG           CommBufPA;  // Phys.address of the scatter/gather area

    /*-----------------------------------------------------------------------
     * scatter/gather area size in bytes, used only to free it, in the dtor
     */
    ULONG           CommBufSize;

    ULONG           MaxMapRegs;

    /************************************************************************
     * NON-CONST members - change on each new DMA mapping
     */

    /*
     * any user-defined values
     */
    ULONG           UserUlong1;
    ULONG           UserUlong2;
//    ULONG           UserUlong3;
    ULONG           UserUlong4;

//
// BUGBUG
//
//#ifdef DBG

    /*-----------------------------------------------------------------------
     * Tracks the number of map regs claimed by user calling
     * DmaCtxAllocateAdapterChannel. The PROCESSOR is not expected to claim more
     * map regs than indicated by MaxMapRegs, so ASSERTs check this condition.
     */
    ULONG           MapRegsClaimed;

    /*-----------------------------------------------------------------------
     * Counter of real mapping overlappings occured in this DMA context, when
     * some map regs are still in use but a new portion of map regs is given to
     * the user to make a new mapping. I want to see this behavior and need to
     * check it. The overlapping occurences are registered by AdapterControlRoutine
     */
    ULONG           OverlapCnt;
    ULONG           MapsInProgressCnt;

//#endif

} DMA_CTX, *PDMA_CTX;

/****************************************************************************
 * PUBLIC methods for PROCBASE. PROCBASE calls them (except dtor) before
 * passing this context to the PROCESSOR, to fully prepare the context.
 */
void
DmaCtxConstructor(
    IN OUT PDMA_CTX     pThis
    );

void
DmaCtxDestructor(
    IN OUT PDMA_CTX     pThis
    );

BOOLEAN
DmaCtxGetAdapter(
    IN OUT PDMA_CTX     pThis,
    IN ULONG            MapRegsDesired,
    IN ULONG            MaxXferSize
    );

/****************************************************************************
 * PUBLIC methods for PROCESSOR. PROCESSOR calls them when it needs to map
 * physical RAM addresses to "logical" DMA addresses which are then loaded
 * into 7751 scatter/gather descriptors.
 */

#define DmaCtxGetSgCtlAreaPtr( pThis)       ( pThis)->pCommBuf
#define DmaCtxGetSgCtlAreaPhysAddr( pThis)  ( pThis)->CommBufPA
#define DmaCtxGetMaxMapRegs( pThis)         ( pThis)->MaxMapRegs
#define DmaCtxIsPartialTransfers( pThis)    FALSE

#endif /* __DMA_CTX_BSD_H__ */
