#include "osindependence.h"
#include "driverdef.h"
#include "pwrcryptioctl.h"
#include "km_interface.h"
#include "kmi_private.h"
#include "calcreqpp.h"  // for CalcReqCheckAndPrepare()
#include "globobj.h"
#include <linux/module.h> // only needed here

#define PWRCRYPT_DEVICE_STRING  "PowerCrypt (Hi/fn 7751 based) board"

#define CDEV_MAJOR 127 	// XXX this should probably be dynamic

BOOLEAN bCanWork = FALSE;   // switched to TRUE on successful driver init.
int     nMaxDevicesToAcquire;

/****************************************************************************
****************************************************************************/

#define UNIT(x) minor(x)

/****************************************************************************
****************************************************************************/

/* Defines and stuff for our entry points. */


static int PwrCryptOpen(struct inode *, struct file *);
static int PwrCryptIoctl(struct inode *, struct file *, unsigned int, unsigned long);



static int PwrCryptClose(struct inode *, struct file *);
int init_module(void);
void cleanup_module(void);

static int PwrCryptPciProbe(void);
static void PwrCryptPciAttach(struct pci_dev *, int);

static struct wait_queue *waitq = NULL;

/****************************************************************************
=STD= for PCI devices

    We call this function from init_module() to find and init all our
    boards. We leave it to the appropriate functions to set up things like 
    interrupts and such. The return value is the number of boards we found.


****************************************************************************/

int 
PwrCryptPciProbe
   ( 
    )
{
	int goodboard = 0;
 	struct pci_dev *dev = NULL;
	
	if(pci_present) {
		while(dev = pci_find_device(PWRCRYPT_PCI_VENDOR_ID,
						PWRCRYPT_PCI_DEVICE_ID,
						dev)) {
				
				goodboard++;
				PwrCryptPciAttach(dev, goodboard - 1);
			} 	
		
	printk("pwrcrypt: found %d 7751s", goodboard);
	return goodboard;
	
	}
else return 0; /* No PCI -- no boards. */									 
}

/****************************************************************************
****************************************************************************/

/****************************************************************************
=STD= for PCI devices
    Called by system to ask the driver to initialize its device.
    If multiple devices are installed, called multiple times and "unit"
    selects the device instance.

    Notice that the func has no return value, so when the system calls it,
    this is not the question whether the driver will agree to service
    the device. That has already been decided, based, probably, on the call
    to the Probe().

****************************************************************************/
void
PwrCryptPciAttach
    (
	struct pci_dev *CfgID,
        int         unit
    )
{
    PPROCBASE pNewProcBase;

    /************************************************************************
     * Create 1 PROC_BASE object for each device.
     *
     * Notice - configuration limitation on the number of devices is placed by
     * the Probe(). Here we only check if we are going to run over the array
     * bounds which shouldn't happen if each successful Probe() induces no more
     * than one Attach().
     */
    
    
    ASSERT( glob.pProcBase[ glob.NumOfValidProcBases] == NULL);

    pNewProcBase = (PPROCBASE)kmalloc( sizeof( *pNewProcBase), GFP_KERNEL);

    ProcBaseConstructor( pNewProcBase, &glob.Config);

    if ( ProcBaseAcquireHardware( pNewProcBase,
                                  &glob.Dispatcher,
                                  CfgID))
    {
        glob.pProcBase[ glob.NumOfValidProcBases] = pNewProcBase;
        glob.NumOfValidProcBases ++;
	
    }
    else
    {
        if ( !ProcBaseReleaseHardware( pNewProcBase))
        {
            DPF(KERN_ERR "PwrCryptPciAttach: release failed!");
        }
        ProcBaseDestructor( pNewProcBase);
        kfree( pNewProcBase);
	
    }
}

/****************************************************************************
=STD=
    Device open handler
****************************************************************************/

static
int
PwrCryptOpen(
    struct inode *inode,
    struct file *filp
    )
{
    DPF("PwrCryptOpen");
    if ( !bCanWork)
    {
        return ENXIO;
    }

    MOD_INC_USE_COUNT;
    return 0;
}

/****************************************************************************
=STD=
    Device close handler
****************************************************************************/

/*---------------------------------------------------------------------------
PCFreeSlotCallback()

This function is called by the SlotTrackerEnumSlotsByKey() for each slot
which associated "key" matches the "key" passed to the routine.
This function frees the specifieed slot.
---------------------------------------------------------------------------*/
static
void
PCFreeSlotCallback(
    IN PVOID    Context,
    IN ULONG    SlotNumber
    )
{
    struct file *pProc;

    BOOLEAN     bFreeSuccess;
    IOCTL_PWRCRYPT_FREE_SLOT_INPUT_OUTPUT FreeSlotPack;
    BSDSTATUS   status;

    DPF("callback: free slot %08lx, context=0x%x\n", SlotNumber, (unsigned int) Context);

    pProc = (struct file *) Context;

    /* Tell the SlotTracker to verify the passed slot number against
     * the key (struct proc ptr) and free its remembered association
     * between the reserved slot number and the struct proc for which
     * the IRP has been issued.
     *
     * Note! In this callback, the slot number will always be correct,
     * since it is given by SlotTracker itself. And it will always
     * be associated with the Context, since we use it as the key,
     * when asking the SlotTracker to enumerate slots by key.
     */
    bFreeSuccess = SlotTrackerUnregisterSlot(
                                    &glob.SlotTracker,
                                    SlotNumber,
                                    (ULONG) pProc,
                                    &FreeSlotPack.nSlotNumberToFree);

    DPF("callback: slotman freed slot %ld\n", FreeSlotPack.nSlotNumberToFree);

    /*
     * FreeSlot failure can only happen if the slot's RefCnt is not zero,
     * but this must not be the case here, because we are called from
     * IRP_MJ_CLOSE handler when there can't be outstanding IRPs for the
     * corresponding file object and hence its associated slots.
     */
    ASSERT( bFreeSuccess);

    /*
     * After SlotTrackerFreeSlot, nobody has access to the slot until it is
     * reserved again, but before that can happen, the slot must be freed via
     * the following call to the PwrCrypt service.
     */
    status = PwrcryptFreeSlot( &FreeSlotPack);

    ASSERT( BSD_SUCCESS( status));
}

/*---------------------------------------------------------------------------
The close() itself
---------------------------------------------------------------------------*/
static int close_cnt = 0;

static
int
PwrCryptClose(
    struct inode *inode,
    struct file *filp
    )
{
    close_cnt++;
    DPF("PwrCryptClose #%d. pProc %08x\n", close_cnt, (unsigned int)filp);

    /*
     * Free all slots associated with the given file object.
     * Note the trick below - we call the SlotTracket to enumerate slots by
     * key (file object ptr) and give it its own slot freeing method as
     * callback.
     */
    SlotTrackerEnumSlotsByKey(
            &glob.SlotTracker,
            PCFreeSlotCallback,     // callback address
            (PVOID) filp,              // Context
            (ULONG) filp);             // Key
    
    MOD_DEC_USE_COUNT;

    return 0;
}

/****************************************************************************

CalculateCompletionRoutine()

Description:
    Called by the core PwrCrypt driver when it completes the execution of the
    calc.request in PwrcryptCalculate().

Arguments:

    nComplRtnParm1 - ptr to the event to be signaled
    nComplRtnParm2 - not used

****************************************************************************/
static
void
CalculateCompletionRoutine
    (
        ULONG   nComplRtnParm1,
        ULONG   nComplRtnParm2
    )
{
    DPF( "CalculateCompletionRoutine - signalling the event\n");
    ASSERT( nComplRtnParm2 == 0x12345678);
    PCrptSignalEvent((PPCRPT_EVENT) nComplRtnParm1) ;
}

/****************************************************************************
=STD=
    Device IOCTL handler.

    Remember, that all our IOCTL codes are defined via IOWR(), so the system
    makes a work copy of the user request in non-paged memory, then calls us
    and gives ptr to it in arg, then, when we return, copies the work copy
    back over the original user request and returns to the user

****************************************************************************/

/*
 * CALC_REQ_CONTAINER
 *
 * It's funny, but the copy of the IOCTL packet (*arg) that the
 * system makes is actually placed on the process's kernel mode stack. The
 * DPF() below can demonstrate this - values of "arg" and the address of the
 * local var "status" are very close.
 *
 * When I learnt about that, I had understood why the size of the IOCTL
 * packet is limited to, if I remember it right, 4KB. The process's kernel
 * mode stack is used for handling page faults. Then it must be non-pageable
 * and, hence, small.
 *
 * Since the kernel mode stack is process-specific and is mapped into the
 * process's private address space, it can't be accessed in arbitrary process
 * context. Since we are indeed planning to touch the IOCTL packet at
 * interrupt time, we unfortunately have to make another copy of it in the
 * globally accessed KM memory obtained from the standard allocator.
 *
 * More in 4.4BSD p.86
 *
 * Also, an event is involved in stopping the current process until the
 * calc.request is fulfilled. The event is signalled at interrupt time too,
 * so it goes to the "container" together with the copy of the calc.request.
 */

typedef struct _CALC_REQ_CONTAINER
{
    IOCTL_PWRCRYPT_CALCULATE_INPUT_OUTPUT   CalcReq;
    PCRPT_EVENT                             CompletionEvent;
}
CALC_REQ_CONTAINER, *PCALC_REQ_CONTAINER;

static
int
PwrCryptIoctl
    (
	struct inode *inode,
	struct file *pProc,
	unsigned int CtlCode,
	unsigned long arg
    )
{
    int         status;  

    DPF( "PwrCryptIoctl: pProc=0x%x CtlCode=%08x Arg=%08lx size=%08d\n", 
	     (unsigned int) pProc, CtlCode, arg, _IOC_SIZE(CtlCode));

    if ( !bCanWork)
    {
        return ENXIO;
    }

    if ( (void*) arg == NULL)
    {
        return EINVAL;
    }

    status = STATUS_SUCCESS;

    DPF("about to go to ioctl switch()\n");

    switch (CtlCode)
    {
        case IOCTL_PWRCRYPT_CALCULATE:
        {
            PCALC_REQ_CONTAINER                     pCalcReqContainer;

            PIOCTL_PWRCRYPT_CALCULATE_INPUT_OUTPUT  pCalcReq;
	    PIOCTL_PWRCRYPT_CALCULATE_INPUT_OUTPUT pCalcReqOrig;
            PPWRCRYPT_CALCULATE_KM_WORKAREA         pWA;

	    DPF("CALCULATE:\n");
		
	     
	    pCalcReqOrig = (PIOCTL_PWRCRYPT_CALCULATE_INPUT_OUTPUT)kmalloc
			(sizeof(*pCalcReqOrig), GFP_KERNEL); 
 
            copy_from_user(
		(PIOCTL_PWRCRYPT_CALCULATE_INPUT_OUTPUT)pCalcReqOrig, 
		(PIOCTL_PWRCRYPT_CALCULATE_INPUT_OUTPUT)arg, 
		sizeof(*pCalcReqOrig));

	    
            DPF("    src data at 0x%x, dest data at 0x%x\n", 
		   (unsigned int)pCalcReqOrig->pSrcData,
		   (unsigned int)pCalcReqOrig->pDstBuffer);
	    DPF("    src count=%ld, dest count=%ld\n", 
		   pCalcReqOrig->nTotalSrcCnt,
		   pCalcReqOrig->nTotalDstCnt); 
                    
	    DPF( "    making container %08x(%08x:%08x)\n",
                 sizeof(*pCalcReqContainer),
                 sizeof(*pCalcReq),
                 sizeof(pCalcReqContainer->CompletionEvent)
               );


            pCalcReqContainer =
                ( PCALC_REQ_CONTAINER)
                kmalloc( sizeof(*pCalcReqContainer), GFP_KERNEL);
	    
	    pCalcReq = &pCalcReqContainer->CalcReq;

            memcpy( pCalcReq, pCalcReqOrig, sizeof(*pCalcReq));

            /*
             * Setup the pointer to the KM work area inside of the new copy
             * of the calc.req in the "container". Note that this routine
             * touches only KMWorkArea in the CalcReq, other members are filled
             * by the calling app.
             */
            pWA = (PPWRCRYPT_CALCULATE_KM_WORKAREA) &pCalcReq->KMWorkArea;

            /*
             * Check Command-related members of
             * the CalcReq and fill Command-related members of CalcReq.KMWorkArea
             */
            if ( CalcReqCheckAndPrepare( pCalcReq, sizeof( *pCalcReq)))
            {
                if ( SlotTrackerReferenceSlot( &glob.SlotTracker,
                                               pCalcReq->nSlotNumber,
                                               (unsigned long) pProc))
                {
                    /*
                     * The IOCTL input buffer contains ptrs to 2 data buffers.
                     * These ptrs point to the application's private address space.
                     * Indicate this fact by setting appropriate flags, so that the
                     * driver would know how to handle them (it will attempt to
                     * locks them and build MDLs).
                     *
                     * Also, since our caller is a user-mode app, the buffers
                     * must be verified on access from the user mode
                     */
                    pWA->flBufsDescType = ( PWRCRYPT_CALCULATE_IN_PAGE |
                                            PWRCRYPT_CALCULATE_IN_USER |
                                            PWRCRYPT_CALCULATE_OUT_PAGE |
                                            PWRCRYPT_CALCULATE_OUT_USER
                                          );

                    /*
                     * It's known that the calc.service is called in the requesting
                     * process context, so we set this var to TRUE, so that the
                     * process's quota would be charged when the PwrCrypt driver
                     * will create MDLs.
                     */
                    pWA->bChargeQuota = TRUE;

                    /* Report the completion routine address and its two parms
                     * to the service. It doesn't interpret the parms, just
                     * calls the routine and passes them to it, upon request
                     * completion. We need to see the CompletionEvent in our
                     * completion routine and set this event.
                     */
                    pWA->pComplRtn = CalculateCompletionRoutine;
                    pWA->nComplRtnParm1 = (ULONG) &pCalcReqContainer->CompletionEvent;
                    pWA->nComplRtnParm2 = 0x12345678;

                    /* Initialize event _before_ calling the calculating
                     * service - the service will call our callback routine
                     * at any time and the routine will signal the event.
                     */
                    PCrptInitEvent(&pCalcReqContainer->CompletionEvent);

                    /*
                     * Call the PwrCrypt calculating service
                     */
                    status = PwrcryptCalculate( pCalcReq);

                    /* By not returning SUCCESS, the service refuses to process the
                     * request. This may be caused by insufficient resources or bad
                     * user data buffers.
                     *
                     * It it DOES return SUCCESS, it will DO call the completion
                     * routine and we must not touch the IRP anymore, since the
                     * completion routine _is_ to complete this IRP and, in fact,
                     * it is possible that it has been already completed by the
                     * time when the service returns. The status is checked at
                     * the bottom of this IOCTL dispatcher routine to decide
                     * whether to complete the IRP.
                     */
                    if ( status == STATUS_SUCCESS)
                    {
                        /*
                         * The service has accepted the calc.request -
                         * yield the CPU and wait for the event signaling
                         */
                        PCrptWaitOnEvent( &pCalcReqContainer->CompletionEvent);

                        /*
                         * The event has been signalled - the calc.request has
                         * been fulfilled. Copy the two OUT fields from the copy
                         * of the IOCTL packet inside of our "container" to the
                         * original system copy on the stack.
                         *
                         * Attention - this copy operation's calculations are
                         * based on the fact that the nResultSize and Result
                         * are the first fields of the CalcReq.
                         * We need to copy back to user the nResultSize + the
                         * effective part of the Result field, which size is
                         * indicated by the nResultSize.
                         */
                        ASSERT( FIELD_OFFSET(
                                    IOCTL_PWRCRYPT_CALCULATE_INPUT_OUTPUT,
                                    nResultSize) == 0);

                        copy_to_user(
                            (PIOCTL_PWRCRYPT_CALCULATE_INPUT_OUTPUT)arg,
                            (PIOCTL_PWRCRYPT_CALCULATE_INPUT_OUTPUT)pCalcReq,
                            FIELD_OFFSET(
                                IOCTL_PWRCRYPT_CALCULATE_INPUT_OUTPUT,
                                Result) +
                            pCalcReq->nResultSize);
                    }

                    if ( !SlotTrackerDereferenceSlot(
                                &glob.SlotTracker,
                                pCalcReq->nSlotNumber,
                                (unsigned long) pProc))
                    {
                        // BUGBUG: unexpected condition - slot remains "busy"
                        DPF("Unexpected error - slot dereferencing has failed. slot 0x%04lx\n", pCalcReq->nSlotNumber);
                    }
                } /* if ReferenceSlot */
                else
                {
                    DPF("Invalid slot number\n");
                    status = EINVAL;
                }
            } /* if IOCTL parms are valid */
            else
            {
                DPF( "Invalid calculating request\n");
                status = EINVAL;
            }

            kfree( pCalcReqContainer);
	    kfree( pCalcReqOrig);
            break;
        }

        case IOCTL_PWRCRYPT_RESERVE_SLOT:
        {
		
           PIOCTL_PWRCRYPT_RESERVE_SLOT_INPUT_OUTPUT pResSlotReq; 
		
           DPF( "RESERVE: arg=0x%lx, pResSlotReq=0x%x\n", 
		     arg, 
		     (unsigned int)pResSlotReq);

           pResSlotReq = kmalloc(sizeof (*pResSlotReq), GFP_KERNEL);
	   
	   copy_from_user(
		(PIOCTL_PWRCRYPT_RESERVE_SLOT_INPUT_OUTPUT)pResSlotReq,
		(PIOCTL_PWRCRYPT_RESERVE_SLOT_INPUT_OUTPUT)arg,
		sizeof(*pResSlotReq));

            /* 
             * Call the PwrCrypt slot reservation service first.
             */

            status = PwrcryptReserveSlot( pResSlotReq);
            DPF("pwrcrypt: PwrCryptReserveSlot returned %d\n", status);
            if ( BSD_SUCCESS( status))

            {
                BOOLEAN bRegSuccess;

                /*
                 * Tell the SlotTracker to remember the association between
                 * the reserved slot number and the file object for which
                 * the IRP has been issued.
                 * Note that we also substitute a Slot Tracker index
                 * in the place of the PwrCrypt core slot number in the
                 * request packet, so the user mode caller receives
                 * Slot Tracker's index instead of PwrCrypt's slot number.
                 */

		DPF("     calling SlotTrackerRegisterSlot for slot %d\n",
		       pResSlotReq->nReservedSlotNumber);
                bRegSuccess = SlotTrackerRegisterSlot(
                                    &glob.SlotTracker,
                                    pResSlotReq->nReservedSlotNumber,
                                    (ULONG) pProc,
                                    &pResSlotReq->nReservedSlotNumber);


            	copy_to_user(
		(PIOCTL_PWRCRYPT_RESERVE_SLOT_INPUT_OUTPUT)arg,
		(PIOCTL_PWRCRYPT_RESERVE_SLOT_INPUT_OUTPUT)pResSlotReq,
		sizeof(*pResSlotReq));
		
		kfree(pResSlotReq);

		DPF("     SlotTrackerRegisterSlot returned %d\n", bRegSuccess);
                ASSERT( bRegSuccess);
            }
            break;
        }

        case IOCTL_PWRCRYPT_FREE_SLOT:
        {
            PIOCTL_PWRCRYPT_FREE_SLOT_INPUT_OUTPUT    pFreeSlotReq;

	    pFreeSlotReq = kmalloc(sizeof(*pFreeSlotReq), GFP_KERNEL);
	    
            copy_from_user(
		(PIOCTL_PWRCRYPT_FREE_SLOT_INPUT_OUTPUT)&pFreeSlotReq, 
		(PIOCTL_PWRCRYPT_FREE_SLOT_INPUT_OUTPUT)arg,
		sizeof(*pFreeSlotReq));

            DPF( "FREE\n");

	    DPF( "    pFreeSlotReq->nSlotNumberToFree = %ld\n", 
		 pFreeSlotReq->nSlotNumberToFree); 
            /* Tell the SlotTracker to verify the passed slot number against
             * the key (file obj ptr) and free its remembered association
             * between the reserved slot number and the key (file object ptr
             * for which the IRP has been issued).
             * Note that we also substitute the PwrCrypt core slot number
             * in the place of the Slot Tracker index in the request packet, so the user mode caller receives
	     * before calling to the PwrCrypt with that packet.
             */
            if ( SlotTrackerUnregisterSlot(
                                    &glob.SlotTracker,
                                    pFreeSlotReq->nSlotNumberToFree,
                                    (ULONG) pProc,
                                    &pFreeSlotReq->nSlotNumberToFree))
            {
                /*
                 * SlotTracker has said "Ok" - call the PwrCrypt service.
                 */
	    	copy_to_user(
		(PIOCTL_PWRCRYPT_FREE_SLOT_INPUT_OUTPUT)arg,
		(PIOCTL_PWRCRYPT_FREE_SLOT_INPUT_OUTPUT)pFreeSlotReq, 
		sizeof(*pFreeSlotReq));
		
                status = PwrcryptFreeSlot( pFreeSlotReq);

                if ( !BSD_SUCCESS( status))
                {
                    DPF("Slot freeing failed\n");
                }
            }
            else
            {
                DPF("Attempt to free a wrong slot\n");
                status = EINVAL;
            }

	    copy_to_user(
		(PIOCTL_PWRCRYPT_FREE_SLOT_INPUT_OUTPUT)arg,
		(PIOCTL_PWRCRYPT_FREE_SLOT_INPUT_OUTPUT)pFreeSlotReq, 
		sizeof(*pFreeSlotReq));
		
	    kfree(pFreeSlotReq);
            break;
        }

        case IOCTL_PWRCRYPT_GET_CAPS:
        {
            PIOCTL_PWRCRYPT_GET_CAPS_INPUT_OUTPUT  pGetCapsReq = NULL;

           copy_from_user(
		 (PIOCTL_PWRCRYPT_GET_CAPS_INPUT_OUTPUT)&pGetCapsReq, 
		 (PIOCTL_PWRCRYPT_GET_CAPS_INPUT_OUTPUT)  arg,
		 sizeof(*pGetCapsReq));

            DPF( "GETCAPS\n");

            /*
             * Call the PwrCrypt service.
             */
            status = PwrcryptGetCaps( pGetCapsReq);
//            ASSERT( status != STATUS_PENDING);

           copy_to_user(
		 (PIOCTL_PWRCRYPT_GET_CAPS_INPUT_OUTPUT)  arg,
		 (PIOCTL_PWRCRYPT_GET_CAPS_INPUT_OUTPUT)pGetCapsReq, 
		 sizeof(*pGetCapsReq));

            break;
        }

        default:
        {
            DPF( "INVALID\n");
            status = ENOTTY;
            break;
        }
    }
    DPF("    returning a status of %d\n", status);
    return -status;
}


/* Our file_operations struct. We need this to register open, close and
   ioctl -- it's roughly equivalent to cdevsw[] in FreeBSD. */

struct file_operations pwrcrypt_fops = {
	NULL, 		/* llseek */
	NULL,		/* read */
	NULL,		/* write */
	NULL,		/* readdir */
	NULL,		/* poll */
       	&PwrCryptIoctl, /* ioctl */
	NULL,		/* mmap */
        &PwrCryptOpen,  /* open */
	NULL,		/* flush */
        &PwrCryptClose, /* release/close */
	NULL, 		/* fsync */
	NULL,		/* fasync */
	NULL,		/* check_media_change */
	NULL,		/* revalidate */
	NULL		/* lock */
};


/****************************************************************************
init_module ( called from the linux kernel when the driver module is loaded)

Tries to locate and initialize our PCI devices and relevant data objects
and reports the result to the caller. We do the probing and attaching else-
where in this version of the driver; what we do here is call the probe and
attach stuff, then bail out if the driver didn't find anything.

****************************************************************************/

int
init_module ()
{
    BSDSTATUS   status;

    /* The "get capabilities" request sent to PwrCrypt
     */
    IOCTL_PWRCRYPT_GET_CAPS_INPUT_OUTPUT    GetCapsPack;


    printk(KERN_INFO "PowerCrypt (Hi/fn 7751) device driver for Linux 2.0.x\n");
    printk(KERN_INFO "(C) 1999 Global Technologies Group Inc.\n");
    printk(KERN_INFO "(C) 1999 Ashley-Laurent Inc.\n");
    printk(KERN_INFO "All Rights Reserved\n\n");


    /************************************************************************
     * Initialize the driver's permanent/global data.
     */
    GlobalObjectsConstructor( &glob);

    status = GlobalObjectsInit( &glob);
    if ( !BSD_SUCCESS( status))
    {
        printk("Unable to initialize global objects. status=0x%08x\n", status);
        GlobalObjectsDestructor( &glob);
	cleanup_module();
        return status;
    }

    /*
     * Now, the glob.Config is alive and can be used
     */

    /*
     * This may be useful for configuration flexibility.
     */
    nMaxDevicesToAcquire = ConfigGetDeviceCntLimit( &glob.Config);

	if(!PwrCryptPciProbe()) {

        printk(KERN_WARNING "pwrcrypt: no 7751s found\n");
        GlobalObjectsDestructor( &glob);
        return ENXIO;
    }
    else
    {
        DPF("    PwrCryptPciProbe succeeded\n");
        DELAY(1000);
	if ( glob.NumOfValidProcBases != 0)
        {
            printk(KERN_INFO "    %ld devices found, enabling KM interface\n", glob.NumOfValidProcBases);
            KmInterfaceEnable();
        }
        else
        {
            printk(KERN_WARNING "pwrcrypt: no devices initialized\n");
            GlobalObjectsDestructor( &glob);
            return ENXIO;
        }
    
    	status = register_chrdev(CDEV_MAJOR, "pwrcrypt", &pwrcrypt_fops);

    	if(status < 0) {
		printk(KERN_WARNING "pwrcrypt: could not get major number\n");
		return status;
    	}
    
    }

    /************************************************************************
     * UM interface part initialization.
     * We don't destruct the core driver's global objects even on error -
     * they are not ours to destruct
     */
    {
        ULONG   SlotsToTrack;

        /************************************************************************
         * Now that we know we have proper access to PwrCrypt driver, call its
         * service to retrieve the maximum number of slots it exposes to clients.
         * Though we check the return status, the service should never normally
         * fail. Then, using the number of slots obtained from PwrCrypt,
         * initialize SlotTracker
         */

            status = PwrcryptGetCaps( &GetCapsPack);
            if ( !BSD_SUCCESS( status))
            {
                DPF( "Can't get PwrCrypt capabilities. Status = %x \n", status);
                GetCapsPack.nNumOfDESSlots = 0;
                GetCapsPack.nNumOfRC4Slots = 0;
                GetCapsPack.nNumOfCompHistSlots = 0;
            }

        SlotsToTrack = GetCapsPack.nNumOfDESSlots +
                       GetCapsPack.nNumOfRC4Slots +
                       GetCapsPack.nNumOfCompHistSlots;

        ASSERT( SlotsToTrack != 0);

        if ( SlotsToTrack != 0)
        {
            SlotTrackerInit( &glob.SlotTracker, SlotsToTrack);
            if ( SlotTrackerIsValid( &glob.SlotTracker))
            {
                /*
                 * "Unlock" the driver for user-mode requests
                 */
                bCanWork = TRUE;
            }
            else
            {
                printk("Can't initialize SlotTracker.\n");
            }
        }
    }

    if ( !bCanWork)
    {
        GlobalObjectsDestructor( &glob);
        return EACCES;
    }

    return STATUS_SUCCESS;
}


/****************************************************************************
=LKM=
cleanup_module (called from the linux kernel)

Removes the driver from the system. With Linux 2.0.x, we can't return a status,
so we have to get this as clean as possible.

****************************************************************************/


void cleanup_module() 
{
   

    KmInterfaceDisable();

   /* Linux's device registry isn't as complex as BSD's, and it looks like we 
      don't need to unmap the PCI hardware here (we can do it up in the 
      destructor), so we just call that and then unmap our char device entry. */ 

    GlobalObjectsDestructor(&glob);
    
    if(unregister_chrdev(CDEV_MAJOR, "pwrcrypt") < 0) {
	printk(KERN_EMERG "pwrcrypt: couldn't unregister major!\n");
	printk(KERN_EMERG "pwrcrypt: you may want to consider rebooting now.\n");
   }

    /* "You're still here? It's over. Go home." */
   
}


