#ifndef __PROCBASE_H__
#define __PROCBASE_H__

#include "processor.h"      // for PROCESSOR
#include "config.h"         // for PCONFIG
#include "dma_ctx.h"        // for DMA_CTX

typedef struct _ADAPTER_MEM_RANGE {

    ULONG RangeStart;             // Original device memory physical address
    ULONG MappedRangeStart;       // virt.addr.of the mapped memory
    ULONG       RangeLength;            // Length of thee memory range
    BOOLEAN ResourceMapped;             // Flag: resource is mapped

} ADAPTER_MEM_RANGE, *PADAPTER_MEM_RANGE;

/****************************************************************************
 * ProcBase class
 *
 *  The role of PROCBASE is:
 *
 *      - perform O/S-dependent part of PCI resources acquisition (assignments
 *        in the Registry, mapping memory ranges, registering an ISR)
 *
 *      - launching the embedded PROCESSOR to perform all the further data
 *        processing
 *
 *      - stopping the PROCESSOR when it is time to unload the driver
 *
 *      - returning acquired resources to the system
 */

typedef struct _PROCBASE
{
    ULONG           Type;

    /*-----------------------------------------------------------------------
     * The main object that is constructed and launched by this O/S-dependent
     * PROCBASE object
     */
    PROCESSOR           Proc;

    /*-----------------------------------------------------------------------
     * The configuration storage that holds some initialization parameters for
     * both PROCBASE and its PROCESSOR
     */
    PCONFIG             pConfig;

    /*-----------------------------------------------------------------------
     * 2 DMA contexts - initialized by PROCBASE, used by PROCESSOR
     */
    DMA_CTX             SrcDmaCtx;
    DMA_CTX             DstDmaCtx;

    /*-----------------------------------------------------------------------
     * "Adapter Object"-related stuff
     */
    struct pci_dev *    CfgID;      // PCI config.ID for the board
    ULONG               MaxXferLen; // Maximum transfer length for adapter

    /*-----------------------------------------------------------------------
     * These flags/vars indicate which system resources we have acquired that
     * require to be returned back.
     * The list order of flags is the order of acquisition.
     */
    BOOLEAN             bProcessorPlugged;
    BOOLEAN             bISRInstalled;

    // pCommonBuf != NULL means HalGetCommonBuffer succeeded

    // MemRanges[].ResourceMapped indicates that the range was mapped and
    // RangesCnt shows number of the array's elements to scan

    ULONG               RangesCnt;              // Number of I/O and memory resources
    ADAPTER_MEM_RANGE   MemRanges[PCI_TYPE0_ADDRESSES];   // PCI base address info
    PCRPT_SPIN_LOCK	InterruptSpinlock;

} PROCBASE, *PPROCBASE;

/**********************************************************************
 * PUBLIC methods
 */

/*---------------------------------------------------------------------
 * Creator-called
 */

void
ProcBaseConstructor
    (
        IN OUT PPROCBASE pThis,
        IN PCONFIG       pConfig
    );

BOOLEAN
ProcBaseAcquireHardware
    (
        IN PPROCBASE        pThis,
        IN PDISPATCHER      pDispatcher,
	IN struct pci_dev * CfgID
    );

BOOLEAN
ProcBaseReleaseHardware
    (
        IN PPROCBASE       pThis
    );

void
ProcBaseDestructor
    (
        IN PPROCBASE pThis
    );

/*---------------------------------------------------------------------
 * Internally and PROCESSOR-called
 */

#ifdef _NT_DRIVER_BUILD_
    void
    ProcBaseRequestInt
        (
            IN PPROCBASE       pThis
        );
#endif

typedef
BOOLEAN
(*PPROCBASE_SYNCHRONIZE_ROUTINE) (
    IN PVOID SynchronizeContext
    );

BOOLEAN
ProcBaseSynchronizeExecution
    (
        IN PPROCBASE       pThis,
        IN PPROCBASE_SYNCHRONIZE_ROUTINE pSynchronizeRoutine,
        IN PVOID  SynchronizeContext
    );

/**********************************************************************
 * PRIVATE methods
 */

BOOLEAN
TranslateAndMapAddrs
    (
        IN PPROCBASE    pThis
    );

static
void
DeviceISR
    (
        int 	irq,
	void    *arg,
	struct pt_regs *regs
    );

#endif
