/**********************************************************************

    Module: testio.c

        Product: kernel mode driver for "PowerCrypt" encryption
        acceleration board based on Hi/fn 7751 chip
        by Ashley Laurent, Inc.

    ---------------------------------------------------------------

    Copyright:

        Ashley Laurent, Inc., 1999

        All Rights Reserved.

    ---------------------------------------------------------------

    Functional Description:

        Test program for the FreeBSD  and Linux PowerCrypt drivers.
        Opens the driver device and sends IOCTL requests.

    ---------------------------------------------------------------

    Environment:

        FreeBSD user mode

    ---------------------------------------------------------------

    Revision history:

        *   05-Mar-99   initial revision in NT
        *   30-Apr-99   changed system calls to their Unix variants

**********************************************************************/
/*
 * This program tests the driver's ability to perform DES encryption
 *
 * Then it tries to open SLOTS_TO_OPEN slots and then closes all the opened
 * slots.
 */

#include <stdio.h>
#include <sys/errno.h>
#include <sys/fcntl.h>
#include <sys/ioctl.h>

/*
 * IOCTL interface to the driver
 */
#include "pwrcryptioctl.h"

/*
 * The encode command that we're going to send
 */
#include "descomm.h"

/*
 * Make C more human-readable
 */
#define bool_t  int

#ifndef FALSE
#   define FALSE    (0)
#endif
#ifndef TRUE
#   define TRUE (1)
#endif
#ifndef NULL
#   define NULL 0
#endif

#define LOWORD(l)   ((unsigned short)(l))

#ifdef TRIPLE_DES
#define ALGO CRYPCOM_ALGO_3DES
#else
#define ALGO CRYPCOM_ALGO_DES
#endif

#ifdef USE_ECB
#define DESMODE CRYPCOM_DES_MODE_ECB
#else /* use CBC by default */
#define DESMODE CRYPCOM_NEW_IV | CRYPCOM_DES_MODE_CBC
#endif

#ifndef TRIPLE_DES
static  unsigned char   DESKey[8]   = { 0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77, 0x88 };
// static  unsigned char   DESKey[8]   = { 0xC2, 0x38, 0xFA, 0x75, 0x28, 0x24, 0x7C, 0xB2 };

#else
// static  unsigned char   DESKey[24]   = { 0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77, 0x88, 0x99, 0xAA, 0xBB, 0xCC, 0xDD, 0xEE, 0xFF, 0, 1, 0x23, 0x45, 0x67, 0x89,0xAB, 0xCD, 0xEF };
static unsigned char DESKey[24] = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };  
#endif

static  unsigned char   DESIV[8]   = { 0x01, 0x23, 0x45, 0x67, 0x89, 0xab, 0xcd, 0xef};


int main (int argc, char *argv[])
{
    /* We will issue 3 different IOCTLs - here are their data packets
     */
    IOCTL_PWRCRYPT_RESERVE_SLOT_INPUT_OUTPUT    ReserveSlotInOutPack;
    IOCTL_PWRCRYPT_FREE_SLOT_INPUT_OUTPUT       FreeSlotInOutPack;
    IOCTL_PWRCRYPT_CALCULATE_INPUT_OUTPUT       CalcInOutPack;

    /* Slot number that we reserve for our single calculation request
     */
    u_long   CalcSlotNum;

    /* Statistics counter
     */
    u_long   nReqsSent;

    /* Ptrs to 2 parts of the command that we use
     */
    PDES_COMMAND    pCom;
    PDES_RESULT     pRlt;

    /* We will try to open as many slots as we can - here are the
     * storage for returned slot numbers
     */
    u_long      SlotNumber[ 5];
    u_long      nSlotsOpened;
    u_long	nBufferSizeToAllocate;
    u_long	nRepCount;
    bool_t      bFailure;

    int     fd;                 // file descriptor for PcrptUMI device

    unsigned char    *pBuf, *pBuf2;
    int     i;
    int     chKey;

#ifdef TRIPLE_DES
    printf("This version of stress compiled for Triple DES testing\n");
#endif
    
    
    printf("%d", argc);
    if(argc < 2) {
	printf("usage: %s packet-size reps\n", argv[0]);
	exit(1);   
    }
    
    nBufferSizeToAllocate = atoi(argv[1]);
    nRepCount = atoi(argv[2]);
   
    /*
     * Allocate a single data buffer
     */
    pBuf = (unsigned char*) malloc( nBufferSizeToAllocate);
    if ( pBuf == NULL)
    {
        printf( "malloc failed\n");
        return 1;
    }

    pBuf2 = (unsigned char*) malloc( nBufferSizeToAllocate);
    if ( pBuf2 == NULL)
    {
        printf( "malloc failed\n");
        return 1;
    
    }

    memset(pBuf2, NULL, nBufferSizeToAllocate);

    printf("testio:buffer %08x:%08x\n", pBuf, nBufferSizeToAllocate);
    printf("testio:buffer %08x:%08x\n", pBuf2, nBufferSizeToAllocate);

    /*
     * Open a device created by the driver
     */
    fd = open( "/dev/"PWRCRYPT_UM_INTERFACE_DEVICE_NAME"",
               O_RDWR,
               0);

    if ( fd == -1)
    {
        perror( "CreateFile 1 failed");
        free( pBuf);
	free( pBuf2);
        return 1;
    }

    /*
     * Reserve a slot to be able to make a calculation request
     */
    memset(&ReserveSlotInOutPack, 0, sizeof(IOCTL_PWRCRYPT_RESERVE_SLOT_INPUT_OUTPUT));
    bFailure = ioctl( fd,
                      IOCTL_PWRCRYPT_RESERVE_SLOT,
                      &ReserveSlotInOutPack) == -1;

    if ( !bFailure)
    {
        /* Remember the reserved slot number
         */
        CalcSlotNum = ReserveSlotInOutPack.nReservedSlotNumber;
        printf( "Reserved calc slot number 0x%04x\n", CalcSlotNum);
    }
    else
    {
        perror( "Calc slot reservation has failed");
        free( pBuf);
	free( pBuf2);
        return 1;
    }

    /*
     * Fill the allocated buffer with some eye-catching data
     */
    for ( i = 0; i < nBufferSizeToAllocate; i++)
    {
        pBuf[ i] = (char) i;
    }

    nReqsSent = 0;
    while( nReqsSent < nRepCount) //nReqsSent < 1)
    {
        printf( "Requests sent %d\n", nReqsSent);
        nReqsSent += 2;

        /*
         * Output some data from the buffer
         */

        /************************************************************************
         * Assemble the calc.request
         */
        CalcInOutPack.nSlotNumber = CalcSlotNum;

        CalcInOutPack.pSrcData = pBuf;
        CalcInOutPack.nTotalSrcCnt = nBufferSizeToAllocate;

        CalcInOutPack.pDstBuffer = pBuf2;
        CalcInOutPack.nTotalDstCnt = CalcInOutPack.nTotalSrcCnt;

        /*-----------------------------------------------------------------------
         * Assemble the 7751 instruction in the allocated memory
         */

        /* These two ptrs are needed because we don't use the two areas'
         * structures as they are defined in the IOCTL buffer. This is
         * the proper use, however, since the definition contains mere
         * placeholders which are _intended_ to be filled differently
         * depending on required instruction
         */
        pCom = ( PDES_COMMAND) &CalcInOutPack.Command;
        pRlt = ( PDES_RESULT) &CalcInOutPack.Result;

        /*
         * Fill in the Command part of the 7751 instruction.
         */
        BASECOM_FILL(
            &pCom->base,
            BASECOM_SCRAMBLE | BASECOM_CRYP,
            CalcInOutPack.nTotalSrcCnt,
            CalcInOutPack.nTotalDstCnt);

        /*
         * (0 is the size of data to skip at the head of buffer. We skip nothing)
         */
        CRYPCOM_FILL(
            &pCom->cryp,
            CRYPCOM_NEW_KEY | CRYPCOM_DONT_STORE_CTX | ALGO | DESMODE,
            0,
            CalcInOutPack.nTotalSrcCnt);

        /*
         * Finally, copy enc.key and IV into DES_COMMAND
         */
        memcpy( &pCom->EncKey[0], &DESKey[0], sizeof( pCom->EncKey));
        memcpy( &pCom->EncIV[0],  &DESIV[0],  sizeof( pCom->EncIV));

        /*
         * Send the calc.req.
         */
        if ( ioctl( fd,
                    IOCTL_PWRCRYPT_CALCULATE,
                    &CalcInOutPack) == -1)
        {
            perror( "Encode failed");
            printf( "Press a key...");
            chKey = getchar();
        }

        printf("\n");

        /************************************************************************

         * Now decode the same data
         *
         * Fill in the Command part of the 7751 instruction.
         */
        BASECOM_FILL(
            &pCom->base,
            BASECOM_UNSCRAMBLE | BASECOM_CRYP,
            CalcInOutPack.nTotalSrcCnt,
            CalcInOutPack.nTotalDstCnt);

        /*
         * (0 is the size of data to skip at the head of buffer. We skip nothing)
         */
        CRYPCOM_FILL(
            &pCom->cryp,
            CRYPCOM_NEW_KEY | CRYPCOM_DONT_STORE_CTX | ALGO | DESMODE, 
            0,
            CalcInOutPack.nTotalSrcCnt);

	/* Switch the input and output */
	
	CalcInOutPack.pSrcData = pBuf2;
	CalcInOutPack.pDstBuffer = pBuf;

	/* Zero out the buffer so we can see if it came back ok */

	memset(pBuf, 0, CalcInOutPack.nTotalDstCnt);
 
        /*
         * Send the calc.req.
         */
        if ( ioctl( fd,
                    IOCTL_PWRCRYPT_CALCULATE,
                    &CalcInOutPack) == -1)
        {
            perror( "Decode failed");
            printf( "Press a key...");
            chKey = getchar();
        }

    }
//    return 0;
    /************************************************************************
     * Free the testing data buffer
     */
    free( pBuf);
    free( pBuf2);
	
    /* Now close fd and get outta Dodge */

    close(fd);
    return 0;
}
