/**********************************************************************

    Module: testio.c

        Product: kernel mode driver for "PowerCrypt" encryption
        acceleration board based on Hi/fn 7751 chip
        by Ashley Laurent, Inc.

    ---------------------------------------------------------------

    Copyright:

        Ashley Laurent, Inc., 1999

        All Rights Reserved.

    ---------------------------------------------------------------

    Functional Description:

        Test program for the FreeBSD  and Linux PowerCrypt drivers.
        Opens the driver device and sends IOCTL requests.

    ---------------------------------------------------------------

    Environment:

        FreeBSD user mode

    ---------------------------------------------------------------

    Revision history:

        *   05-Mar-99   initial revision in NT
        *   30-Apr-99   changed system calls to their Unix variants

**********************************************************************/
/*
 * This program tests the driver's ability to perform DES encryption
 *
 * Then it tries to open SLOTS_TO_OPEN slots and then closes all the opened
 * slots.
 */

#include <stdio.h>
#include <sys/errno.h>
#include <sys/fcntl.h>
#include <sys/ioctl.h>

/*
 * IOCTL interface to the driver
 */
#include "pwrcryptioctl.h"

/*
 * The encode command that we're going to send
 */
#include "descomm.h"

/*
 * Make C more human-readable
 */
#define bool_t  int

#ifndef FALSE
#   define FALSE    (0)
#endif
#ifndef TRUE
#   define TRUE (1)
#endif
#ifndef NULL
#   define NULL 0
#endif

#define LOWORD(l)   ((unsigned short)(l))

/*
 * Define how many slots we will try to open, the data buffer size and
 * keys for the encode/decode operations 
 */
#define SLOTS_TO_OPEN          1 
#define BUFFER_SIZE_TO_ALLOCATE 8192 

typedef struct _LZS_COMMAND {
	BASE_COMMAND base;
	COMP_COMMAND comp;
} LZS_COMMAND, *PLZS_COMMAND;
	
typedef struct _LZS_RESULT {
	BASE_RESULT base;
	COMP_RESULT comp;
} LZS_RESULT, *PLZS_RESULT;

int main (int argc, char *argv[])
{
    /* We will issue 3 different IOCTLs - here are their data packets
     */
    IOCTL_PWRCRYPT_RESERVE_SLOT_INPUT_OUTPUT    ReserveSlotInOutPack;
    IOCTL_PWRCRYPT_FREE_SLOT_INPUT_OUTPUT       FreeSlotInOutPack;
    IOCTL_PWRCRYPT_CALCULATE_INPUT_OUTPUT       CalcInOutPack;

    /* Slot number that we reserve for our single calculation request
     */
    u_long   CalcSlotNum;

    /* Statistics counter
     */
    u_long   nReqsSent;

    /* Ptrs to 2 parts of the command that we use
     */
    PLZS_COMMAND    pCom;
    PLZS_RESULT     pRlt;

    /* We will try to open as many slots as we can - here are the
     * storage for returned slot numbers
     */
    u_long      SlotNumber[ SLOTS_TO_OPEN];
    u_long      nSlotsOpened;
    bool_t      bFailure;

    int     fd, fd2;                 // file descriptor for PcrptUMI device

    unsigned char    *pBuf, *pBuf2;
    int     i;
    int     chKey;
    int     performance; /* performance level for the compression */ 

#ifdef TRIPLE_DES
    printf("This version of testio compiled for Triple DES testing\n");
#endif
    
    /*
     * Allocate a single data buffer
     */
    pBuf = (unsigned char*) malloc( BUFFER_SIZE_TO_ALLOCATE);
    if ( pBuf == NULL)
    {
        printf( "malloc failed\n");
        return 1;
    }

    pBuf2 = (unsigned char*) malloc( BUFFER_SIZE_TO_ALLOCATE);
    if ( pBuf2 == NULL)
    {
        printf( "malloc failed\n");
        return 1;
    
    }

    memset(pBuf2, NULL, BUFFER_SIZE_TO_ALLOCATE);

    printf("testio:buffer %08x:%08x\n", pBuf, BUFFER_SIZE_TO_ALLOCATE);
    printf("testio:buffer %08x:%08x\n", pBuf2, BUFFER_SIZE_TO_ALLOCATE);

    /*
     * Open a device created by the driver
     */
    fd = open( "/dev/"PWRCRYPT_UM_INTERFACE_DEVICE_NAME"",
               O_RDWR,
               0);

    if ( fd == -1)
    {
        perror( "could not open device "PWRCRYPT_UM_INTERFACE_DEVICE_NAME"");

        free( pBuf);
	free( pBuf2);
        return 1;
    }

#if 0
    if(argc != 2) {
	printf("usage: %s comp-level\n", argv[0]);
	return 1;
        }
    else
	performance = atoi(argv[1]);
#endif
   
    printf("performance = %d\n", performance);
 
    /*
     * Reserve a slot to be able to make a calculation request
     */
    bFailure = ioctl( fd,
                      IOCTL_PWRCRYPT_RESERVE_SLOT,
                      &ReserveSlotInOutPack) == -1;

    if ( !bFailure)
    {
        /* Remember the reserved slot number
         */
        CalcSlotNum = ReserveSlotInOutPack.nReservedSlotNumber;
        printf( "Reserved calc slot number 0x%04x\n", CalcSlotNum);
    }
    else
    {
        perror( "Calc slot reservation has failed");
        free( pBuf);
	free( pBuf2);
        return 1;
    }

    /* Open our test file and read it in */
    
    fd2 = open("test.file", O_RDONLY);
    if(fd2 == -1) {
		perror("could not open test.file");
		return 1;
    }

    /* XXX this reads exactly the number of bytes specified. if test.file
       isn't that big, you lose.  Autosizing would be nice, but it's not 
       important right now. */

    read(fd2, pBuf, BUFFER_SIZE_TO_ALLOCATE);
    close(fd2); // we don't need it after this point
    
    nReqsSent = 0;
    while( nReqsSent < 2)
    {
        printf( "Requests sent %d\n", nReqsSent);
        nReqsSent += 2;

        /*
         * Output some data from the buffer
         */
        printf("Source data\n");
        for ( i = 0; i < 16; i++)
        {
            printf( "%02x ", pBuf[ i]);
        }
        printf("\n");

        /***********************************************************************
         * Assemble the calc.request
         */
        CalcInOutPack.nSlotNumber = CalcSlotNum;

        CalcInOutPack.pSrcData = pBuf;
        CalcInOutPack.nTotalSrcCnt = BUFFER_SIZE_TO_ALLOCATE;

        CalcInOutPack.pDstBuffer = pBuf2;
        CalcInOutPack.nTotalDstCnt = CalcInOutPack.nTotalSrcCnt;

        /*-----------------------------------------------------------------------
         * Assemble the 7751 instruction in the allocated memory
         */

        /* These two ptrs are needed because we don't use the two areas'
         * structures as they are defined in the IOCTL buffer. This is
         * the proper use, however, since the definition contains mere
         * placeholders which are _intended_ to be filled differently
         * depending on required instruction
         */

        pCom = ( PLZS_COMMAND) &CalcInOutPack.Command;
        pRlt = ( PLZS_RESULT) &CalcInOutPack.Result;

        /*
         * Fill in the Command part of the 7751 instruction.
         */
        BASECOM_FILL(
            &pCom->base,
            BASECOM_SCRAMBLE | BASECOM_COMP,
            CalcInOutPack.nTotalSrcCnt,
            CalcInOutPack.nTotalDstCnt);

        /*
         * (0 is the size of data to skip at the head of buffer. We skip nothing)
         */
	COMPCOM_FILL(
	    &pCom->comp,
	    COMPCOM_ALGO_LZS,
	    0, /* use best quality compression */
            0,
            CalcInOutPack.nTotalSrcCnt);

        /*
         * Send the calc.req.
         */
        if ( ioctl( fd,
                    IOCTL_PWRCRYPT_CALCULATE,
                    &CalcInOutPack) == -1)
        {
            perror( "Encode failed");
            printf( "Press a key...");
            chKey = getchar();
        }

        /*
         * Output some data from the buffer
         */
        printf("Encoded data dest count sz=%d\n", CalcInOutPack.nTotalDstCnt);
        for ( i = 0; i < 16; i++)
        {
            printf( "%02x ", pBuf2[ i]);
        }
        printf("\n");

        /************************************************************************

         * Now decode the same data
         *
         * Fill in the Command part of the 7751 instruction.
         */
        BASECOM_FILL(
            &pCom->base,
            BASECOM_UNSCRAMBLE | BASECOM_COMP,
            CalcInOutPack.nTotalSrcCnt,
            CalcInOutPack.nTotalDstCnt);

        /*
         * (0 is the size of data to skip at the head of buffer. We skip nothing)
         */

	COMPCOM_FILL(
	    &pCom->comp,
	    COMPCOM_ALGO_LZS,
	    0, /* use best quality compression */
            0,
            CalcInOutPack.nTotalSrcCnt);

	/* Switch the input and output */
	
	CalcInOutPack.pSrcData = pBuf2;
	CalcInOutPack.pDstBuffer = pBuf;

	/* Zero out the buffer so we can see if it came back ok */

	memset(pBuf, 0, CalcInOutPack.nTotalDstCnt);
 
        /*
         * Send the calc.req.
         */
        if ( ioctl( fd,
                    IOCTL_PWRCRYPT_CALCULATE,
                    &CalcInOutPack) == -1)
        {
            perror( "Decode failed");
            printf( "Press a key...");
            chKey = getchar();
        }

        /*
         * Output some data from the buffer
         */
        printf("Decoded data rlt sz=%d\n", CalcInOutPack.nResultSize);
        for ( i = 0; i < 16; i++)
        {
            printf( "%02x ", pBuf[ i]);
        }
        printf("\n");
    }
    
    /************************************************************************
     * Free the testing data buffer
     */
    free( pBuf);
    free( pBuf2);

    close( fd);
    return 0;
}
