/**********************************************************************

    Module: testio.c

        Product: kernel mode driver for "PowerCrypt" encryption
        acceleration board based on Hi/fn 7751 chip
        by Ashley Laurent, Inc.

    ---------------------------------------------------------------

    Copyright:

        Ashley Laurent, Inc., 1999

        All Rights Reserved.

    ---------------------------------------------------------------

    Functional Description:

        Test program for the FreeBSD  and Linux PowerCrypt drivers.
        Opens the driver device and sends IOCTL requests.

    ---------------------------------------------------------------

    Environment:

        FreeBSD user mode

    ---------------------------------------------------------------

    Revision history:

        *   05-Mar-99   initial revision in NT
        *   30-Apr-99   changed system calls to their Unix variants

**********************************************************************/
/*
 * This program tests the driver's ability to perform DES encryption
 *
 * Then it tries to open SLOTS_TO_OPEN slots and then closes all the opened
 * slots.
 */

#include <stdio.h>
#include <sys/errno.h>
#include <sys/fcntl.h>
#include <sys/ioctl.h>

/*
 * IOCTL interface to the driver
 */
#include "pwrcryptioctl.h"

/*
 * The encode command that we're going to send
 */
#include "descomm.h"

/*
 * Make C more human-readable
 */
#define bool_t  int

#ifndef FALSE
#   define FALSE    (0)
#endif
#ifndef TRUE
#   define TRUE (1)
#endif
#ifndef NULL
#   define NULL 0
#endif

#define LOWORD(l)   ((unsigned short)(l))


/*
 * Define how many slots we will try to open, the data buffer size and
 * keys for the encode/decode operations 
 */
#define SLOTS_TO_OPEN          1 
#define BUFFER_SIZE_TO_ALLOCATE 32768

static  unsigned char   DESkey[64]   = { 0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77, 0x88, 0x99, 0xAA, 0xBB, 0xCC, 0xDD, 0xEE, 0xFF, 0, 1, 0x23, 0x45, 0x67, 0x89,0xAB, 0xCD, 0xEF, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17, 0x18, 0x19, 0x21, 0x23, 0x24, 0x25, 0x26, 0x27, 0x28, 0x29, 0x31, 0x32, 0x34, 0x35, 0x36, 0x37, 0x38, 0x39, 0x41, 0x42, 0x43, 0x45, 0x46, 0x47, 0x48, 0x49, 0x51, 0x52, 0x53, 0x54, 0x56, 0x57, 0x58, 0x59 };

static  unsigned char   DESIV[20];   

typedef struct _SHA_COMMAND
{
	BASE_COMMAND	base;	
	MAC_COMMAND		mac;
//	unsigned char	TRIALkey[64];
} SHA_COMMAND, *PSHA_COMMAND; 
              
typedef struct _SHA_RESULT
{
	BASE_RESULT		base;
	MAC_RESULT		mac;
} SHA_RESULT, *PSHA_RESULT; 
            
int main (int argc, char *argv[])
{
    /* We will issue 3 different IOCTLs - here are their data packets
     */
    IOCTL_PWRCRYPT_RESERVE_SLOT_INPUT_OUTPUT    ReserveSlotInOutPack;
    IOCTL_PWRCRYPT_FREE_SLOT_INPUT_OUTPUT       FreeSlotInOutPack;
    IOCTL_PWRCRYPT_CALCULATE_INPUT_OUTPUT       CalcInOutPack;

    /* Slot number that we reserve for our single calculation request
     */
    u_long   CalcSlotNum;

    /* Statistics counter
     */
    u_long   nReqsSent;

    /* Ptrs to 2 parts of the command that we use
     */
	PSHA_COMMAND		pCom;
	PSHA_RESULT			pRlt;

    /* We will try to open as many slots as we can - here are the
     * storage for returned slot numbers
     */
    u_long      SlotNumber[ SLOTS_TO_OPEN];
    u_long      nSlotsOpened;
    bool_t      bFailure;

    int     fd, fd2;                 // file descriptor for PcrptUMI device

    unsigned char    *pBuf, *pBuf2;
    int     i, x;
    int     chKey;

    
    /*
     * Allocate a single data buffer
     */
    pBuf = (unsigned char*) malloc( BUFFER_SIZE_TO_ALLOCATE);
    if ( pBuf == NULL)
    {
        printf( "malloc failed\n");
        return 1;
    }

    pBuf2 = (unsigned char*) malloc( BUFFER_SIZE_TO_ALLOCATE);
    if ( pBuf2 == NULL)
    {
        printf( "malloc failed\n");
        return 1;
    
    }

    memset(pBuf2, 0, BUFFER_SIZE_TO_ALLOCATE);

    printf("testio:buffer %08x:%08x\n", pBuf, BUFFER_SIZE_TO_ALLOCATE);
    printf("testio:buffer %08x:%08x\n", pBuf2, BUFFER_SIZE_TO_ALLOCATE);

    /*
     * Open a device created by the driver
     */
    fd = open( "/dev/"PWRCRYPT_UM_INTERFACE_DEVICE_NAME"",
               O_RDWR,
               0);

    if ( fd == -1)
    {
        perror( "CreateFile 1 failed");
        free( pBuf);
	free( pBuf2);
        return 1;
    }

    /*
     * Reserve a slot to be able to make a calculation request
     */
    bFailure = ioctl( fd,
                      IOCTL_PWRCRYPT_RESERVE_SLOT,
                      &ReserveSlotInOutPack) == -1;

    if ( !bFailure)
    {
        /* Remember the reserved slot number
         */
        CalcSlotNum = ReserveSlotInOutPack.nReservedSlotNumber;
        printf( "Reserved calc slot number 0x%04x\n", CalcSlotNum);
    }
    else
    {
        perror( "Calc slot reservation has failed");
        free( pBuf);
	free( pBuf2);
        return 1;
    }

#if 0
    /*
     * Fill the allocated buffer with some eye-catching data
     */
    for ( i = 0 ; i < BUFFER_SIZE_TO_ALLOCATE; i++)
    {
        pBuf[ i] = (unsigned char) (i ^ 256) ;
    }
#endif

    fd2 = open("test.file", O_RDONLY);
    if(fd2 == -1) {
	perror("couldn't read input buffer file");
	exit(1);
    }
    
    read(fd2, pBuf, BUFFER_SIZE_TO_ALLOCATE);
    
    nReqsSent = 0;
    while( nReqsSent < 2) //nReqsSent < 1)
    {
        printf( "Requests sent %d\n", nReqsSent);
        nReqsSent += 2;

        /*
         * Output some data from the buffer
         */
        printf("Source data\n");
        for ( i = 0; i < 20; i++)
        {
            printf( "%02x ", pBuf[ i]);
        }
        printf("\n");

        /************************************************************************
         * Assemble the calc.request
         */
  
	memset(&CalcInOutPack, 0, sizeof( CalcInOutPack ) );

        CalcInOutPack.nSlotNumber = CalcSlotNum;

        CalcInOutPack.pSrcData = pBuf;
        CalcInOutPack.nTotalSrcCnt = BUFFER_SIZE_TO_ALLOCATE;

        CalcInOutPack.pDstBuffer = pBuf2;
        CalcInOutPack.nTotalDstCnt = CalcInOutPack.nTotalSrcCnt;

        /*-----------------------------------------------------------------------
         * Assemble the 7751 instruction in the allocated memory
         */

        /* These two ptrs are needed because we don't use the two areas'
         * structures as they are defined in the IOCTL buffer. This is
         * the proper use, however, since the definition contains mere
         * placeholders which are _intended_ to be filled differently
         * depending on required instruction
         */
        pCom = ( PSHA_COMMAND ) &CalcInOutPack.Command;
		pRlt = ( PSHA_RESULT ) &CalcInOutPack.Result;
		memset( &CalcInOutPack.Result, '\0', sizeof( CalcInOutPack.Result ) );

        /*
         * Fill in the Command part of the 7751 instruction.
         */
        BASECOM_FILL(
            &pCom->base,
	    BASECOM_SCRAMBLE | BASECOM_MAC,
            CalcInOutPack.nTotalSrcCnt,
            CalcInOutPack.nTotalDstCnt);

        /*
         * (0 is the size of data to skip at the head of buffer. We skip nothing)
         */
        MACCOM_FILL(
            &pCom->mac,
            MACCOM_NEW_KEY | MACCOM_MODE_HMAC | MACCOM_MAC_RESULT | MACCOM_ALGO_MD5, 
            0,
            CalcInOutPack.nTotalSrcCnt);

	printf("Value of MAC command field: 0x%04x\n", pCom->mac.Command); 
	 //return 0;
        /*
         * Finally, copy enc.key and IV into DES_COMMAND
         */
        // memcpy( pCom->TRIALkey, DESkey, sizeof( pCom->TRIALkey ));
	// memset(pCom->TRIALkey, 0, sizeof( pCom->TRIALkey));
/*		for(i = 0; i < 64; i++)
		{
			printf("%02x ", pCom->TRIALkey[i]);
		}
*/
		printf("\n");
        /*
         * Send the calc.req.
         */
        if ( ioctl( fd,
                    IOCTL_PWRCRYPT_CALCULATE,
                    &CalcInOutPack) == -1)
        {
            perror( "Encode failed");
            printf( "Press a key...");
            chKey = getchar();
        }

        /*
         * Output some data from the buffer
         */
		printf("premature result %d\n", pRlt->mac.PrematureStop);
		printf("miscompare %d\n", pRlt->mac.Miscompare);
        for ( i = 0; i < 20; i++)
        {
            printf( "%x ", pRlt->mac.MAC[i]);        
        }
        printf("\n");
     }

    /************************************************************************
     * Free the testing data buffer
     */
    free( pBuf);
    free( pBuf2);


    /*
     * Close the device handle
     */
    close( fd);

    return 0;
}
