/*++

Copyright (c) 1997-1998  Microsoft Corporation

Module Name:

    findpci.c

Abstract:

    This file contains generic routines to locate a PCI device.

Author:

    Steve Dziok (SteveDz)

Environment:

    Kernel mode

Revision History:


--*/

#include <ntddk.h>
#include "debug.h"
#include "findpci.h"


NTSTATUS
FindPciDevice(
    IN USHORT   VendorId,
    IN USHORT   DeviceId,
    OUT PPCI_DEVICE_LIST PciDeviceList
    )

/*++

Routine Description:

    Find all the PCI devices for the VendorId and DeviceId specified by the
    caller.  The located devices are in the returned PCI_DEVICE_LIST.

    If the caller specifies VendorId == PC_INVALID_VENDORID, then return all
    valid PCI devices on the system.

Arguments:

    VendorId - The manufacturer of the device.  Assigned by the PCI SIG.  If
               this value is PC_INVALID_VENDORID (0xFFFF), then all valid PCI
               devices on the machine will be located.

    DeviceId - The particular device created by the manufacturer.  Assigned by
               the manufacturer.

    PciDeviceList - A list of all the PCI devices that match the VendorId and
                    DeviceId.  The list simply indicates the bus number and slot
                    number for each location.  A FoundDevCnt field in the list indicates
                    how many devices were found.

Return Value:

    NTSTATUS

--*/
{
    ULONG                   busNumber ;
    ULONG                   devNumber;
    ULONG                   funcNumber;

    ULONG                   FoundDevCnt;          // counter of found devices
    NTSTATUS                status;         // func.return value

    PCI_SLOT_NUMBER         slotNumber;
    PCI_COMMON_CONFIG       pciData;
    PPCI_DEVICE_LOCATION    pciDeviceLocation;

    DebugPrint((3, "FindPciDevice \n"));

    //
    //
    // typedef struct _PCI_SLOT_NUMBER {
    //     union {
    //         struct {
    //             ULONG   DeviceNumber:5;
    //             ULONG   FunctionNumber:3;
    //             ULONG   Reserved:24;
    //         } bits;
    //         ULONG   AsULONG;
    //     } u;
    // } PCI_SLOT_NUMBER, *PPCI_SLOT_NUMBER;
    //

    status = STATUS_DEVICE_DOES_NOT_EXIST;
    busNumber = 0;
    FoundDevCnt = 0;

    slotNumber.u.AsULONG = 0;

    //
    // This list member not only counts devices but also points
    // to the first empty device entry in the list
    //
    PciDeviceList->FoundDevCnt = 0;

    //
    // Scan each bus.
    //
    for ( busNumber = 0; busNumber < MAX_PCI_BUSES; busNumber++ )
    {
        //
        // Scan each device.
        //
        for ( devNumber = 0; devNumber < PCI_MAX_DEVICES; devNumber++ )
        {

            slotNumber.u.bits.DeviceNumber = devNumber;

            //
            // Scan each function.
            //
            for ( funcNumber = 0; funcNumber < PCI_MAX_FUNCTION; funcNumber++ )
            {
                //
                // Point to the first empty entry in the device list
                //
                pciDeviceLocation = &PciDeviceList->List[ FoundDevCnt];

                slotNumber.u.bits.FunctionNumber = funcNumber;

                //
                // Check what's in the current slot.
                //

                if (!HalGetBusData(PCIConfiguration,
                                   busNumber,
                                   slotNumber.u.AsULONG,
                                   &pciData,
                                   sizeof(ULONG)
                                   ) ) {

                    //
                    // The specified PCI bus does not exist.  We are done
                    // with this bus.  Set the device number to the maximum
                    // and break out of the function loop.  This will scan
                    // the next system bus.  There is no guarantee that the
                    // system buses are sequentially ordered.
                    //

                    devNumber = PCI_MAX_DEVICES;

                    break;

                }

                if (pciData.VendorID == PCI_INVALID_VENDORID ) {

                    //
                    // Although the current function may show an invalid
                    // device, we really have to keep checking for functions
                    // on the current device.  This is required because the
                    // PCI specification is unclear whether function numbers
                    // in a PCI device must be numerically sequential.  For
                    // example, a particular PCI device could respond to
                    // functions 1, 3, and 4.
                    //

                    //
                    // Check next function (see ifdef below)
                    //

                    continue;

                }

                DebugPrint((3,
                            "PCI device found: bus = 0x%x  slot = 0x%02x\n",
                            busNumber,
                            slotNumber
                            ));

                //
                // Check if we need to return all devices or only those specified
                // by the caller.
                //

                if ( ( VendorId != PCI_INVALID_VENDORID ) &&
                        ( pciData.VendorID != VendorId || pciData.DeviceID != DeviceId )) {

                    DebugPrint((3,
                               "VendorId (0x%04x) or DeviceId (0x%04x) not as requested. \n",
                                pciData.VendorID,
                                pciData.DeviceID
                                ));

                    //
                    // Check next function (see ifdef below)
                    //

                    continue;

                }


#if 0
            // OSR says:
 //
 // If this function 0 and this is not a multifunction device, move on
 // to the next device and do not check other function, otherwise
            // a crash may happen
 //
 if ( (functionNumber == 0) &&
 !(configInfo->HeaderType & PCI_MULTIFUNCTION) ) {

 break;
#endif

                DebugPrint((3, "VendorId and DeviceId match! \n"));

                //
                // At this point, we've found a valid PCI device.
                //

                //
                // Save the found information.
                //

                pciDeviceLocation->BusNumber = busNumber;
                pciDeviceLocation->SlotNumber = slotNumber;

                //
                // Increment the PCI device found count to indicate new device
                // and to point to the next location in the device list
                //
                FoundDevCnt++;

                //
                // Indicate at least one device found.
                //

                status = STATUS_SUCCESS;


            }   // funcNumber

        }   // devNumber

    }   // busNumber


    DebugPrint((1, "%d matching PCI devices found. \n", FoundDevCnt));

    //
    // Update the number of devices in the found list.
    //

    PciDeviceList->FoundDevCnt = FoundDevCnt;

    return status;

}   // FindPciDevice
