/**********************************************************************

    Module: briefcomm.c

        Product: kernel mode driver for "PowerCrypt" encryption
        acceleration board based on Hi/fn 7751 chip
        by Ashley Laurent, Inc.

    ---------------------------------------------------------------

    Copyright:

        Ashley Laurent, Inc., 1999
        All Rights Reserved.

    ---------------------------------------------------------------

    Functional Description:

        Implementation of routines which transfer information (bits)
        between the natural 7751 instruction and its "brief" representation,
        used by dispatching code to decide where to route the instruction.

    ---------------------------------------------------------------

    Environment:

        NT or BSD kernel mode

    ---------------------------------------------------------------

    Author:

        Stan Shkolnyy

    ---------------------------------------------------------------

    Revision history:

        *   19-Mar-99   initial revision

**********************************************************************/

#include "osindependence.h"
#include "pwrcryptioctl.h"
#include "km_interface.h"
#include "debug.h"
#include "br_comm_private.h"

/****************************************************************************
    BASE_COMMAND    base;
    COMP_COMMAND    comp;
    PAD_COMMAND     pad;
    MAC_COMMAND     mac;
    CRYP_COMMAND    cryp;
****************************************************************************/
void
BriefCommandCreateByFullCommand
    (
        OUT     PBRIEF_COMMAND     pBriefComm,
        IN      PBASE_COMMAND      pBC
    )
{
    PUCHAR          pComm;

    /*
     * Point to the byte below the base command
     */
    pComm = (PUCHAR) pBC + sizeof( BASE_COMMAND);

    /*
     * WORKWORK - currently, this copy is redundant.
     */
    pBriefComm->SlotNumber = pBC->SlotNo;

    /*
     * Fill flags according to the specified command
     */
    pBriefComm->flags = 0;

    if ( BASECOM_IS_COMP( pBC))
    {
        pBriefComm->flags |= COMP_ENGAGED;
        if ( COMPCOM_IS_NEW_CTX(( PCOMP_COMMAND) pComm))
        {
            pBriefComm->flags |= COMP_NEW_CTX;
        }
        pComm += sizeof( COMP_COMMAND);
    }

    if ( BASECOM_IS_PAD( pBC))
    {
        pComm += sizeof( PAD_COMMAND);
    }

    if ( BASECOM_IS_MAC( pBC))
    {
        pBriefComm->flags |= MAC_ENGAGED;
        if ( MACCOM_IS_NEW_CTX(( PMAC_COMMAND) pComm))
        {
            pBriefComm->flags |= MAC_NEW_CTX;
        }
        pComm += sizeof( MAC_COMMAND);
    }

    if ( BASECOM_IS_CRYP( pBC))
    {
        pBriefComm->flags |= CRYP_ENGAGED;
        if ( CRYPCOM_IS_NEW_CTX(( PCRYP_COMMAND) pComm))
        {
            pBriefComm->flags |= CRYP_NEW_CTX;
        }

        if ( CRYPCOM_IS_DONT_STORE_CTX(( PCRYP_COMMAND) pComm))
        {
            pBriefComm->flags |= CRYP_DONT_STORE_CTX;
        }

        pComm += sizeof( CRYP_COMMAND);
    }
}

/****************************************************************************
    BASE_COMMAND    base;
    COMP_COMMAND    comp;
    PAD_COMMAND     pad;
    MAC_COMMAND     mac;
    CRYP_COMMAND    cryp;
****************************************************************************/
void
BriefCommandUpdateFullCommand
    (
        IN      PBRIEF_COMMAND     pBriefComm,
        IN OUT  PBASE_COMMAND      pBC
    )
{
    PUCHAR          pComm;
    UCHAR           updateflags;

#ifdef _NO_DST
    /*
     * Piggiback our flag to the Base Command
     */
    pBC->Command |= BASECOM_DISABLE_DEST_FIFO;
#endif

    /*
     * Point to the byte below the base command
     */
    pComm = (PUCHAR) pBC + sizeof( BASE_COMMAND);

    /*
     * We already were here.
     */
    ASSERT( pBriefComm->SlotNumber == pBC->SlotNo);

    /*
     * See which flags were added in newflags compared to flags - these
     * changes need to be reflected in the 7751 command
     */
    updateflags = ( pBriefComm->flags ^ pBriefComm->newflags);

    if ( BASECOM_IS_COMP( pBC))
    {
        if ( updateflags & COMP_NEW_CTX)
        {
            ((PCOMP_COMMAND) pComm)->Command |= COMPCOM_CLEAR_HISTORY;
        }
        pComm += sizeof( COMP_COMMAND);
    }

    if ( BASECOM_IS_PAD( pBC))
    {
        pComm += sizeof( PAD_COMMAND);
    }

    if ( BASECOM_IS_MAC( pBC))
    {
        if ( updateflags & MAC_NEW_CTX)
        {
            ((PMAC_COMMAND) pComm)->Command |= MACCOM_NEW_KEY;
        }
        pComm += sizeof( MAC_COMMAND);
    }

    if ( BASECOM_IS_CRYP( pBC))
    {
        if ( updateflags & CRYP_NEW_CTX)
        {
            ((PCRYP_COMMAND) pComm)->Command |= CRYPCOM_NEW_KEY;
        }

        /*
         * Enc. IV is always sent for DES/3DES and non-EBC mode
         */
        if ( (
               ( ((PCRYP_COMMAND) pComm)->Command & CRYPCOM_ALGO_BITMASK) == CRYPCOM_ALGO_DES ||
               ( ((PCRYP_COMMAND) pComm)->Command & CRYPCOM_ALGO_BITMASK) == CRYPCOM_ALGO_3DES
             ) &&
             ( ((PCRYP_COMMAND) pComm)->Command & CRYPCOM_MODE_BITMASK) != CRYPCOM_DES_MODE_ECB)
        {
            ((PCRYP_COMMAND) pComm)->Command |= CRYPCOM_NEW_IV;
        }

        pComm += sizeof( CRYP_COMMAND);
    }
}
