/**********************************************************************

    Module: config.c

        Product: kernel mode driver for "PowerCrypt" encryption
        acceleration board based on Hi/fn 7751 chip
        by Ashley Laurent, Inc.

    ---------------------------------------------------------------

    Copyright:

        Ashley Laurent, Inc., 1999
        All Rights Reserved.

    ---------------------------------------------------------------

    Functional Description:

        Implementation of the CONFIG object that supplies other objects with
        permanent configuration information, needed on initilization time.
        In NT, the information is withdrawn from the Registry.
        
    ---------------------------------------------------------------

    Environment:

        NT or BSD kernel mode

    ---------------------------------------------------------------

    Author:

        Stan Shkolnyy

    ---------------------------------------------------------------

    Revision history:

        *   08-Mar-99   initial revision

**********************************************************************/

#include "osindependence.h"
#include "debug.h"
#include "config.h"
#include "driverdef.h"

#ifdef ALLOC_PRAGMA
    #pragma alloc_text(INIT, ConfigConstructor)
    #pragma alloc_text(INIT, ConfigInit)
#endif

/****************************************************************************
****************************************************************************/

void
ConfigConstructor(
    IN OUT PCONFIG      pThis
    )
{
    DebugPrint((2, "ConfigConstructor\n"));

    pThis->Type = CONFIG_SIGN;

    pThis->NumOfCtxMemSlots = 0;
    pThis->bValidConfig = FALSE;
}

/****************************************************************************
****************************************************************************/

#ifdef _NT_DRIVER_BUILD_

void
ConfigInit(
    IN OUT PCONFIG      pThis,
    IN PUNICODE_STRING  RegistryPath
    )

#endif

#ifdef _BSD_DRIVER_BUILD_

void
ConfigInit(
    IN OUT PCONFIG      pThis
    )

#endif

{
    DebugPrint((2, "ConfigInit\n"));
    ASSERT( pThis->Type == CONFIG_SIGN);

    //
    // WORKWORK - read all this from the Registry
    //

    pThis->NumOfCtxMemSlots = 16;

#ifdef _SLOW_WORK
    pThis->nPollFreq = 0xff;
    pThis->nPollScalar = 7;
#else
    pThis->nPollFreq = 1;
    pThis->nPollScalar = 0;
#endif

    pThis->bValidConfig = TRUE;

    pThis->DeviceCntLimit = 1; //MAXIMUM_DEVICES_SUPPORTED;

    /*
     * Unlock key.
     * For reference hardware, the key is 0.
     * For the real thing, it must match the value in the onboard EEPROM
     */
    PCrptZeroMemory( &pThis->UnlockKey.Level1Value[ 0],
                            sizeof( pThis->UnlockKey.Level1Value));
    PCrptZeroMemory( &pThis->UnlockKey.Level2Appendix[ 0],
                            sizeof( pThis->UnlockKey.Level2Appendix));

    /*
     * Unlock level.
     */
    pThis->UnlockLevel = FULLY_UNLOCKED;
}

/****************************************************************************
****************************************************************************/
void
ConfigDestructor(
    IN OUT PCONFIG      pThis
    )
{
    DebugPrint((2, "ConfigDestructor\n"));
    ASSERT( pThis->Type == CONFIG_SIGN);
}

/****************************************************************************
****************************************************************************/
void
ConfigGetUnlockKey
    (
        IN OUT  PCONFIG     pThis,
        OUT     PUNLOCK_KEY pUnlockKey
    )
{
    ASSERT( pThis->Type == CONFIG_SIGN);

    *pUnlockKey = pThis->UnlockKey;
}
