/**********************************************************************

    Module: config.h

        Product: kernel mode driver for "PowerCrypt" encryption
        acceleration board based on Hi/fn 7751 chip
        by Ashley Laurent, Inc.

    ---------------------------------------------------------------

    Copyright:

        Ashley Laurent, Inc., 1999
        All Rights Reserved.

    ---------------------------------------------------------------

    Functional Description:

        Declaration of the CONFIG object that supplies other objects with
        permanent configuration information, needed on initilization time.
        In NT, the information is withdrawn from the Registry.
        
    ---------------------------------------------------------------

    Environment:

        NT or BSD kernel mode

    ---------------------------------------------------------------

    Author:

        Stan Shkolnyy

    ---------------------------------------------------------------

    Revision history:

        *   08-Mar-99   initial revision

**********************************************************************/

#ifndef __CONFIG_H__
#define __CONFIG_H__

#include "driverdef.h"      // for UNLOCK_KEY

/* Undefine PCONFIG if it's defined. We need this because sys/param.h on 
   FreeBSD defines it as 32. */

#ifdef PCONFIG
#undef PCONFIG
#endif /* PCONFIG */


/****************************************************************************
 * Configuration storage
 */

typedef struct _CONFIG
{
    ULONG           Type;

    /*
     * After calling ConfigInit(), tells whether the configuration was
     * successfully read. Use ConfigIsValid() to read this value.
     */
    BOOLEAN         bValidConfig;

    /*
     * The 7751 processing unit configuration in the format of the PU Config
     * Register, except polling speed fields that are defined by the next two
     * members.
     */
    ULONG           nPUConfigBits;
    ULONG           nPollFreq;
    ULONG           nPollScalar;

    /* WORKWORK
     * 7751 context memory size. The current version of the driver is unable to
     * detect the size of the context memory installed. It is told about it from
     * outside, through this member. Setting this value above the actual size
     * may lead to a behavior that I have not analysed yet :-)
     */
    ULONG           CtxMemSize;

    /*
     * Limits the number of context memory slots that the driver uses.
     * This number is also limited by the size of the context memory and
     * the proc.unit configuration.
     * Should be used to force the driver to operate with smaller number of
     * slots than is allowed by the above mentioned factors.
     */
    ULONG           NumOfCtxMemSlots;

    /*
     * Limits the number of 7751 boards that the driver uses.
     * Must be <= MAXIMUM_DEVICES_SUPPORTED.
     * Should be used to force the driver to acquire and use less number of
     * devices than installed in the machine.
     */
    ULONG           DeviceCntLimit;     // don't acquire more then this many devices

    /*
     * "Unlock key" used to unlock advanced capabilities of the 7751 chip.
     * This key is vendor-specific and must much the contents of the EEPROM
     * installed on the board.
     */
    UNLOCK_KEY      UnlockKey;
    SecurLevel      UnlockLevel;

#ifdef _NT_DRIVER_BUILD_

    /*-----------------------------------------------------------------------
     */

#endif

} CONFIG, *PCONFIG;


/****************************************************************************
 * PUBLIC methods for DriverEntry. DriverEntry calls them to construct the
 * object and fill the configuration. After that, it passes this object to
 * PROCBASE which then makes it available to its PROCESSOR
 */

void
ConfigConstructor(
    IN OUT PCONFIG      pThis
    );

#ifdef _NT_DRIVER_BUILD_

void
ConfigInit(
    IN OUT PCONFIG      pThis,
    IN PUNICODE_STRING  RegistryPath
    );

#endif

#ifdef _BSD_DRIVER_BUILD_


void
ConfigInit(
    IN OUT PCONFIG      pThis
    );

#endif

void
ConfigDestructor(
    IN OUT PCONFIG      pThis
    );
      
#define ConfigIsValid( pThis)   ( (pThis)->bValidConfig)

/****************************************************************************
 * PUBLIC methods for objects that need configuration info. These objects can
 * be found by inclusion of "config.h" into their header files.
 * ( PROCESSOR, SLOTMAN, ...)
 */

#define ConfigGetPUConfigBits( pThis)           ( (pThis)->nPUConfigBits)
#define ConfigGetPollFreq( pThis)               ( (pThis)->nPollFreq)
#define ConfigGetPollScalar( pThis)             ( (pThis)->nPollScalar)

#define ConfigGetCtxMemSize( pThis)             ( (pThis)->CtxMemSize)
#define ConfigGetNumOfCtxMemSlots( pThis)       ( (pThis)->NumOfCtxMemSlots)
#define ConfigGetUnlockSecurityLevel( pThis)    ( (pThis)->UnlockLevel)
#define ConfigGetDeviceCntLimit( pThis)         ( (pThis)->DeviceCntLimit)

void
ConfigGetUnlockKey
    (
        IN OUT  PCONFIG     pThis,
        OUT     PUNLOCK_KEY pUnlockKey
    );

#endif
