#include "osindependence.h"
#include "debug.h"
#include "processor.h"
#include "globobj.h"
#include "dispatcher.h"

#ifdef ALLOC_PRAGMA
#pragma alloc_text(INIT, DispatcherConstructor)
#pragma alloc_text(INIT, DispatcherRegisterProcessor)
#endif

static
void
StartRequests
    (
        IN OUT  PDISPATCHER pThis
    );

static
int
FindLeastBusyProcessor
    (
        IN OUT  PDISPATCHER pThis,
        IN PBRIEF_COMMAND   pBriefComm
    );

/****************************************************************************
****************************************************************************/
void
DispatcherConstructor
    (
        IN OUT  PDISPATCHER pThis
    )
{
    ULONG   i;

    DebugPrint((2, "DispatcherConstructor\n"));

    pThis->Type = DISPATCHER_SIGN;

    PCrptListInit( &pThis->Queue);
    PCrptSpinLockInit( &pThis->SpinLock);

    pThis->nNumOfProcs = 0;

    for ( i = 0; i < MAXIMUM_DEVICES_SUPPORTED; i++)
    {
        pThis->PCD[ i].pProc = NULL;
    }

    pThis->nReqs = 0;
    pThis->nMaxReqs = 0;
    pThis->nTotalReqs = 0;
}

/****************************************************************************
****************************************************************************/
void
DispatcherDestructor
    (
        IN OUT  PDISPATCHER pThis
    )
{
    DebugPrint((2, "DispatcherDestructor\n"));
    ASSERT( pThis->Type == DISPATCHER_SIGN);
}

/****************************************************************************
    Called only at initialization time, when no work is being done by the
    DISPATCHER
****************************************************************************/
ULONG
DispatcherRegisterProcessor
    (
        IN OUT  PDISPATCHER pThis,
        IN OUT  struct _PROCESSOR  *pProc
    )
{
    ULONG   i;

    DebugPrint((2, "DispatcherRegisterProcessor\n"));
    ASSERT( pThis->Type == DISPATCHER_SIGN);

    for ( i = 0; i < MAXIMUM_DEVICES_SUPPORTED; i++)
    {
        if ( pThis->PCD[ i].pProc == NULL)
        {
            pThis->PCD[ i].pProc = pProc;
            pThis->PCD[ i].pProc->nReqsInQueue = 0;
            pThis->PCD[ i].pProc->bInactive = FALSE;
            pThis->nNumOfProcs++;
            return i;
        }
    }
    return  (ULONG) -1L;
}

/****************************************************************************
        lock
        QueuePut( Pack)
    ---------------------------
        StartRequests()
    ---------------------------
        unlock
****************************************************************************/
void
DispatcherStartRequest
    (
        IN OUT  PDISPATCHER pThis,
        IN      PIOCTL_PWRCRYPT_CALCULATE_INPUT_OUTPUT
                            pCalcReq
    )
{
    ASSERT( pThis->Type == DISPATCHER_SIGN);

    /*
     * Acquire the spinlock, because we are going to put a new request
     * into the queue and call StartRequests() which performs re-scheduling
     */
    PCrptSpinLockAcquire( &pThis->SpinLock);

    /*
     * Put the incoming new request at the tail of the dispatcher queue
     *
     * We use pCalcReq->Result field as the space for list entry, since this
     * field is not used until the "instruction" is de-queued and gets executed
     * by a 7751. After the instruction is de-queued, we don't ever need to
     * queue it again, so the Result is free to be filled by the 7751
     */
    ASSERT( sizeof( pCalcReq->Result) >= sizeof ( PPCRPT_LIST_ENTRY));

    PCrptListInsertTail( &pThis->Queue, ( PPCRPT_LIST_ENTRY) &pCalcReq->Result);

    pThis->nTotalReqs++;
    pThis->nReqs++;
    if ( pThis->nReqs > pThis->nMaxReqs)
    {
        pThis->nMaxReqs = pThis->nReqs;
    }

    DebugPrint((2, "DispatcherStartRequest #%d\n", pThis->nTotalReqs));

    /*
     * Perform requests' re-scheduling, since we've added a new request
     */
    StartRequests( pThis);

    /*
     * Release the spinlock, because we are exiting the dispatcher.
     */
    PCrptSpinLockRelease( &pThis->SpinLock);
}

/****************************************************************************
    Called by a PROCESSOR when it's completed a request and removed it from its
    queue. The DISPATCHER then decrements the req.cnt for this PROCESSOR and
    knows that this PROCESSOR can accept a new request. Therefore, it calls
    StartRequests()

        lock
        CompletedProcessor.Busy--
    ---------------------------
        StartRequests()
    ---------------------------
        unlock
****************************************************************************/
void
DispatcherStartNextRequest
    (
        IN OUT  PDISPATCHER pThis,
        IN OUT  struct _PROCESSOR  *pProc
    )
{
    DebugPrint((2, "DispatcherStartNextRequest\n"));
    ASSERT( pThis->Type == DISPATCHER_SIGN);

    /*
     * Acquire the spinlock, because we are going to call StartRequests()
     * which performs re-scheduling
     */
    PCrptSpinLockAcquire( &pThis->SpinLock);

    /*
     * Indicate that a request has left the processor queue
     */
    pProc->nReqsInQueue--;

    /*
     * Perform requests' re-scheduling, since a request has been completed
     * and a place become free in the queue of one of the PROCESSORs.
     */
    StartRequests( pThis);

    /*
     * Release the spinlock, because we are exiting the dispatcher.
     */
    PCrptSpinLockRelease( &pThis->SpinLock);
}

/****************************************************************************

    This function is called with the DISPATCHER spinlock held - any queue
    manipulations are safe.

        while ( can start request AND ( Pack = QueueGet()))
            ReadyProcessor.Busy++
            unlock
            ReadyProcessor.StartIO( Pack)
            lock
        end

****************************************************************************/
static
void
StartRequests
    (
        IN OUT  PDISPATCHER pThis
    )
{
    PPCRPT_LIST_ENTRY   pRemovedListEntry;
    PIOCTL_PWRCRYPT_CALCULATE_INPUT_OUTPUT pCalcReq;
    PBRIEF_COMMAND      pBriefComm;
    int                 iLeastBusy;
    BOOLEAN             bMore;

    ASSERT( pThis->Type == DISPATCHER_SIGN);

    /*
     * While the req.queue is not empty and a processor ready to accept
     * a new request can be found, withdraw requests from the queue and
     * start them on the found processor.
     */
    bMore = TRUE;
    while( bMore && !PCrptIsListEmpty( &pThis->Queue))
    {
        /*
         * Remove a request from the list, because we need it in the
         * FindLeastBusyProcessor().
         */
        ASSERT( sizeof( pCalcReq->Result) >= sizeof ( *pRemovedListEntry));

        pRemovedListEntry = PCrptListRemoveHead( &pThis->Queue);
        pCalcReq = CONTAINING_RECORD( pRemovedListEntry,
                                      IOCTL_PWRCRYPT_CALCULATE_INPUT_OUTPUT,
                                      Result);

        pBriefComm = &((PPWRCRYPT_CALCULATE_KM_WORKAREA) pCalcReq->KMWorkArea)->BriefComm;

        /*
         * Try to find a processor ready to accept a new request. The index of
         * a processor in the dispatcher array is returned. Ret.value of -1
         * means failure - no processor is ready to accept a request.
         */
        iLeastBusy = FindLeastBusyProcessor( pThis, pBriefComm);
        if ( iLeastBusy >= 0)
        {
            PPROCESSOR  pProc;

            /*
             * The request can be started and we know on which
             * PROCESSOR. Now we can made the changes in the slot affinities
             * that are caused by the Command and finish the BriefCommand
             * so that it can be used by the PROCESSOR to form the final
             * hardware-sen command/context block
             */
            SlotmanUpdateSlotAffinitiesByCommand( &glob.SlotMan, pBriefComm, iLeastBusy);

            /*
             * This is the selected processor
             */
            pProc = pThis->PCD[ iLeastBusy].pProc;

            /*
             * Indicate that we are placing a new request into the processor's queue.
             * This action (performed with the spinlock held), "reserves" a place
             * for the request in the processor's queue. This guarantees that
             * ProcessorStartRequest() can be safely called later.
             */
            pProc->nReqsInQueue++;

            /*
             * Place the request into the processor's queue.
             *
             * Note! The starting of the request on a processor is splitted
             * into two operations - queueing and actual starting. This is
             * because we need to preserve the order of the requests, so we
             * need to queue them while the dispatcher spinlock is held, but
             * we can't start them with the spinlock held, because that's a
             * potentially long operation. So, this inherently monolitic
             * operation was splitted.
             */
            ProcessorEnqueueRequest( pProc, pCalcReq);

            //statistics
            pThis->nReqs--;

            /*
             * Release the spinlock, because we've completed the dispatcher state
             * transition
             */
            PCrptSpinLockRelease( &pThis->SpinLock);

            /*
             * Start a request that's been queued earlier.
             *
             * Note! There is a window between the release of the spinlock and
             * request starting, where the order of the requests could be
             * changed would we not queue the request in advance. Of course,
             * the window still exist, but it is not dangerous, since the
             * requests are ordered properly in the processor's queue and the
             * following call simply starts the right request, which is
             * the first to start in the processor queue (you see that the
             * request ptr is not among the parms of this call)
             */
            ProcessorStartQueuedRequest( pProc);

            /*
             * Acquire the spinlock again, because we are going to begin the next
             * iteration - try to start more requests
             */
            PCrptSpinLockAcquire( &pThis->SpinLock);
        }
        else
        {
            /*
             * No processor is ready to accept the request - we've withdrawn it
             * from the queue in vain - put it back (at the top where it was),
             * stop the loop and quit. As soon as any PROCESSOR will complete a
             * request, it will call the DISPATCHER back via DispatcherStartNextRequest
             * and this func will be executed again and be able to start a request
             *
             * We use pCalcReq->Result field as the space for list entry, since this
             * field is not used until the "instruction" is de-queued and gets executed
             * by a 7751. After the instruction is de-queued, we don't ever need to
             * queue it again, so the Result is free to be filled by the 7751
             */
            ASSERT( sizeof( pCalcReq->Result) >= sizeof ( PPCRPT_LIST_ENTRY));
            PCrptListInsertTail( &pThis->Queue, ( PPCRPT_LIST_ENTRY) &pCalcReq->Result);
            bMore = FALSE;
        }
    }
}

/****************************************************************************

    This function is called with the DISPATCHER spinlock held

****************************************************************************/
static
int
FindLeastBusyProcessor
    (
        IN OUT  PDISPATCHER pThis,
        IN PBRIEF_COMMAND   pBriefComm
    )
{
    int     i;

//
// BUGBUG: This algo simply chooses the first processor which has a place in
// its internal request queue and is not inactive. It might be improved in
// the following ways:
//
//  - use affinities (through SlotMan and BriefComm),
//  - use the amount of data remaining to process on the processors (this will
//    require significant co-operation from the processor side, because it will
//    need to track the progress of processing using additional interrupts)
//
//   SLOT_AFFINITY

    ASSERT( pThis->Type == DISPATCHER_SIGN);
    ASSERT( pThis->nNumOfProcs != 0);

    for ( i = 0; i < pThis->nNumOfProcs; i++)
    {
        ASSERT( pThis->PCD[ i].pProc != NULL);

        if ( pThis->PCD[ i].pProc->nReqsInQueue < PROCESSOR_QUEUE_LENGTH &&
             !pThis->PCD[ i].pProc->bInactive)
        {
            /*
             * This processor index is what the caller is after
             */
            return i;
        }
    }
    return -1;
}

/****************************************************************************
DispSetProcessorInactiveState()

Called by a PROCESSOR when it needs some rest and time to recover from
a fatal error. Before changing the PROCESSOR state, acquires the spinlock,
because it intersect with the FindLeastBusyProcessor()


****************************************************************************/
void
DispSetProcessorInactiveState
    (
        IN OUT  PDISPATCHER pThis,
        IN OUT  struct _PROCESSOR  *pProc,
        IN      BOOLEAN     bNewState
    )
{
    DebugPrint((2, "DispSetProcessorInactiveState(%d)\n", bNewState));
    ASSERT( pThis->Type == DISPATCHER_SIGN);

    /*
     * Acquire the spinlock, because we are going to intersect with
     * FindLeastBusyProcessor()
     */
    PCrptSpinLockAcquire( &pThis->SpinLock);
    pProc->bInactive = bNewState;
    PCrptSpinLockRelease( &pThis->SpinLock);
}
