/**********************************************************************

    Module: dispatcher.h

        Product: kernel mode driver for "PowerCrypt" encryption
        acceleration board based on Hi/fn 7751 chip
        by Ashley Laurent, Inc.

    ---------------------------------------------------------------

    Copyright:

        Ashley Laurent, Inc., 1999
        All Rights Reserved.

    ---------------------------------------------------------------

    Functional Description:

        Declaration of the DISPATCHER object that performs the buffering and
        dispatching of calculating requests( 7751 instructions).

    ---------------------------------------------------------------

    Environment:

        NT or BSD kernel mode

    ---------------------------------------------------------------

    Author:

        Stan Shkolnyy

    ---------------------------------------------------------------

    Revision history:

        *   18-Mar-99   initial revision

**********************************************************************/

#ifndef __DISPATCHER_H__
#define __DISPATCHER_H__

#include "driverdef.h"      // for MAXIMUM_DEVICES_SUPPORTED
#include "pwrcryptioctl.h"  // for IOCTL_PWRCRYPT_CALCULATE_INPUT_OUTPUT
#include "briefcomm.h"      // for BRIEF_COMMAND

/**********************************************************************
 * The task of the dispatcher is to buffer incoming calculating requests and
 * route them to the 7751 processors in such a way that to maximize the
 * throughput.
 *
 * The dispatcher code executes every time when the current work situation
 * changes. The following events change the situation:
 *
 *  - a new calculating request has come,
 *  - a request has been completed by a 7751 processor.
 *
 * On each of these events, a thread enters the dispatcher, modifies its
 * current state and performs re-scheduling according to the new state.
 *
 * The actions performed on the situation change, are slightly different
 * depending on the event.
 *
 * On a new request, the dispatcher examines the current state of each of the
 * 7751 processors and puts the request to the private queue of the least busy
 * one. The private queue can only contain 2 requests and both of them are in
 * progress - being processed by the hardware. Then the dispatcher calls a
 * processor's method to wake up the processor, in case it is stopped. If none
 * of the processor private queues have a space for a new request, the
 * dispatcher queues it into its own queue. The PwrcryptCalculate() service
 * originates the "new request" events by calling DispatcherStartRequest()
 * method with a ptr to the new request.
 *
 * On a request completion, the dispatcher checks its own request queue. If it
 * is empty, there is nothing to do for the dispatcher. If not, the dispatcher
 * withdraws a request from the queue and then behaves exactly as in the "new
 * request" situation described above. (Note, that, unlike the above situation,
 * it is now guaranteed that at least one 7751 processor (that one that has
 * just completed a request) has a free space in its private queue). The
 * PROCESSOR objects originate the "request completed" events by calling
 * DispatcherStartNextRequest() method.
 *
 * The idea of this request processing mechanism is borrowed from NT driver
 * StartIO mechanism (PROCESSOR = DEVICE_OBJECT, Calculating Request = IRP),
 * with the following major distinctions:
 *
 *  - there may be more than 1 request in progress on a given PROCESSOR
 *    (in place of BOOLEAN DeviceQueue.Busy, a semaphore-like counter is used),
 *
 *  - the mechanism works on a group of PROCESSORs - for instance, there is no
 *    direct correspondence between the PROCESSOR that has just completed a
 *    request and the PROCESSOR that will process the next request.
 *

    Original StartIO mechanism
    --------------------------

    Synchronization is based on Busy flag - if it is set, nobody can proceed
    with calling StartIO. The flag "locks" the device.

    =========================================================================
    StartPacket( Pack)
    begin
        lock
        if ( !Busy)
            Busy = true
            unlock
            StartIO( Pack)
        else
            QueuePut( Pack)
            unlock
        end
    end
  
    =========================================================================
    StartNextPacket( )
    begin
        lock
        assert( Busy) ???
        Pack = QueueGet()
        if Pack
            unlock
            StartIO( Pack)
        else
            Busy = false
            unlock
        end
    end




    My modified mechanism
    ---------------------

    Synchronization is based on "can start request" condition - if it not true,
    nobody can proceed with calling StartIO. The condition "locks" the device.

    =========================================================================
    Dispatcher.StartPacket( Pack)
    begin
        lock
        QueuePut( Pack)
    ---------------------------
        while ( can start request AND ( Pack = QueueGet()))
            transit closer to can't start request
            ReadyProcessor.Busy++
            unlock
            ReadyProcessor.StartIO( Pack)
            lock
        end
    ---------------------------
        unlock
    end
  
    =========================================================================
    Dispatcher.StartNextPacket( CompletedProcessor)
    begin
        lock
        transit further from can't start request
        CompletedProcessor.Busy--
    ---------------------------
        while ( can start request AND ( Pack = QueueGet()))
            transit closer to can't start request
            ReadyProcessor.Busy++
            unlock
            ReadyProcessor.StartIO( Pack)
            lock
        end
    ---------------------------
        unlock
    end

    =========================================================================
    Processor.DPC
    begin
        lock
        CompletedPack = this.QueueWithdraw()
        unlock
        Dispatcher.StartNextPacket( this)
        Complete( CompletedPack)
    end

    =========================================================================
    Processor.StartIO( Pack)
    begin
        lock
        this.QueueDeposit( Pack)
        unlock
        this.StartPacket( Pack)
    end

 */

/*
 * DISPATCHER has the following control record for each PROCESSOR it governs
 */
struct _PROCESSOR;

typedef struct _PROC_CTL_DATA
{
    struct _PROCESSOR   *pProc;

} PROC_CTL_DATA, *PPROC_CTL_DATA;

typedef struct _DISPATCHER
{
    ULONG           Type;

    PCRPT_LIST      Queue;

    /*
     * For statistics only
     */
    ULONG           nReqs;
    ULONG           nMaxReqs;
    ULONG           nTotalReqs;

    /*
     * This spinlock is held while the dispatcher registers a work situation
     * change.
     */
    PCRPT_DECLARE_SPINLOCK( SpinLock);

    /*
     * The array of processor descriptors
     */
    int             nNumOfProcs;
    PROC_CTL_DATA   PCD[ MAXIMUM_DEVICES_SUPPORTED];

} DISPATCHER, *PDISPATCHER;

/**********************************************************************
 * PUBLIC methods
 */
void
DispatcherConstructor
    (
        IN OUT  PDISPATCHER pThis
    );

void
DispatcherDestructor
    (
        IN OUT  PDISPATCHER pThis
    );

ULONG
DispatcherRegisterProcessor
    (
        IN OUT  PDISPATCHER pThis,
        IN OUT  struct _PROCESSOR  *pProc
    );

void
DispSetProcessorInactiveState
    (
        IN OUT  PDISPATCHER pThis,
        IN OUT  struct _PROCESSOR  *pProc,
        IN      BOOLEAN     bNewState
    );

#define DispatcherDeactivateProcessor( pThis, pProc) \
                            DispSetProcessorInactiveState( pThis, pProc, TRUE)

#define DispatcherActivateProcessor( pThis, pProc) \
                            DispSetProcessorInactiveState( pThis, pProc, FALSE)

/*---------------------------------------------------------------------------
DispatcherStartRequest calls a processor's StartIo routine with the given request or
inserts the request into the processor's queue if the processor is already busy.

pProcessor
pRequest

*/

void
DispatcherStartRequest
    (
        IN OUT  PDISPATCHER pThis,
        IN      PIOCTL_PWRCRYPT_CALCULATE_INPUT_OUTPUT
                            pCalcReq
    );

/*---------------------------------------------------------------------------
DispatcherStartNextRequest dequeues the next request, if any, from a given
processor's queue and calls the processor's StartIo routine.

pProcessor

*/

void
DispatcherStartNextRequest
    (
        IN OUT  PDISPATCHER pThis,
        IN OUT  struct _PROCESSOR  *pProc
    );

#endif
