/**********************************************************************

    Module: dma_ctx.c

        Product: kernel mode driver for "PowerCrypt" encryption
        acceleration board based on Hi/fn 7751 chip
        by Ashley Laurent, Inc.

    ---------------------------------------------------------------

    Copyright:

        Ashley Laurent, Inc., 1999
        All Rights Reserved.

    ---------------------------------------------------------------

    Functional Description:

        Implementation of the DMA_CTX object that presents the interface
        through which PROCESSOR uses computer DMA facility

    ---------------------------------------------------------------

    Environment:

        NT or BSD kernel mode

    ---------------------------------------------------------------

    Author:

        Stan Shkolnyy

    ---------------------------------------------------------------

    Revision history:

        *   05-Mar-99   initial revision

**********************************************************************/

#include "osindependence.h"
#include "debug.h"
#include "driverdef.h"
#include "dma_ctx.h"
#include "xfer.h"

#ifdef ALLOC_PRAGMA
#pragma alloc_text(INIT, DmaCtxConstructor)
#pragma alloc_text(INIT, DmaCtxCreateDevice)
#pragma alloc_text(INIT, DmaCtxInit)
#pragma alloc_text(INIT, DmaCtxAllocateDmaMemory)
#endif

/****************************************************************************
****************************************************************************/
void
DmaCtxConstructor(
    PDMA_CTX            pThis,
    BOOLEAN             bWriteToDevice
    )
{
    DebugPrint((2, "DmaCtxConstructor\n"));

    pThis->Type = DMA_CTX_SIGN;

    pThis->pDevice = NULL;
    pThis->pAdapter = NULL;
    pThis->pCommBuf = NULL;
    pThis->bWriteToDevice = bWriteToDevice;

#ifdef DBG
    pThis->bDeviceQueued = FALSE;
    pThis->MapRegsClaimed = 0;
    pThis->OverlapCnt = 0;
    pThis->MapsInProgressCnt = 0;
#endif

}

/****************************************************************************
****************************************************************************/
void
DmaCtxDestructor(
    PDMA_CTX            pThis
    )
{
    DebugPrint((2, "DmaCtxDestructor\n"));
    ASSERT( pThis->Type == DMA_CTX_SIGN);

    ASSERT( !pThis->bDeviceQueued);

    /*
     * Free common buffer.
     */
    if ( pThis->pCommBuf != NULL)
    {
        DebugPrint((3, "    Freeing common buffer at 0x%08x\n", pThis->pCommBuf));
        HalFreeCommonBuffer( pThis->pAdapter,
                             pThis->CommBufSize,
                             pThis->CommBufPA,
                             pThis->pCommBuf,
                             FALSE          // means "not cached"
                             );
    }

    //
    // WORKWORK - I've never seen a manual that tells this should be done,
    // however it looks sensible for me and doesn't crash on my machine
    //
    // Update - this shouldn't be done. At least, it doesn't work as I expected
    // on the system DMA adapter object
    //
//    if ( pThis->pAdapter != NULL)
//    {
//        ASSERT(0);
//        ObDereferenceObject( pThis->pAdapter);
//    }

    if ( pThis->pDevice != NULL)
    {
        IoDeleteDevice( pThis->pDevice);
    }
}

/****************************************************************************
****************************************************************************/
NTSTATUS
DmaCtxCreateDevice(
    PDMA_CTX            pThis,
    IN PDRIVER_OBJECT   DriverObject,
    IN PUNICODE_STRING  pusDeviceName
    )
{
    NTSTATUS status;

    ASSERT( pThis->Type == DMA_CTX_SIGN);

    status = IoCreateDevice( DriverObject,
                             0,             // no dev.ext. sizeof( WORK_DEVICE_EXTENSION),
                             pusDeviceName,
                             FILE_DEVICE_UNKNOWN,
                             0,                      // No device-specific info
                             FALSE,                  // Not an exclusive device
                             &pThis->pDevice
                             );
    if ( !NT_SUCCESS( status))
    {
        DebugPrint((1, "Unable to create dma device object. status=0x%08x\n", status));
    }
    return status;
}

/****************************************************************************
****************************************************************************/
NTSTATUS
DmaCtxInit(
    IN OUT PDMA_CTX     pThis,
    IN ULONG            MapRegsDesired,
    IN ULONG            MaxXferSize
    )
{
    NTSTATUS         status;
    DEVICE_DESCRIPTION      DevDesc;
    ULONG                   MapRegsAllowed;

    DebugPrint((2, "DmaCtxGetAdapter\n"));
    ASSERT( pThis->Type == DMA_CTX_SIGN);
    ASSERT( pThis->pCommBuf == NULL);

    /*
     * dev.obj. is not needed for this func. to do its job, but it _should_ be
     * created by this time
     */
    ASSERT( pThis->pDevice);

    /*
     * Set up the adapter-specific information.
     *
     * On my Intel machine, HalGetAdapter cuts returned map regs number
     * according to DevDesc.MaximumLength, when I try to request more map
     * registers than it's needed to transfer DevDesc.MaximumLength bytes.
     * Since I still need this much map regs, I use PAGE_SIZE * MapRegsDesired
     * instead of MaxXferSize.
     */
    RtlZeroMemory( &DevDesc, sizeof( DevDesc));

    DevDesc.Version             = DEVICE_DESCRIPTION_VERSION;
    DevDesc.Master              = TRUE;
    DevDesc.ScatterGather       = TRUE;
    DevDesc.Dma32BitAddresses   = TRUE;
    DevDesc.BusNumber           = pThis->BusNumber;
    DevDesc.InterfaceType       = PCIBus;
    DevDesc.MaximumLength       = PAGE_SIZE * MapRegsDesired;   //MaxXferSize

    //
    // The following adapter-specific information is not required.
    //

    //        DevDesc.DemandMode = FALSE;
    //        DevDesc.AutoInitialize = FALSE;
    //        DevDesc.IgnoreCount = xxx;
    //        DevDesc.DmaChannel = xxx;
    //        DevDesc.DmaWidth = xxx;
    //        DevDesc.DmaSpeed = xxx;
    //        DevDesc.DmaPort = xxx;

    /*
     * We're asking for the max.number of DMA maping registers that we _can_
     * utilize, but our request is not guaranteed to be satisfied.
     * Alphas, for instance, have real mapping registers implemented in
     * hardware, so the returned allowed number may be smaller.
     */
    MapRegsAllowed = MapRegsDesired;
    pThis->pAdapter = HalGetAdapter( &DevDesc, &MapRegsAllowed);

    status = ( pThis->pAdapter != NULL && MapRegsAllowed != 0) ?
                            STATUS_SUCCESS :
                            STATUS_DEVICE_CONFIGURATION_ERROR;

    if ( NT_SUCCESS( status))
    {
        pThis->MaxMapRegs = MapRegsAllowed;

        pThis->bPartialTransfers = ( MapRegsAllowed < MapRegsDesired);

        if ( pThis->bPartialTransfers)
        {
            DebugPrint(( 2, "HalGetAdapter has cut our abilities! 0x%04x map regs granted\n",
                         MapRegsAllowed));
            ASSERT(0);

            //
            // BUGBUG: delete this if the partial MDL stuff is implemented
            //
            return STATUS_DEVICE_CONFIGURATION_ERROR;
        }
    }
    else
    {
        DebugPrint((1, "HalGetAdapter failed\n"));
    }
    return status;
}

/****************************************************************************
****************************************************************************/
PVOID
DmaCtxAllocateDmaMemory(
    IN OUT  PDMA_CTX    pThis,
    IN      ULONG       SgCtlAreaSize,
    OUT     ULONG       *pDmaAddr
    )
{
    DebugPrint((2, "DmaCtxAllocateDmaMemory size=0x%08x\n", SgCtlAreaSize));
    ASSERT( pThis->pCommBuf == NULL);

    #ifdef HW_BUG_SEARCH_BUILD
        pThis->CommBufSize = 4096;
    #else
        pThis->CommBufSize = SgCtlAreaSize;
    #endif

    pThis->pCommBuf = HalAllocateCommonBuffer(
                                    pThis->pAdapter,
                                    pThis->CommBufSize,
                                    &pThis->CommBufPA,
                                    FALSE      // means "not cached"
                                    );
    if ( pThis->pCommBuf == NULL)
    {
        DebugPrint((1, "HalAllocateCommonBuffer failed\n"));
    }
    else
    {
        DebugPrint((3, "    Allocated common buffer at VA=0x%08x PA=0x%08x Size=0x%04x\n", pThis->pCommBuf, pThis->CommBufPA.LowPart, pThis->CommBufSize));
        *pDmaAddr = pThis->CommBufPA.LowPart;
    }
    return pThis->pCommBuf;

}
/****************************************************************************
****************************************************************************/
void 
DmaCtxAllocateAdapterChannel(
    IN OUT PDMA_CTX     pThis,
    IN ULONG            NumberOfMapRegisters,
    IN PMDL             pMdl,
    IN PMAP_CONTROL     MapControlRoutine,
    IN PVOID            MapControlRoutineContext
    )
{
    NTSTATUS status;
    KIRQL    CallerIRQL;
    KIRQL    UnraisedIRQL;

    CallerIRQL = KeGetCurrentIrql();
    ASSERT( NumberOfMapRegisters == 0x41);

    DebugPrint((2, "DmaCtxAllocateAdapterChannel called at IRQL = %d\n", CallerIRQL));
    DebugPrint((2, "    mdl=0x%08x adap=0x%08x dev=0x%08x #mapregs= 0x%02x\n", pMdl, pThis->pAdapter, pThis->pDevice, NumberOfMapRegisters));
    ASSERT( pThis->Type == DMA_CTX_SIGN);

    /*
     * IoAllocateAdapterChannel must be called at DISPATCH_LEVEL -
     * raise it if necessary
     */
    if ( CallerIRQL < DISPATCH_LEVEL)
    {
        DebugPrint((2, "    raising IRQL to DISPATCH_LEVEL\n"));
        KeRaiseIrql( DISPATCH_LEVEL, &UnraisedIRQL);
        ASSERT( UnraisedIRQL == CallerIRQL);
    }

#ifdef DBG
    if ( pThis->MapRegsClaimed != 0)
    {
        pThis->OverlapCnt++;
    }
#endif

    /*
     * Copy these two callback coordinates - they'll be used in
     * AdapterControlRoutine to call the specified callback.
     */
    pThis->MapControlRoutine        = MapControlRoutine;
    pThis->MapControlRoutineContext = MapControlRoutineContext;

#ifdef DBG
    pThis->MapRegsClaimed += NumberOfMapRegisters;
#endif

    /*
     * The driver algorithms are not supposed to claim more map regs than
     * it is available. In practice, this would probably lead
     * to pDevice to be queued despite the fact that nobody is executing
     * its AdapterControl, due to map regs shortage. Again, I don't know
     * for sure. BUGBUG - however, I don't rely on this in the driver, so
     * it doesn't matter.
     */
    ASSERT( pThis->MapRegsClaimed <= pThis->MaxMapRegs);

    /* pDevice must not be queued twice!
     */
    ASSERT( !pThis->bDeviceQueued);

#ifdef DBG
    pThis->bDeviceQueued = TRUE;
#endif

    /*
     * Flush processors' caches
     */
    KeFlushIoBuffers(
        pMdl,
        !pThis->bWriteToDevice, // if the driver is reading information from the device to system memory
        TRUE);                  // if DMA transfer, not PIO

    pThis->MapsInProgressCnt++;

    /* IoAllocateAdapterChannel either queues pDevice to pAdapter queue or
     * calls the AdapterControlRoutine callback immediately, before returning.
     */
    status = IoAllocateAdapterChannel(
                pThis->pAdapter,
                pThis->pDevice,
                NumberOfMapRegisters,
                AdapterControlRoutine,  //IN PDRIVER_CONTROL
                pThis);                 //IN PVOID  Context

    /*
     * DDK says that it can fail with STATUS_INSUFFICIENT_RESOURCES only if
     * NumberOfMapRegisters is larger than the number acquired by
     * HalGetAdapter, so, in this driver, it's a programmer's error if it fails.
     */
    ASSERT( NT_SUCCESS( status));

    /* bDeviceQueued is set to FALSE by the callback, check it now to see
     * whether the callback was queued or called "inline". Of course, there
     * is a "race" window here, but this is for debugging only, to catch
     * my errors.
     */
    DebugPrint(( 3,
                 ( pThis->bDeviceQueued ?
                 "AdapterControlRoutine has been queued\n" :
                 "AdapterControlRoutine has been called ""inline""\n")
              ));

    /*
     * Lower the IRQL if it was raised earlier
     */
    if ( CallerIRQL < DISPATCH_LEVEL)
    {
        DebugPrint((2, "    Lowering IRQL back to %d\n", UnraisedIRQL));
        KeLowerIrql( UnraisedIRQL);
    }
}

/****************************************************************************
****************************************************************************/
IO_ALLOCATION_ACTION
AdapterControlRoutine(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp,
    IN PVOID MapRegisterBase,
    IN PVOID Context            // pThis is passed through this parameter
    )
{
    PDMA_CTX pThis;

    DebugPrint((2, "(DmaCtx)AdapterControlRoutine\n"));
    DebugPrint((2, "    mapregbase=0x%08x\n", MapRegisterBase));

    ASSERT( KeGetCurrentIrql() == DISPATCH_LEVEL);

    /*
     * Go to the proper DMA_CTX object by Context, which contains PDMA_CTX
     * as it was arranged by DmaCtxAllocateAdapterChannel
     */
    pThis = ( PDMA_CTX) Context;
    ASSERT( pThis->Type == DMA_CTX_SIGN);

    /*
     * the device is already queued twice!
     */
    ASSERT( pThis->bDeviceQueued);

    /*
     * When this routine is called, I think, the pDevice is already
     * de-queued from pAdapter queue. Do I think right??????
     */
#ifdef DBG
    pThis->bDeviceQueued = FALSE;
#endif

    /* Call the routine registered by our client. In that routine, it will
     * map its phys.RAM range to log.DMA range by calling our DmaCtxMapTransfer
     * one or more times
     */
    (pThis->MapControlRoutine)( pThis->MapControlRoutineContext, MapRegisterBase);

    /* Perform the required busmaster charm, which means we want to make this
     * ADAPTER_OBJECT available for new IoAllocateAdapterChannel calls but
     * also want to keep the mapping registers that we currently own and free
     * them later via IoFreeMapRegisters
     */
    return DeallocateObjectKeepRegisters;
}

/****************************************************************************
****************************************************************************/
ULONG
DmaCtxMapTransfer(
    IN OUT PDMA_CTX     pThis,
    IN PVOID            MapRegisterBase,
    IN PMDL             pMdl,
    IN PVOID            CurrentVa,
    IN OUT PULONG       pLength
    )
{
    PHYSICAL_ADDRESS    LogAddr;

    DebugPrint((2, "DmaCtxMapTransfer\n"));
    ASSERT( pThis->Type == DMA_CTX_SIGN);

//    DebugPrint((2, "    mdl=0x%08x mapregbase=0x%08x adap=0x%08x\n", pMdl, MapRegisterBase, pThis->pAdapter));

    ASSERT( KeGetCurrentIrql() == DISPATCH_LEVEL);

    LogAddr = IoMapTransfer(
                    pThis->pAdapter,
                    pMdl,
                    MapRegisterBase,
                    CurrentVa,
                    pLength,
                    pThis->bWriteToDevice);

    return LogAddr.LowPart;
}

/****************************************************************************
****************************************************************************/
VOID
DmaCtxCompleteTransfer(
    IN OUT PDMA_CTX     pThis,
    IN PVOID            MapRegisterBase,
    IN ULONG            NumberOfMapRegisters,
    IN PMDL             pMdl
//    IN PVOID            TransferVa,
//    IN ULONG            TransferLength
    )
{
    BOOLEAN bFlushOk;

    DebugPrint((2, "DmaCtxCompleteTransfer\n"));
    ASSERT( pThis->Type == DMA_CTX_SIGN);

    DebugPrint((2, "    mdl=0x%08x adap=0x%08x #mapregs= 0x%02x mapregbase=0x%08x\n", pMdl, pThis->pAdapter, NumberOfMapRegisters, MapRegisterBase));
    ASSERT( NumberOfMapRegisters == 0x41);

    ASSERT( KeGetCurrentIrql() == DISPATCH_LEVEL);

    ASSERT( pThis->MapsInProgressCnt <= DMA_CTX_MAX_PARALLEL_MAPS &&
            pThis->MapsInProgressCnt > 0);

    pThis->MapsInProgressCnt--;
    #ifdef DBG
        pThis->MapRegsClaimed -= NumberOfMapRegisters;
    #endif

    bFlushOk = IoFlushAdapterBuffers(
                    pThis->pAdapter,
                    pMdl,
                    MapRegisterBase,
                    MmGetMdlVirtualAddress( pMdl),  //TransferVa,
                    MmGetMdlByteCount( pMdl),       //TransferLength,
                    pThis->bWriteToDevice);

    ASSERT( bFlushOk);

    IoFreeMapRegisters( pThis->pAdapter, MapRegisterBase, NumberOfMapRegisters);
}
