/**********************************************************************

    Module: dma_ctx.h

        Product: kernel mode driver for "PowerCrypt" encryption
        acceleration board based on Hi/fn 7751 chip
        by Ashley Laurent, Inc.

    ---------------------------------------------------------------

    Copyright:

        Ashley Laurent, Inc., 1999
        All Rights Reserved.

    ---------------------------------------------------------------

    Functional Description:

        Declaration of the DMA_CTX object that presents the interface
        through which PROCESSOR uses computer DMA facility


    ---------------------------------------------------------------

    Environment:

        NT or BSD kernel mode

    ---------------------------------------------------------------

    Author:

        Stan Shkolnyy

    ---------------------------------------------------------------

    Revision history:

        *   05-Mar-99   initial revision

**********************************************************************/

#ifndef __DMA_CTX_H__
#define __DMA_CTX_H__

/****************************************************************************
 * DMA_CTX user-defined routine that is called as a consequence of a call to
 * DmaCtxAllocateAdapterChannel. In this routine, the user can call
 * DmaCtxMapTransfer to map its phys.RAM range to a number of logical DMA 
 * ranges.
 *
 * The callback must store away the MapRegisterBase value passed to it.
 * It will need this value to pass it to DmaCtxFreeMapRegisters later, when
 * the transfer is over.
 *
 * We chose not to store this and other values internally, because there may be
 * two simultaneous transfers in progress in a single DMA_CTX and we would have
 * to store 2 sets of such values and introduce "handles" to give the user, etc.
 *
 * They are valid only until the user returns from the callback routine.
 * After the routine returns, a new transfer may be initiated in the DMA
 * context, since DMA_CTX code frees "adapter channel" after calling the
 * routine.
 */
typedef void
(*PMAP_CONTROL)(
    IN PVOID        Context,
    IN PVOID        MapRegBase
    );

/****************************************************************************
 * Include O/S-specific stuff
 */
#ifdef _NT_DRIVER_BUILD_
    #include "dma_ctx_nt.h"
#endif
#ifdef _BSD_DRIVER_BUILD_
    #include "dma_ctx_bsd.h"
#endif

/****************************************************************************
 * PUBLIC methods for PROCESSOR. PROCESSOR calls them when it needs to map
 * physical RAM addresses to "logical" DMA addresses which are then loaded
 * into 7751 scatter/gather descriptors.
 */

PWRCRYPT_STATUS
DmaCtxInit(
    IN OUT PDMA_CTX     pThis,
    IN ULONG            MapRegsDesired,
    IN ULONG            MaxXferSize
    );

PVOID
DmaCtxAllocateDmaMemory(
    IN OUT  PDMA_CTX    pThis,
    IN      ULONG       SgCtlAreaSize,
    OUT     ULONG       *pDmaAddr
    );

void 
DmaCtxAllocateAdapterChannel(
    IN OUT PDMA_CTX     pThis,
    IN ULONG            NumberOfMapRegisters,
    IN PMDL             pMdl,
    IN PMAP_CONTROL     MapControlRoutine,
    IN PVOID            MapControlRoutineContext
    );

//PHYSICAL_ADDRESS
ULONG
DmaCtxMapTransfer(
    IN OUT PDMA_CTX     pThis,
    IN PVOID            MapRegisterBase,
    IN PMDL             pMdl,
    IN PVOID            CurrentVa,
    IN PULONG           pLength
    );

VOID
DmaCtxCompleteTransfer(
    IN OUT PDMA_CTX     pThis,
    IN PVOID            MapRegisterBase,
    IN ULONG            NumberOfMapRegisters,
    IN PMDL             pMdl
//    IN PVOID            TransferVa,
//    IN ULONG            TransferLength
    );

#endif /* __DMA_CTX_H__ */
