/**********************************************************************

    Module: dma_ctx_nt.h

        Product: kernel mode driver for "PowerCrypt" encryption
        acceleration board based on Hi/fn 7751 chip
        by Ashley Laurent, Inc.

    ---------------------------------------------------------------

    Copyright:

        Ashley Laurent, Inc., 1999
        All Rights Reserved.

    ---------------------------------------------------------------

    Functional Description:

        Declaration of the NT-specific part of the DMA_CTX object that
        presents the interface through which PROCESSOR uses computer
        DMA facility

    ---------------------------------------------------------------

    Environment:

        NT kernel mode

    ---------------------------------------------------------------

    Author:

        Stan Shkolnyy

    ---------------------------------------------------------------

    Revision history:

        *   05-Mar-99   initial revision

**********************************************************************/

#ifndef __DMA_CTX_NT_H__
#define __DMA_CTX_NT_H__

/****************************************************************************
 * This many transfers can be simultaneously mapped in a single DMA context
 */
#define DMA_CTX_MAX_PARALLEL_MAPS   2

/****************************************************************************
 * DMA context
 *
 * The purpose of DMA_CTX is to hide as much NT specific details as possible
 * from the PROCESSOR. Each PROCESSOR uses 2 DMA_CTX instances, for two
 * transfer directions.
 *
 * The context is half-duplex - the direction of transfers is specified in a
 * constructor parm and can't be changed.
 *
 * Though it was stated above that _two_ transfers can be in progress in a
 * single DMA_CTX simultaneously, this parallelism has a limitation.
 *
 * Two transfers can be simultaneously in progress, but not two MAP_CONTROL
 * routines. A new MAP_CONTROL routine can't be scheduled by the
 * DmaCtxAllocateAdapterChannel until the previuosly scheduled routine is called.
 *
 * "Two transfers in progress" should be understood as two MAP_CONTROL routines
 * scheduled and called without overlapping and two map register sets used for
 * DMA mapping, each mapping made in its corresponding MAP_CONTROL routine.
 * This way, two map reg sets are used simultaneously.
 */

typedef struct _DMA_CTX
{
    /************************************************************************
     * CONST members - filled once and never change
     */

    ULONG           Type;

    /*-----------------------------------------------------------------------
     * These vars are constant during the DMA_CTX life. They are passed to
     * DMA mapping O/S interfaces internally
     */
    PDEVICE_OBJECT  pDevice;
    PADAPTER_OBJECT pAdapter;
    ULONG           BusNumber;  // System bus number
    ULONG           MaxMapRegs;
    BOOLEAN         bWriteToDevice;

    /*-----------------------------------------------------------------------
     * Returned by HalGetCommonBuffer - valid if ( pSgCtlArea != NULL)
     */
    PVOID           pCommBuf;       // VA of the scatter/gather area
    PHYSICAL_ADDRESS    CommBufPA;  // Phys.address of the scatter/gather area

    /*-----------------------------------------------------------------------
     * scatter/gather area size in bytes, used only to free it, in the dtor
     */
    ULONG           CommBufSize;

    /*-----------------------------------------------------------------------
     * indicates whether the system has cut our request for map registers
     * and partial transfers are possible.
     */
    BOOLEAN         bPartialTransfers;

    /************************************************************************
     * NON-CONST members - change on each new DMA mapping
     */

    /*-----------------------------------------------------------------------
     * The address and a parm of a routine that must be called from our
     * AdapterControlRoutine. PROCESSOR specifies these values when it calls
     * DmaCtxAllocateAdapterChannel method.
     */
    PMAP_CONTROL    MapControlRoutine;
    PVOID           MapControlRoutineContext;

    /*
     * any user-defined values
     */
    ULONG           UserUlong1;
    ULONG           UserUlong2;
    ULONG           UserUlong4;

#ifdef DBG

    /*-----------------------------------------------------------------------
     * DEVICE_OBJECT is queued to ADAPTER_OBJECT queue by IoAllocateAdapterChannel
     * One dev.obj. can't be queued twice at the same time, since it contains only
     * 1 embedded WAIT_CONTEXT_BLOCK used for this purpose.
     * This flag watches the "queued" state of pDevice
     */
    BOOLEAN         bDeviceQueued;

    /*-----------------------------------------------------------------------
     * Tracks the number of map regs claimed by user calling
     * DmaCtxAllocateAdapterChannel. The PROCESSOR is not expected to claim more
     * map regs than indicated by MaxMapRegs, so ASSERTs check this condition.
     */
    ULONG           MapRegsClaimed;

    /*-----------------------------------------------------------------------
     * Counter of real mapping overlappings occured in this DMA context, when
     * some map regs are still in use but a new portion of map regs is given to
     * the user to make a new mapping. I want to see this behavior and need to
     * check it. The overlapping occurences are registered by AdapterControlRoutine
     */
    ULONG           OverlapCnt;
    ULONG           MapsInProgressCnt;

#endif

} DMA_CTX, *PDMA_CTX;

/****************************************************************************
 * PUBLIC methods for PROCBASE. PROCBASE calls them (except dtor) before
 * passing this context to the PROCESSOR, to fully prepare the context.
 */

void
DmaCtxConstructor(
    IN OUT PDMA_CTX     pThis,
    BOOLEAN             bWriteToDevice
    );

void
DmaCtxDestructor(
    IN OUT PDMA_CTX     pThis
    );

#define DmaCtxAssignBusNumber( pThis, Number) (pThis)->BusNumber = Number

NTSTATUS
DmaCtxCreateDevice(
    IN OUT PDMA_CTX     pThis,
    IN PDRIVER_OBJECT   DriverObject,
    IN PUNICODE_STRING  pusDeviceName
    );

/****************************************************************************
 * PUBLIC methods for PROCESSOR. PROCESSOR calls them when it needs to map
 * physical RAM addresses to "logical" DMA addresses which are then loaded
 * into 7751 scatter/gather descriptors.
 */

#define DmaCtxGetSgCtlAreaPtr( pThis)       ( pThis)->pCommBuf
#define DmaCtxGetSgCtlAreaPhysAddr( pThis)  ( pThis)->CommBufPA.LowPart
#define DmaCtxGetMaxMapRegs( pThis)         ( pThis)->MaxMapRegs
#define DmaCtxIsPartialTransfers( pThis)    ( pThis)->bPartialTransfers

/****************************************************************************
 * PRIVATE methods
 */

/* This prototype is NT-defined and can't be changed
 */
IO_ALLOCATION_ACTION
AdapterControlRoutine(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp,
    IN PVOID MapRegisterBase,
    IN PVOID Context            // (PDMA_CTX pThis) is passed through this parameter
    );

#endif /* __DMA_CTX_NT_H__ */
