/**********************************************************************

    Module: driver.c

        Product: kernel mode driver for "PowerCrypt" encryption
        acceleration board based on Hi/fn 7751 chip
        by Ashley Laurent, Inc.

    ---------------------------------------------------------------

    Copyright:

        Ashley Laurent, Inc., 1999
        All Rights Reserved.

    ---------------------------------------------------------------

    Functional Description:

        The main file for the NT driver. Contains the usual stuff NT requires
        any driver to have - DriverEntry and necessary dispatch routines.

    ---------------------------------------------------------------

    Environment:

        NT or BSD kernel mode

    ---------------------------------------------------------------

    Author:

        Stan Shkolnyy

    ---------------------------------------------------------------

    Revision history:

        *   05-Mar-99   initial revision

**********************************************************************/

#include "osindependence.h"
#include "debug.h"
#include "myunistr.h"
#include "driver.h"
#include "globobj.h"

#include "findpci.h"

/* We need only the PWRCRYPT_LINK_DEVICE_UNICODE_NAME from the km_interface.h,
 * but km_interface.h needs two .h-files before it
 */
#include "pwrcryptioctl.h"
#include "kmi_private.h"

#include "memtrack.h"

#ifdef DBG

#define DEBUG_BUFFER_SIZE   512

ULONG DebugPrintLevel = 3;
UCHAR DebugBuffer[DEBUG_BUFFER_SIZE];


#endif

/****************************************************************************
 * Driver's temporary data used by DriverEntry
 */
typedef struct _DRIVER_TEMPS
{
    UNICODE_STRING          usLinkDeviceName;
    UNICODE_STRING          usWorkDeviceNameX;
    PPCI_DEVICE_LIST        pciDevList;

} DRIVER_TEMPS, *PDRIVER_TEMPS;

NTSTATUS
DriverTempsConstructor
    (
        PDRIVER_TEMPS pThis
    );

/*
 * Make sure the initialization code is removed from memory after use.
 */
#ifdef ALLOC_PRAGMA
#pragma alloc_text(INIT, DriverEntry)
#pragma alloc_text(INIT, DriverTempsConstructor)
#endif

/****************************************************************************
****************************************************************************/
NTSTATUS
DriverTempsConstructor( PDRIVER_TEMPS pThis)
{
    pThis->pciDevList = NULL;

    //
    // Allocate storage for the list of PCI devices (take into account that the
    // definition of PCI_DEVICE_LIST already includes 1 PCI_DEVICE_LOCATION).
    //

    pThis->pciDevList = (PPCI_DEVICE_LIST) MemAllocatePool( NonPagedPool,
                                                     sizeof(PCI_DEVICE_LIST) +
                                                        ( MAX_PCIDEVICELIST_ENTRIES - 1) *
                                                        sizeof(PCI_DEVICE_LOCATION));
    if ( pThis->pciDevList == NULL)
    {
        DebugPrint((1, "Unable to allocate storage for pciDeviceList \n"));
        return STATUS_INSUFFICIENT_RESOURCES;
    }

    /* Prepare a unicode name for the future linking device object.
     */
    RtlInitUnicodeString( &pThis->usLinkDeviceName, PWRCRYPT_LINK_DEVICE_UNICODE_NAME);

    /* Prepare a unicode name for the future working device object.
     */
    RtlInitUnicodeString( &pThis->usWorkDeviceNameX, PWRCRYPT_WORK_DEVICE_UNICODE_NAME_X);

    return STATUS_SUCCESS;
}

/****************************************************************************
****************************************************************************/
void
DriverTempsDestructor( PDRIVER_TEMPS pThis)
{
    if ( pThis->pciDevList != NULL)
    {
        MemFreePool( pThis->pciDevList);
    }
}


/****************************************************************************
****************************************************************************/
NTSTATUS
DriverEntry(
    IN PDRIVER_OBJECT DriverObject,
    IN PUNICODE_STRING RegistryPath
    )
/*++

Routine Description:

    Determine if a PCI device exists, setup relevant structures, 
    claim resources, connect interrupts, and report status back to
    the system.  Most of the work done is in routines called by
    DriverEntry.
    
Arguments:

    DriverObject - Pointer to the driver object created by the I/O manager.
    
    RegistryPath - Pointer to the driver specific key 
                   \Registry
                      \Machine
                         \System
                            \CurrentControlSet
                               \Services
                                  \<DriverName>
    
Return Value:

    NTSTATUS
    
Remarks:
    
    The routine also creates a device object. The sole purpose of the device,
    as hinted by its name, is to support retention of this driver in memory
    while somebody uses it. Users are kernel-mode drivers.

    Clients use the driver by calling its exported DLL entry points. However,
    this driver can be unloaded while others still think it is present.
    To solve the problem, a device is created by this driver, and the client
    must call IoGetDeviceObjectPointer to obtain a pointer to the corresponding
    file object (IRP_MJ_CREATE is sent to us). The client must store this pointer
    until it uses this driver. Then, it must call ObDereferenceObject on this
    file object pointer (IRP_MJ_CLOSE is sent).

    NET STOP effect. If NET STOP is issued for this driver and clients are using
    it, it won't be stopped immediately. However, then, no new clients will be
    able to obtain a pointer to the file object for its device. When the last of
    former clients issues its ObDereferenceObject, this driver is unloaded
    automatically.

    DLL design consequences. Since this driver is also a DLL, it is loaded
    automatically when a dynamically linked client is loaded. However, NT
    doesn't treat it as a driver in this case and DriverEntry is not called.
    After that, it remains in memory even after the client is unloaded and
    neither NET STOP nor NET START work on it.

--*/
{
    NTSTATUS        status;
    DRIVER_TEMPS    temp;   // data that used locally but needs cleanup

    USHORT          i;
    int             nMaxDevicesToAcquire;

    DebugPrint((3, "DriverEntry routine. DriverObject at 0x%x \n", DriverObject));

    /*
     * Initialize Russinovich's memory tracking library
     */
    MemTrackInit();

    /************************************************************************
     * Initialize the driver's permanent/global and
     * temp (for this routine only) data. The reason for this is that some
     * of this data requires memory deallocations and we don't want to spread
     * them throughout this routine.
     */
    status = GlobalObjectsConstructor( &glob, RegistryPath);
    if ( !NT_SUCCESS( status))
    {
        DebugPrint((1, "Unable to construct global objects. status=0x%08x\n", status));
        GlobalObjectsDestructor( &glob, DriverObject);
        return status;
    }

    status = DriverTempsConstructor( &temp);
    if ( !NT_SUCCESS( status))
    {
        DebugPrint((1, "Unable to construct temporary data. status=0x%08x\n", status));
        GlobalObjectsDestructor( &glob, DriverObject);
        DriverTempsDestructor( &temp);
        return status;
    }

    status = GlobalObjectsInit( &glob, RegistryPath);
    if ( !NT_SUCCESS( status))
    {
        DebugPrint((1, "Unable to initialize global objects. status=0x%08x\n", status));
        GlobalObjectsDestructor( &glob, DriverObject);
        return status;
    }

    /*
     * Now, the glob.Config is alive and can be used
     */

    /************************************************************************
     * Fill the driver's function table.
     * It provides:
     *      Unload - to be able to unload.
     *      Create/Close - for it "link" device to be opened by KM clients
     */
    DriverObject->DriverUnload = PwrCryptUnload;
    DriverObject->MajorFunction[IRP_MJ_CREATE] = PwrCryptCreate;
    DriverObject->MajorFunction[IRP_MJ_CLOSE] = PwrCryptClose;

    /************************************************************************
     * Find all the PCI devices specified by *_IDs and put their locations
     * into the list.
     * If the func returns success, at least 1 device has been found
     */
    status = FindPciDevice( PWRCRYPT_PCI_VENDOR_ID,
                            PWRCRYPT_PCI_DEVICE_ID,
                            temp.pciDevList);
    if ( !NT_SUCCESS( status))
    {
        DebugPrint((1, "Unable to find a needed PCI device\n"));
        GlobalObjectsDestructor( &glob, DriverObject);
        DriverTempsDestructor( &temp);
        return status;
    }
    else
    {
        DebugPrint((3, "Found %d needed PCI device(s)\n", temp.pciDevList->FoundDevCnt));
    }

    /************************************************************************
     * Create the "link" device object needed for interfacing to clients.
     */
    status = IoCreateDevice( DriverObject,
                             0,                 // no dex.ext. //sizeof( LINK_DEVICE_EXTENSION),
                             &temp.usLinkDeviceName,
                             FILE_DEVICE_UNKNOWN,
                             0,                      // No device-specific info
                             FALSE,                  // Not an exclusive device
                             &glob.pLnkDev
                             );
    if ( !NT_SUCCESS( status))
    {
        DebugPrint((1, "Unable to create link device object. status=0x%08x\n", status));
        GlobalObjectsDestructor( &glob, DriverObject);
        DriverTempsDestructor( &temp);
        return status;
    }

    /*
     * Limit the number of the devices to acquire by the user-specified
     * number. If more devices are plugged into the computer, the rest will
     * be ignored.
     */
    nMaxDevicesToAcquire = ConfigGetDeviceCntLimit( &glob.Config);

    /************************************************************************
     * Create 1 PROC_BASE object for each device found.
     *
     * Note that some of devices may fail to initialize - they will be dropped
     * and the initialization will continue
     */
    for ( i = 0;
          i < nMaxDevicesToAcquire &&
          i < temp.pciDevList->FoundDevCnt &&
          glob.NumOfValidProcBases < MAXIMUM_DEVICES_SUPPORTED;
          i++ )
    {
        PPCI_DEVICE_LOCATION    pciDevInfo;
    
        //
        // Get current PCI device location.
        //
        pciDevInfo = &temp.pciDevList->List[ i];

        ASSERT( glob.pProcBase[ i] == NULL);
        DebugPrint((3, "    allocating 0x%04x bytes for PROCBASE[%d]\n", sizeof( PROCBASE), i));

        glob.pProcBase[ i] = (PPROCBASE) MemAllocatePool( NonPagedPool,
                                                        sizeof( PROCBASE));
        if ( glob.pProcBase[ i] != NULL)
        {
            ProcBaseConstructor( glob.pProcBase[ i], &glob.Config);

            /* Customize device name for this processor
             */
            ReplaceOrderNum( &temp.usWorkDeviceNameX, i);

            /* Customize device registry path for this processor
             */
            ReplaceOrderNum( &glob.usProcRegPathCopyX, i);

            status = ProcBaseAcquireHardware( glob.pProcBase[ i],
                                              &glob.Dispatcher,
                                              &glob.usProcRegPathCopyX,
                                              pciDevInfo->BusNumber,
                                              pciDevInfo->SlotNumber.u.AsULONG,
                                              DriverObject,
                                              &temp.usWorkDeviceNameX);
            if ( NT_SUCCESS( status))
            {
                glob.NumOfValidProcBases ++;
            }
            else
            {
                ASSERT(0);
                status = ProcBaseReleaseHardware( glob.pProcBase[ i],
                                                  &glob.usProcRegPathCopyX,
                                                  DriverObject);
                ASSERT( NT_SUCCESS( status));
                ProcBaseDestructor( glob.pProcBase[ i]);
                MemFreePool( glob.pProcBase[ i]);
                glob.pProcBase[ i] = NULL;
            }
        }
    } //for

    /************************************************************************
     * Check whether some PROCBASEs were created and finally decide
     * whether to load
     */
    status = ( glob.NumOfValidProcBases != 0 ? STATUS_SUCCESS : STATUS_NO_SUCH_DEVICE);

    /* If "to load", enable the KMInterface, else destruct everything
     */
    if ( NT_SUCCESS( status))
    {
        KmInterfaceEnable();
    }
    else
    {
        DebugPrint((1, "No work devices were created\n"));
        GlobalObjectsDestructor( &glob, DriverObject);
    }

    DriverTempsDestructor( &temp);
    return status;

} // DriverEntry


/****************************************************************************

NTSTATUS
PwrCryptCreate(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp
    )

Description:
    
Arguments:

    DeviceObject - Object representing a particular adapter.
    
    Irp - IRP_MJ_CREATE I/O request.
    
Return Value:

    NTSTATUS
****************************************************************************/
NTSTATUS
PwrCryptCreate(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp
    )
{
    NTSTATUS status;

    DebugPrint((3, "PwrCryptCreate\n"));

    /*
     * Only "link" device is supposed to be opened
     */
    if ( DeviceObject == glob.pLnkDev)
    {
        status = STATUS_SUCCESS;
    }
    else
    {
        DebugPrint(( 1, "Error - IRP_MJ_OPEN sent to a work device\n"));
        status = STATUS_ACCESS_DENIED;
    }

    Irp->IoStatus.Status = status;
    Irp->IoStatus.Information = 0;

    IoCompleteRequest (Irp, IO_NO_INCREMENT);

    return status;

}   // PwrCryptCreate



/****************************************************************************

NTSTATUS
PwrCryptClose(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp
    )

Description:

Arguments:

    DeviceObject - Object representing a particular adapter.

    Irp - IRP_MJ_CLOSE I/O request.

Return Value:

    NTSTATUS
****************************************************************************/
NTSTATUS
PwrCryptClose(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp
    )
{
    NTSTATUS status;

    DebugPrint((3, "PwrCryptClose\n"));

    /*
     * Only "link" device is supposed to be opened and closed
     */
    if ( DeviceObject == glob.pLnkDev)
    {
        status = STATUS_SUCCESS;
    }
    else
    {
        DebugPrint(( 1, "Error - IRP_MJ_CLOSE sent to a work device\n"));
        status = STATUS_ACCESS_DENIED;
    }

    return status;
    Irp->IoStatus.Status = status;
    Irp->IoStatus.Information = 0;

    IoCompleteRequest (Irp, IO_NO_INCREMENT);

    return status;

}   // PwrCryptDmaClose

/****************************************************************************

VOID
PwrCryptUnload(
    IN PDRIVER_OBJECT DriverObject
    )

Description:

    Prepare the driver for unload.  For all the device objects created
    for this driver, loop through each one, disconnect interrupts, unmap
    memory and I/O spaces, free memory, etc.

Arguments:

    DriverObject - Pointer to the driver object created by the I/O manager.

Return Value:

    None
****************************************************************************/
VOID
PwrCryptUnload(
    IN PDRIVER_OBJECT DriverObject
    )
{
    DebugPrint((3, "PwrCryptUnload\n"));

    GlobalObjectsDestructor( &glob, DriverObject);
}
