/**********************************************************************

    Module: globobj.c

        Product: kernel mode driver for "PowerCrypt" encryption
        acceleration board based on Hi/fn 7751 chip
        by Ashley Laurent, Inc.

    ---------------------------------------------------------------

    Copyright:

        Ashley Laurent, Inc., 1999
        All Rights Reserved.

    ---------------------------------------------------------------

    Functional Description:

        Construction/init/destruction of the driver's global objects and data

    ---------------------------------------------------------------

    Environment:

        NT or BSD kernel mode

    ---------------------------------------------------------------

    Author:

        Stan Shkolnyy

    ---------------------------------------------------------------

    Revision history:

        *   05-Mar-99   initial revision

**********************************************************************/

#include "osindependence.h"
#include "debug.h"
#include "myunistr.h"
#include "globobj.h"

/*
 * Make sure the initialization code is removed from memory after use.
 */
#ifdef ALLOC_PRAGMA
#pragma alloc_text(INIT, GlobalObjectsConstructor)
#pragma alloc_text(INIT, GlobalObjectsInit)
#endif

/****************************************************************************
 * Driver's global objects and data
 */
GLOBAL_OBJECTS glob;

/****************************************************************************
****************************************************************************/
NTSTATUS
GlobalObjectsConstructor( PGLOBAL_OBJECTS pThis, PUNICODE_STRING pusDriverRegPath)
{
    NTSTATUS    status;
    ULONG       nIndex;

    WatchConstructor( &pThis->Watch);

    ConfigConstructor( &pThis->Config);
    SlotmanConstructor( &pThis->SlotMan);
    pThis->usProcRegPathCopyX.Buffer = NULL;
    pThis->NumOfValidProcBases = 0;
    DispatcherConstructor( &pThis->Dispatcher);

    /*
     * No PROC_BASEs exist yet
     */
    for ( nIndex = 0; nIndex < MAXIMUM_DEVICES_SUPPORTED; nIndex++)
    {
        pThis->pProcBase[ nIndex] = NULL;
    }

    /*
     * No link device exists yet
     */
    pThis->pLnkDev = NULL;

    /*
     * Allocate memory and create copy of the reg.path or fail the function.
     * Note that the copy is longer enough to be appended by "\ProcessorX" later
     */
    status = MyCloneUnicodeString( &pThis->usProcRegPathCopyX,
                                   pusDriverRegPath,
                                   sizeof( PROCESSOR_REG_SUB_PATH)    // delta
                                   );
    if ( !NT_SUCCESS( status))
    {
        return status;
    }

    /*
     * Append the copy with "\ProcessorX"
     */
    status = RtlAppendUnicodeToString( &pThis->usProcRegPathCopyX, PROCESSOR_REG_SUB_PATH);

    if ( !NT_SUCCESS( status))
    {
        return status;
    }

    return STATUS_SUCCESS;
}

/****************************************************************************
****************************************************************************/
NTSTATUS
GlobalObjectsInit( PGLOBAL_OBJECTS pThis, PUNICODE_STRING pusDriverRegPath)
{
    WatchInit( &pThis->Watch);

    /*
     * First, try to retrieve driver configuration, because it is needed
     * for use by other objects
     */
    ConfigInit( &pThis->Config, pusDriverRegPath);
    if ( !ConfigIsValid( &pThis->Config))
    {
        return STATUS_DEVICE_CONFIGURATION_ERROR;
    }

    /*
     * Init. slot manager - it allocates memory, so can fail
     */
    SlotmanInit( &pThis->SlotMan, &pThis->Config);
    if ( !SlotmanIsValid( &pThis->SlotMan))
    {
        return STATUS_INSUFFICIENT_RESOURCES;
    }

    return STATUS_SUCCESS;
}

/****************************************************************************
****************************************************************************/
void
GlobalObjectsDestructor( PGLOBAL_OBJECTS pThis, PDRIVER_OBJECT DriverObject)
{
    NTSTATUS    status;
    ULONG       i;

    /*
     * Destruct all existing PROC_BASEs and free their memory
     */
    for ( i = 0; i < MAXIMUM_DEVICES_SUPPORTED; i++)
    {
        if ( pThis->pProcBase[ i] != NULL)
        {
            /* Customize reg.path for this processor
             */
            ReplaceOrderNum( &pThis->usProcRegPathCopyX, (USHORT) i);

            status = ProcBaseReleaseHardware( pThis->pProcBase[ i],
                                              &pThis->usProcRegPathCopyX,
                                              DriverObject);
            ASSERT( NT_SUCCESS( status));
            ProcBaseDestructor( pThis->pProcBase[ i]);
            MemFreePool( pThis->pProcBase[ i]);
            pThis->pProcBase[ i] = NULL;
        }
    }

    /*
     * Delete link device
     */
    if ( pThis->pLnkDev != NULL)
    {
        IoDeleteDevice( pThis->pLnkDev);
        pThis->pLnkDev = NULL;
    }

    if ( pThis->usProcRegPathCopyX.Buffer != NULL)
    {
        MemFreePool( pThis->usProcRegPathCopyX.Buffer);
    }

    SlotmanDestructor( &pThis->SlotMan);
    DispatcherDestructor( &pThis->Dispatcher);
    ConfigDestructor( &pThis->Config);
    WatchDestructor( &pThis->Watch);
}
