/**********************************************************************

    Module: proc-dma.h

        Product: kernel mode driver for "PowerCrypt" encryption
        acceleration board based on Hi/fn 7751 chip
        by Ashley Laurent, Inc.

    ---------------------------------------------------------------

    Copyright:

        Ashley Laurent, Inc., 1999
        All Rights Reserved.

    ---------------------------------------------------------------

    Functional Description:

        Layout and bits of the 7751 DMA unit control registers.

        The regs are mapped into virtual memory, so they are defined as array.
        Not all significant bits are defined, only those used by this driver.

    ---------------------------------------------------------------

    Environment:

        NT or BSD kernel mode

    ---------------------------------------------------------------

    Author:

        Stan Shkolnyy

    ---------------------------------------------------------------

    Revision history:

        *   15-Feb-99   initial revision

**********************************************************************/

#ifndef __PROC_DMA_H__
#define __PROC_DMA_H__

/****************************************************************************
 * 7751 DMA memory mapped registers layout.
 */
typedef struct _DMA_REGISTERS {

    ULONG   REG_ARRAY00[ 0x03]; // 00-0B - not defined by hardware
    ULONG   COM_RING_ADR;       // 0C-0F
    ULONG   REG_ARRAY10[ 0x03]; // 10-1B - not defined by hardware
    ULONG   SRC_RING_ADR;       // 1C-1F
    ULONG   REG_ARRAY20[ 0x03]; // 20-2B - not defined by hardware
    ULONG   RLT_RING_ADR;       // 2C-2F
    ULONG   REG_ARRAY30[ 0x03]; // 30-3B - not defined by hardware
    ULONG   DST_RING_ADR;       // 3C-3F
    ULONG   STATUS_AND_CTL;     // 40-43
    ULONG   INT_ENABLE;         // 44-47
    ULONG   CONFIG;             // 48-4B
    ULONG   REG_ARRAY4C[ 0x13]; // 4C-97 - not defined by hardware
    ULONG   REV_ID;             // 98-9B
    ULONG   REG_ARRAY9C[ 0x16]; // 9C-F3 - not defined by hardware
    ULONG   START_SIGNATURE;    // F4-F7
    ULONG   REG_F8;             // F8-FB - not defined by hardware
    ULONG   UNLOCK_SIGNATURE;   // FC-FF

} DMA_REGISTERS, *PDMA_REGISTERS;

/****************************************************************************
 * DMA Configuration reg bits
 */

#define D_CONF_MASK_MASTER_RESET    0x00000001L
#define D_CONF_MASK_DMA_RESET       0x00000002L
#define D_CONF_CPU_CONTROLS_LAST    0x00000010L
#define D_CONF_UNLOCK               0x00000800L

#define D_CONF_POLLFREQ_MASK        0x00ff0000L
#define D_CONF_VAL_TO_POLLFREQ_BITS( val)  ( (val << 16) & D_CONF_POLLFREQ_MASK)

#define D_CONF_POLLSCALAR_MASK      0x00000700L
#define D_CONF_VAL_TO_POLLSCALAR_BITS( val)  ( (val << 8) & D_CONF_POLLSCALAR_MASK)

#define D_CONF_POLLFREQ_BITS_TO_VAL( reg)  ( (reg & D_CONF_POLLFREQ_MASK) >> 16)

/* Fixed DMA unit configuration that is set once at init.time
 *
 * Polling frequency (bits 16..23) = 08
 * Invalid poll scalar (bits 8..10) = 3
 * "Last" descriptor is determined by the CPU (bit 4 is set)
 * And bit 2 must be set
 */
#define D_CONF_CONST_CONFIG     ( 0x00000004L |                 \
                                  D_CONF_CPU_CONTROLS_LAST |    \
                                  D_CONF_MASK_MASTER_RESET |    \
                                  D_CONF_MASK_DMA_RESET)

#define D_INT_ABORTS_BITS      0x20202020L
#define D_INT_PU_BIT           0x00000001L

#define D_INT_COMRING_LAST_BIT 0x00000008L
#define D_INT_SRCRING_LAST_BIT 0x00000800L
#define D_INT_RLTRING_LAST_BIT 0x00080000L
#define D_INT_DSTRING_LAST_BIT 0x08000000L

#define D_INT_SRCRING_DONE_BIT 0x00001000L
#define D_INT_DSTRING_DONE_BIT 0x10000000L

#define D_CONF_DISAB_SCHED     0x00000080L  //WORKWORK - what is this???

/*
 * Fixed DMA unit interrupt configuration that is set once at init.time
 */
#ifdef _NO_DST

#ifdef HW_BUG_SEARCH_BUILD

    #define D_INT_CONST_INTS    (   D_INT_ABORTS_BITS | \
                                    D_INT_RLTRING_LAST_BIT | \
                                    D_INT_PU_BIT)
#else
    #define D_INT_CONST_INTS    (   D_INT_ABORTS_BITS | \
                                    D_INT_PU_BIT)

#endif

#else
    #define D_INT_CONST_INTS    (   D_INT_ABORTS_BITS | \
                                    D_INT_SRCRING_DONE_BIT | \
                                    D_INT_DSTRING_DONE_BIT | \
                                    D_INT_COMRING_LAST_BIT | \
                                    D_INT_RLTRING_LAST_BIT | \
                                    D_INT_PU_BIT)
#endif

/****************************************************************************
 * DMA Status and Control reg bits
 */

/*
 * Enable/disable DMA engines to poll descriptors
 * (The 0x40404040 is observed in the reg after the board is powered on)
 */
#define D_STATCTL_DISABLE_ALL_DMA   0x40404040
#define D_STATCTL_ENABLE_ALL_DMA    0x80808080

#endif
