/**********************************************************************

    Module: proc_req.c

        Product: kernel mode driver for "PowerCrypt" encryption
        acceleration board based on Hi/fn 7751 chip
        by Ashley Laurent, Inc.

    ---------------------------------------------------------------

    Copyright:

        Ashley Laurent, Inc., 1999
        All Rights Reserved.

    ---------------------------------------------------------------

    Functional Description:

        Implementation of the PROCESSOR object that performs all the
        work of the 7751 programming.

    ---------------------------------------------------------------

    Environment:

        NT or BSD kernel mode

    ---------------------------------------------------------------

    Author:

        Stan Shkolnyy

    ---------------------------------------------------------------

    Revision history:

        *   15-Feb-99   initial revision

**********************************************************************/

#include "osindependence.h"
#include "debug.h"
#include "globobj.h"
#include "processor.h"
#include "procbase.h"
#include "kmi_private.h"    // for KmInterfaceCompleteCalcRequest()
#include "br_comm_private.h"      // for BriefCommUpdateFullCommand()

/****************************************************************************
ProcessorEnqueueRequest()

Abstract:
    Adds a request to the PROCESSOR's internal queue.
    Doesn't start the processing of the request

Remarks:
    Called by the DISPATCHER and with its spinlock held

****************************************************************************/
void
ProcessorEnqueueRequest(
    IN PPROCESSOR                               pThis,
    IN PIOCTL_PWRCRYPT_CALCULATE_INPUT_OUTPUT   pCalcReq
    )
{
    PPROC_QUEUE_ENTRY pQEntry;

    ASSERT( pThis->Type == PROCESSOR_SIGN);

    /*
     * Acquire the spinlock, because we are going to put a new request
     * into the queue and change some indexes that indicate the queue state
     * and can be changed by other routines as well.
     *
     * Note! The agent that guarantees that our queue has free place is
     * the DISPATCHER, which controls nReqsInQueue and knows PROCESSOR_QUEUE_LENGTH.
     * So, the following code _assumes_ that there is a free place.
     */
    PCrptSpinLockAcquire( &pThis->SpinLock);

    DebugPrint((2, "\n\nProcessorQueueRequest 0x%08x\n", pCalcReq));

    /*
     * The DISPATCHER always increments the counter _before_ calling this func.
     */
    ASSERT( pThis->nReqsInQueue > 0 && pThis->nReqsInQueue <= PROCESSOR_QUEUE_LENGTH);
    CheckQueueIndexes( pThis);

    /*
     * By incrementing the newest entry index, point to the first free entry
     */
    IncrementQueueIndex( pThis->iNewestQueueEntry);
    pQEntry = &pThis->Queue[ pThis->iNewestQueueEntry];

    /*
     * Some enumerations need the counter of queued requests
     */
    pThis->BusyEntryCnt++;

    DebugPrint((2, "    queueing to entry %d\n", pThis->iNewestQueueEntry));

    /*
     * Remember calc.req. ptr in this entry.
     * For this func, it is enough to just remember the ptr.
     * The starting of the request processing and the rest of work is done
     * in ProcessorStartQueuedRequest() when it will see bStarted = FALSE
     */
    pQEntry->pCalcReq = pCalcReq;
    pQEntry->bStarted = FALSE;

    /*
     * They become !=0 only when the request is started and are decremented
     * as the DMA transfers are progressing
     */
    ASSERT( pQEntry->XferProgress[ SRC_XFER].DataRemainCnt == 0);
    ASSERT( pQEntry->XferProgress[ DST_XFER].DataRemainCnt == 0);

    /*
     * Bump statistics counter
     */
    pThis->nReqsTotalReceived++;

    /*
     * Release the spinlock, because we've completed queue modification
     */
    PCrptSpinLockRelease( &pThis->SpinLock);
}

/****************************************************************************
ProcessorStartQueuedRequest()

Abstract:
    Picks the oldest pending request in the PROCESSOR's internal queue and
    starts its processing. The pending requests are those added earlier by the
    ProcessorEnqueueRequest().

    - initialize the rest of the target queue entry fields that are initialized
      by the calc.request data

    - copy Command and Context into DMA memory, update some "new key" fields
      in the Command copy in that memory

    - set "size" field in the target Command ring descriptor and start the
      Command transfer.

    - start the SrcData transfer, if possible, or arrange for it to start
      as soon as the current transfer completes.

    - start the DstData transfer, if possible, or arrange for it to start
      as soon as the current transfer completes.

    - set "size" field in the target Result ring descriptor and start the
      Result transfer.

****************************************************************************/
void
ProcessorStartQueuedRequest(
    IN PPROCESSOR   pThis
    )
{
    PPROC_QUEUE_ENTRY                  pQEntry;
    PIOCTL_PWRCRYPT_CALCULATE_INPUT_OUTPUT  pCalcReq;
    PPWRCRYPT_CALCULATE_KM_WORKAREA         pWA;

    PUCHAR  pContextPtr;
    USHORT  nContextSize;

    UCHAR   i;

    ASSERT( pThis->Type == PROCESSOR_SIGN);

#ifdef _NT_DRIVER_BUILD_
    WatchRecalc( &glob.Watch);
#endif

    /*-----------------------------------------------------------------------
     * Find the first non-started request and mark it as started. We do this
     * under the spinlock, because we are going to walk the queue and modify
     * bStarted flag in a queue entry.
     *
     * Note, that the loop is guaranteed to finish (to find !bStarted), because
     * the DISPATCHER calls this func exactly once for each request that it
     * enqueues, and each invocation starts exactly 1 request.
     */
    PCrptSpinLockAcquire( &pThis->SpinLock);

    DebugPrint((2, "\n\nProcessorStartQueuedRequest\n"));

    /*
     * The DISPATCHER always has this counter incremented _before_ calling this
     * func. This function can't be called if there are no requests in the queue
     */
    ASSERT( pThis->nReqsInQueue > 0 && pThis->nReqsInQueue <= PROCESSOR_QUEUE_LENGTH);
    CheckQueueIndexes( pThis);

    i = pThis->iOldestQueueEntry;
    pQEntry = NULL;
    do
    {
        if ( !pThis->Queue[ i].bStarted)
        {
            DebugPrint((2, "    checking queue entry %d - starting\n", i));
            pQEntry = &pThis->Queue[ i];
        }
        else
        {
            DebugPrint((2, "    checking queue entry %d - skipping\n", i));
            IncrementQueueIndex( i);
        }
    }
    while ( pQEntry == NULL);

    /*
     * By the found queue entry that points to the request to start, setup ptrs to:
     *  - the request itself,
     *  - the KM work area in the request
     */
    pCalcReq    = pQEntry->pCalcReq;
    pWA         = (PPWRCRYPT_CALCULATE_KM_WORKAREA) &pCalcReq->KMWorkArea;

    DebugPrint((2, "    starting 0x%08x\n", pCalcReq));

    /*
     * initialize the rest of the target queue entry fields that are initialized
     * by the calc.request data
     */
    pQEntry->XferProgress[ SRC_XFER].pFullMdl        = (PMEM_DESC_LIST) pCalcReq->pSrcData;
    pQEntry->XferProgress[ SRC_XFER].DataRemainCnt   = CNT_FROM_MDL( pQEntry->XferProgress[ SRC_XFER].pFullMdl);
    pQEntry->XferProgress[ SRC_XFER].DataCurrVa      = VIRT_ADDR_FROM_MDL( pQEntry->XferProgress[ SRC_XFER].pFullMdl);

    pQEntry->XferProgress[ DST_XFER].pFullMdl        = (PMEM_DESC_LIST) pCalcReq->pDstBuffer;
    pQEntry->XferProgress[ DST_XFER].DataRemainCnt   = CNT_FROM_MDL( pQEntry->XferProgress[ DST_XFER].pFullMdl);
    pQEntry->XferProgress[ DST_XFER].DataCurrVa      = VIRT_ADDR_FROM_MDL( pQEntry->XferProgress[ DST_XFER].pFullMdl);

    #ifndef _SKIP_MODE
    #ifdef DBG
    {
        int j;
        for ( j = 0; j < sizeof(BIGGEST_COMMAND); j++)
        {
            ASSERT( pQEntry->pCommand[ j] == 0x33);
        }
        for ( j = 0; j < sizeof(BIGGEST_RESULT); j++)
        {
            ASSERT( pQEntry->pResult[ j] == 0x77);
        }
    }
    #endif
    #endif

    /*-----------------------------------------------------------------------
     * set "size" field in the target Result ring descriptor and enable the
     * Result transfer by setting the VALID bit.
     *
     * As experiments show, this is Ok to set size of biggest result in the
     * result descriptor. The 7751 changes this value to the actual size of
     * the generated Result. Quite convenient!
     *
     * Note also, that though we set the VALID bit in the Result descriptor,
     * this causes nop action, since we have not set the VALID bit in the
     * Command descriptor yet. The 7751 won't start execution of the
     * instruction until the Command descriptor is activated
     */
    SgDescSetDataSize( pQEntry->pResultDesc, sizeof( BIGGEST_RESULT));
#ifndef _SKIP_MODE
    SgDescAddFlags( pQEntry->pResultDesc, SG_DESC_FLAG_VALID | SG_DESC_FLAG_LAST);
#endif

    /*
     * We've loaded the descriptor, validate the "jump" descriptor at the end
     * of this ring, to ensure that the 7751 won't stop. Actually, this is only
     * needed if the current descriptor is last but one in the ring but it would
     * require additional unnecessary calculations here and everything we need
     * is to make sure the jump is always "valid".
     */
    RingValidateJump( pThis->pRltRing);

    /*
     * Mark it as "started". This does two things:
     *
     * - protects it from being started again by another invocation of this func.
     *
     * - indicates for ProcTryToCompleteRequests that this req. was started
     *   and thus should be considered as a candidate for completion - examined
     *   for having the VALID bit in its Result descriptor cleared and possibly
     *   completed, if the bit is cleared. However, the bit can't be cleared
     *   until the code below prepares and starts the Command, and the 7751
     *   executes the instruction and enters the Result phase. Until that time,
     *   this QEntry will remain busy.
     */
    pQEntry->bStarted = TRUE;

    /*-----------------------------------------------------------------------
     * Finally, after the entry has been fully filled, release the spinlock
     * and let races to start :-) From now, the entry may be involved in DMA
     * mapping for source or destination data buffers at any time, if some
     * currently active mapping finishes. Non-zero DataRemainCnt vars will be
     * seen by the ProcTryToStartDMATransfer and it will attempt to start a DMA
     * transfer for these data.
     *
     * Meantime, we continue to initialize the Command and the Command SG
     * descriptor. Without this, even if the SrcData and DstData transfers are
     * mapped, they won't start until the Command is read, so won't the Result.
     */
#ifndef _SKIP_MODE
    PCrptSpinLockRelease( &pThis->SpinLock);
#endif

    /*-----------------------------------------------------------------------
     * copy Command and Context into DMA memory, update some "new key" fields
     * in the Command copy in that memory
     *
     * set "size" field in the target Command ring descriptor and start the
     * Command transfer.
     */

    /*
     * Copy Command into DMA memory and update it by "brief command".
     * (The update may set some "new key" flags)
     */
    DebugPrint((2, "copying 0x%04x bytes of Command to comm.buf memory\n", pWA->nCommandSize));
        
    PCrptCopyMemory( pQEntry->pCommand, &pCalcReq->Command, pWA->nCommandSize);
    BriefCommandUpdateFullCommand( &pWA->BriefComm, ( PBASE_COMMAND) pQEntry->pCommand);

    /*
     * Copy context data (if any) next to the Command.
     *
     * (Remember, the pQEntry->pCommand points to BIGGEST_COMMAND-sized area.
     * BIGGEST_COMMAND includes the biggest Context as well, so there can be no
     * problems with space).
     *
     * The key order in the Context: MacKey, EncKey, EncIV
     */
    pContextPtr = pQEntry->pCommand + pWA->nCommandSize;
    nContextSize = 0;

    /*
     * Check newflags that reflect which keys are to be sent and copy
     * those keys
     */
    if ( ( pWA->BriefComm.newflags & MAC_ENGAGED) &&
         ( pWA->BriefComm.newflags & MAC_NEW_CTX))
    {
        DebugPrint((2, "%x bytes of MAC key added\n", pWA->nMACKeySize));
        PCrptCopyMemory( pContextPtr, &pWA->pMACKey[0], pWA->nMACKeySize);
        pContextPtr += pWA->nMACKeySize;
        nContextSize += pWA->nMACKeySize;
    }

    if ( pWA->BriefComm.newflags & CRYP_ENGAGED)
    {
        if ( pWA->BriefComm.newflags & CRYP_NEW_CTX)
        {
            DebugPrint((2, "%x bytes of Enc key added\n", pWA->nEncKeySize));
            PCrptCopyMemory( pContextPtr, &pWA->pEncKey[0], pWA->nEncKeySize);
            pContextPtr += pWA->nEncKeySize;
            nContextSize += pWA->nEncKeySize;
        }

        /*
         * Enc.IV is always sent, if present
         */
        if ( pWA->pEncIV != NULL)
        {
            ASSERT( _7751_3DES_IV_SIZE == _7751_DES_IV_SIZE);
            ASSERT( pWA->nEncIVSize == _7751_DES_IV_SIZE);
    
            DebugPrint((2, "%x bytes of Enc IV added\n", pWA->nEncIVSize));
            PCrptCopyMemory( pContextPtr, &pWA->pEncIV[0], pWA->nEncIVSize);
            pContextPtr += pWA->nEncIVSize;
            nContextSize += pWA->nEncIVSize;
        }
    }

    //
    // BUGBUG - delete this
    //
//    pThis->IntsState.DUInts = D_INT_RLTRING_LAST_BIT;
//    PCrptSpinLockRelease( &pThis->SpinLock);
//    ProcBaseRequestInt( pThis->pProcBase);
//    return;

    /*-----------------------------------------------------------------------
     * Finally, set the size of Command+Context in the Command descriptor
     * and "valid" bit - the 7751 starts reading the command pointed to by the
     * descriptor.
     *
     * (It's interesting to try to put different values of the comm.size into
     * the descriptor and watch how it behaves in terms of switching the proc.
     * unit states during command reading - just a good place to do this :-)
     */
    SgDescSetDataSize( pQEntry->pCommandDesc, pWA->nCommandSize + nContextSize);

#ifndef _SKIP_MODE
    SgDescAddFlags( pQEntry->pCommandDesc, SG_DESC_FLAG_VALID | SG_DESC_FLAG_LAST);
#endif

    DebugPrint((2, "    after setting VALID in ComDesc\n"));
    #ifdef DBG
        DmaDumpRegs( pThis);
    #endif

    /*
     * We've loaded the descriptor, validate the "jump" descriptor at the end
     * of this ring, to ensure that the 7751 won't stop. Actually, this is only
     * needed if the current descriptor is last but one in the ring but it would
     * require additional unnecessary calculations here and everything we need
     * is to make sure the jump is always "valid".
     */
    RingValidateJump( pThis->pComRing);

#ifdef _SKIP_MODE
    PCrptSpinLockRelease( &pThis->SpinLock);
#endif

    /*-----------------------------------------------------------------------
     * Start the SrcData and DstData transfers, if possible, or arrange for
     * them to start as soon as the current transfer completes. The "start"
     * includes loading SG descriptors with data addresses and setting VALID
     * bits, so that the 7751 begins data transfers.
     */
    ProcTryToStartDMATransfer( pThis, SRC_XFER);

    #ifdef DBG
        DebugPrint((2, "    after starting SRC xfer\n"));
        DmaDumpRegs( pThis);
    #endif

    ProcTryToStartDMATransfer( pThis, DST_XFER);

    #ifdef DBG
        DebugPrint((2, "    after starting DST xfer\n"));
        DmaDumpRegs( pThis);
    #endif

    #ifdef _NT_DRIVER_BUILD_

        #ifdef HW_BUG_SEARCH_BUILD

            WatchCheck( &glob.Watch);

            DebugPrint((2, "    waiting for result\n"));
            while ( SgDescGetFlags( pQEntry->pResultDesc) & SG_DESC_FLAG_VALID)
            {
            }
            //
            // generate a fake "int" as so the data transfer has been finished
            //
            pThis->IntsState.DUInts = D_INT_RLTRING_LAST_BIT | D_INT_SRCRING_DONE_BIT | D_INT_DSTRING_DONE_BIT;
            ProcBaseRequestInt( pThis->pProcBase);
        #endif

        #ifdef _SKIP_MODE
            //
            // generate a fake "int" as so the data transfer has been finished
            //
            pThis->IntsState.DUInts = D_INT_RLTRING_LAST_BIT | D_INT_SRCRING_DONE_BIT | D_INT_DSTRING_DONE_BIT;
            ProcBaseRequestInt( pThis->pProcBase);
        #endif
    #endif
}

/****************************************************************************
ProcessorConfirmAndRegisterInterrupt()

NT:         called from ISR, executes under IRQ handler conditions
FreeBSD:    called from ISR, executes under IRQ handler conditions

Abstract:
    Confirms/disables/register active interrupts on the 7751.

Return:
    TRUE, if it is our device that interrupted

Remarks:
    Invocations of this procedure are serialized, so it owns the PROCESSOR
    exclusively.
    By returning TRUE, the func induces calling ProcServRegInts()
    at some time later. There is no garantee, however, for one-to-one
    correspondence between each returned TRUE and ServRegInts(). It may
    happen, that ServRegInts() will be called only once for several returned
    TRUEs. It is guaranteed, however, that as soon as this func returns
    TRUE, ServRegInts() will be eventually called, and it may be called
    at _any_ time.

****************************************************************************/
BOOLEAN
ProcessorConfirmAndRegisterInterrupt(
    IN PPROCESSOR   pThis
    )
{
    ULONG   DUStatus, DUInts;
    ULONG   PUStatus, PUInts;

    DebugPrint((3, "\n\n**** Interrupt\n"));
    ASSERT( pThis->Type == PROCESSOR_SIGN);

    /*
     * The set bits in the DMA status and control reg indicate some conditions -
     * compare the bits with interrupt enable bits that we set when we
     * were programming the DMA Unit, to see whether some bits match -
     * those are our interrupts that we need to remember and must disable.
     */
    DUStatus = READ_DMA32( STATUS_AND_CTL);
    DUInts = ( DUStatus & pThis->DUEnabledInts);

    if ( DUInts == 0)
    {
        /*
         * This is not our interrupt
         */
        return FALSE;
    }

    /*
     * Yes, that may be our chip what has caused this interrupt - it shows
     * some interrupting conditions. Acknowledge and disable the detected ints.
     */
    DebugPrint((3, "****           DMA Unit status=0x%08x ints=0x%08x\n", DUStatus, DUInts));
    #ifdef DBG
        PuDumpRegs( pThis);
        DmaDumpRegs( pThis);
        ProcDumpXferStatus( pThis, SRC_XFER);
        ProcDumpXferStatus( pThis, DST_XFER);
    #endif

    /*
     * First, check for proc.unit interrupts and acknowledge them.
     * They go outside through a DMA STATUS_AND_CTL register bit, so we
     * first acknowledge them in the proc.unit and then (after they have
     * been deactivated) - acknowledge the bit in the DMA STATUS_AND_CTL.
     *
     * To acknowledge the proc.unit ints, write 1 into the corresponding bits
     * in the INT_STATUS - then they become cleared. So, we write back
     * exactly the same bits that we've detected as the cause of the interrupt.
     */
    if ( DUInts & D_INT_PU_BIT)
    {
        /*
         * Detect which proc.unit interrupt bits have caused the interrupt
         */
        PUStatus = READ_PROC32( INT_STATUS);
        PUInts = ( PUStatus & pThis->PUEnabledInts);

        DebugPrint((3, "****           ProcUnit int.status=0x%08x ints=0x%08x\n", PUStatus, PUInts));
        ASSERT(0);

        /*
         * Write them back into the same reg
         */
        WRITE_PROC32( INT_STATUS, PUInts);

        pThis->IntsState.PUInts |= PUInts;
    }

    /*
     * Second, acknowledge DMA unit interrupts.
     * To acknowledge the DMA unit ints, write 1 into the corresponding bits
     * in the STATUS_AND_CTL - then they become cleared. So, we write back
     * exactly the same bits that we detected as the cause of the interrupt.
     */
    WRITE_DMA32( STATUS_AND_CTL, DUInts);

    /*
     * This is an important place - we need to remember what we've found.
     * We register interrupts by simple accumulation of the interrupt bits.
     * When ProcServRegInts() gets executed, it knows to take and use the
     * accumulated bits and to clear them. Note, that if two identical
     * ints happen before a ProcServRegInts() has have a chance to work, the
     * corresponding bit will be set but it's not known how many ints occurred,
     * so the ProcServRegInts() can't use the int bits to determine the number
     * of ints, only the _kind_ of ints that have been triggered.
     */
    pThis->IntsState.DUInts |= DUInts;

#ifdef _NO_DST
    //
    // This could be needed if the 7751 wouldn't walk the dst ring if the dst data
    // output is disabled. However, it do walk the desc.array as usual and do
    // generate the DSTRING_DONE int, so, it is not necessary.
    //
    //if ( DUInts & D_INT_SRCRING_DONE_BIT)
    //{
    //    pThis->IntsState.DUInts |= D_INT_DSTRING_DONE_BIT;
    //}

    //
    // Since we have src done and dst done ints disabled, artifically set
    // these bits on "result last" interrupt
    //
    if ( DUInts & D_INT_RLTRING_LAST_BIT)
    {
        pThis->IntsState.DUInts |= ( D_INT_SRCRING_DONE_BIT | D_INT_DSTRING_DONE_BIT);
    }
#endif

    return TRUE;
}

/****************************************************************************

ProcIntSyncRoutine()

    This routine is used to safely read and reset the PROCESSOR's
    registry of detected interrupts. This registry is filled by the
    ProcessorConfirmAndRegisterInterrupt() called from the 7751 ISR.

    Calls to this routine are arranged by the
    ProcessorServiceRegisteredInterrupts() to read and reset that registry.

    !!! During execution of this routine the ISR is disabled !!!

    The routine receives a ptr to an INT_STATE_XFER_BLOCK as a parameter.
    This block contains a ptr to the PROCESSOR which int.registry is to be
    copied and reset, and an area where that registry is to be copied.

****************************************************************************/

typedef struct _INT_STATE_XFER_BLOCK
{
    IN  PPROCESSOR              pProc;
    OUT PROCESSOR_INTS_STATE    IntsSave;

} INT_STATE_XFER_BLOCK, *PINT_STATE_XFER_BLOCK;

static
BOOLEAN
ProcIntSyncRoutine
    (
        IN PVOID SynchronizeContext
    )
{
    PINT_STATE_XFER_BLOCK pIntXferBlk;

    DebugPrint((3, "ProcIntSyncRoutine\n"));

    pIntXferBlk = (PINT_STATE_XFER_BLOCK) SynchronizeContext;
    ASSERT( pIntXferBlk->pProc->Type == PROCESSOR_SIGN);

    pIntXferBlk->IntsSave = pIntXferBlk->pProc->IntsState;
    IntsStateReset( pIntXferBlk->pProc);
    return TRUE;
}
 
/****************************************************************************
ProcessorServiceRegisteredInterrupts()

NT:         called from DPC handler, executes at DISPATCH_LEVEL
FreeBSD:    called from ISR, executes under IRQ handler conditions

****************************************************************************/
static
void
ProcRecoverFromFatalError(
    IN PPROCESSOR   pThis,
    IN ULONG        DUInts,
    IN ULONG        PUInts
    );

void
ProcessorServiceRegisteredInterrupts(
    IN PPROCESSOR   pThis
    )
{
    INT_STATE_XFER_BLOCK IntXferBlk;

    ULONG       DUInts, PUInts;
    LONG        SrcMapIntCnt, DstMapIntCnt;
    BOOLEAN     bFatalErrorMode;

    DebugPrint((3, "\n"));
    DebugPrint((3, "ProcessorServiceRegisteredInts\n"));
    ASSERT( pThis->Type == PROCESSOR_SIGN);

    /*
     * Read and reset the PROCESSOR's interrupt registry which is filled in
     * by the ProcessorConfirmAndRegisterInterrupt(). As this routine does
     * this, it takes the responsibility to react to the interrupt bits it
     * has discovered in the registry. Withdrawal of the registry is
     * synchronized amongs the invocations of this routine by acquisition of
     * the interrupt spinlock in the ProcBaseSynchronizeExecution().
     * The function returns BOOLEAN which is always TRUE as returned by the
     * ProcIntSyncRoutine().
     */
    IntXferBlk.pProc = pThis;
    ProcBaseSynchronizeExecution( pThis->pProcBase,
                                  ProcIntSyncRoutine,
                                  &IntXferBlk);

    DUInts       = IntXferBlk.IntsSave.DUInts;
    PUInts       = IntXferBlk.IntsSave.PUInts;
    SrcMapIntCnt = IntXferBlk.IntsSave.MapIntCnt[ SRC_XFER];
    DstMapIntCnt = IntXferBlk.IntsSave.MapIntCnt[ DST_XFER];

    /*
     * Due to the policy explained above, it may happen that there is no work
     * for this routine. It was scheduled, true, but another instance of this
     * routine's execution has already took the responsibility for the job.
     */
    if ( DUInts == 0 && PUInts == 0 && SrcMapIntCnt == 0 && DstMapIntCnt == 0)
    {
        DebugPrint((3, "    No work found\n"));
        return;
    }

    /*
     * On a fatal error, the DispatcherDeactivateProcessor() causes
     * the DISPATCHER to temporarily forget about this PROCESSOR, until the
     * reverse routine is called. This guaranties that, to the function bottom,
     * the fatal request becomes the "oldest" request in the PROCESSOR's queue
     * because any "good" older requests will be completed in the
     * ProcTryToCompleteRequests().
     * Zero or more valid requests may still remain pending in the queue,
     * however.
     *
     * At the other hand, (bFatalErrorMode == TRUE) causes this function's
     * code below not to initiate any new DMA transfers but only complete
     * those that have finished to this time.
     */
    if ( ( DUInts & D_INT_ABORTS_BITS) ||
         ( PUInts & ( P_INT_DATA_ERROR | P_INT_INVAL_COMM)))
    {
        DebugPrint((3, "7751 fatal error\n"));
        DispatcherDeactivateProcessor( pThis->pDispatcher, pThis);
        bFatalErrorMode = TRUE;
    }
    else
    {
        bFatalErrorMode = TRUE;
    }

    #if DBG
        if ( DUInts & D_INT_COMRING_LAST_BIT)
        {
            /*
             * The 7751 has finished reading the Command
             */
            DebugPrint((3, "LAST in Command ring - do nothing\n"));
        }
    #endif

    /*
     * On any of data ring ints - complete the finished transfer.
     * Also, if necessary, try to map more data
     *
     * XxxMapIntCnt is caused by MapControl routine to indicate that it has
     * freed the adapter channel, so that a new mapping can be attempted.
     */
    if ( DUInts & D_INT_SRCRING_DONE_BIT)
    {
        DebugPrint((3, "DONE in SrcData ring\n"));
        ProcTryToCompleteDMATransfers( pThis, SRC_XFER);
    }

    if ( ( SrcMapIntCnt != 0 || ( DUInts & D_INT_SRCRING_DONE_BIT)) &&
         !bFatalErrorMode)
    {
        DebugPrint((3, "0x%04x delayed Src mappings\n", SrcMapIntCnt));
        ProcTryToStartDMATransfer( pThis, SRC_XFER);
    }

    if ( DUInts & D_INT_DSTRING_DONE_BIT)
    {
        DebugPrint((3, "DONE in DstData ring\n"));
        ProcTryToCompleteDMATransfers( pThis, DST_XFER);
    }

    if ( ( DstMapIntCnt != 0 || ( DUInts & D_INT_DSTRING_DONE_BIT)) &&
         !bFatalErrorMode)
    {
        DebugPrint((3, "0x%04x delayed Dst mappings\n", DstMapIntCnt));
        ProcTryToStartDMATransfer( pThis, DST_XFER);
    }

    /*
     * On Result ring int - complete the oldest request
     */
    if ( DUInts & D_INT_RLTRING_LAST_BIT)
    {
        DebugPrint((3, "LAST in Result ring - instruction has been completed\n"));
        ProcTryToCompleteRequests( pThis);
    }
}

/****************************************************************************
ProcRecoverFromFatalError()

Aborts the faulty request, restarts the PROCESSOR and good requests.

****************************************************************************/
static
void
ProcRecoverFromFatalError(
    IN PPROCESSOR   pThis,
    IN ULONG        DUInts,
    IN ULONG        PUInts
    )
{
    DebugPrint((3, "ProcRecoverFromFatalError\n"));
    ASSERT( pThis->Type == PROCESSOR_SIGN);
    return;
}

/****************************************************************************
****************************************************************************/
void
ProcTryToCompleteRequests(
    IN PPROCESSOR   pThis
    )
{
    BOOLEAN     bContinue;

    DebugPrint((3, "ProcTryToCompleteRequests\n"));
    ASSERT( pThis->Type == PROCESSOR_SIGN);

    /*
     * Acquire the spinlock, because we are going to remove the completed
     * request from the queue and move an index
     */
    PCrptSpinLockAcquire( &pThis->SpinLock);

    bContinue = TRUE;
    while ( pThis->BusyEntryCnt != 0 && bContinue)
    {
        PPROC_QUEUE_ENTRY                  pQEntry;

        CheckQueueIndexes( pThis);

        /*
         * Setup ptr to the queue entry that contains the oldest request
         */
        pQEntry = &pThis->Queue[ pThis->iOldestQueueEntry];

        /*
         * Check to see if the entry was started (because we could run into a
         * completely fresh entry that has just been added by the DISPATCHER)
         * and the Result descriptor has its VALID bit cleared. If both these
         * conditions hold true, the request being checked has been finished
         * by the 7751 and we can complete it.
         */
        if ( pQEntry->bStarted &&
             !(SgDescGetFlags( pQEntry->pResultDesc) & SG_DESC_FLAG_VALID))
        {
            PIOCTL_PWRCRYPT_CALCULATE_INPUT_OUTPUT  pCompletedCalcReq;
            PPWRCRYPT_CALCULATE_KM_WORKAREA pWA;
            USHORT nResultSize;

            pCompletedCalcReq = pQEntry->pCalcReq;
            pWA = (PPWRCRYPT_CALCULATE_KM_WORKAREA) &pCompletedCalcReq->KMWorkArea;

            /*
             * Copy the generated Result from the DMA memory into the CalcReq
             * and also set its size in the KM workarea member.
             *
             * After the 7751 has finished its work, the Result descriptor
             * contains the size of the generated Result and pResult points
             * to the DMA memory where the actual Result resides.
             */
            nResultSize = SgDescGetDataSize( pQEntry->pResultDesc);
            if ( nResultSize == 0 ||
                 nResultSize > sizeof( BIGGEST_RESULT))
            {
                DebugPrint((1, "Severe error: 7751 has generated 0x%04x bytes of Result, reporting 0\n", nResultSize));
                nResultSize = 0;
            }
            else
            {
                PCrptCopyMemory( &pCompletedCalcReq->Result, pQEntry->pResult, nResultSize);
            }

            /*
             * Report the size of the Result that has been copied into the
             * CalcReq so that the client would know how many bytes in the
             * Result member are to be deemed actual Result.
             */
            pCompletedCalcReq->nResultSize = nResultSize;

#ifdef DBG
            PCrptFillMemory( pQEntry->pCommand, sizeof(BIGGEST_COMMAND), 0x33);
            PCrptFillMemory( pQEntry->pResult, sizeof(BIGGEST_RESULT), 0x77);
#endif

            /*
             * They become !=0 only when the request is started and are decremented
             * as the DMA transfers are progressing. As we've just finished the
             * transfer, they should be 0. However, it may happen, that RESULT_DONE
             * int is triggered before all the user-supplied data is transferred,
             * if a data buffer is too big and the 7751 has decided that the work
             * is over before we've mapped all of the data.
             *
             * BUGBUG - need to handle this somehow
             */
#ifndef _SKIP_MODE
            ASSERT( pQEntry->XferProgress[ SRC_XFER].DataRemainCnt == 0);
            ASSERT( pQEntry->XferProgress[ DST_XFER].DataRemainCnt == 0);
#endif

            /*
             * Sentinel value preventing ProcTryToStartTransfer from thinking
             * that this entry needs DMA mapping, when this entry is not
             * in use. These vars should go to 0 by themselves, if the request
             * submitted is accurate.
             */
            pQEntry->XferProgress[ SRC_XFER].DataRemainCnt = 0;
            pQEntry->XferProgress[ DST_XFER].DataRemainCnt = 0;

            /*
             * By incrementing the oldest entry index, make the next entry oldest
             * and put the completed request entry out of view - it is not in the
             * queue anymore, and the only pointer to it is pCompletedCalcReq in
             * the stack.
             */
            IncrementQueueIndex( pThis->iOldestQueueEntry);

            pThis->BusyEntryCnt --;

            DebugPrint((3, "    completing req=0x%08x\n", pCompletedCalcReq));

            /*
             * Release the spinlock, because we've completed the queue modification
             * and now we need to call the DISPATCHER which may call us back
             * causing us to acquire this same spinlock in another place.
             */
            PCrptSpinLockRelease( &pThis->SpinLock);

            /*
             * Tell the DISPATCHER it can start the next request - this may lead
             * to placing a new request in the processor queue (that is, the
             * DISPATCHER may call the processor's methods that start requests.
             */
            DispatcherStartNextRequest( pThis->pDispatcher, pThis);

            /*
             * Tell the KM interface that it can complete this calc.request
             * by notifying the client, etc.
             */
            KmInterfaceCompleteCalcRequest( pCompletedCalcReq);

#ifdef _NT_DRIVER_BUILD_
            WatchCheck( &glob.Watch);
#endif

            /*
             * Acquire the spinlock, because we are going to continue the loop
             * and check to see whether we can complete another request.
             */
            PCrptSpinLockAcquire( &pThis->SpinLock);
        }
        else
        {
            /*
             * The oldest request is not completed yet by the 7751 - terminate
             * the loop since there is nothing we can do here
             */
            bContinue = FALSE;

            //
            // Rare case
            //
//            ASSERT(0);
            DebugPrint((3, "    VALID is still set in req=0x%08x\n", pQEntry->pCalcReq));
        }
    }

    /*
     * Release the spinlock, because we've completed the queue modification
     */
    PCrptSpinLockRelease( &pThis->SpinLock);

}

/****************************************************************************
****************************************************************************/
void
IntsStateReset(
    IN OUT PPROCESSOR   pThis
    )
{
    ASSERT( pThis->Type == PROCESSOR_SIGN);

    pThis->IntsState.DUInts = 0;
    pThis->IntsState.PUInts = 0;
    pThis->IntsState.MapIntCnt[ SRC_XFER] = 0;
    pThis->IntsState.MapIntCnt[ DST_XFER] = 0;
}
