#include "osindependence.h"
#include "debug.h"
#include "myunistr.h"
#include "procbase.h"

#ifdef ALLOC_PRAGMA
#pragma alloc_text(INIT, ProcBaseConstructor)
#pragma alloc_text(INIT, ProcBaseAcquireHardware)
#pragma alloc_text(INIT, ParseResListCopyAssignments)
#pragma alloc_text(INIT, TranslateAndMapAddrs)
#pragma alloc_text(INIT, InstallInterruptHandler)
#endif

/****************************************************************************
Description:
    Initializes important flags needed for proper resource cleanup.
    Calls Processor's ctor
****************************************************************************/
void
ProcBaseConstructor(
    IN OUT PPROCBASE pThis,
    IN PCONFIG       pConfig
    )
{
    DebugPrint((2, "ProcBaseConstructor\n"));

    pThis->Type = PROCBASE_SIGN;

    pThis->pConfig = pConfig;
    pThis->bISRInstalled = FALSE;
    pThis->bProcessorPlugged = FALSE;
    pThis->bRegistryResourcesAssigned = FALSE;
    pThis->RangesCnt = 0;
    pThis->pInterruptObject = NULL;

    KeInitializeDpc( &pThis->DPC,       // IN PKDPC  Dpc
                     DeviceDPC,         // IN PKDEFERRED_ROUTINE DeferredRoutine
                     (PVOID) pThis);    // IN PVOID  DeferredContext

#ifdef _SKIP_MODE
    KeSetImportanceDpc( &pThis->DPC, LowImportance);
#endif

    DmaCtxConstructor( &pThis->SrcDmaCtx, TRUE);    // WriteToDevice
    DmaCtxConstructor( &pThis->DstDmaCtx, FALSE);   // not WriteToDevice

    ProcessorConstructor( &pThis->Proc, pThis, pConfig);
}

/****************************************************************************
Description:
****************************************************************************/
void
ProcBaseDestructor(
    IN PPROCBASE pThis
    )
{
    DebugPrint((2, "ProcBaseDestructor\n"));

    ASSERT( pThis->Type == PROCBASE_SIGN);

    ProcessorDestructor( &pThis->Proc);

    DmaCtxDestructor( &pThis->SrcDmaCtx);
    DmaCtxDestructor( &pThis->DstDmaCtx);

    ASSERT( !pThis->bRegistryResourcesAssigned);
    ASSERT( !pThis->bProcessorPlugged);
    ASSERT( !pThis->bISRInstalled);
}

/****************************************************************************
pDispatcher -
    needed merely to be passed to the PROCESSOR when it is plugged.
****************************************************************************/
NTSTATUS
ProcBaseAcquireHardware(
    IN PPROCBASE        pThis,
    IN PDISPATCHER      pDispatcher,
    IN PUNICODE_STRING  pusProcBaseRegPath,
    IN ULONG            BusNumber,
    IN ULONG            SlotNumber,
    IN PDRIVER_OBJECT   DriverObject,
    IN PUNICODE_STRING  pusDeviceName
    )
{
    NTSTATUS            status;             // func.ret.value
    PCM_RESOURCE_LIST   pAllocResList;

    DebugPrint((2, "ProcBaseAcquireHardware\n"));
    ASSERT( pThis->Type == PROCBASE_SIGN);

    /*
     * Copy processors hardware coordinates
     */
    pThis->BusNumber = BusNumber;
    pThis->SlotNumber = SlotNumber;

    /*
     * They need to know the BusNumber to properly fulfill the
     * DmaCtxInit() eventually called by the PROCESSOR.
     */
    DmaCtxAssignBusNumber( &pThis->SrcDmaCtx, BusNumber);
    DmaCtxAssignBusNumber( &pThis->DstDmaCtx, BusNumber);

    /*
     * Create devices
     */
    {
        UNICODE_STRING      usCurDevName;
        USHORT              NakedNameLen;

        /*
         * Allocate memory and create copy of the device name. It is made longer
         * to accept "src" or "dst" suffix later.
         */
        status = MyCloneUnicodeString( &usCurDevName,
                                       pusDeviceName,
                                       3 * sizeof( USHORT)    // delta
                                       );
        if ( !NT_SUCCESS( status))
        {
            DebugPrint((1, "DMA device name alloc. failed. status=0x%08x\n", status));
            return status;
        }

        /* Remember the naked (without suffix) name string length, to restore it
         * later, when we'll need to add another suffix
         */
        NakedNameLen = usCurDevName.Length;

        /*
         * Create 2 work device objects, needed for src data and dst data DMA
         */
        status = RtlAppendUnicodeToString( &usCurDevName, L"src");
        ASSERT( NT_SUCCESS( status));   // because the string _must_ be long enough
        status = DmaCtxCreateDevice( &pThis->SrcDmaCtx, DriverObject, &usCurDevName);
        if ( !NT_SUCCESS( status))
        {
            MemFreePool( usCurDevName.Buffer);
            return status;
        }

        /* Set the old string length as so we have not added suffix
         */
        usCurDevName.Length = NakedNameLen;

        status = RtlAppendUnicodeToString( &usCurDevName, L"dst");
        ASSERT( NT_SUCCESS( status));   // because the string _must_ be long enough
        status = DmaCtxCreateDevice( &pThis->DstDmaCtx, DriverObject, &usCurDevName);
        if ( !NT_SUCCESS( status))
        {
            MemFreePool( usCurDevName.Buffer);
            return status;
        }

        /*
         * Devices have been created - don't need device name string anymore
         */
        MemFreePool( usCurDevName.Buffer);
    }

    /* Assign resources, write the assignment to the Registry and create
     * resource list that describes allocated resources - I/O or memory
     * ranges and an interrupt (we must free the list storage)
     *
     * NULL as a class name forces using of "OtherDrivers" class.
     *
     * DriverObject is used to find the driver's name - this name is used
     * as the name of the key under "OtherDrivers" key
     *
     * DeviceObject is used to find the device's name - this name is used
     * as the name of the value under the above driver's key.
     * We use "src DMA" device obj only because we need some device obj here,
     * "dst DMA" dev obj could be used with the same success.
     */
    pAllocResList = NULL;
    status = HalAssignSlotResources( pusProcBaseRegPath,
                                     NULL,          // PUNICODE_STRING ClassName
                                     DriverObject,
                                     pThis->SrcDmaCtx.pDevice,
                                     PCIBus,
                                     pThis->BusNumber,
                                     pThis->SlotNumber,
                                     &pAllocResList      // output
                                     );
    if ( NT_SUCCESS( status))
    {
        ASSERT( pAllocResList != NULL);
        ASSERT( pAllocResList->Count == 1);

        pThis->bRegistryResourcesAssigned = TRUE;

        /*
         * Parse the list of allocated resources and copy mem.ranges and
         * interrupt info into the PROCBASE storage for later use
         */
        ParseResListCopyAssignments( pThis, &pAllocResList->List[0]);

        /*
         * We have copied res.list info - must free its storage
         */
        ExFreePool( pAllocResList);

        /* 7751 has 2 ranges of mem-mapped registers
         */
        if ( pThis->RangesCnt == 2)
        {
            status = TranslateAndMapAddrs( pThis);
            if ( NT_SUCCESS( status))
            {
                ASSERT( pThis->MemRanges[0].ResourceMapped);
                ASSERT( pThis->MemRanges[1].ResourceMapped);

                /* Plug the processor into the processor base by letting
                 * it know about DMA memory, hardware registers and the
                 * DISPATCHER.
                 */
                if ( NT_SUCCESS(
                        ProcessorPlug(
                            &pThis->Proc,
                            &pThis->SrcDmaCtx,
                            &pThis->DstDmaCtx,
                            pDispatcher,
                            pThis->MemRanges[1].MappedRangeStart, // dma regs VA
                            pThis->MemRanges[0].MappedRangeStart // proc regs VA
                        )))
                {
                    pThis->bProcessorPlugged = TRUE;

                    /* Now that the PROCESSOR is operational, install ISR
                     */
                    status = InstallInterruptHandler( pThis);
                    if ( NT_SUCCESS( status))
                    {
                        pThis->bISRInstalled = TRUE;
                    }
                }
                else
                {
                    DebugPrint((1, "ProcessorPlug failed\n"));
                    status = STATUS_DEVICE_CONFIGURATION_ERROR;
                }
            }
        }
        else
        {
            DebugPrint((1, "Detected %d memory ranges instead of 2\n", pThis->RangesCnt));
            status = STATUS_DEVICE_CONFIGURATION_ERROR;
        }
    }
    else
    {
        DebugPrint((1, "HalAssignResources failed. Status = %x \n", status));
    }

    /*
     * Tell the dispatcher about this processor only if everything is Ok
     */
    if ( NT_SUCCESS(status))
    {
        DispatcherRegisterProcessor( pDispatcher, &pThis->Proc);
    }

    return status;
}

/****************************************************************************

    This method is always called on driver unloading or
    if ProcBaseAcquireHardware() has failed, to release all the hardware
    resources.

****************************************************************************/
NTSTATUS
ProcBaseReleaseHardware(
    IN PPROCBASE       pThis,
    IN PUNICODE_STRING  pusProcBaseRegPath,
    IN PDRIVER_OBJECT   DriverObject
    )
{
    NTSTATUS status;                    // func.ret.value

    ULONG               nIndex;         // walks mem.ranges array
    PADAPTER_MEM_RANGE  pCurRange;      // walks mem.ranges array

    DebugPrint((2, "ProcBaseReleaseHardware\n"));
    ASSERT( pThis->Type == PROCBASE_SIGN);

    /*
     * Check the resources acquisition flags and return all the resources
     * that must be returned.
     *
     * Some of the resources may not be acquired, if the earlier call
     * to ProcBaseAcquireHardware() has failed - this is why the flags
     * are needed
     */

    status = STATUS_SUCCESS;

    if ( pThis->bISRInstalled)
    {
        IoDisconnectInterrupt( pThis->pInterruptObject);
        pThis->bISRInstalled = FALSE;
    }

    if ( pThis->bProcessorPlugged)
    {
        ProcessorUnplug( &pThis->Proc);
        pThis->bProcessorPlugged = FALSE;
    }

    /* Walk the mem.ranges array and un-map mapped ranges.
     * (MmUnmapIoSpace doesn't have a return value)
     */
    for ( nIndex = 0; nIndex < pThis->RangesCnt; nIndex++)
    {
        pCurRange = &pThis->MemRanges[ nIndex];
        if ( pCurRange->ResourceMapped)
        {
            DebugPrint((3, "    Unmapping memory at 0x%08x\n", pCurRange->MappedRangeStart));
            MmUnmapIoSpace( pCurRange->MappedRangeStart, pCurRange->RangeLength);
        }
    }

    if ( pThis->bRegistryResourcesAssigned)
    {
        /*
         * Free resources claimed in the registry.
         */
        PCM_RESOURCE_LIST   pDummyResList;

        DebugPrint((3, "    Deassigning resources in the Registry\n"));

        /*
         * NULL as a class name forces using of "OtherDrivers" class.
         *
         * DriverObject is used to find the driver's name - this name is used
         * as the name of the key under "OtherDrivers" key
         *
         * DeviceObject is used to find the device's name - this name is used
         * as the name of the value under the above driver's key. For this
         * function, we must use the same dev obj that was used in
         * HalAssignSlotResources to assign these resources.
         */
        pDummyResList = NULL;
        status = IoAssignResources( pusProcBaseRegPath,
                                    NULL,               // PUNICODE_STRING ClassName
                                    DriverObject,
                                    pThis->SrcDmaCtx.pDevice,
                                    NULL,               // RequestedResources
                                    &pDummyResList
                                    );
        ASSERT( pDummyResList == NULL);

        if ( !NT_SUCCESS( status))
        {
            DebugPrint((1, "IoAssignResources failed on release. Status = %x \n", status));
        }

        pThis->bRegistryResourcesAssigned = FALSE;
    }

    return status;
}

/****************************************************************************
Description:

    Translates a bus-specific address into the corresponding system logical
    address. Also, this will map a physical range to nonpaged system space if
    required.

    Most PCI devices respond to either an I/O range or a memory range, but
    not both.  For simplicity, this function simply maps all the resources
    for a given PCI device.  In reality, only one of the ranges might be
    used, so some cleanup may be needed here.

Arguments:

    pThis            - ptr to PROCBASE

****************************************************************************/
NTSTATUS
TranslateAndMapAddrs(
    IN PPROCBASE   pThis
    )
{
    NTSTATUS            status;         // func.ret.value

    ULONG               nIndex;         // walks mem.ranges array
    PADAPTER_MEM_RANGE  pCurRange;      // walks mem.ranges array

    ULONG               AddrSpaceType;
    PHYSICAL_ADDRESS    TranslatedAddr;

    ASSERT( pThis->Type == PROCBASE_SIGN);

    //
    // Loop through each of the raw I/O ports and memory ranges.
    //

    status = STATUS_SUCCESS;

    for ( nIndex = 0;
          nIndex < pThis->RangesCnt && NT_SUCCESS( status);
          nIndex++)
    {
        pCurRange = &pThis->MemRanges[ nIndex];

        //
        // Translate the bus-specific address to a system logical address.
        //

        AddrSpaceType = pCurRange->AddrSpaceType;

        if ( HalTranslateBusAddress( PCIBus,
                                     pThis->BusNumber,
                                     pCurRange->RangeStart,
                                     &AddrSpaceType,
                                     &TranslatedAddr))
        {
            //
            // Save the (possibly changed) I/O-memory indicator.  Remember that
            // zero indicates a memory address, and one indicates I/O space.  On
            // some RISC systems, there is no I/O space.  So an AddressSpace
            // indicating I/O space input to HalTranslateBusAddress may be
            // changed to memory space on return.
            //

            pCurRange->MappedAddrSpaceType = AddrSpaceType;

            //
            // Map the device base address into the virtual address space if the
            // address is in memory space.  Note that HalTranslateBusAddress may
            // have changed the address space indicator on return.
            //

            if ( AddrSpaceType == 0)
            {
                pCurRange->MappedRangeStart = MmMapIoSpace( TranslatedAddr,
                                                            pCurRange->RangeLength,
                                                            FALSE   // means "not cached"
                                                            );
                if ( pCurRange != NULL)
                {
                    DebugPrint((3, "    Memory mapped at 0x%08x\n", pCurRange->MappedRangeStart));
                    pCurRange->ResourceMapped = TRUE;
                }
                else
                {
                    DebugPrint((2, "MmMapIoSpace failed for 0x%x \n", pCurRange->RangeStart));
                    status = STATUS_DEVICE_CONFIGURATION_ERROR;
                }
            }
            else
            {
                pCurRange->MappedRangeStart = (PVOID) TranslatedAddr.LowPart;
            }
        }
        else
        {
            DebugPrint((2, "HalTranslateBusAddress failed for 0x%x \n", pCurRange->RangeStart));
            status = STATUS_DEVICE_CONFIGURATION_ERROR;
        }
    }//for

    return status;
}

/****************************************************************************
Description:

    Given a full resource description, store away the relevant information
    for the adapter in the PROCBASE object.

Arguments:

    pThis            - ptr to PROCBASE

    AdapterResDesc   - The CM_FULL_RESOURCE_DESCRIPTOR list for this adapter.

****************************************************************************/
VOID
ParseResListCopyAssignments(
    IN PPROCBASE                   pThis,
    IN PCM_FULL_RESOURCE_DESCRIPTOR pAdapterResDesc
    )
{
    PADAPTER_MEM_RANGE              pCurMemRange;   // walks our mem.ranges array
    ULONG                           ResDescIndex;   // walks our mem.ranges array

    PCM_PARTIAL_RESOURCE_DESCRIPTOR pCurPartDesc;   // walks passed res.desc.array
    ULONG                           CurRangeIndex;  // walks passed res.desc.array

    BOOLEAN             bInterruptSet;

    DebugPrint((3, "ParseResourceList \n"));
    ASSERT( pThis->Type == PROCBASE_SIGN);

    bInterruptSet = FALSE;

    /* Begin filling mem.range descriptors in the PROCBASE from the first one
     * and increment this index as they are filled
     */
    CurRangeIndex = 0;

    for ( ResDescIndex = 0;
          ResDescIndex < pAdapterResDesc->PartialResourceList.Count;
          ResDescIndex++)
    {
        /* Point to the current partial desc. to parse
         */
        pCurPartDesc = &pAdapterResDesc->PartialResourceList.PartialDescriptors[ ResDescIndex];

        /* Point to the current empty memory range descriptor in the PROCBASE
         */
        pCurMemRange = &pThis->MemRanges[ CurRangeIndex];

        switch ( pCurPartDesc->Type)
        {
        case CmResourceTypePort:

            //
            // Verify that we haven't exceeded the maximum resources for a
            // PCI device.
            //

            if ( CurRangeIndex < PCI_TYPE0_ADDRESSES)
            {
                DebugPrint((3,
                            "  I/O range:     Base = %08x : %08x   Length = %08x \n",
                            pCurPartDesc->u.Port.Start.HighPart,
                            pCurPartDesc->u.Port.Start.LowPart,
                            pCurPartDesc->u.Port.Length
                            ));

                /* Copy I/O resource information.
                 */
                pCurMemRange->RangeStart = pCurPartDesc->u.Port.Start;
                pCurMemRange->RangeLength = pCurPartDesc->u.Port.Length;
                pCurMemRange->AddrSpaceType = 1;    // I/O space
                pCurMemRange->ResourceMapped = FALSE;

                /* Inc.index to point the next empty mem.range descriptor
                 */
                CurRangeIndex++;
            }
            break;

        case CmResourceTypeMemory:

            //
            // Verify that we haven't exceeded the maximum resources for a
            // PCI device.
            //

            if (CurRangeIndex < PCI_TYPE0_ADDRESSES)
            {
                DebugPrint((3,
                            "  Memory range:  Base = %08x : %08x   Length = %08x \n",
                            pCurPartDesc->u.Port.Start.HighPart,
                            pCurPartDesc->u.Port.Start.LowPart,
                            pCurPartDesc->u.Port.Length
                            ));

                //
                // This driver will only look at read/write memory.
                //

                if ( pCurPartDesc->Flags == CM_RESOURCE_MEMORY_READ_WRITE)
                {
                    //
                    // Save memory resource information.
                    //

                    pCurMemRange->RangeStart = pCurPartDesc->u.Port.Start;
                    pCurMemRange->RangeLength = pCurPartDesc->u.Port.Length;
                    pCurMemRange->AddrSpaceType = 0;    // memory space
                    pCurMemRange->ResourceMapped = FALSE;

                    /* Inc.index to point the next empty mem.range descriptor
                     */
                    CurRangeIndex++;
                }
                else
                {
                    DebugPrint(( 3,
                                 "  ** Ignoring memory: flags = %08x \n",
                                 pCurPartDesc->Flags
                                 ));
                }
            }
            break;

        case CmResourceTypeInterrupt:

            DebugPrint((3,
                        "  Interrupt:  Level = %x    Vector = %x    Mode = %x \n",
                        pCurPartDesc->u.Interrupt.Level,
                        pCurPartDesc->u.Interrupt.Vector,
                        pCurPartDesc->Flags
                        ));

            //
            // Save interrupt information.
            //

            pThis->InterruptLevel = pCurPartDesc->u.Interrupt.Level;
            pThis->InterruptVector = pCurPartDesc->u.Interrupt.Vector;
            pThis->InterruptMode =
                ( pCurPartDesc->Flags == CM_RESOURCE_INTERRUPT_LEVEL_SENSITIVE) ?
                    LevelSensitive :
                    Latched;
            break;

        case CmResourceTypeDma:

            //
            // Don't do anything with DMA type resources.  This doesn't
            // really make sense for a PCI busmaster device.
            //

            DebugPrint((3, "CmResourceTypeDma not supported \n"));
            break;

        default:

            //
            // Skip unknown resource types.
            //

            DebugPrint((3, "Unknown resource type 0x%04x\n", pCurPartDesc->Type));
            break;

        }   // switch
    }

    //
    // Save the count of the number of resources found.
    //

    pThis->RangesCnt = CurRangeIndex;

}   // ParseResourceList

/****************************************************************************
****************************************************************************/
NTSTATUS
InstallInterruptHandler
    (
        IN PPROCBASE      pThis
    )
/*++

Routine Description:

    Initialize the ISR.
                                                                     
Arguments:

Return Value:

    NTSTATUS

--*/
{
    NTSTATUS            status;
    ULONG               vector;
    KIRQL               irql;
    KAFFINITY           affinity;

    DebugPrint((2, "ProcBase.InstallInterruptHandler\n"));
    ASSERT( pThis->Type == PROCBASE_SIGN);

    /*
     * Check the validity of pThis->InterruptLevel and pThis->InterruptVector
     */
    ASSERT( pThis->bRegistryResourcesAssigned);

    /*
     * Set up the device interrupt.  Need to set up the interrupt after 
     * setting up the memory and I/O resources in case the card suddenly 
     * generates an interrupt and the driver's ISR is called.  If this 
     * happens, the device registers will be properly mapped so we can 
     * safely clear the interrupt.
     */
    irql = (UCHAR) pThis->InterruptLevel;
    vector = HalGetInterruptVector(
                PCIBus,
                pThis->BusNumber,
                pThis->InterruptLevel,
                pThis->InterruptVector,
                &irql,
                &affinity);

    DebugPrint((2, "    HalGetInterruptVector -> vector=%x irql=%x affinity=%x\n", vector, irql, affinity));

    status = IoConnectInterrupt(
                &pThis->pInterruptObject,
                DeviceISR,
                pThis,                  // Context passed to the ISR - PPROCBASE
                NULL,                   // No spinlock - single vector supported
                vector,
                irql,
                irql,                   // Synchronization IRQL
                pThis->InterruptMode,   // LevelSensitive or Latched
                TRUE,                   // Shareable interrupts
                affinity,
                FALSE);                 // Don't save FP stack

    if ( !NT_SUCCESS(status))
    {
        DebugPrint((1, "Unable to connect interrupt. Status = %x \n", status));
        pThis->pInterruptObject = NULL;
    }

    return status;
    
}   // InstallInterruptHandler

/****************************************************************************
    The standard NT interrupt service routine
****************************************************************************/
BOOLEAN
DeviceISR(
    IN PKINTERRUPT pInterrupt,
    IN PVOID ServiceContext
    )
{
    PPROCBASE   pThis;

    /*
     * We arranged this in the InstallInterruptHandler func.
     */
    pThis = (PPROCBASE) ServiceContext;
    ASSERT( pThis->Type == PROCBASE_SIGN);
    ASSERT( pThis->pInterruptObject == pInterrupt);

    /*
     * Call PROCESSOR's interrupt method which return FALSE if it sees
     * that its 7751 board hasn't generated this interrupt. Then we don't need
     * to queue the DPC and should quickly return FALSE too.
     */
    if ( !ProcessorConfirmAndRegisterInterrupt( &pThis->Proc))
    {
        return FALSE;
    }

    /*
     * It was our device that interrupted - queue our DPC
     */
    ProcBaseRequestInt( pThis);

    return TRUE;

}   //PwrCryptISR

/****************************************************************************
    The standard NT DPC routine

VOID
(*PKDEFERRED_ROUTINE)(
    IN PKDPC Dpc;
    IN PVOID DeferredContext;
    IN PVOID SystemArgument1;
    IN PVOID SystemArgument2;
    );
****************************************************************************/
VOID
DeviceDPC(
    IN PKDPC pDpc,
    IN PVOID DeferredContext,
    IN PVOID SystemArgument1,
    IN PVOID SystemArgument2
    )
{
    PPROCBASE   pThis;

    DebugPrint((3, "**** DPC 0x%08x start\n", pDpc));
    ASSERT( (ULONG)SystemArgument1 == 0x1234 && (ULONG)SystemArgument2 == 0x5678);

    /*
     * We arranged this in the ctor.
     */
    pThis = (PPROCBASE) DeferredContext;
    ASSERT( pThis->Type == PROCBASE_SIGN);
    ASSERT( &pThis->DPC == pDpc);

    ProcessorServiceRegisteredInterrupts( &pThis->Proc);

    DebugPrint((3, "finish DPC 0x%08x ****\n", pDpc));
}

/****************************************************************************
****************************************************************************/
void
ProcBaseRequestInt(
    IN PPROCBASE       pThis
    )
{
    BOOLEAN     bDPCQueued;

    DebugPrint((3, "ProcBaseRequestInt **** KeInsertQueueDPC( 0x%08x )\n", &pThis->DPC));
    ASSERT( pThis->Type == PROCBASE_SIGN);

    bDPCQueued = KeInsertQueueDpc( &pThis->DPC, (PVOID) 0x1234, (PVOID) 0x5678);

    DebugPrint((3, "ProcBaseRequestInt **** DPC inserting status - %s\n",
                   bDPCQueued ? "Ok" : "Failure"));
}

/****************************************************************************
****************************************************************************/
BOOLEAN
ProcBaseSynchronizeExecution
    (
        IN PPROCBASE       pThis,
        IN PPROCBASE_SYNCHRONIZE_ROUTINE pSynchronizeRoutine,
        IN PVOID  SynchronizeContext
    )
{
    DebugPrint((3, "ProcBaseSynchronizeExecution\n"));
    ASSERT( pThis->Type == PROCBASE_SIGN);

    return KeSynchronizeExecution(
                pThis->pInterruptObject,
                pSynchronizeRoutine,
                SynchronizeContext);
}
