#ifndef __PROCBASE_H__
#define __PROCBASE_H__

#include "processor.h" // for PROCESSOR
#include "dma_ctx.h"            // for DMA_CTX

typedef struct _ADAPTER_MEM_RANGE {

    PHYSICAL_ADDRESS RangeStart;    // Original device physical address
    ULONG   RangeLength;            // Length of I/O or memory range
    ULONG   AddrSpaceType;          // 0 - mem, 1 - I/O (as defined by NT)

    PVOID   MappedRangeStart;       // Mapped I/O or memory range
    ULONG   MappedAddrSpaceType;    // 0 - mem, 1 - I/O (as defined by NT)

    BOOLEAN ResourceMapped;         // Flag: resource is mapped (i.e. MmMapIoSpace called)

} ADAPTER_MEM_RANGE, *PADAPTER_MEM_RANGE;

/****************************************************************************
 * ProcBase class
 *
 *  The role of PROCBASE is:
 *
 *      - perform O/S-dependent part of PCI resources acquisition (assignments
 *        in the Registry, mapping memory ranges, registering an ISR)
 *
 *      - launching the embedded PROCESSOR to perform all the further data
 *        processing
 *
 *      - stopping the PROCESSOR when it is time to unload the driver
 *
 *      - returning acquired resources to the system
 */

typedef struct _PROCBASE
{
    ULONG           Type;

    /*-----------------------------------------------------------------------
     * The main object that is constructed and launched by this O/S-dependent
     * PROCBASE object
     */
    PROCESSOR           Proc;

    /*-----------------------------------------------------------------------
     * The configuration storage that holds some initialization parameters for
     * both PROCBASE and its PROCESSOR
     */
    PCONFIG             pConfig;

    /*-----------------------------------------------------------------------
     * 2 DMA contexts - initialized by PROCBASE, used by PROCESSOR
     */
    DMA_CTX             SrcDmaCtx;
    DMA_CTX             DstDmaCtx;

    /*-----------------------------------------------------------------------
     * "Adapter Object"-related stuff
     */
    ULONG               BusNumber;  // System bus number
    ULONG               SlotNumber; // PCI slot number
    ULONG               MaxXferLen; // Maximum transfer length for adapter

    /*-----------------------------------------------------------------------
     * These flags/vars indicate which system resources we have acquired that
     * require to be returned back.
     * The list order of flags is the order of acquisition.
     */
    BOOLEAN             bRegistryResourcesAssigned;
    BOOLEAN             bProcessorPlugged;
    BOOLEAN             bISRInstalled;

    // pCommonBuf != NULL means HalGetCommonBuffer succeeded

    // MemRanges[].ResourceMapped indicates that the range was mapped and
    // RangesCnt shows number of the array's elements to scan

    /*-----------------------------------------------------------------------
     * Assigned by HalAssignSlotResources - valid if ( bResourcesAssigned)
     */
    ULONG               InterruptLevel;         // Unmapped system interrupt level
    ULONG               InterruptVector;        // Unmapped system interrupt vector
    KINTERRUPT_MODE     InterruptMode;          // LevelSensitive or Latched

    ULONG               RangesCnt;              // Number of I/O and memory resources
    ADAPTER_MEM_RANGE   MemRanges[PCI_TYPE0_ADDRESSES];   // PCI base address info

    /*-----------------------------------------------------------------------
     */
    PKINTERRUPT         pInterruptObject;        // Returned by IoConnectInterrupt
    KDPC                DPC;

} PROCBASE, *PPROCBASE;

/**********************************************************************
 * PUBLIC methods
 */

/*---------------------------------------------------------------------
 * Creator-called
 */

void
ProcBaseConstructor
    (
        IN OUT PPROCBASE pThis,
        IN PCONFIG       pConfig
    );

NTSTATUS
ProcBaseAcquireHardware
    (
        IN PPROCBASE        pThis,
        IN PDISPATCHER      pDispatcher,
        IN PUNICODE_STRING  pusProcBaseRegPath,
        IN ULONG            BusNumber,
        IN ULONG            SlotNumber,
        IN PDRIVER_OBJECT   DriverObject,
        IN PUNICODE_STRING  pusDeviceName
    );

NTSTATUS
ProcBaseReleaseHardware
    (
        IN PPROCBASE       pThis,
        IN PUNICODE_STRING  pusProcBaseRegPath,
        IN PDRIVER_OBJECT   DriverObject
    );

void
ProcBaseDestructor
    (
        IN PPROCBASE pThis
    );

/*---------------------------------------------------------------------
 * Internally and PROCESSOR-called
 */

void
ProcBaseRequestInt
    (
        IN PPROCBASE       pThis
    );


typedef PKSYNCHRONIZE_ROUTINE   PPROCBASE_SYNCHRONIZE_ROUTINE;

BOOLEAN
ProcBaseSynchronizeExecution
    (
        IN PPROCBASE       pThis,
        IN PPROCBASE_SYNCHRONIZE_ROUTINE pSynchronizeRoutine,
        IN PVOID  SynchronizeContext
    );

/**********************************************************************
 * PRIVATE methods
 */
VOID
ParseResListCopyAssignments
    (
        IN PPROCBASE    pThis,
        IN PCM_FULL_RESOURCE_DESCRIPTOR pAdapterResDesc
    );

NTSTATUS
TranslateAndMapAddrs
    (
        IN PPROCBASE    pThis
    );

NTSTATUS
InstallInterruptHandler
    (
        IN PPROCBASE    pThis
    );

BOOLEAN
DeviceISR
    (
        IN PKINTERRUPT Interrupt,
        IN PVOID ServiceContext
    );

VOID
DeviceDPC
    (
        IN PKDPC Dpc,
        IN PVOID DeferredContext,
        IN PVOID SystemArgument1,
        IN PVOID SystemArgument2
    );

#endif
