#include "osindependence.h"
#include "debug.h"
#include "driverdef.h"
#include "sg_ctlarea.h"
#include "instrset.h"

#ifdef ALLOC_PRAGMA
#pragma alloc_text(INIT, SgCtlAreaConstructor)
#endif

/****************************************************************************
 * SgCtlAreaConstructor -
 *
 * Description:
 *
 *  Performs first-time initialization of 7751 DMA
 *  control area that contains scatter/gather descriptors for 7751 DMA engines.
 *
 *  - Calls ctors of all 4 rings that it contains.
 *  Note, that SG_DESC arrays that follow rings are deemed parts of respecting
 *  rings and are initialized by rings' ctors (the 2nd parm od the ring ctor
 *  tells it how much s/g descriptors it has. See also ring.h)
 *
 *  - Puts addresses of commands and results to their respectful descriptors.
 *  This connection is permanent. This way, the processor loops through
 *  the commands' and results' arrays. 
 *
 * pPhysThis - the address of the SG_CTL_AREA as seen by DMA engines
 */
void
SgCtlAreaConstructor(
    PSG_CTL_AREA    pThis,
    ULONG           PhysThis,
    ULONG           SmallRingDataFragmentSize
    )
{
    ULONG   i;

    DebugPrint((3, "SgCtlAreaConstructor pThis=0x%08x PhysThis=0x%08x\n", pThis, PhysThis));

    pThis->Type = SG_CTL_AREA_SIGN;

    /* Small ring data area and its fragments must be 32-bit aligned.
     */
    ASSERT( ((ULONG) ( &pThis->SmallRingData[ 0]) & 0x00000003L) == 0);
    ASSERT( ( SmallRingDataFragmentSize & 0x00000003L) == 0);

    pThis->PhysThis = PhysThis;
    pThis->SmallRingDataSize = ( COM_AND_RES_RING_LEN - 1) * SmallRingDataFragmentSize;

    /* Construct both rings - this fills "jump" descriptors at the end of a ring
     */
    RingConstructor( &pThis->SmallRing,
                     COM_AND_RES_RING_LEN,
                     PhysThis + FIELD_OFFSET( SG_CTL_AREA, SmallRing));

    RingConstructor( &pThis->BigRing,
                     DATA_RING_LEN,
                     PhysThis + FIELD_OFFSET( SG_CTL_AREA, BigRing));

    /* Put physical addresses of SmallRingData fragments into the descriptors
     * of the small ring. Phys.addrs in those descriptors are never changed
     * during the driver's work. Commands or Results are merely put into the
     * SmallRingData fragments and descriptors are made "valid", when it's
     * time to process those Commands or Results.
     */
    for( i = 0; i < COM_AND_RES_RING_LEN - 1; i++)
    {
        SgDescSetDataPhysAddr(  &pThis->SmallRing.Descs[ i],
                                PhysThis + FIELD_OFFSET( SG_CTL_AREA,
                                                         SmallRingData[ i * SmallRingDataFragmentSize]));
    }
}
