#include "osindependence.h"
#include "debug.h"
#include "driverdef.h"
#include "sg_ring.h"

#ifdef ALLOC_PRAGMA
#pragma alloc_text(INIT, RingConstructor)
#endif

/****************************************************************************
 * RingConstructor()
 *
 * Description:
 *   the first-time initialization of the RING - zeroes out all the descriptors
 *   and sets the FirstFreeDescIndex to point to the first one (notice, that
 *   it doesn't matter, actually, which descriptor to make the first, since this
 *   is a ring, but SG_CTL_AREA relies on ). The last step is to assemble a "jump" descriptor at the array's
 *   tail to point to the first one.
 *
 * DescriptorCnt -
 *  indicates the number of descriptors in the ring, including the last (jump)
 *  descriptor
 *
 * ThisPhysAddr -
 *  address of the RING, as seen by the DMA engine
 */

void
RingConstructor(
    PRING   pThis,
    ULONG   DescCnt,
    ULONG   ThisPhysAddr
    )
{
    DebugPrint((3, "RingConstructor pThis=0x%08x PhysThis=0x%08x\n", pThis, ThisPhysAddr));

    pThis->Type = SG_RING_SIGN;

    /* We always have 1 "jump" descriptor, let's have data descriptors too.
     */
    ASSERT( DescCnt > 1);

    /* Descriptors must be 32-bit aligned.
     */
    ASSERT( ((ULONG) ( &pThis->Descs[ 0]) & 0x00000003L) == 0);

    pThis->DescCnt = DescCnt;

    /*
     * Fill "data address" in the last s/g descriptor so that it would
     * point to the first s/g descriptor thus forming a descriptor "ring"
     */
    SgDescSetDataSize(
        &pThis->Descs[ DescCnt - 1], 0);

    SgDescSetDataPhysAddr(
        &pThis->Descs[ DescCnt - 1], ThisPhysAddr +
                                     FIELD_OFFSET( RING, Descs[0]));
}

/****************************************************************************
 * RingInvalidateDescriptors()
 *
 *   Zeroes out the SG descriptors' flags to make them "invalid" and
 *   prevent a DMA engine from processing them. The last descriptor is made
 *   the "jump" descriptor by setting the appropriate flags.
 *
 *   Declares the ring's first SG descriptor the first descriptor to be used
 *   for an eventual DMA arrangement.
 *
 *   Returns logical DMA address of the first descriptor of the ring,
 *   suitable to be loaded into a "Ring Address Register" of the 7751.
 *   It is taken from the last, "jump", descriptor of the ring.
 */
ULONG
RingResetAndGetPhysAddr(
    PRING   pThis
    )
{
    ASSERT( pThis->Type == SG_RING_SIGN);

    pThis->FirstFreeDescIndex = 0;

    {
        ULONG   i;

        for ( i = 0; i < pThis->DescCnt - 1; i++)
        {
            SgDescSetFlags(  &pThis->Descs[ i], 0);
        }
    }

    SgDescSetFlags(
        &pThis->Descs[ pThis->DescCnt - 1], SG_DESC_FLAG_JUMP | SG_DESC_FLAG_MASKDONE);

    return SgDescGetDataPhysAddr( &pThis->Descs[ pThis->DescCnt - 1]);
}
 
/****************************************************************************
 */
void
RingValidateJump(
    PRING   pThis
    )
{
    ASSERT( pThis->Type == SG_RING_SIGN);
    SgDescAddFlags(
        &pThis->Descs[ pThis->DescCnt - 1], SG_DESC_FLAG_VALID);
}

/****************************************************************************
 * RingAdvanceFirstFreeDescIndex()
 *
 * Description:
 *  Advances FirstFreeDescIndex to the next descriptor with "end-of-array"
 *  checking and proper index "wrapping" to the first descriptor in the array
 */
void
RingAdvanceFirstFreeDescIndex(
    PRING   pThis
    )
{
    ASSERT( pThis->Type == SG_RING_SIGN);
    /*
     * Since this func always skips the last ("jump") descriptor,
     * it can't be the First Free Descriptor
     */
    ASSERT( pThis->FirstFreeDescIndex != pThis->DescCnt - 1);                     

    /*
     * Advance index to the next descriptor
     */
    pThis->FirstFreeDescIndex++;

    /*
     * If this next descriptor is the last - it is a "jump".
     *  - mark it as "valid"
     *  - wrap the index to the first descriptor in the array
     */
    if ( pThis->FirstFreeDescIndex == pThis->DescCnt - 1)
    {
        /*
         * A "jump" must be "jump" and not "valid"
         */
#ifndef _SKIP_MODE
        ASSERT( !( SgDescGetFlags( &pThis->Descs[ pThis->FirstFreeDescIndex]) &
                   SG_DESC_FLAG_VALID));
#endif

        ASSERT( SgDescGetFlags( &pThis->Descs[ pThis->FirstFreeDescIndex]) &
                SG_DESC_FLAG_JUMP);

        SgDescAddFlags( &pThis->Descs[ pThis->FirstFreeDescIndex],
                        SG_DESC_FLAG_VALID);

        /*
         * Wrap the index to the first descriptor
         */
        pThis->FirstFreeDescIndex = 0;
    }
}
