/**********************************************************************

    Module: SlotTracker.h

        Product: kernel mode driver for "PowerCrypt" encryption
        acceleration board based on Hi/fn 7751 chip
        by Ashley Laurent, Inc.

    ---------------------------------------------------------------

    Copyright:

        Ashley Laurent, Inc., 1999
        All Rights Reserved.

    ---------------------------------------------------------------

    Functional Description:

        Declaration of the SLOTTRACKER object that performs
        the control over the 7751 internal context memory slot usage to
        allow multiple clients to use the slots without conflicts.

    ---------------------------------------------------------------

    Environment:

        NT or BSD kernel mode

    ---------------------------------------------------------------

    Author:

        Stan Shkolnyy

    ---------------------------------------------------------------

    Revision history:

        *   15-Feb-99   initial revision

**********************************************************************/

#ifndef __SLOTTRACKER_H__
#define __SLOTTRACKER_H__

#include "osindependence.h"

/**********************************************************************
 * Slot Tracker maintains a descriptor for each PwrCrypt slot.
 * The descriptor associates the slot with a unique key, specified by the
 * SlotTracker client. This allows the client to free a group of slots
 * associated with the same key, if it knows the key. The descriptor also
 * contains a reference counter, used to prohibit the slot freeing when
 * somebody uses it
 */
typedef struct _SLOT_DESC
{
    ULONG   nKey;
    ULONG   RefCnt;
    ULONG   CoreSlotNumber;

} SLOT_DESC, *PSLOT_DESC;

/**********************************************************************
 * SlotTracker creates and maintains an array of slot descriptors
 */

typedef struct _SLOTTRACKER
{
    ULONG           Type;

    /*
     * Number of ctx mem slots managed
     */
    ULONG               NumOfSlots;

    /*
     * Ptr to an array of slot descriptors, of NumOfSlots length
     */
    SLOT_DESC           *pSlots;

    /*
     * Spinlock acquired by the methods that use or free slots
     */
    PCRPT_DECLARE_SPINLOCK( SpinLock);

} SLOTTRACKER, *PSLOTTRACKER;

/****************************************************************************
 * SLOTTRACKER user-defined routine that is called several times as a
 * consequence of a call to SlotTrackerEnumSlotsByKey.
 * The above mentioned method traverses the slot tracking array and calls the
 * specified callback for each slot number which key matches the key passed
 * to the method.
 */
typedef void
(*PSLOT_ENUM_CALLBACK)
    (
        IN PVOID        Context,
        IN ULONG        SlotMunber
    );

/**********************************************************************
 * PUBLIC methods
 */
void
SlotTrackerConstructor
    (
        IN OUT  PSLOTTRACKER pThis
    );

void
SlotTrackerInit
    (
        IN OUT  PSLOTTRACKER pThis,
        IN      ULONG        NumOfSlots
    );

void
SlotTrackerDestructor
    (
        IN PSLOTTRACKER pThis
    );

BOOLEAN
SlotTrackerRegisterSlot
    (
        IN PSLOTTRACKER pThis,
        IN ULONG        CoreSlotNumber,
        IN ULONG        nKey,
        OUT ULONG       *pSlotNumber
    );

BOOLEAN
SlotTrackerUnregisterSlot
    (
        IN PSLOTTRACKER pThis,
        IN ULONG        SlotNumber,
        IN ULONG        nKey,
        OUT ULONG       *pCoreSlotNumber
    );

BOOLEAN
SlotTrackerReferenceSlot
    (
        IN PSLOTTRACKER pThis,
        IN ULONG        SlotNumber,
        IN ULONG        nKey
    );

BOOLEAN
SlotTrackerDereferenceSlot
    (
        IN PSLOTTRACKER pThis,
        IN ULONG        SlotNumber,
        IN ULONG        nKey
    );

void
SlotTrackerEnumSlotsByKey
    (
        IN PSLOTTRACKER         pThis,
        IN PSLOT_ENUM_CALLBACK  pCallBack,
        IN PVOID                Context,
        IN ULONG                nKey
    );

#define SlotTrackerIsValid( pThis)   ( (pThis)->pSlots != NULL)

#endif
