/**********************************************************************

    Module: userbuf.h

        Product: kernel mode driver for "PowerCrypt" encryption
        acceleration board based on Hi/fn 7751 chip
        by Ashley Laurent, Inc.

    ---------------------------------------------------------------

    Copyright:

        Ashley Laurent, Inc., 1999
        All Rights Reserved.

    ---------------------------------------------------------------

    Functional Description:

        NT implementation of functions for locking pageable buffer with MDL
        creation and for unlocking it with MDL desctruction.

    ---------------------------------------------------------------

    Environment:

        NT kernel mode

    ---------------------------------------------------------------

    Author:

        Stan Shkolnyy

    ---------------------------------------------------------------

    Revision history:

        *   15-Feb-99   initial revision

**********************************************************************/

#include "osindependence.h"
#include "debug.h"
#include "userbuf.h"

ULONG   BytesLocked = 0;

/****************************************************************************

UserBuffersLock()

Description:
    Locks the user buffer in memory and creates an MDL for it

Arguments:

Return:


/////////////////////////////
//
// BUGBUG Incorporate this if 
//
VOID 
    MmBuildMdlForNonPagedPool(
        IN OUT PMDL  MemoryDescriptorList
        );
MmBuildMdlForNonPagedPool fills in the corresponding physical page array of
a given MDL for a buffer in nonpaged system space (pool).

Parameters
MemoryDescriptorList
Points to an MDL that supplies a virtual address, byte offset, and length.
 
Comments
The physical page portion of the MDL is updated as the pages are locked
in memory. The input MDL virtual address must be within the nonpaged portion
of system space.

Callers of MmBuildMdlForNonPagedPool must be running at IRQL <= DISPATCH_LEVEL.


****************************************************************************/

PWRCRYPT_STATUS
LockBufferCreateMdl (
    IN  PVOID            pBuf,
    IN  ULONG            BufSize,
    IN  BOOLEAN          bChargeQuota,
    IN  PWRCRYPT_PROCESSOR_MODE  AccessMode,
    IN  BOOLEAN          bWriteAccessRequired,
    OUT PMDL             *ppMdl
    )
{
    PWRCRYPT_STATUS status;        // func.ret.value
    PMDL            pMdl;

    DebugPrint((3, "LockUserBuffer\n"));

    *ppMdl = NULL;

    status = PWRCRYPT_STATUS_SUCCESS;

    pMdl = IoAllocateMdl( pBuf,         // VirtualAddress
                          BufSize,      // Length
                          FALSE,        // SecondaryBuffer
                          bChargeQuota,
                          NULL);        // Irp

    if ( pMdl != NULL)
    {
        //
        // Now probe the buffer described by the Mdl.  If we get an exception,
        // deallocate the Mdl and return the appropriate status.
        //
        try
        {
            MmProbeAndLockPages(
                pMdl,
                AccessMode,
                bWriteAccessRequired ? IoModifyAccess : IoReadAccess
                );

            /* If we are here, the call has worked - return MDL* to the caller
             *
             * Now, as it can be seen
             * pMdl->MdlFlags = MDL_PAGES_LOCKED | MDL_WRITE_OPERATION
             */
            *ppMdl = pMdl;

            BytesLocked += MmGetMdlByteCount(pMdl);

            // here I verify my understanding of things
            ASSERT( pBuf == MmGetMdlVirtualAddress( pMdl));
            ASSERT( BufSize == MmGetMdlByteCount( pMdl));
            DebugPrint((3, "    pbuf=%08x size=%08x mdl=%08x\n", pBuf, BufSize, pMdl));
        }
        except( EXCEPTION_EXECUTE_HANDLER)
        {
            ASSERT(0);
            status = GetExceptionCode();
            DebugPrint((2, "    MmProbeAndLockPages exception 0x%08x\n", status));
            IoFreeMdl( pMdl);
            status = PWRCRYPT_STATUS_INVALID_USER_BUFFER;
        }
    }
    else
    {
        DebugPrint((2, "    Mdl alloc.failed\n"));
        status = PWRCRYPT_STATUS_NO_MEMORY;
        ASSERT(0);
    }

    return status;
}


/****************************************************************************

UnlockBufferDelMdl()

Description:
    Unlocks the buffer specified by the MDL and deletes MDL

Arguments:

Return:
    None

****************************************************************************/
void
UnlockBufferDelMdl (
    IN PMDL    pMdl
    )
{
    DebugPrint((3, "UnlockUserBuffer pbuf=%08x size=%08x mdl=%08x\n", MmGetMdlVirtualAddress( pMdl), MmGetMdlByteCount(pMdl), pMdl));

    BytesLocked -= MmGetMdlByteCount(pMdl);

    MmUnlockPages( pMdl);
    IoFreeMdl( pMdl);
}
