#include "osindependence.h"
#include <pwrcryptioctl.h>
#include <km_interface.h>
#include "crypto.h"

#include "descomm.h"

/****************************************************************************
 * Some forward declarations of common utility funcs used by crypto funcs
 */
BOOLEAN
ExecuteInstructionSync
    (
        PIOCTL_PWRCRYPT_CALCULATE_INPUT_OUTPUT pInstr
    );

/****************************************************************************

Note! Enc.key and IV are deemed the "read-only" property of the function until
it returns. They _must_ be allocated in the SYSTEM NON-PAGED memory.

****************************************************************************/
ULONG
DESEncrypt
    (
    IN  PVOID                   pPlainPayload,
    IN  ULONG                   ulPlainPayloadSize,
    IN  PVOID                   pCipherPayload,
    IN  PKEY                    pEncryptionKey,
    IN  PIV                     pEncryptionIV
    )
{
    ULONG           nRltDataSz;         // func.ret.value

    PIOCTL_PWRCRYPT_CALCULATE_INPUT_OUTPUT  pInstr;
    PPWRCRYPT_CALCULATE_KM_WORKAREA         pWA;

    PDES_COMMAND    pCom;
    PDES_RESULT     pRlt;

    USHORT          SlotNumber;

    ASSERT( pEncryptionIV->IVLength == _7751_DES_IV_SIZE);
    ASSERT( pEncryptionKey->KeyLength == _7751_DES_KEY_SIZE);
    ASSERT( pEncryptionKey->KeyNumber == 1);

    /* Assume we'll process nothing
     */
    nRltDataSz = 0;

    /* BUGBUG: Assume slot number 0, since it is not passed to this function.
     * Also, we specify CRYPCOM_DONT_STORE_CTX in the crypto subcommand - this
     * means the 7751 context memory will remain unchanged (the key and IV
     * won't be written into it).
     */
    SlotNumber = 0;

    /* Allocate instruction storage
     * BUGBUG: Consider some other method of memory allocation which would
     * cause less memory fragmentation
     */
    pInstr = (PIOCTL_PWRCRYPT_CALCULATE_INPUT_OUTPUT) PCrptNonPagedMAlloc( sizeof( *pInstr)); //DES_INSTRUCTION_SIZE);
    if ( pInstr != NULL)
    {
        /*
         * Setup a pointer to the KM work area
         */
        pWA = (PPWRCRYPT_CALCULATE_KM_WORKAREA) &pInstr->KMWorkArea;

        /*
         * Zero the work area for not to assign zeroes to its members
         */
        PCrptZeroMemory( pWA, sizeof( *pWA));

        /*
         * The total size of the command with all its subcommands
         */
        pWA->nCommandSize = sizeof( BASE_COMMAND) + sizeof( CRYP_COMMAND);

        /*-------------------------------------------------------------------
         * Assemble the instruction in the allocated memory
         */

        /*
         * Fill in the data buffer pointers and lengths
         */
        pInstr->nSlotNumber     = SlotNumber;
        pInstr->pSrcData        = pPlainPayload;
        pInstr->nTotalSrcCnt    = ulPlainPayloadSize;
        pInstr->pDstBuffer      = pCipherPayload;
        pInstr->nTotalDstCnt    = ulPlainPayloadSize;

        /* These two ptrs are needed because we don't use the two areas'
         * structures as they are defined in the IOCTL buffer. This is
         * the proper use, however, since the definition contains mere
         * placeholders which are _intended_ to be filled differently
         * depending on required instruction
         */
        pCom = ( PDES_COMMAND) &pInstr->Command;
        pRlt = ( PDES_RESULT) &pInstr->Result;

        /*
         * Fill in the Command part of the 7751 instruction and its length
         */
        BASECOM_FILL(
            &pCom->base,
            BASECOM_SCRAMBLE | BASECOM_CRYP,
            ulPlainPayloadSize,
            ulPlainPayloadSize);

        CRYPCOM_FILL(
            &pCom->cryp,
            CRYPCOM_NEW_IV | CRYPCOM_NEW_KEY | CRYPCOM_DONT_STORE_CTX | CRYPCOM_ALGO_DES | CRYPCOM_DES_MODE_CBC,
            0,
            ulPlainPayloadSize);

        /*
         * Set the coordinates of the parts of Context part of the 7751 instruction
         */
        pWA->pEncKey = &pEncryptionKey->KeyValue[ 0][ 0];
        pWA->nEncKeySize = 8;
        pWA->pEncIV = &pEncryptionIV->IVValue[ 0];
        pWA->nEncIVSize  = 8;

        /*
         * Fill in the buffer passing methods in the KM work area
         * Pay attention to which flags to specify, as described in the
         * km_interface.h.
         * (Since this func is presumable KM-called, and the data buffers are
         * presumably located in the KM addr.space, we don't specify
         * PWRCRYPT_CALCULATE_XX_USER flags)
         *
         * Also, if the buffers are in paged memory, the call to the
         * PwrcryptCalculate() _must_ be made at PASSIVE_LEVEL
         */
        pWA->flBufsDescType = ( PWRCRYPT_CALCULATE_IN_PAGE |
                                PWRCRYPT_CALCULATE_OUT_PAGE
                              );

        /*-------------------------------------------------------------------
         * If it's known that this func is called in the requesting process context,
         * this var should be set to TRUE, so that the process's quota would be
         * charged when the PwrCrypt driver will create MDLs.
         * BUGBUG: how this can be known??? Only by adding a parm to this function.
         */
        pWA->bChargeQuota = FALSE;  // I don't think the NDIS stuff executes in the requestor context :-)

        /*-------------------------------------------------------------------
         * Execute the instruction synchronously - on return, the result is
         * known and ready to be used
         */
        if ( ExecuteInstructionSync( pInstr))
        {
            /* Check and print the Result part of the 7751 instruction
             */
//            USHORT  pRlt->base.Reserved      :16;
//            USHORT  pRlt->base.SessNo        :12;
//            USHORT  pRlt->base.DstCntHi      :2;
//            USHORT  pRlt->base.SrcCntHi      :2;
//            USHORT  pRlt->base.SrcCntLo      :16;
//            USHORT  pRlt->base.DstCntLo      :16;
//
//            USHORT  pRlt->cryp.PrematureStop :1;     //EncSrcCnt went to 0, but TotalSrcCnt still != 0

            /* By setting this value we report success to the caller
             */
            nRltDataSz = ulPlainPayloadSize;
        }

        /* Free instruction storage
         */
        PCrptNonPagedMFree( pInstr);
    }
    return nRltDataSz;
}

/****************************************************************************

CalculateCompletionRoutine()

Description:
    Called by the PwrCrypt driver when it completes the execution of the
    instruction in PwrcryptCalculate()

Arguments:

    nComplRtnParm1 - ptr to the event to be signaled
    nComplRtnParm2 - not used

****************************************************************************/
void
static
CalculateCompletionRoutine
    (
        ULONG   nComplRtnParm1,
        ULONG   nComplRtnParm2
    )
{
    ASSERT( nComplRtnParm2 == 0x12345678);
    PCrptSignalEvent( (PPCRPT_EVENT) nComplRtnParm1);
}

/****************************************************************************
****************************************************************************/
BOOLEAN
static
ExecuteInstructionSync
    (
        PIOCTL_PWRCRYPT_CALCULATE_INPUT_OUTPUT pInstr
    )
{
    PPWRCRYPT_CALCULATE_KM_WORKAREA pWA;
    PCRPT_EVENT                     CompletionEvent;
    PWRCRYPT_STATUS                 status;

    /*
     * Setup a pointer to the KM work area
     */
    pWA = (PPWRCRYPT_CALCULATE_KM_WORKAREA) &pInstr->KMWorkArea;

    /* Report the completion routine address and its two parms to
     * the service. It doesn't interpret the parms, just calls the
     * routine and passes them to it, upon request completion.
     * We need to see the CompletionEvent in our completion routine and
     * set this event.
     */
    pWA->pComplRtn = CalculateCompletionRoutine;
    pWA->nComplRtnParm1 = (ULONG) &CompletionEvent;
    pWA->nComplRtnParm2 = 0x12345678;

    /*-------------------------------------------------------------------
     * Send the instruction to the PwrCrypt driver for execution
     */

    /* Initialize event _before_ calling the calculating service -
     * the service will call our callback routine at any time and the
     * routine will signal the event.
     */
    PCrptInitEvent( &CompletionEvent);

    /*
     * Call the PwrCrypt calculating service.
     */
    status = PwrcryptCalculate( pInstr);

    /*-------------------------------------------------------------------
     * If the driver has accepted the instruction, wait for its completion
     */

    /* By not returning PENDING, the service refuses to process the
     * request. This may be caused by insufficient resources, bad
     * user data buffers or bad instruction.
     *
     * It it DOES return PENDING, it will DO call the completion
     * routine and we must wait for it.
     */
    if ( status != PWRCRYPT_STATUS_PENDING)
    {
        return FALSE;
    }

    if ( PCrptCanBlock())
    {
        /* Yield the CPU and wait for event signaling
         */
        PCrptWaitOnEvent( &CompletionEvent);
    }
    else
    {
        /*
         * Forgive me and suggest something better :-)
         * Spin in a loop checking for event signaling
         */
        while ( !PCrptIsEventSignaled( &CompletionEvent))
        {
        }
    }
    return TRUE;
}
