/**********************************************************************

    Module: driver.c

        Product: kernel mode driver providing user mode interface
        to the PwrCrypt driver for encryption acceleration board
        based on Hi/fn 7751 chip by Ashley Laurent, Inc.

    ---------------------------------------------------------------

    Copyright:

        Ashley Laurent, Inc., 1999
        All Rights Reserved.

    ---------------------------------------------------------------

    Functional Description:

        Implements standard driver entry points

    ---------------------------------------------------------------

    Environment:

        NT or BSD kernel and user mode

    ---------------------------------------------------------------

    Author:

        Stan Shkolnyy

    ---------------------------------------------------------------

    Revision history:

        *   16-Mar-99   initial revision

**********************************************************************/

#include <stdio.h>      // Needed for sprintf() routine.
#include <ntddk.h>
#include "debug.h"
#include "driver.h"
#include "instrchk.h"

#include <pwrcryptioctl.h>
#include <km_interface.h>
#include "calcreqpp.h"  // for CalcReqCheckAndPrepare()

#ifdef DBG

#define DEBUG_BUFFER_SIZE   512

ULONG DebugPrintLevel = 3;
UCHAR DebugBuffer[DEBUG_BUFFER_SIZE];

#endif

//
// Make sure the initialization code is removed from memory after use.
//

#ifdef ALLOC_PRAGMA
#pragma alloc_text(INIT, DriverEntry)
#endif

/****************************************************************************
 * Driver's global data
 */
DRIVER_GLOBALS  glob;

/****************************************************************************
****************************************************************************/
NTSTATUS
DriverGlobalsConstructor( PDRIVER_GLOBALS pThis, PUNICODE_STRING pusDriverRegPath)
{
    NTSTATUS    status;

    pThis->pDevObj = NULL;
    pThis->pPwrCryptFileObj = NULL;

    SlotTrackerConstructor( &pThis->SlotTracker);
    SymlinkConstructor( &pThis->SymLink);

    status = STATUS_SUCCESS;
    return status;
}

/****************************************************************************
****************************************************************************/
void
DriverGlobalsDestructor( PDRIVER_GLOBALS pThis, PDRIVER_OBJECT DriverObject)
{
    DebugPrint((3, "PwrCryptUnload\n"));

    SymlinkDestructor( &pThis->SymLink);

    if ( pThis->pDevObj)
    {
        IoDeleteDevice( pThis->pDevObj);
    }

    /* Free the PwrCrypt device
     */
    if ( pThis->pPwrCryptFileObj)
    {
        ObDereferenceObject( pThis->pPwrCryptFileObj);
    }

    SlotTrackerDestructor( &pThis->SlotTracker);
}


/****************************************************************************
****************************************************************************/
NTSTATUS
DriverEntry(
    IN PDRIVER_OBJECT DriverObject,
    IN PUNICODE_STRING RegistryPath
    )
/*++

Routine Description:

    Determine if a PCI device exists, setup relevant structures, 
    claim resources, connect interrupts, and report status back to
    the system.  Most of the work done is in routines called by
    DriverEntry.
    
Arguments:

    DriverObject - Pointer to the driver object created by the I/O manager.
    
    RegistryPath - Pointer to the driver specific key 
                   \Registry
                      \Machine
                         \System
                            \CurrentControlSet
                               \Services
                                  \<DriverName>
    
Return Value:

    NTSTATUS
    
--*/
{
    NTSTATUS                status;

    UCHAR                   szDeviceName[MAXIMUM_FILENAME_LENGTH];
    ANSI_STRING             ansiDeviceName;

    UNICODE_STRING          usDeviceName;   // used twice, temporarily, for dev.names

    PDEVICE_OBJECT          pPwrCryptDevObj;
    ULONG                   SlotsToTrack;

    /* The "get capabilities" request sent to PwrCrypt
     */
    IOCTL_PWRCRYPT_GET_CAPS_INPUT_OUTPUT    GetCapsPack;

    DebugPrint((0, "DriverEntry routine\n"));
    DebugPrint((3, "DriverObject 0x%x \n", DriverObject));
    
    /*
     * Initialize Russinovich's memory tracking library
     */
    MemTrackInit();

    /************************************************************************
     * Initialize the driver's permanent/global and
     * temp (for this routine only) data. The reason for this is that some
     * of this data requires memory deallocations and we don't want to spread
     * them throughout this routine.
     */
    status = DriverGlobalsConstructor( &glob, RegistryPath);
    if ( !NT_SUCCESS( status))
    {
        DebugPrint((1, "Unable to construct global objects. status=0x%08x\n", status));
        DriverGlobalsDestructor( &glob, DriverObject);
        return status;
    }

    /************************************************************************
     * Try to connect to the driver on which services we are dependent.
     * If its special linking device object can't be found, we can't continue.
     * There are 2 causes why it can't be found:
     *  - that driver's file is absent,
     *  - that driver was not loaded.
     * (Notice, that it can be loaded automatically for us, since this driver
     * is linked with that driver's import library. However, in this case that
     * driver's DriverEntry is not called and it is not known as a driver to
     * the O/S and so on. This is not how it is intended to operate and it
     * won't be able to create the device we are looking for. All this plays
     * for us to ensure the proper load order.
     */
    RtlInitUnicodeString( &usDeviceName, PWRCRYPT_LINK_DEVICE_UNICODE_NAME);
    status = IoGetDeviceObjectPointer( &usDeviceName,
                                       FILE_ALL_ACCESS, 
                                       &glob.pPwrCryptFileObj,
                                       &pPwrCryptDevObj);
    if ( !NT_SUCCESS( status))
    {
        DebugPrint((1, "Can't get ptr to PwrCrypt link device. Status = %x \n", status));
        DriverGlobalsDestructor( &glob, DriverObject);
        return status;
    }

    /* We don't need file object, so free it but reference the device object
     * before that
     */
//    status = ObReferenceObjectByPointer( pPwrCryptDevObj,
//                                         FILE_ALL_ACCESS,
//                                         NULL,
//                                         KernelMode);
//    ObDereferenceObject( pPwrCryptFileObj);
//    if ( !NT_SUCCESS( status))
//    {
//        DebugPrint((1, "Can't reference PwrCrypt link device by ptr. Status = %x \n", status));
//        return status;
//    }

    /************************************************************************
     * Now that we know we have proper access to PwrCrypt driver, call its
     * service to retrieve the maximum number of slots it exposes to clients.
     * Though we check the return status, the service should never normally
     * fail. Then, using the number of slots obtained from PwrCrypt,
     * initialize SlotTracker
     */
    if ( PwrcryptGetCaps( &GetCapsPack) != PWRCRYPT_STATUS_SUCCESS)
    {
        DebugPrint((1, "Can't get PwrCrypt capabilities. Status = %x \n", status));
        DriverGlobalsDestructor( &glob, DriverObject);
        return status;
    }

    SlotsToTrack = GetCapsPack.nNumOfDESSlots +
                   GetCapsPack.nNumOfRC4Slots +
                   GetCapsPack.nNumOfCompHistSlots;
    ASSERT( SlotsToTrack != 0);

    SlotTrackerInit( &glob.SlotTracker, SlotsToTrack);

    if ( !SlotTrackerIsValid( &glob.SlotTracker))
    {
        DebugPrint((1, "Can't initialize SlotTracker.\n"));
        DriverGlobalsDestructor( &glob, DriverObject);
        return STATUS_INSUFFICIENT_RESOURCES;
    }

    /************************************************************************
     * Set up the device driver entry points.
     */
    DriverObject->DriverUnload = PwrCryptUnload;

    DriverObject->MajorFunction[IRP_MJ_CREATE] = PwrCryptCreate;
    DriverObject->MajorFunction[IRP_MJ_CLOSE] = PwrCryptClose;
    DriverObject->MajorFunction[IRP_MJ_DEVICE_CONTROL] = PwrCryptIoctl;

    /************************************************************************
     * Create a unicode name for the device object and the object itself.
     * No numerical suffix is appended to the name, since we always provide
     * only 1 device.
     *
     * Note, that symbolic link to the device will be created later, to prevent
     * premature requests from clients
     */
    sprintf( szDeviceName, "\\Device\\"PWRCRYPT_UM_INTERFACE_DEVICE_NAME"");
    RtlInitAnsiString( &ansiDeviceName, szDeviceName);
    status = RtlAnsiStringToUnicodeString( &usDeviceName, &ansiDeviceName, TRUE);
    if ( !NT_SUCCESS( status))
    {
        DebugPrint((1, "Unable to create UNICODE device name. Status = %x \n", status));
        DriverGlobalsDestructor( &glob, DriverObject);
        return status;
    }

    // Remember on a return! Now allocated:
    //  - device name unicode string

    status = IoCreateDevice( DriverObject,
                             0,                     // no dev.ext
                             &usDeviceName,
                             FILE_DEVICE_UNKNOWN,
                             0,                      // No device-specific info
                             FALSE,                  // Not an exclusive device
                             &glob.pDevObj
                             );

    if ( !NT_SUCCESS( status))
    {
        DebugPrint((1, "Can't create device object. Status = %x \n", status));
        DriverGlobalsDestructor( &glob, DriverObject);
        RtlFreeUnicodeString( &usDeviceName);
        return status;
    }

    /************************************************************************
     * Create symbolic link to the device object to make it visible to user-mode
     * applications
     */
    status = SymlinkCreateSymbolicLink(
                    &glob.SymLink,
                    "\\DosDevices\\"PWRCRYPT_UM_INTERFACE_DEVICE_NAME"",
                    &usDeviceName);

    if ( !NT_SUCCESS( status))
    {
        DebugPrint((1, "Error creating symbolic link.  Status = %x \n", status));
        DriverGlobalsDestructor( &glob, DriverObject);
        RtlFreeUnicodeString( &usDeviceName);
        return status;
    }

    /************************************************************************
     * The device and the link are created - don't need the device name string
     */
    RtlFreeUnicodeString( &usDeviceName);

    return status;

} // DriverEntry




/****************************************************************************

NTSTATUS
PwrCryptCreate(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp
    )

Description:
    
Arguments:

    DeviceObject - Object representing a particular adapter.
    
    Irp - IRP_MJ_CREATE I/O request.
    
Return Value:

    NTSTATUS
****************************************************************************/
NTSTATUS
PwrCryptCreate(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp
    )
{
    PIO_STACK_LOCATION  pStackLoc;
    PFILE_OBJECT        pFileObj;

    pStackLoc = IoGetCurrentIrpStackLocation( Irp);
    pFileObj = pStackLoc->FileObject;

    DebugPrint((3, "PwrCryptCreate. File Object 0x%08x\n", pFileObj));

    Irp->IoStatus.Status = STATUS_SUCCESS;
    Irp->IoStatus.Information = 0;

    IoCompleteRequest (Irp, IO_NO_INCREMENT);

    return STATUS_SUCCESS;

}   // PwrCryptCreate



/****************************************************************************

NTSTATUS
PwrCryptClose(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp
    )

Description:

Arguments:

    DeviceObject - Object representing a particular adapter.

    Irp - IRP_MJ_CLOSE I/O request.

Return Value:

    NTSTATUS
****************************************************************************/
void
PCFreeSlotCallback(
    IN PVOID    Context,
    IN ULONG    SlotNumber
    )
{
    PFILE_OBJECT    pFileObj;

    BOOLEAN         bFreeSuccess;

    IOCTL_PWRCRYPT_FREE_SLOT_INPUT_OUTPUT FreeSlotPack;
    NTSTATUS        status;

    pFileObj = (PFILE_OBJECT) Context;

    /* Tell the SlotTracker to verify the passed slot number against
     * the key (file obj ptr) and free its remembered the association
     * between the reserved slot number and the file object for which
     * the IRP has been issued.
     *
     * Note! In this callback, the slot number will always be correct,
     * since it is given by SlotTracker itself. And it will always
     * be associated with the Context, since we use it as the key,
     * when asking the SlotTracker to enumerate slots by key.
     */
    bFreeSuccess = SlotTrackerUnregisterSlot(
                                    &glob.SlotTracker,
                                    SlotNumber,
                                    (ULONG) pFileObj,
                                    &FreeSlotPack.nSlotNumberToFree);
    /*
     * FreeSlot failure can only happen if the slot's RefCnt is not zero,
     * but this must not be the case here, because we are called from
     * IRP_MJ_CLOSE handler when there can't be outstanding IRPs for the
     * corresponding file object and hence its associated slots.
     */
    ASSERT( bFreeSuccess);

    /*
     * After SlotTrackerFreeSlot, nobody has access to the slot until it is
     * reserved again, but before that can happen, the slot must be freed via
     * the following call to the PwrCrypt service.
     */
    status = PwrcryptFreeSlot( &FreeSlotPack);

    ASSERT( NT_SUCCESS( status));
}

NTSTATUS
PwrCryptClose(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp
    )
{
    PIO_STACK_LOCATION  pStackLoc;
    PFILE_OBJECT        pFileObj;

    pStackLoc = IoGetCurrentIrpStackLocation( Irp);
    pFileObj = pStackLoc->FileObject;

    DebugPrint((3, "PwrCryptClose. File Object 0x%08x\n", pFileObj));

    /*
     * Free all slots associated with the given file object.
     * Note the trick below - we call the SlotTracket to enumerate slots by
     * key (file object ptr) and give it its own slot freeing method as
     * callback. 
     */
    SlotTrackerEnumSlotsByKey(
            &glob.SlotTracker,
            PCFreeSlotCallback,     // callback address
            (PVOID) pFileObj,       // Context
            (ULONG) pFileObj);      // Key

    Irp->IoStatus.Status = STATUS_SUCCESS;
    Irp->IoStatus.Information = 0;

    IoCompleteRequest (Irp, IO_NO_INCREMENT);

    return STATUS_SUCCESS;

}   // PwrCryptClose

/****************************************************************************

NTSTATUS
PwrCryptIoctl(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp
    )

Description:

    IOCTL dispather - the only entry for all of the work requests.

Arguments:

    DeviceObject - Object representing a particular adapter.

    Irp - IRP_MJ_CREATE or IRP_MJ_CLOSE I/O request.

Return Value:

    NTSTATUS

Remarks:

    IOCTL_PWRCRYPT_CALCULATE is the only request that may not be completed
    synchronously in this dispatching routine.

****************************************************************************/
NTSTATUS
PwrCryptIoctl(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp
    )
{
    PIO_STACK_LOCATION  pStackLoc;
    PFILE_OBJECT        pFileObj;

    ULONG               CtlCode;
    ULONG               InputLen, OutputLen;
    NTSTATUS            status;
    PMDL                pSrcMdl, pDstMdl;

    pStackLoc = IoGetCurrentIrpStackLocation( Irp);
    CtlCode = pStackLoc->Parameters.DeviceIoControl.IoControlCode;
    InputLen = pStackLoc->Parameters.DeviceIoControl.InputBufferLength;
    OutputLen = pStackLoc->Parameters.DeviceIoControl.OutputBufferLength;

    DebugPrint((3, "PwrCryptIoctl 0x%08x\n", CtlCode));

    pFileObj = pStackLoc->FileObject;
    DebugPrint((3, "File Object 0x%08x\n", pFileObj));

    /*
     * First, set up it so that no data would be copied to the IOCTL output
     * buffer, assume a failure. It'll be set to a non-zero, if the work will
     * complete successfully
     */
    Irp->IoStatus.Information = 0;

    /*
     * For future, remember that ALL our IOCTLs are BUFFERED - the input buffer
     * has been copied to a system-space non-paged buffer and we're working with
     * this buffer which is named "request" in the comments below. Since it is
     * in the system memory space, it is visible in an arbitrary context.
     * Irp->AssociatedIrp.SystemBuffer points to it.
     */

    switch ( CtlCode)
    {
        case IOCTL_PWRCRYPT_CALCULATE:
        {
            PIOCTL_PWRCRYPT_CALCULATE_INPUT_OUTPUT pCalcReq;
            PPWRCRYPT_CALCULATE_KM_WORKAREA pWA;

            pCalcReq = ( PIOCTL_PWRCRYPT_CALCULATE_INPUT_OUTPUT) Irp->AssociatedIrp.SystemBuffer;

            /*
             * Setup a pointer to the KM work area. Note that this routine
             * touches only KMWorkArea in the CalcReq, other members are filled
             * by the calling app.
             */
            pWA = (PPWRCRYPT_CALCULATE_KM_WORKAREA) &pCalcReq->KMWorkArea;

            /*
             * Verify IOCTL buffers' lengths, check Command-related members of
             * the CalcReq and fill Command-related members of CalcReq.KMWorkArea
             */
            if ( InputLen >= sizeof( *pCalcReq) &&
                 OutputLen <= InputLen &&
                 OutputLen >= sizeof( BIGGEST_RESULT) &&
                 CalcReqCheckAndPrepare( pCalcReq, InputLen))
            {
                if ( SlotTrackerReferenceSlot( &glob.SlotTracker,
                                               pCalcReq->nSlotNumber,
                                               (ULONG) pFileObj))
                {
                    /*
                     * Sounds good, we will start this request. We mark it pending
                     * and return STATUS_PENDING, so that I/O Mgr wouldn't care
                     * about it.
                     *
                     * The IRP is completed either in this driver's completion
                     * routine that is called by the PwrCrypt, or we complete
                     * it now, in this routine, if the PwrCrypt doesn't return
                     * STATUS_PENDING from its PwrcryptCalculate() service
                     */
                    IoMarkIrpPending( Irp);

                    /*
                     * The IOCTL input buffer contains ptrs to 2 data buffers.
                     * Generally, this driver can't know for certain whether
                     * these buffers are already locked, so it assumes they are
                     * not and sets the following flags. This causes the
                     * underlying PwrCrypt to _always_ lock them
                     */
                    pWA->flBufsDescType = ( PWRCRYPT_CALCULATE_IN_PAGE |
                                            PWRCRYPT_CALCULATE_OUT_PAGE);

                    /*
                     * It is of importance to correctly detect whether the
                     * buffers belong to the user-mode VA space. Otherwise, the
                     * locking of the buffers may unproperly fail or kernel
                     * memory be damaged by the application.
                     *
                     * We also use the same condition to decide whether to
                     * charge the current process quota, though it may not
                     * always be absolutely correct.
                     */
                    if ( Irp->RequestorMode == UserMode)
                    {
                        pWA->flBufsDescType |= ( PWRCRYPT_CALCULATE_IN_USER |
                                                 PWRCRYPT_CALCULATE_OUT_USER);
                        pWA->bChargeQuota = TRUE;
                    }
                    else
                    {
                        pWA->bChargeQuota = FALSE;
                    }

                    /*
                     * Report the completion routine address and its two parms
                     * to the service. It doesn't interpret the parms, just
                     * calls the routine and passes them to it, upon request
                     * completion. We find it convenient to see the Irp and
                     * CalcReq in our completion routine.
                     */
                    pWA->pComplRtn = CalculateCompletionRoutine;
                    pWA->nComplRtnParm1 = (ULONG) Irp;
                    pWA->nComplRtnParm2 = (ULONG) pCalcReq;

                    /*
                     * Call the PwrCrypt calculating service
                     */
                    status = PwrcryptCalculate( pCalcReq);

                    /* By not returning PENDING, the service refuses to process the
                     * request. This may be caused by insufficient resources or bad
                     * user data buffers.
                     *
                     * It it DOES return PENDING, it will DO call the completion
                     * routine and we must not touch the IRP anymore, since the
                     * completion routine _is_ to complete this IRP and, in fact,
                     * it is possible that it has been already completed by the
                     * time when the service returns. The status is checked at
                     * the bottom of this IOCTL dispatcher routine to decide
                     * whether to complete the IRP.
                     *
                     * If the status is not PENDING, our completion routine will
                     * not be called, so we must dereference the slot been
                     * referenced earlier for this failed operation, which is
                     * normally done by the completion routine.
                     */
                    if ( status != STATUS_PENDING)
                    {
                        if ( !SlotTrackerDereferenceSlot( &glob.SlotTracker,
                                                          pCalcReq->nSlotNumber,
                                                          (ULONG) pFileObj))
                        {
                            // BUGBUG: unexpected condition - slot remains "busy"
                            DebugPrint((3, "Unexpected error - slot dereferencing has failed. slot 0x04x\n", pCalcReq->nSlotNumber));
                        }
                    }
                }//if ReferenceSlot
                else
                {
                    DebugPrint((2, "Invalid slot number\n"));
                    status = STATUS_INVALID_PARAMETER;
                }
            }//if IOCTL parms are valid
            else
            {
                DebugPrint((2, "Invalid calculating request\n"));
                status = STATUS_INVALID_PARAMETER;
            }
            break;
        }

        case IOCTL_PWRCRYPT_RESERVE_SLOT:
        {
            PIOCTL_PWRCRYPT_RESERVE_SLOT_INPUT_OUTPUT pResSlotReq;

            pResSlotReq = ( PIOCTL_PWRCRYPT_RESERVE_SLOT_INPUT_OUTPUT) Irp->AssociatedIrp.SystemBuffer;

            /* Verify:
             *  - IOCTL buffers' lengths
             */
            if ( InputLen >= sizeof( *pResSlotReq) &&
                 InputLen == OutputLen)
            {
                /*
                 * Call the PwrCrypt slot reservation service first.
                 */
                status = PwrcryptReserveSlot( pResSlotReq);
                ASSERT( status != STATUS_PENDING);
                if ( NT_SUCCESS( status))
                {
                    BOOLEAN bRegSuccess;

                    /*
                     * Tell the SlotTracker to remember the association between
                     * the reserved slot number and the file object for which
                     * the IRP has been issued.
                     * Note that we also substitute a Slot Tracker index
                     * in the place of the PwrCrypt core slot number in the
                     * request packet, so the user mode caller receives
                     * Slot Tracker's index instead of PwrCrypt's slot number.
                     */
                    bRegSuccess = SlotTrackerRegisterSlot(
                                        &glob.SlotTracker,
                                        pResSlotReq->nReservedSlotNumber,
                                        (ULONG) pFileObj,
                                        &pResSlotReq->nReservedSlotNumber);

                    ASSERT( bRegSuccess);

                    /*
                     * Tell the I/O Manager to copy "reserve slot" request
                     * packet back to the user output buffer
                     */
                    Irp->IoStatus.Information = InputLen;
                }
            }
            else
            {
                DebugPrint((2, "Invalid IOCTL buffer(s') size(s)\n"));
                status = STATUS_INVALID_PARAMETER;
            }
            break;
        }

        case IOCTL_PWRCRYPT_FREE_SLOT:
        {
            PIOCTL_PWRCRYPT_FREE_SLOT_INPUT_OUTPUT pFreeSlotReq;

            pFreeSlotReq = ( PIOCTL_PWRCRYPT_FREE_SLOT_INPUT_OUTPUT) Irp->AssociatedIrp.SystemBuffer;

            /* Verify:
             *  - IOCTL buffers' lengths
             */
            if ( InputLen >= sizeof( *pFreeSlotReq) &&
                 InputLen == OutputLen)
            {
                ULONG CoreSlotNumber;

                /* Tell the SlotTracker to verify the passed slot number against
                 * the key (file obj ptr) and free its remembered association
                 * between the reserved slot number and the key (file object ptr
                 * for which the IRP has been issued).
                 * Note that we also substitute the PwrCrypt core slot number
                 * in the place of the Slot Tracker index in the request packet, so the user mode caller receives
                 * before calling to the PwrCrypt with that packet.
                 */
                if ( SlotTrackerUnregisterSlot(
                                    &glob.SlotTracker,
                                    pFreeSlotReq->nSlotNumberToFree,
                                    (ULONG) pFileObj,
                                    &pFreeSlotReq->nSlotNumberToFree))
                {
                    /*
                     * SlotTracker has said "Ok" - call the PwrCrypt service.
                     */
                    status = PwrcryptFreeSlot( pFreeSlotReq);
                    ASSERT( status != STATUS_PENDING);
                    if ( NT_SUCCESS( status))
                    {
                        /*
                         * Tell the I/O Manager not to copy "free slot" request
                         * packet back to the user output buffer since we
                         * return nothing.
                         */
                        Irp->IoStatus.Information = 0; //InputLen;
                    }
                }
                else
                {
                    DebugPrint((2, "Attempt to free a wrong slot\n"));
                    status = STATUS_INVALID_PARAMETER;
                }
            }
            else
            {
                DebugPrint((2, "Invalid IOCTL buffer(s') size(s)\n"));
                status = STATUS_INVALID_PARAMETER;
            }
            break;
        }

        case IOCTL_PWRCRYPT_GET_CAPS:
        {
            PIOCTL_PWRCRYPT_GET_CAPS_INPUT_OUTPUT pGetCapsReq;

            pGetCapsReq = ( PIOCTL_PWRCRYPT_GET_CAPS_INPUT_OUTPUT) Irp->AssociatedIrp.SystemBuffer;

            /*
             * Call the PwrCrypt service.
             */
            status = PwrcryptGetCaps( pGetCapsReq);
            ASSERT( status != STATUS_PENDING);
            if ( NT_SUCCESS( status))
            {
                /*
                 * Tell the I/O Manager to copy "get caps" request packet back
                 * to the user output buffer.
                 */
                Irp->IoStatus.Information = InputLen;
            }
            break;
        }

        default:
        {
            status = STATUS_INVALID_DEVICE_REQUEST;
            break;
        }
    }//switch

    /*
     * Remember, only the calculating service may return PENDING
     */
    if ( status != STATUS_PENDING)
    {
        Irp->IoStatus.Status = status;
        IoCompleteRequest (Irp, IO_NO_INCREMENT);
    }
    else
    {
        ASSERT( CtlCode == IOCTL_PWRCRYPT_CALCULATE);
    }

    return status;

}   // PwrCryptIoctl

/****************************************************************************

CalculateCompletionRoutine()

Description:

Arguments:

    nComplRtnParm1 - ptr to the IRP to complete
    nComplRtnParm2 - ptr to the Calc.Request.

Return Value:

****************************************************************************/
void
CalculateCompletionRoutine(
    ULONG   nComplRtnParm1,
    ULONG   nComplRtnParm2
    )
{
    PIO_STACK_LOCATION  pStackLoc;
    PFILE_OBJECT        pFileObj;

    ULONG               OutputLen;

    PIRP                Irp;
    PIOCTL_PWRCRYPT_CALCULATE_INPUT_OUTPUT pCalcReq;
    PPWRCRYPT_CALCULATE_KM_WORKAREA pWA;

    Irp = (PIRP) nComplRtnParm1;
    pCalcReq = (PIOCTL_PWRCRYPT_CALCULATE_INPUT_OUTPUT) nComplRtnParm2;

    pStackLoc = IoGetCurrentIrpStackLocation( Irp);
    pFileObj = pStackLoc->FileObject;

    /*
     * Setup a pointer to the KM work area for we need to know the size of Result
     */
    pWA = (PPWRCRYPT_CALCULATE_KM_WORKAREA) &pCalcReq->KMWorkArea;

    DebugPrint((3, "CalculateCompletionRoutine. Req=0x%08x FileObj=0x%08x\n", pCalcReq, pFileObj));

    /*
     * Dereference the slot referenced for the completed operation
     */
    if ( !SlotTrackerDereferenceSlot( &glob.SlotTracker,
                                      pCalcReq->nSlotNumber,
                                      (ULONG) pFileObj))
    {
        // BUGBUG: unexpected condition - slot remains "busy"
        DebugPrint((3, "Unexpected error - slot dereferencing has failed. slot 0x04x\n", pCalcReq->nSlotNumber));
    }

    /*
     * Obtain IOCTL inbuf len from the cur IRP stack loc.
     * We need it below.
     */
    OutputLen = pStackLoc->Parameters.DeviceIoControl.OutputBufferLength;

    /*
     * Command I/O Mgr to copy the 7751 Result which is located at the head
     * of the CalcReq copy in the system non-paged buffer, so that the client
     * application would see it
     */
    ASSERT( pCalcReq->nResultSize <= OutputLen);

    /*
     * Attention - this calculation is based on the fact that the nResultSize
     * and Result are the first fields of the CalcReq.
     * We need to copy back to user the nResultSize + the effective
     * part of the Result field, which size is indicated by the nResultSize.
     */
    ASSERT( FIELD_OFFSET(
                IOCTL_PWRCRYPT_CALCULATE_INPUT_OUTPUT,
                nResultSize) == 0);

    Irp->IoStatus.Information =
                FIELD_OFFSET( IOCTL_PWRCRYPT_CALCULATE_INPUT_OUTPUT, Result) +
                pCalcReq->nResultSize;

    if ( Irp->IoStatus.Information > OutputLen)
    {
        Irp->IoStatus.Information = OutputLen;
    }

    /*
     * BUGBUG - investigate the proper value of INCREMENT here
     */
    Irp->IoStatus.Status = STATUS_SUCCESS;
    IoCompleteRequest (Irp, IO_NO_INCREMENT);
}

/****************************************************************************

VOID
PwrCryptUnload(
    IN PDRIVER_OBJECT DriverObject
    )

Description:

    Prepare the driver for unload.  For all the device objects created
    for this driver, loop through each one, disconnect interrupts, unmap
    memory and I/O spaces, free memory, etc.

Arguments:

    DriverObject - Pointer to the driver object created by the I/O manager.

Return Value:

    None
****************************************************************************/
VOID
PwrCryptUnload(
    IN PDRIVER_OBJECT DriverObject
    )
{
    /*
     * Destruct global data
     */
    DriverGlobalsDestructor( &glob, DriverObject);
}
