/**********************************************************************

    Module: osindependence.h

        Product: kernel mode driver for "PowerCrypt" encryption
        acceleration board based on Hi/fn 7751 chip
        by Ashley Laurent, Inc.

    ---------------------------------------------------------------

    Copyright:

        Ashley Laurent, Inc., 1999
        All Rights Reserved.

    ---------------------------------------------------------------

    Functional Description:

        O/S-independent wrappers for O/S dependent API functions that
        O/S-independent code needs.

    ---------------------------------------------------------------

    Environment:

        NT or BSD kernel mode

    ---------------------------------------------------------------

    Author:

        Stan Shkolnyy

    ---------------------------------------------------------------

    Revision history:

        *   15-Feb-99   initial revision

**********************************************************************/

#ifndef __OSINDEPENDENCE_H__
#define __OSINDEPENDENCE_H__

/*
 * One and only of these two symbols must be defined by makefile to force
 * this header to include O/S-specific stuff
 * _NT_DRIVER_BUILD_
 * _BSD_DRIVER_BUILD_
 */

#define LOWORD(l)           ((unsigned short)(l))
#define HIWORD(l)           ((unsigned short)(((unsigned long)(l) >> 16) & 0xFFFF))

/* Autodetect stuff to make lee's life easier */

#if !defined(_NT_DRIVER_BUILD_) && !defined(_BSD_DRIVER_BUILD_)

#ifdef _WIN32
#define	_NT_DRIVER_BUILD_ 1
#elif defined(__FreeBSD__)
#define _BSD_DRIVER_BUILD_ 1
#elif defined(linux)
#define _BSD_DRIVER_BUILD 1
#define LINUX_NOT_BSD
#else
#error Can\'t seem to recognise your OS
#endif

#endif



/******************
 * NT
 */
#ifdef  _NT_DRIVER_BUILD_

#include <ntddk.h>
#include "memtrack.h"

/*---------------------------------------------------------------------------
 * Status codes defined by O/S and returned by the driver.
 * This driver doesn't define its own status codes, just aliases and uses
 * system ones, since the most of the errors are originated from system calls
 * and some system codes are suitable to describe our errors.
 */
typedef NTSTATUS                                    PWRCRYPT_STATUS;

typedef KPROCESSOR_MODE                             PWRCRYPT_PROCESSOR_MODE;

#define PWRCRYPT_SUCCESS( status)                   NT_SUCCESS( status)

#define PWRCRYPT_STATUS_SUCCESS                     STATUS_SUCCESS
#define PWRCRYPT_STATUS_PENDING                     STATUS_PENDING
#define PWRCRYPT_STATUS_INSUFFICIENT_RESOURCES      STATUS_INSUFFICIENT_RESOURCES
#define PWRCRYPT_STATUS_NO_MEMORY                   STATUS_NO_MEMORY
#define PWRCRYPT_STATUS_INVALID_USER_BUFFER         STATUS_INVALID_USER_BUFFER
#define PWRCRYPT_STATUS_INVALID_PARAMETER           STATUS_INVALID_PARAMETER
#define PWRCRYPT_STATUS_DEVICE_NOT_CONNECTED        STATUS_DEVICE_NOT_CONNECTED
#define PWRCRYPT_STATUS_NOT_IMPLEMENTED             STATUS_NOT_IMPLEMENTED

/*---------------------------------------------------------------------------
 * Memory
 */
#define PCrptZeroMemory( adr, len)              RtlZeroMemory( adr, len)
#define PCrptFillMemory( adr, len, val)         RtlFillMemory( adr, len, val)
#define PCrptCopyMemory( dstadr, srcadr, len)   RtlCopyMemory( dstadr, srcadr, len)
#define PCrptNonPagedMAlloc( size)              MemAllocatePool( NonPagedPool, size)
#define PCrptNonPagedMFree( adr)                MemFreePool( adr)

/*---------------------------------------------------------------------------
 * Doubly-linked list - unsafe, needs protection
 */
typedef LIST_ENTRY  PCRPT_LIST_ENTRY, *PPCRPT_LIST_ENTRY;

typedef struct _PCRPT_LIST
{
    LIST_ENTRY  ListHead;

} PCRPT_LIST, *PPCRPT_LIST;

void
PCrptListInit(
    IN OUT  PPCRPT_LIST pThis
    );

BOOLEAN
PCrptIsListEmpty(
    IN OUT  PPCRPT_LIST pThis
    );

void
PCrptListInsertTail(
    IN OUT  PPCRPT_LIST pThis,
    PPCRPT_LIST_ENTRY   ListEntry
    );

void
PCrptListInsertHead(
    IN OUT  PPCRPT_LIST pThis,
    PPCRPT_LIST_ENTRY   ListEntry
    );

PPCRPT_LIST_ENTRY
PCrptListRemoveHead(
    IN OUT  PPCRPT_LIST pThis
    );

/*---------------------------------------------------------------------------
 * Event synchronization
 *
 * Watch for PLARGE_INTEGER Timeout.
 * MS monkeys were not able to explain intelligibly that
 *    ( Timeout == NULL) means "wait for indefinite time" and
 *    ( Timeout != NULL && *Timeout == 0) means "return current event
 *    status and don't wait"
 */ 

typedef KEVENT          PCRPT_EVENT, *PPCRPT_EVENT;

/*
 * Initialize an event to non-signaled state
 */
#define PCrptInitEvent(pEvent)  KeInitializeEvent( pEvent, NotificationEvent, FALSE)

/*
 * Check event state
 */
#define PCrptIsEventSignaled( pEvent)   ( KeReadStateEvent( pEvent) != 0)

/*
 * Wait for event indefinitely
 */
#define PCrptWaitOnEvent( pEvent) \
\
    KeWaitForSingleObject( \
        pEvent, \
        Suspended,   /* KWAIT_REASON */ \
        KernelMode,  /* KPROCESSOR_MODE */ \
        FALSE,       /* BOOLEAN Alertable */ \
        NULL         /* PLARGE_INTEGER Timeout */ \
        )

/*
 * Signal an event
 */
#define PCrptSignalEvent( pEvent) \
\
    KeSetEvent( \
        pEvent, \
        IO_NO_INCREMENT,   /* priority increment */ \
        FALSE              /* no, we are not going to sleep after that */ \
        )

/*---------------------------------------------------------------------------
 * Spinlock synchronization
 */

typedef struct _PCRPT_SPIN_LOCK
{
    KSPIN_LOCK  SpinLock;
    KIRQL       OldIrql;

} PCRPT_SPIN_LOCK, *PPCRPT_SPIN_LOCK;

#define PCRPT_DECLARE_SPINLOCK( VarName)  PCRPT_SPIN_LOCK VarName

#define PCrptSpinLockInit( pThis) \
    KeInitializeSpinLock( &(pThis)->SpinLock)
#define PCrptSpinLockAcquire( pThis) \
    KeAcquireSpinLock( &(pThis)->SpinLock, &(pThis)->OldIrql)
#define PCrptSpinLockAcquireAtDpcLevel( pThis) \
    KeAcquireSpinLockAtDpcLevel( &(pThis)->SpinLock)
#define PCrptSpinLockRelease( pThis) \
    KeReleaseSpinLock( &(pThis)->SpinLock, (pThis)->OldIrql)
#define PCrptSpinLockReleaseFromDpcLevel( pThis) \
    KeReleaseSpinLockFromDpcLevel( &(pThis)->SpinLock)

/*---------------------------------------------------------------------------
 * Indicates whether the caller can wait on dispatcher objects
 */
#define PCrptCanBlock()   ( KeGetCurrentIrql() < DISPATCH_LEVEL) 

/*---------------------------------------------------------------------------
 * Make a delay specified in ms. Used only for making delays during hardware
 * initialization, not at normal operation time
 */
#define delay( ms)     KeStallExecutionProcessor( 1000*(ms))

/*---------------------------------------------------------------------------
 * memory-mapped register accessors
 */
#define READ_MEM_REG32( RegAddr)         READ_REGISTER_ULONG( RegAddr)
#define WRITE_MEM_REG32( RegAddr, Value) WRITE_REGISTER_ULONG( RegAddr, Value)

/*
 * The list describing physical memory that backs up a locked virtual memory
 * and some corresponding routines
 */
typedef MDL MEM_DESC_LIST, *PMEM_DESC_LIST;

#ifndef HW_BUG_SEARCH_BUILD

#define CNT_FROM_MDL( pMdl)             MmGetMdlByteCount( pMdl)
#define VIRT_ADDR_FROM_MDL( pMdl)       MmGetMdlVirtualAddress( pMdl)

#else

#define VIRT_ADDR_FROM_MDL( pMdl)       ((PUCHAR)MmGetMdlVirtualAddress( pMdl)+8*1024)
#define CNT_FROM_MDL( pMdl)             (8*1024)

#endif /* HW_BUG_SEARCH_BUILD */

#define ALLOCATE_MDL( nPages)           IoAllocateMdl( NULL,                \
                                                       nPages * PAGE_SIZE,  \
                                                       FALSE,               \
                                                       FALSE,               \
                                                       NULL)

#define FREE_MDL( pMdl)                 IoFreeMdl( pMdl)

#define PCrptBuildPartialMdl( pSrcMdl, pTargetMdl, StartVA, Length)         \
                IoBuildPartialMdl( pSrcMdl, pTargetMdl, StartVA, Length)

#define PCrptPrepareMdlForReuse( pMdl)  MmPrepareMdlForReuse( pMdl)

#endif /* _NT_DRIVER_BUILD_ */

/******************
 * BSD
 */
#ifdef  _BSD_DRIVER_BUILD_

#include <bsdddk.h>

//
// define the same macros as defined above for NT
//

/*---------------------------------------------------------------------------
 * Status codes defined by O/S and returned by the driver.
 * This driver doesn't define its own status codes, just aliases and uses
 * system ones, since the most of the errors are originated from system calls
 * and some system codes are suitable to describe our errors.
 */
typedef int                                         PWRCRYPT_STATUS;

typedef KPROCESSOR_MODE                             PWRCRYPT_PROCESSOR_MODE;

#define PWRCRYPT_SUCCESS( status)                   ( status == STATUS_SUCCESS)

#define PWRCRYPT_STATUS_SUCCESS                     STATUS_SUCCESS
#define PWRCRYPT_STATUS_PENDING                     STATUS_SUCCESS
#define PWRCRYPT_STATUS_INSUFFICIENT_RESOURCES      EAGAIN
#define PWRCRYPT_STATUS_NO_MEMORY                   ENOMEM
#define PWRCRYPT_STATUS_INVALID_USER_BUFFER         EFAULT
#define PWRCRYPT_STATUS_INVALID_PARAMETER           EINVAL
#define PWRCRYPT_STATUS_DEVICE_NOT_CONNECTED        ENXIO
#define PWRCRYPT_STATUS_NOT_IMPLEMENTED             ENODEV

/*---------------------------------------------------------------------------
 * Memory
 */
#define PCrptZeroMemory( pBuf, len)             bzero( pBuf, len)
#define PCrptFillMemory( pBuf, len, val) \
{ \
    int i; \
    for ( i = 0; i < len; i++) \
    { \
        ((unsigned char*)pBuf)[ i] = val; \
    } \
}

#define PCrptCopyMemory( dstadr, srcadr, len)   memcpy( dstadr, srcadr, len)
#define PCrptNonPagedMAlloc( size)              malloc( size, M_DEVBUF, M_NOWAIT)
#define PCrptNonPagedMFree( pBuf)               free( pBuf, M_DEVBUF)

/*---------------------------------------------------------------------------
 * Doubly-linked list - unsafe, needs protection
 */
typedef LIST_ENTRY  PCRPT_LIST_ENTRY, *PPCRPT_LIST_ENTRY;

typedef struct _PCRPT_LIST
{
    LIST_ENTRY  ListHead;

} PCRPT_LIST, *PPCRPT_LIST;

void
PCrptListInit(
    IN OUT  PPCRPT_LIST pThis
    );

BOOLEAN
PCrptIsListEmpty(
    IN OUT  PPCRPT_LIST pThis
    );

void
PCrptListInsertTail(
    IN OUT  PPCRPT_LIST pThis,
    PPCRPT_LIST_ENTRY   ListEntry
    );

void
PCrptListInsertHead(
    IN OUT  PPCRPT_LIST pThis,
    PPCRPT_LIST_ENTRY   ListEntry
    );

PPCRPT_LIST_ENTRY
PCrptListRemoveHead(
    IN OUT  PPCRPT_LIST pThis
    );

/*---------------------------------------------------------------------------
 * Event synchronization
 *
 * Watch for PLARGE_INTEGER Timeout.
 * MS monkeys were not able to explain intelligibly that
 *    ( Timeout == NULL) means "wait for indefinite time" and
 *    ( *Timeout == 0)  means "return current event status and don't wait"
 */

typedef struct _PCRPT_EVENT
{
    LONG   SignalState;

} PCRPT_EVENT, *PPCRPT_EVENT;

/*
 * Initialize an event to non-signaled state
 */
#define PCrptInitEvent( pEvent)         (pEvent)->SignalState = 0

/*
 * Check event state but _don't_ wait for it
 */
#define PCrptIsEventSignaled( pEvent)   ( (pEvent)->SignalState == 0)

/*
 * Wait for event indefinitely
 *
 * BUGBUG - clarify "priority".
 *
 * If the PCATCH flag is ORed to priority parameter, tsleep will return
 * EINTR otr ERESTART if a signal comes. Then the interrupted syscall
 * should return that error code to the caller (4.4BSD p.85).
 *
 * Looking at the tsleep() code in kern_synch.c I concluded that it returns
 * after sleeping at the same hardware priority level (spl) as it was entered.
 */
#define PCrptWaitOnEvent( pEvent) \
{ \
    long    OldSpl; \
    OldSpl = splhigh(); \
    while ( (pEvent)->SignalState == 0) \
    { \
        int nSleepResult; \
        nSleepResult = tsleep( pEvent, PRIBIO, "PcEvWt", 0); \
    } \
    splx( OldSpl); \
}

/*
 * Signal an event
 */
#define PCrptSignalEvent( pEvent) \
{ \
    (pEvent)->SignalState++; \
    wakeup( pEvent); \
}

/*---------------------------------------------------------------------------
 * Spinlock synchronization for FreeBSD 2.2.6
 *
 * This FreeBSD version is identical to Windows in the part of spinlocks -
 * it runs on a single processor and doesn't need them. Also, the kernel code
 * in both systems is never preempted, except if it induces paging or waits
 * on a sync.object.
 *
 * This permits this driver to implement spinlocks as "no-op" for this FreeBSD.
 * The driver never induces paging (FreeBSD kernel is non-pageable and driver's
 * own data too) or waits with a spinlock held. Such behavior is imposed by NT
 * and the only data that may be pageable are user data buffers.
 */
#define PCRPT_DECLARE_SPINLOCK( VarName)
#define PCrptSpinLockInit( pThis)
#define PCrptSpinLockAcquire( pThis)
#define PCrptSpinLockAcquireAtDpcLevel( pThis)
#define PCrptSpinLockRelease( pThis)
#define PCrptSpinLockReleaseFromDpcLevel( pThis)

/*---------------------------------------------------------------------------
 * Make a delay specified in ms. Used only for making delays during hardware
 * initialization, not at the normal operation time.
 */
#define delay( ms)     DELAY( 1000*(ms))

/*---------------------------------------------------------------------------
 * memory-mapped register accessors
 */
#define READ_MEM_REG32( RegAddr)         *( ( unsigned long*)RegAddr)
#define WRITE_MEM_REG32( RegAddr, Value) *( ( unsigned long*)RegAddr) = Value

/*
 * The list describing physical memory that backs up a locked virtual memory
 * and some corresponding routines
 */
typedef MDL MEM_DESC_LIST, *PMEM_DESC_LIST;

#define CNT_FROM_MDL( pMdl)             MmGetMdlByteCount( pMdl)
#define VIRT_ADDR_FROM_MDL( pMdl)       MmGetMdlVirtualAddress( pMdl)

#if 0

#define ALLOCATE_MDL( nPages)           AllocateMdl( NULL,                \
                                                       nPages * PAGE_SIZE,  \
                                                       FALSE,               \
                                                       FALSE,               \
                                                       NULL)

#define FREE_MDL( pMdl)                 FreeMdl( pMdl)

#define PCrptBuildPartialMdl( pSrcMdl, pTargetMdl, StartVA, Length)         \
                IoBuildPartialMdl( pSrcMdl, pTargetMdl, StartVA, Length)

#define PCrptPrepareMdlForReuse( pMdl)  MmPrepareMdlForReuse( pMdl)

#else

#define ALLOCATE_MDL( nPages)   (void*)printf("BUGBUG")
#define FREE_MDL( pMdl)            printf("BUGBUG")
#define PCrptBuildPartialMdl( pSrcMdl, pTargetMdl, StartVA, Length)    printf("BUGBUG")
#define PCrptPrepareMdlForReuse( pMdl)    printf("BUGBUG")

#endif /* HW_BUG_SEARCH_BUILD */

#endif /* _BSD_DRIVER_BUILD_ */

#endif /* __OSINDEPENDENCE_H__ */
