/**********************************************************************

    Module: symlink.c

        Product: kernel mode driver providing user mode interface
        to the PwrCrypt driver for encryption acceleration board
        based on Hi/fn 7751 chip by Ashley Laurent, Inc.

    ---------------------------------------------------------------

    Copyright:

        Ashley Laurent, Inc., 1999
        All Rights Reserved.

    ---------------------------------------------------------------

    Functional Description:

        Implements the SymLink object that represents the symbolic link
        created by the driver under DosDevices.

    ---------------------------------------------------------------

    Environment:

        NT or BSD kernel and user mode

    ---------------------------------------------------------------

    Author:

        Stan Shkolnyy

    ---------------------------------------------------------------

    Revision history:

        *   16-Mar-99   initial revision

**********************************************************************/

#include <ntddk.h>
#include "debug.h"
#include "driverdef.h"
#include "symlink.h"

#ifdef ALLOC_PRAGMA
#pragma alloc_text(INIT, SymlinkConstructor)
#endif

/****************************************************************************
void
SymlinkConstructor(
    IN PSYMLINK pThis
    )

Description:

    Constructs interface object

Arguments:

    pThis - ptr to uninitialized object

****************************************************************************/

void
SymlinkConstructor(
    IN PSYMLINK pThis
    )
{
    DebugPrint((2, "SymlinkConstructor\n"));
    pThis->Type = SYMLINK_SIGN;
    RtlInitUnicodeString( &pThis->usLinkName, NULL);
}

/****************************************************************************
NTSTATUS
SymlinkCreateSymbolicLink(
    IN PSYMLINK pThis,
    IN UCHAR *szLinkName,
    IN PUNICODE_STRING pusDeviceName
    )

Description:

    Allocates and remembers unicode string with symlink name.
    Creates symlink.
    Either both action succeed or both fail (no link - no remembered string)

Arguments:

    pThis - ptr to uninitialized object

****************************************************************************/

NTSTATUS
SymlinkCreateSymbolicLink(
    IN PSYMLINK pThis,
    IN UCHAR *szLinkName,
    IN PUNICODE_STRING pusDeviceName
    )
{
    NTSTATUS        status;             // func.ret.value
    ANSI_STRING     ansiLinkName;

    DebugPrint((2, "SymlinkCreateSymbolicLink %s\n", szLinkName));
    ASSERT( pThis->Type == SYMLINK_SIGN);

    RtlInitAnsiString( &ansiLinkName, szLinkName);

    /*
     * Allocate buffer (TRUE) for Unicode string.
     * Convert ANSI string to Unicode.
     */
    status = RtlAnsiStringToUnicodeString( &pThis->usLinkName, &ansiLinkName, TRUE);

    if ( NT_SUCCESS( status))
    {
        /*
         * New unicode string has been allocated and stored in the Symlink for future.
         * Create symbolic link
         */
        status = IoCreateSymbolicLink( &pThis->usLinkName, pusDeviceName);
        if ( !NT_SUCCESS( status))
        {
            DebugPrint((1, "Unable to create symbolic link.  Status = %x \n", status));

            /*
             * If link creation fails, don't forget to free string.
             */
            RtlFreeUnicodeString( &pThis->usLinkName);
            RtlInitUnicodeString( &pThis->usLinkName, NULL);
        }
    }
    else
    {
        DebugPrint((1, "Unable to allocate/convert unicode string for symbolic link.  Status = %x \n",
                    status
                    ));
        RtlInitUnicodeString( &pThis->usLinkName, NULL);
    }

    return status;
}

/****************************************************************************
void
SymlinkDeleteSymbolicLink(
    IN PSYMLINK pThis
    )

Description:

    Deletes symlink using link name remembered in the object.
    De-allocates the rememberd unicode string.

Arguments:

    pThis - ptr to Symlink object

****************************************************************************/

void
SymlinkDeleteSymbolicLink(
    IN PSYMLINK pThis
    )
{
    NTSTATUS        status;

    DebugPrint((2, "SymlinkDeleteSymbolicLink\n"));
    ASSERT( pThis->Type == SYMLINK_SIGN);

    ASSERT( pThis->usLinkName.Buffer != NULL);
    status = IoDeleteSymbolicLink( &pThis->usLinkName);
    ASSERT( NT_SUCCESS( status));
    RtlFreeUnicodeString( &pThis->usLinkName);
    RtlInitUnicodeString( &pThis->usLinkName, NULL);
}

/****************************************************************************
void
SymlinkDestructor(
    IN PSYMLINK pThis
    )

Description:

    Destructs interface object.
    (if the symlink name is found non-empty, deletes also symlink)

****************************************************************************/

void
SymlinkDestructor(
    IN PSYMLINK pThis
    )
{
    ASSERT( pThis->Type == SYMLINK_SIGN);

    if ( pThis->usLinkName.Buffer != NULL)
    {
        DebugPrint((2, "SymlinkDestructor - automatically calling SymlinkDeleteSymbolicLink\n"));
        SymlinkDeleteSymbolicLink( pThis);
    }
    else
    {
        DebugPrint((2, "SymlinkDestructor\n"));
    }
}
