/**********************************************************************

    Module: calcreqpp.c

        Product: kernel mode driver providing user mode interface
        to the PwrCrypt driver for encryption acceleration board
        based on Hi/fn 7751 chip by Ashley Laurent, Inc.

    ---------------------------------------------------------------

    Copyright:

        Ashley Laurent, Inc., 1999
        All Rights Reserved.

    ---------------------------------------------------------------

    Functional Description:

        Preprocessing of the user mode calc.request before forwarding
        it to the core driver via KM interface.

    ---------------------------------------------------------------

    Environment:

        NT or BSD kernel and user mode

    ---------------------------------------------------------------

    Author:

        Stan Shkolnyy

    ---------------------------------------------------------------

    Revision history:

        *   16-Mar-99   initial revision

**********************************************************************/

#include <osindependence.h>
#include "debug.h"
#include "instrchk.h"
#include <km_interface.h>
#include "calcreqpp.h"

/****************************************************************************
    Note! This routine _assumes_ that CalcReq is long enough to contain
    at least SMALLEST_COMMAND. The caller must ensure this.

    This procedure doesn't allocate resources, it merely fills ptrs in
    pCalcReq->KMWorkArea. So, it easily returns with FALSE in any place when
    it detects that the contents of the CalcReq is invalid.

****************************************************************************/
BOOLEAN
CalcReqCheckAndPrepare
    (
        IN OUT  PIOCTL_PWRCRYPT_CALCULATE_INPUT_OUTPUT  pCalcReq,
        IN      ULONG       nReqLen
    )
{
    USHORT  nCommSize;
    USHORT  CommCodeBits;
    BOOLEAN bCalcCommand;
    ULONG   nCtxOffset;

    PPWRCRYPT_CALCULATE_KM_WORKAREA pWA;

    DebugPrint((3, "CalcReqCheckAndPrepare 0x%08x, 0x%08x\n", pCalcReq, nReqLen));

    /*
     * Basic checks on data buffer lengths and alignment
     */
    if ( ( pCalcReq->nTotalSrcCnt & 0x00000003L) != 0 ||
         ( pCalcReq->nTotalDstCnt & 0x00000003L) != 0)
    {
        DebugPrint(( 1, "   not aligned data counter(s)\n"));
        return FALSE;
    }

    if ( pCalcReq->nTotalSrcCnt > PROCESSOR_MAX_XFER_SIZE ||
         pCalcReq->nTotalDstCnt > PROCESSOR_MAX_XFER_SIZE)
    {
        DebugPrint(( 1, "   too large data counter(s)\n"));
        return FALSE;
    }

    if ( pCalcReq->nTotalSrcCnt < 4 ||
         pCalcReq->nTotalDstCnt < 4)
    {
        DebugPrint(( 1, "   too small data counter(s)\n"));
        return FALSE;
    }

    if ( ((ULONG) pCalcReq->pSrcData & 0x00000003L) != 0 ||
         ((ULONG) pCalcReq->pDstBuffer & 0x00000003L) != 0)
    {
        DebugPrint(( 1, "   not aligned data buffer(s)\n"));
        return FALSE;
    }

    /*
     * Check command code and let only encode/decode commands pass
     */
    CommCodeBits = BASECOM_GET_COMCODE_BITS( (PBASE_COMMAND) &pCalcReq->Command);
    bCalcCommand = FALSE;
    switch ( CommCodeBits)
    {
        case BASECOM_UNSCRAMBLE:
            DebugPrint(( 3, "   DECODE command detected\n"));
            bCalcCommand = TRUE;
            break;
        case BASECOM_SCRAMBLE:
            DebugPrint(( 3, "   ENCODE command detected\n"));
            bCalcCommand = TRUE;
            break;
        case BASECOM_READ_RAM:
            DebugPrint(( 1, "   READ RAM command detected - stopping\n"));
            return FALSE;
            break;
        case BASECOM_WRITE_RAM:
            DebugPrint(( 1, "   WRITE RAM command detected - stopping\n"));
            return FALSE;
            break;
        default:
            DebugPrint(( 1, "   unknown command detected - stopping\n"));
            return FALSE;
            break;
    }

    /*
     * Check command bits - at least one of them must be set for calculation
     * command
     */
    if ( !( BASECOM_IS_COMP( (PBASE_COMMAND) &pCalcReq->Command)) &&
         !( BASECOM_IS_PAD( (PBASE_COMMAND) &pCalcReq->Command)) &&
         !( BASECOM_IS_MAC( (PBASE_COMMAND) &pCalcReq->Command)) &&
         !( BASECOM_IS_CRYP( (PBASE_COMMAND) &pCalcReq->Command)) &&
         bCalcCommand)
    {
        DebugPrint(( 1, "   no command bits set - stopping\n"));
        return FALSE;
    }

    /*
     * Find the Command size based on flags in its Base Command part
     * and see whether it fits into the CalcReq
     */
    nCommSize = CommandGetSizeByBaseCommand( (PBASE_COMMAND) &pCalcReq->Command);

    /*
     * Calculate where the Context part will be, if it is needed.
     * (This is not known at this time, but the value is useful
     * for another check)
     */
    nCtxOffset = FIELD_OFFSET(
                    IOCTL_PWRCRYPT_CALCULATE_INPUT_OUTPUT,
                    Command) +
                 nCommSize;

    /*
     * See whether a Command of calculated size fits into the CalcReq.
     * (nCtxOffset gives the first byte after the command)
     */
    if ( nCtxOffset >= nReqLen)
    {
        DebugPrint(( 1, "   CalcReq size is not enough to hold the specified command\n"));
        return FALSE;
    }

    /*
     * Note! The rest of this routine _assumes_ that CalcReq is long enough to
     * contain at least all the subcommands specified by Base Command flags.
     * This is important for CommandGet...ByFullCommand() functions.
     */

    /*
     * Setup a pointer to the KM work area for we need to fill in pointers to
     * MAC and Enc.keys inside of it and zero the work area for not to assign
     * zeroes to its members
     */
    pWA = (PPWRCRYPT_CALCULATE_KM_WORKAREA) &pCalcReq->KMWorkArea;
    PCrptZeroMemory( pWA, sizeof( *pWA));

    /*
     * The total size of the command with all its subcommands
     */
    pWA->nCommandSize = nCommSize;

    /*
     * The code below takes into account that the same enc.key can be placed
     * in differents offset in the context, depending on presence and size of
     * the MAC key. Also, enc.IV may not be present, depending on the enc.method
     *
     * Note, that the Context part is always specified, if MAC or Enc.
     * subcommands are present.
     */
/* */
/* Not needed since the area has been zeroed out earlier */
/*    pWA->pMACKey = NULL; */
/*    pWA->nMACKeySize = 0; */
/*    pWA->pEncKey = NULL; */
/*    pWA->nEncKeySize = 0; */
/*    pWA->pEncIV  = NULL; */
/*    pWA->nEncIVSize  = 0; */

    if ( BASECOM_IS_MAC( (PBASE_COMMAND) &pCalcReq->Command))
    {
        /*
         * If a MAC key is present, it is always first in the context
         */
        pWA->pMACKey = (PUCHAR) pCalcReq + nCtxOffset;

        CommandGetMACKeySizeByFullCommand( (PBASE_COMMAND) &pCalcReq->Command,
                                            &pWA->nMACKeySize);

        /*
         * Check whether it has size and fits into the CalcReq
         */
        if ( pWA->nMACKeySize == 0 ||
             nCtxOffset + pWA->nMACKeySize > nReqLen)
        {
            DebugPrint(( 1, "   CalcReq size is not enough to hold the specified MAC key or invalid command\n"));
            return FALSE;
        }
    }

    if ( BASECOM_IS_CRYP( (PBASE_COMMAND) &pCalcReq->Command))
    {
        /*
         * If an enc.key is present, it always follows a MAC key in the context
         */
        pWA->pEncKey = (PUCHAR) pCalcReq + nCtxOffset + pWA->nMACKeySize;

        CommandGetEncKeySizeByFullCommand( (PBASE_COMMAND) &pCalcReq->Command,
                                           &pWA->nEncKeySize,
                                           &pWA->nEncIVSize);

        /*
         * Check whether it has size and fits into the CalcReq
         */
        if ( pWA->nEncKeySize == 0 ||
             nCtxOffset + pWA->nMACKeySize + pWA->nEncKeySize > nReqLen)
        {
            DebugPrint(( 1, "   CalcReq size is not enough to hold the specified enc.key or invalid command\n"));
            return FALSE;
        }

        /*
         * We are in the encryption branch, set ptr to enc.IV.
         * If it is present, it immediately follows the enc. key
         */
        if ( pWA->nEncIVSize != 0)
        {
            pWA->pEncIV  = pWA->pEncKey + pWA->nEncKeySize;
        }
    }
    return TRUE;
}
