/**********************************************************************

    Module: instrchk.c

        Product: kernel mode driver providing user mode interface
        to the PwrCrypt driver for encryption acceleration board
        based on Hi/fn 7751 chip by Ashley Laurent, Inc.

    ---------------------------------------------------------------

    Copyright:

        Ashley Laurent, Inc., 1999
        All Rights Reserved.

    ---------------------------------------------------------------

    Functional Description:

        Implements utility functions for checking 7751 instructions

    ---------------------------------------------------------------

    Environment:

        NT or BSD kernel and user mode

    ---------------------------------------------------------------

    Author:

        Stan Shkolnyy

    ---------------------------------------------------------------

    Revision history:

        *   16-Mar-99   initial revision

**********************************************************************/

#include <osindependence.h>
#include "debug.h"
#include <instrchk.h>

/****************************************************************************

Description:
    Based on the subcommand flags, calculates the expected total size of the
    command (which includes Base Command and optional subcommands, as indicated
    by the flags).

Return value:
    Calculated size

Remarks:
    The order of Command subcomponents is always as follows, but some
    subcomponents can be omitted.

typedef struct _BIGGEST_COMMAND
{
    BASE_COMMAND    base;
    COMP_COMMAND    comp;
    PAD_COMMAND     pad;
    MAC_COMMAND     mac;
    CRYP_COMMAND    cryp;
    UCHAR           MACKey  [  64];
    union
    {
        struct
        {
            UCHAR           EncKey  [  8];
            UCHAR           EncIV   [  8];
        } DESKeys;

        struct
        {
            UCHAR           EncKey  [ 24];
            UCHAR           EncIV   [  8];
        } TripleDESKeys;

        struct
        {
            UCHAR           EncKey  [ 260];
        } RC4Keys;

    } EncryptionContext;

} BIGGEST_COMMAND, *PBIGGEST_COMMAND;

typedef struct _SMALLEST_COMMAND
{
    BASE_COMMAND    base;
    UCHAR           zerobytes[ 8];

} SMALLEST_COMMAND, *PSMALLEST_COMMAND;

****************************************************************************/
USHORT
CommandGetSizeByBaseCommand
    (
        PBASE_COMMAND   pBC
    )
{
    USHORT  Size;

    Size = 0;

    if ( BASECOM_IS_COMP( pBC))
    {
        Size += sizeof( COMP_COMMAND);
    }

    if ( BASECOM_IS_PAD( pBC))
    {
        Size += sizeof( PAD_COMMAND);
    }

    if ( BASECOM_IS_MAC( pBC))
    {
        Size += sizeof( MAC_COMMAND);
    }

    if ( BASECOM_IS_CRYP( pBC))
    {
        Size += sizeof( CRYP_COMMAND);
    }

    Size += sizeof( BASE_COMMAND);

    /*
     * If only thee base command specified, we will need to append 8 bytes of zero,
     * If base+pad commands specified - 4 bytes of zero.
     * In any of these two cases, the resulting total size of the command
     * will be the SMALLEST_COMMAND size
     */
    if ( Size < sizeof( SMALLEST_COMMAND))
    {
        ASSERT( ( sizeof( SMALLEST_COMMAND) - Size == 4) ||
                ( sizeof( SMALLEST_COMMAND) - Size == 8));

        Size = sizeof( SMALLEST_COMMAND);
    }

    return Size;
}

/****************************************************************************
****************************************************************************/
void
CommandGetMACKeySizeByFullCommand
    (
        IN  PBASE_COMMAND   pBC,
        OUT PUSHORT         pnKeySize
    )
{
    /*
     * MAC key size is fixed
     */
    ASSERT( BASECOM_IS_MAC( pBC));
    *pnKeySize = _7751_MAC_KEY_SIZE;
}

/****************************************************************************

    if *pnKeySize is 0 on return - an invalid cryptocommand was passed

****************************************************************************/
void
CommandGetEncKeySizeByFullCommand
    (
        IN  PBASE_COMMAND   pBC,
        OUT PUSHORT         pnKeySize,
        OUT PUSHORT         pnIVSize
    )
{
    ULONG           nCrypCommOffset;
    PCRYP_COMMAND   pCC;

    ASSERT( BASECOM_IS_CRYP( pBC));

    *pnKeySize  = 0;
    *pnIVSize   = 0;

    /*
     * Calculate the pointer to the Crypto command 
     */
    nCrypCommOffset = sizeof( BASE_COMMAND);

    if ( BASECOM_IS_COMP( pBC))
    {
        nCrypCommOffset += sizeof( COMP_COMMAND);
    }

    if ( BASECOM_IS_PAD( pBC))
    {
        nCrypCommOffset += sizeof( PAD_COMMAND);
    }

    if ( BASECOM_IS_MAC( pBC))
    {
        nCrypCommOffset += sizeof( MAC_COMMAND);
    }

    pCC = (PCRYP_COMMAND) ( (ULONG) pBC + nCrypCommOffset);

    /*
     * See what a Crypto command is this and determine the sizes of the key
     * and IV, which depend on the command.
     */
    switch ( pCC->Command & CRYPCOM_ALGO_BITMASK)
    {
        case CRYPCOM_ALGO_DES:
        {
            *pnKeySize = _7751_DES_KEY_SIZE;

            /*
             * For a DES, see whether an IV is needed
             */
            if ( ( pCC->Command & CRYPCOM_MODE_BITMASK) != CRYPCOM_DES_MODE_ECB)
            {
                *pnIVSize = _7751_DES_IV_SIZE;
            }
            break;
        }

        case CRYPCOM_ALGO_3DES:
        {
            *pnKeySize = _7751_3DES_KEY_SIZE;

            /*
             * For a DES, see whether an IV is needed
             */
            if ( ( pCC->Command & CRYPCOM_MODE_BITMASK) != CRYPCOM_DES_MODE_ECB)
            {
                *pnIVSize = _7751_3DES_IV_SIZE;
            }
            break;
        }

        case CRYPCOM_ALGO_RC4:
        {
            *pnKeySize = _7751_RC4_KEY_SIZE;
            break;
        }
    }

    /*
     * If none of the "cases" above has been chosen, enc.key size remains = 0
     * which is a failure indicator for the caller
     */
}
