/**********************************************************************

    Module: km_interface.c

        Product: kernel mode driver for "PowerCrypt" encryption
        acceleration board based on Hi/fn 7751 chip
        by Ashley Laurent, Inc.

    ---------------------------------------------------------------

    Copyright:

        Ashley Laurent, Inc., 1999
        All Rights Reserved.

    ---------------------------------------------------------------

    Functional Description:

        Implementation of the service entry points exported by the driver
        for kernel-mode clients.

    ---------------------------------------------------------------

    Environment:

        NT or BSD kernel mode

    ---------------------------------------------------------------

    Author:

        Stan Shkolnyy

    ---------------------------------------------------------------

    Revision history:

        *   15-Feb-99   initial revision

**********************************************************************/

#include "osindependence.h"
#include "pwrcryptioctl.h"
#include "km_interface.h"
#include "kmi_private.h"
#include "br_comm_private.h"
#include "debug.h"
#include "userbuf.h"
#include "globobj.h"

/****************************************************************************
    This var is FALSE when the KM interface is disabled. It is important that
    it is initialized at compilation time, so that after the driver loading
    the KM interface would be disabled until explicitly enabled.

    It is enabled in the course of the DriverEntry routine and disabled
    again in Unload. The need of the flag arises due to the DLL model
    of the KM interface. It is used to prevent clients from entering the
    driver until it is ready to work.
****************************************************************************/
BOOLEAN bIsKmInterfaceEnabled = FALSE;

/****************************************************************************
****************************************************************************/
void
KmInterfaceEnable(
    )
{
    bIsKmInterfaceEnabled = TRUE;
}

/****************************************************************************
****************************************************************************/
void
KmInterfaceDisable(
    )
{
    bIsKmInterfaceEnabled = FALSE;
}

/****************************************************************************
****************************************************************************/
#ifdef DBG

BOOLEAN
CheckAlign4( ULONG pBuf, BOOLEAN bMdlPassed)
{
    if ( bMdlPassed)
    {
        pBuf = (ULONG) VIRT_ADDR_FROM_MDL( (PMEM_DESC_LIST) pBuf);
    }
    return ( ( pBuf & 0x00000003L) == 0);
}

#endif
/****************************************************************************

PwrcryptCalculate()

Description:

    Executes the calculation request by processing the user data in the input
    buffer and placing the result data into the output buffer as specified by
    the request.

    Notice, that only this service queues its requests, since only these
    requests require hardware processing. Other services process their requests
    immediately and results are returned to callers without any delay.

Arguments:

    pCalcReq -
        ptr to calculation request, located in non-pageable system memory,
        accessible in arbitrary context

Return Value:

    status code defined by the driver - merely indicates whether the request
    reached the hardware and was executed. It doesn't indicate the total success
    Of the operation but it does indicate whether the needed resources were
    available. Even if the function reports a success, check Result in the
    request itself for a calculation-specific error code. 

Processor IRQL in NT:
    If any of two data buffers is in pageable memory, < DISPATCH_LEVEL,
    otherwise, <= DISPATCH_LEVEL

****************************************************************************/
PWRCRYPT_STATUS
PwrcryptCalculate(
    PIOCTL_PWRCRYPT_CALCULATE_INPUT_OUTPUT  pCalcReq
    )
{
    PWRCRYPT_STATUS     status;        /* func.ret.value */
    PMEM_DESC_LIST      pSrcMdl, pDstMdl;
    PPWRCRYPT_CALCULATE_KM_WORKAREA pWA;
    BOOLEAN bMdl;

    DebugPrint((3, "PwrcryptCalculate. Rq=0x%08x\n", pCalcReq));

    if ( !bIsKmInterfaceEnabled)
    {
        return PWRCRYPT_STATUS_DEVICE_NOT_CONNECTED;
    }

    /*
     * Setup a pointer to the KM work area
     */
    pWA = (PPWRCRYPT_CALCULATE_KM_WORKAREA) &pCalcReq->KMWorkArea;

    /*
     * Let's ensure that the work area fits into the space reserved for
     * it in the IOCTL_PWRCRYPT_CALCULATE_INPUT_OUTPUT packet
     */
    ASSERT( sizeof( pCalcReq->KMWorkArea) >= sizeof( *pWA));

#ifdef DBG
    /**********************************
     * Check the passed data buffers on 4-byte alignment, since they'll be used
     * by the 7751 directly and it requires all buffers to be 32-bit aligned.
     * Note, that there is no need to check the alignment of the 7751
     * instruction passed in the request, because it is copied to an aligned
     * location and the hardware accesses it in that location.
     */
    bMdl = ( ( pWA->flBufsDescType & PWRCRYPT_CALCULATE_IN_MDL) != 0);
    if ( !CheckAlign4( (ULONG) pCalcReq->pSrcData, 
                       bMdl))
    {
        return PWRCRYPT_STATUS_INVALID_PARAMETER;
    }
    bMdl = ( ( pWA->flBufsDescType & PWRCRYPT_CALCULATE_OUT_MDL) != 0);
    if ( !CheckAlign4( (ULONG) pCalcReq->pDstBuffer,
                       bMdl))
    {
        return PWRCRYPT_STATUS_INVALID_PARAMETER;
    }
#endif

    status = PWRCRYPT_STATUS_SUCCESS;

    /* This initialization is needed for the cleanup. The vars indicate whether
     * the MDLs were created in the course of this routine
     */
    pSrcMdl = NULL;
    pDstMdl = NULL;

    /**********************************
     * Create MDLs for those data buffers that are not already represented
     * by MDLs (MDL passing is not indicated as buffer type)
     */

    if ( !( pWA->flBufsDescType & PWRCRYPT_CALCULATE_IN_MDL))
    {
        PWRCRYPT_PROCESSOR_MODE ReqMode;

        ReqMode = ( pWA->flBufsDescType & PWRCRYPT_CALCULATE_IN_USER) ? UserMode : KernelMode;

        status = LockBufferCreateMdl(
                    pCalcReq->pSrcData,
                    pCalcReq->nTotalSrcCnt,
                    pWA->bChargeQuota,
                    ReqMode,
                    FALSE,        /* ask for read access to the buffer */
                    &pSrcMdl);
    }

    if ( status == PWRCRYPT_STATUS_SUCCESS &&
         !( pWA->flBufsDescType & PWRCRYPT_CALCULATE_OUT_MDL))
    {
        PWRCRYPT_PROCESSOR_MODE ReqMode;

        ReqMode = ( pWA->flBufsDescType & PWRCRYPT_CALCULATE_OUT_USER) ? UserMode : KernelMode;

        status = LockBufferCreateMdl(
                    pCalcReq->pDstBuffer,
                    pCalcReq->nTotalDstCnt,
                    pWA->bChargeQuota,
                    ReqMode,
                    TRUE,         /* ask for write access to the buffer */
                    &pDstMdl);

        if ( status != PWRCRYPT_STATUS_SUCCESS && pSrcMdl != NULL)
        {
            UnlockBufferDelMdl( pSrcMdl);
        }
    }

    /**********************************
     * If an MDL creation has failed, return immediately.
     * No resources were allocated in that case, so just return
     */
    if ( status != PWRCRYPT_STATUS_SUCCESS)
    {
        return status;
    }

    /**********************************
     * Now that the user data buffers are Ok - locked in memory, nothing
     * prevents us from passing the request to the hardware for processing.
     */

    /**********************************
     * If any MDLs have been created by this routine, store their addresses in
     * the place of source virtual addresses. Those VAs are therefore re-written
     * but can be always restored by information in corresponding MDLs. On
     * request completion, re-written VAs are restored and MDLs are deleted.
     *
     * After this operation, the buffer representation becomes "normalized",
     * with MDLs for data buffers, instead of virtual addresses.
     */
    if ( pSrcMdl != NULL)
    {
        pCalcReq->pSrcData = pSrcMdl;
    }

    if ( pDstMdl != NULL)
    {
        pCalcReq->pDstBuffer = pDstMdl;
    }
    
    ASSERT( pCalcReq->pSrcData);
    ASSERT( pCalcReq->pDstBuffer);

    /**********************************
     * Do some request pre-processing to tell the DISPATCHER whether the
     * request carries "new key" indication which means that the indicated
     * key(s) must be sent to the processor, or it carries "clear encrypt
     * context" indicator which means the command won't affect enc.context
     * and therefore won't affect slot preferred affinity.
     *
     * This pre-processing results in filling pWA->BriefComm member. The
     * DISPATCHER looks into this field to determine where this request
     * is to go ( see StartRequests() routine in the dispatcher.c).
     *
     * This action is essentially the creation of a "digest" of the Command that
     * contains the information important for the DISPATCHER, thus saving it
     * from parsing the Command.
     */
    BriefCommandCreateByFullCommand( &pWA->BriefComm, (PBASE_COMMAND) &pCalcReq->Command);

    /**********************************
     * Initiate execution of the request
     */
    #ifdef _NO_HW_
        /*
         * For testing only: if we're not to touch the hardware, complete it now
         */
        KmInterfaceCompleteCalcRequest( pCalcReq);
    #else
        /*
         * Forward the request to the DISPATCHER
         */
        DispatcherStartRequest( &glob.Dispatcher, pCalcReq);
    #endif

    /**********************************
     * We must return PENDING status to the caller, to indicate that we've
     * accepted this CalcReq for execution and took the responsibility to
     * call the completion routine specified by the caller in the CalcReq.
     * (KmInterfaceCompleteCalcRequest() calls this user routine).
     * In fact, by this time, the request could have been already completed
     * and the routine called.
     */
    return PWRCRYPT_STATUS_PENDING;
}

/****************************************************************************
    KmInterfaceCompleteRequest indicates the caller has completed all processing
    for the given request and is returning the given Calcreq to the KM interface.
    KM interface performs necessary cleanup and notifies the client.
    This service is PwrCrypt internal and is called by a PROCESSOR (or by the
    PwrcryptCalculate() if we debug).

****************************************************************************/

void
KmInterfaceCompleteCalcRequest
    (
        IN      PIOCTL_PWRCRYPT_CALCULATE_INPUT_OUTPUT  pCalcReq
    )
{
    PPWRCRYPT_CALCULATE_KM_WORKAREA pWA;

    DebugPrint((3, "KMInterfaceCompleteCalcRequest\n"));

    /*
     * Setup a pointer to the KM work area
     */
    pWA = (PPWRCRYPT_CALCULATE_KM_WORKAREA) &pCalcReq->KMWorkArea;

    /*
     * Let's ensure that the structure fits into the space reserved for
     * it in the IOCTL_PWRCRYPT_CALCULATE_INPUT_OUTPUT packet
     */
    ASSERT( sizeof( pCalcReq->KMWorkArea) >= sizeof( *pWA));

    /**********************************
     * If any MDLs were created by this driver (MDL passing is not indicated as
     * buffer type), restore original values of buffer ptrs in the request that
     * we temporarily abused by replacing them with MDL ptrs and deallocate
     * those corresponding MDLs.
     */
    if ( !( pWA->flBufsDescType & PWRCRYPT_CALCULATE_IN_MDL))
    {
         /* We created the source data buf's MDL
          */
         PMEM_DESC_LIST pMdl;

         pMdl = pCalcReq->pSrcData;
         pCalcReq->pSrcData = VIRT_ADDR_FROM_MDL( pMdl);
/*         pCalcReq->nTotalSrcCnt = CNT_FROM_MDL( pMdl); */
         UnlockBufferDelMdl( pMdl);
    }

    if ( !( pWA->flBufsDescType & PWRCRYPT_CALCULATE_OUT_MDL))
    {
         /* We created the destination data buf's MDL
          */
         PMEM_DESC_LIST pMdl;

         pMdl = pCalcReq->pDstBuffer;
         pCalcReq->pDstBuffer = VIRT_ADDR_FROM_MDL( pMdl);
/*         pCalcReq->nTotalDstCnt = CNT_FROM_MDL( pMdl); */
         UnlockBufferDelMdl( pMdl);
    }

    /**********************************
     * Call the user completion routine specified in the request
     */
    (*pWA->pComplRtn)( pWA->nComplRtnParm1, pWA->nComplRtnParm2);
}

/****************************************************************************

PwrcryptReserveSlot()

Description:

    Reserves a slot in the 7751 internal context RAM. This service is
    implemented in software and doesn't involve any hardware operations.

Arguments:

    pReserveSlotReq -
        ptr to slot reservation request

Return Value:

    status code defined by the driver

****************************************************************************/
PWRCRYPT_STATUS
PwrcryptReserveSlot(
    PIOCTL_PWRCRYPT_RESERVE_SLOT_INPUT_OUTPUT   pReserveSlotReq
    )
{
    DebugPrint((3, "PwrcryptReserveSlot: pReserveSlotReq=0x%x\n", pReserveSlotReq));

    
    if ( !bIsKmInterfaceEnabled)
    {
	DebugPrint((3, "        KM not enabled!\n")); 
        return PWRCRYPT_STATUS_DEVICE_NOT_CONNECTED;
    }


    /**********************************
     * Check the slot capabilities the caller wants
     */

    /* */
    /* BUGBUG - not always, only because currently it's hardcoded to use 512-byte slots */
    /* */
    DebugPrint((3,"        checking slot capabilities\n"));
    if ( pReserveSlotReq->fCapabilities & SLOT_CAP_FLAG_STORE_COMP_HISTORY)
    {
	DebugPrint((3,"        sorry, no compression history allowed\n"));
        return PWRCRYPT_STATUS_NOT_IMPLEMENTED;
    }

    /**********************************
     * Ask SlotMan to reserve a slot
     */
    DebugPrint((3,"        calling slotman\n"));

    if ( !SlotmanReserveSlot( &glob.SlotMan, &pReserveSlotReq->nReservedSlotNumber))
    {
        return PWRCRYPT_STATUS_INSUFFICIENT_RESOURCES;
    }

    return PWRCRYPT_STATUS_SUCCESS;
}

/****************************************************************************

PwrcryptFreeSlot()

Description:

    Frees a previously reserved slot in the 7751 internal context RAM. This
    service is implemented in software and doesn't involve any hardware
    operations.

Arguments:

    pFreeSlotReq -
        ptr to slot freeing request

Return Value:

    status code defined by the driver

****************************************************************************/
PWRCRYPT_STATUS
PwrcryptFreeSlot(
    PIOCTL_PWRCRYPT_FREE_SLOT_INPUT_OUTPUT  pFreeSlotReq
    )
{
    DebugPrint((3, "PwrcryptFreeSlot\n"));

    if ( !bIsKmInterfaceEnabled)
    {
        return PWRCRYPT_STATUS_DEVICE_NOT_CONNECTED;
    }

    /**********************************
     * Ask SlotMan to free the slot
     */
    SlotmanFreeSlot( &glob.SlotMan, pFreeSlotReq->nSlotNumberToFree);

    return PWRCRYPT_STATUS_SUCCESS;
}

/****************************************************************************

PwrcryptGetCaps()

Description:

    Reports driver capabilities to implement specific configuration-dependent
    operations.

Arguments:

    pGetCapsReq -
        ptr to capabilities checking request

Return Value:

    status code defined by the driver

****************************************************************************/
PWRCRYPT_STATUS
PwrcryptGetCaps(
    PIOCTL_PWRCRYPT_GET_CAPS_INPUT_OUTPUT  pGetCapsReq
    )
{
    DebugPrint((3, "PwrcryptGetCaps\n"));

    if ( !bIsKmInterfaceEnabled)
    {
        return PWRCRYPT_STATUS_DEVICE_NOT_CONNECTED;
    }

    /* */
    /* BUGBUG */
    /* */
    pGetCapsReq->nNumOfDESSlots         = 0;
    pGetCapsReq->nNumOfRC4Slots         = SlotmanGetNumOfSlots( &glob.SlotMan);
    pGetCapsReq->nNumOfCompHistSlots    = 0;

/*    pGetCapsReq->nNumOfDESSlots         = SlotmanGetNumOfDESSlots( &glob.SlotMan); */
/*    pGetCapsReq->nNumOfRC4Slots         = SlotmanGetNumOfRC4Slots( &glob.SlotMan); */
/*    pGetCapsReq->nNumOfCompHistSlots    = SlotmanGetNumOfCompHistSlots( &glob.SlotMan); */

    return PWRCRYPT_STATUS_SUCCESS;
}
