/**********************************************************************

    Module: osindependence.h

        Product: kernel mode driver for "PowerCrypt" encryption
        acceleration board based on Hi/fn 7751 chip
        by Ashley Laurent, Inc.

    ---------------------------------------------------------------

    Copyright:

        Ashley Laurent, Inc., 1999
        All Rights Reserved.

    ---------------------------------------------------------------

    Functional Description:

        O/S-independent wrappers for O/S dependent API functions that
        O/S-independent code needs.

    ---------------------------------------------------------------

    Environment:

        NT or BSD kernel mode

    ---------------------------------------------------------------

    Author:

        Stan Shkolnyy

    ---------------------------------------------------------------

    Revision history:

        *   15-Feb-99   initial revision

**********************************************************************/

#ifndef __OSINDEPENDENCE_H__
#define __OSINDEPENDENCE_H__

/*
 * One and only of these two symbols must be defined by makefile to force
 * this header to include O/S-specific stuff
 * _NT_DRIVER_BUILD_
 * _BSD_DRIVER_BUILD_
 */

#define LOWORD(l)           ((unsigned short)(l))
#define HIWORD(l)           ((unsigned short)(((unsigned long)(l) >> 16) & 0xFFFF))

/* Autodetect stuff to make lee's life easier */

#if !defined(_NT_DRIVER_BUILD_) && !defined(_BSD_DRIVER_BUILD_)

#ifdef _WIN32
#define	_NT_DRIVER_BUILD_ 1
#elif defined(__FreeBSD__)
#define _BSD_DRIVER_BUILD_ 1
#elif defined(linux)
#define _BSD_DRIVER_BUILD_ 1
#else
#error Could not recognise your OS 
#endif

#endif



/******************
 * NT
 */
#ifdef  _NT_DRIVER_BUILD_

#include <ntddk.h>
#include "memtrack.h"

/*---------------------------------------------------------------------------
 * Status codes defined by O/S and returned by the driver.
 * This driver doesn't define its own status codes, just aliases and uses
 * system ones, since the most of the errors are originated from system calls
 * and some system codes are suitable to describe our errors.
 */
typedef NTSTATUS                                    PWRCRYPT_STATUS;

typedef KPROCESSOR_MODE                             PWRCRYPT_PROCESSOR_MODE;

#define PWRCRYPT_SUCCESS( status)                   NT_SUCCESS( status)

#define PWRCRYPT_STATUS_SUCCESS                     STATUS_SUCCESS
#define PWRCRYPT_STATUS_PENDING                     STATUS_PENDING
#define PWRCRYPT_STATUS_INSUFFICIENT_RESOURCES      STATUS_INSUFFICIENT_RESOURCES
#define PWRCRYPT_STATUS_NO_MEMORY                   STATUS_NO_MEMORY
#define PWRCRYPT_STATUS_INVALID_USER_BUFFER         STATUS_INVALID_USER_BUFFER
#define PWRCRYPT_STATUS_INVALID_PARAMETER           STATUS_INVALID_PARAMETER
#define PWRCRYPT_STATUS_DEVICE_NOT_CONNECTED        STATUS_DEVICE_NOT_CONNECTED
#define PWRCRYPT_STATUS_NOT_IMPLEMENTED             STATUS_NOT_IMPLEMENTED

/*---------------------------------------------------------------------------
 * Memory
 */
#define PCrptZeroMemory( adr, len)              RtlZeroMemory( adr, len)
#define PCrptFillMemory( adr, len, val)         RtlFillMemory( adr, len, val)
#define PCrptCopyMemory( dstadr, srcadr, len)   RtlCopyMemory( dstadr, srcadr, len)
#define PCrptNonPagedMAlloc( size)              MemAllocatePool( NonPagedPool, size)
#define PCrptNonPagedMFree( adr)                MemFreePool( adr)

/*---------------------------------------------------------------------------
 * Doubly-linked list - unsafe, needs protection
 */
typedef LIST_ENTRY  PCRPT_LIST_ENTRY, *PPCRPT_LIST_ENTRY;

typedef struct _PCRPT_LIST
{
    LIST_ENTRY  ListHead;

} PCRPT_LIST, *PPCRPT_LIST;

void
PCrptListInit(
    IN OUT  PPCRPT_LIST pThis
    );

BOOLEAN
PCrptIsListEmpty(
    IN OUT  PPCRPT_LIST pThis
    );

void
PCrptListInsertTail(
    IN OUT  PPCRPT_LIST pThis,
    PPCRPT_LIST_ENTRY   ListEntry
    );

void
PCrptListInsertHead(
    IN OUT  PPCRPT_LIST pThis,
    PPCRPT_LIST_ENTRY   ListEntry
    );

PPCRPT_LIST_ENTRY
PCrptListRemoveHead(
    IN OUT  PPCRPT_LIST pThis
    );

/*---------------------------------------------------------------------------
 * Event synchronization
 *
 * Watch for PLARGE_INTEGER Timeout.
 * MS monkeys were not able to explain intelligibly that
 *    ( Timeout == NULL) means "wait for indefinite time" and
 *    ( Timeout != NULL && *Timeout == 0) means "return current event
 *    status and don't wait"
 */ 

typedef KEVENT          PCRPT_EVENT, *PPCRPT_EVENT;

/*
 * Initialize an event to non-signaled state
 */
#define PCrptInitEvent(pEvent)  KeInitializeEvent( pEvent, NotificationEvent, FALSE)

/*
 * Check event state
 */
#define PCrptIsEventSignaled( pEvent)   ( KeReadStateEvent( pEvent) != 0)

/*
 * Wait for event indefinitely
 */
#define PCrptWaitOnEvent( pEvent) \
\
    KeWaitForSingleObject( \
        pEvent, \
        Suspended,   /* KWAIT_REASON */ \
        KernelMode,  /* KPROCESSOR_MODE */ \
        FALSE,       /* BOOLEAN Alertable */ \
        NULL         /* PLARGE_INTEGER Timeout */ \
        )

/*
 * Signal an event
 */
#define PCrptSignalEvent( pEvent) \
\
    KeSetEvent( \
        pEvent, \
        IO_NO_INCREMENT,   /* priority increment */ \
        FALSE              /* no, we are not going to sleep after that */ \
        )

/*---------------------------------------------------------------------------
 * Spinlock synchronization
 */

typedef struct _PCRPT_SPIN_LOCK
{
    KSPIN_LOCK  SpinLock;
    KIRQL       OldIrql;

} PCRPT_SPIN_LOCK, *PPCRPT_SPIN_LOCK;

#define PCRPT_DECLARE_SPINLOCK( VarName)  PCRPT_SPIN_LOCK VarName

#define PCrptSpinLockInit( pThis) \
    KeInitializeSpinLock( &(pThis)->SpinLock)
#define PCrptSpinLockAcquire( pThis) \
    KeAcquireSpinLock( &(pThis)->SpinLock, &(pThis)->OldIrql)
#define PCrptSpinLockAcquireAtDpcLevel( pThis) \
    KeAcquireSpinLockAtDpcLevel( &(pThis)->SpinLock)
#define PCrptSpinLockRelease( pThis) \
    KeReleaseSpinLock( &(pThis)->SpinLock, (pThis)->OldIrql)
#define PCrptSpinLockReleaseFromDpcLevel( pThis) \
    KeReleaseSpinLockFromDpcLevel( &(pThis)->SpinLock)

/*---------------------------------------------------------------------------
 * Indicates whether the caller can wait on dispatcher objects
 */
#define PCrptCanBlock()   ( KeGetCurrentIrql() < DISPATCH_LEVEL) 

/*---------------------------------------------------------------------------
 * Make a delay specified in ms. Used only for making delays during hardware
 * initialization, not at normal operation time
 */
#define delay( ms)     KeStallExecutionProcessor( 1000*(ms))

/*---------------------------------------------------------------------------
 * memory-mapped register accessors
 */
#define READ_MEM_REG32( RegAddr)         READ_REGISTER_ULONG( RegAddr)
#define WRITE_MEM_REG32( RegAddr, Value) WRITE_REGISTER_ULONG( RegAddr, Value)

/*
 * The list describing physical memory that backs up a locked virtual memory
 * and some corresponding routines
 */
typedef MDL MEM_DESC_LIST, *PMEM_DESC_LIST;

#ifndef HW_BUG_SEARCH_BUILD

#define CNT_FROM_MDL( pMdl)             MmGetMdlByteCount( pMdl)
#define VIRT_ADDR_FROM_MDL( pMdl)       MmGetMdlVirtualAddress( pMdl)

#else

#define VIRT_ADDR_FROM_MDL( pMdl)       ((PUCHAR)MmGetMdlVirtualAddress( pMdl)+8*1024)
#define CNT_FROM_MDL( pMdl)             (8*1024)

#endif /* HW_BUG_SEARCH_BUILD */

#define ALLOCATE_MDL( nPages)           IoAllocateMdl( NULL,                \
                                                       nPages * PAGE_SIZE,  \
                                                       FALSE,               \
                                                       FALSE,               \
                                                       NULL)

#define FREE_MDL( pMdl)                 IoFreeMdl( pMdl)

#define PCrptBuildPartialMdl( pSrcMdl, pTargetMdl, StartVA, Length)         \
                IoBuildPartialMdl( pSrcMdl, pTargetMdl, StartVA, Length)

#define PCrptPrepareMdlForReuse( pMdl)  MmPrepareMdlForReuse( pMdl)

#endif /* _NT_DRIVER_BUILD_ */

/******************
 * BSD
 */
#ifdef  _BSD_DRIVER_BUILD_

#include "bsdddk.h"

/* */
/* define the same macros as defined above for NT */
/* */

/*---------------------------------------------------------------------------
 * Status codes defined by O/S and returned by the driver.
 * This driver doesn't define its own status codes, just aliases and uses
 * system ones, since the most of the errors are originated from system calls
 * and some system codes are suitable to describe our errors.
 */
typedef int 					    PWRCRYPT_STATUS;
typedef KPROCESSOR_MODE                             PWRCRYPT_PROCESSOR_MODE;

#define PWRCRYPT_SUCCESS( status)                   ( status == STATUS_SUCCESS)

#define PWRCRYPT_STATUS_SUCCESS                     STATUS_SUCCESS
#define PWRCRYPT_STATUS_PENDING                     STATUS_SUCCESS
#define PWRCRYPT_STATUS_INSUFFICIENT_RESOURCES      EAGAIN
#define PWRCRYPT_STATUS_NO_MEMORY                   ENOMEM
#define PWRCRYPT_STATUS_INVALID_USER_BUFFER         EFAULT
#define PWRCRYPT_STATUS_INVALID_PARAMETER           EINVAL
#define PWRCRYPT_STATUS_DEVICE_NOT_CONNECTED        ENXIO
#define PWRCRYPT_STATUS_NOT_IMPLEMENTED             ENODEV

/*---------------------------------------------------------------------------
 * Memory
 */

#ifdef linux
#define PCrptZeroMemory( pBuf, len)		memset(pBuf, 0, len)
#else
#define PCrptZeroMemory( pBuf, len)             bzero(pBuf, len)
#endif

#define PCrptFillMemory( pBuf, len, val) \
{ \
    int i; \
    for ( i = 0; i < len; i++) \
    { \
        ((unsigned char*)pBuf)[ i] = val; \
    } \
}

#define PCrptCopyMemory( dstadr, srcadr, len)   memcpy( dstadr, srcadr, len)

#ifdef linux

#define PCrptNonPagedMAlloc( size)              kmalloc( size, GFP_KERNEL)
#define PCrptNonPagedMFree( pBuf)               kfree( pBuf)

#else

#define PCrptNonPagedMAlloc( size)              malloc( size, M_DEVBUF, M_NOWAIT)
#define PCrptNonPagedMFree( pBuf)               free( pBuf, M_DEVBUF)

#endif /* linux */

/*---------------------------------------------------------------------------
 * Doubly-linked list - unsafe, needs protection
 */
typedef LIST_ENTRY  PCRPT_LIST_ENTRY, *PPCRPT_LIST_ENTRY;

typedef struct _PCRPT_LIST
{
    LIST_ENTRY  ListHead;

} PCRPT_LIST, *PPCRPT_LIST;

void
PCrptListInit(
    IN OUT  PPCRPT_LIST pThis
    );

BOOLEAN
PCrptIsListEmpty(
    IN OUT  PPCRPT_LIST pThis
    );

void
PCrptListInsertTail(
    IN OUT  PPCRPT_LIST pThis,
    PPCRPT_LIST_ENTRY   ListEntry
    );

void
PCrptListInsertHead(
    IN OUT  PPCRPT_LIST pThis,
    PPCRPT_LIST_ENTRY   ListEntry
    );

PPCRPT_LIST_ENTRY
PCrptListRemoveHead(
    IN OUT  PPCRPT_LIST pThis
    );

/*---------------------------------------------------------------------------
 * Event synchronization
 *
 * Watch for PLARGE_INTEGER Timeout.
 * MS monkeys were not able to explain intelligibly that
 *    ( Timeout == NULL) means "wait for indefinite time" and
 *    ( *Timeout == 0)  means "return current event status and don't wait"
 */

typedef struct _PCRPT_EVENT
{
    LONG   SignalState;

} PCRPT_EVENT, *PPCRPT_EVENT;

/*
 * Initialize an event to non-signaled state
 */
#define PCrptInitEvent( pEvent)         (pEvent)->SignalState = 0

/*
 * Check event state but _don't_ wait for it
 */
#define PCrptIsEventSignaled( pEvent)   ( (pEvent)->SignalState == 0)

/*
 * Wait for event indefinitely
 *
 * BUGBUG - clarify "priority".
 *
 * If the PCATCH flag is ORed to priority parameter, tsleep will return
 * EINTR otr ERESTART if a signal comes. Then the interrupted syscall
 * should return that error code to the caller (4.4BSD p.85).
 *
 * Looking at the tsleep() code in kern_synch.c I concluded that it returns
 * after sleeping at the same hardware priority level (spl) as it was entered.
 */



#ifdef linux 

#define PCrptWaitOnEvent( pEvent) \
{ \
    int flags; \
    save_flags(flags); \
    cli(); \
    while ( (pEvent)->SignalState == 0) \
	interruptible_sleep_on(&waitq); \
    restore_flags(flags); \
}

/*
 * Signal an event
 */

#define PCrptSignalEvent( pEvent) \
{ \
    (pEvent)->SignalState++; \
    wake_up_interruptible(&waitq); \
}

#else

#define PCrptWaitOnEvent( pEvent) \
{ \
    long    OldSpl; \
    OldSpl = splhigh(); \
    while ( (pEvent)->SignalState == 0) \
    { \
        int nSleepResult; \
        nSleepResult = tsleep( pEvent, PRIBIO, "PcEvWt", 0); \
    } \
    splx( OldSpl); \
}

/*
 * Signal an event
 */
#define PCrptSignalEvent( pEvent) \
{ \
    (pEvent)->SignalState++; \
    wakeup( pEvent); \
}
#endif /* linux */


/*---------------------------------------------------------------------------
 * Spinlock synchronization for FreeBSD 2.2.6
 *
 * Spinlock is a method of synchronization on machines that have multiple
 * CPUs. On such machines, even disabling interrupts does not suffice to
 * protect data, because another CPU still can enter the protected code
 * section. Spinlock causes CPU to enter a tight loop if a variable indicates
 * busy state, and spin in this loop until the state changes to free. At this
 * moment CPU breaks the loop, indicates busy state in the variable and
 * proceeds. Once the critical section is passed, CPU changes the variable to
 * indicate free state.
 *
 * In multiprocessor NT kernel, spinlock acquisition not only indicates busy
 * state in a variable but also raises CPU's IRQL to mask any interrupts below
 * certain level. In singleprocessor NT kernel, spinlock acquisition only
 * raises IRQL, but the spinning is omitted.
 *
 * Raising IRQL and masking certain interrupts is important. In NT, the
 * "weakest" spinlock disables DPC interrupts and preemption. At the same time,
 * code executed at DPC level or higher is disallowed to wait or cause paging.
 * It must continue to run and eventually drop the spinlock. Since waiting is
 * not allowed, in DPC routines (called at IRQL=DPC_LEVEL) spinlocks are the
 * only synchronization mechanism available. If any code executed at IRQL below
 * DPC needs to synchronize with DPC routines and, at the same time, with
 * other executions of itself, it has to use spinlocks which combine multi-CPU
 * protection with disabling DPC interrupts on the current CPU.
 * 
 * Returning to FreeBSD 2.2.6, we notice that this FreeBSD is equal to
 * Win9x with regard to spinlocks - it runs on single CPU machines only and
 * thus doesn't need spinning. Also, the kernel code in both systems is not
 * subject to unvoluntary preemption, so, the only thing of concern is
 * hardware interrupts which can be simply masked. Voluntary preemption
 * occurs in both Win9x and FreeBSD when CPU enters a kind of sleep() function.
 * (In Win9x, entering pageable kernel code or touching pageable data too can
 * cause preemption, but this is voluntary thing too, since driver writer is
 * aware of this).
 *
 * This driver never induces paging (FreeBSD kernel is non-pageable and
 * driver's own data too) or attempts to sleep() with a spinlock held.
 * The only data that driver touches that may be pageable are user data
 * buffers, but they are touched only while CPU is in the calling process
 * context and doesn't hold spinlocks. Such a behavior is imposed by NT. 
 *
 * This permits this driver to implement spinlock acquisition for FreeBSD
 * the way NT does this - by disabling certain ints without any spinning.
 *
 * Now, we need to consider how to disable ints during spinlock acquisitions.
 * Interrupt processing in this driver is performed in two steps. The first one
 * acknowledges the interrupt and arranges for the second step to run later.
 * The driver is structured in such a way that there is no need for any code to
 * synchronize with the first step, except the code of the second step. The
 * rest of the code needs to synchronize with the second step only.
 *
 * In NT, the first step is executed with our HW ints disabled, the second one
 * with DPC ints disabled. This means that spinlocks used by the second step
 * to synchronize with the first step must disable our HW ints. In NT, this is
 * accomplished via a special function that acquires an "interrupt spinlock"
 * and calls a driver-specified routine that is to perform the needed
 * synchronized work. To synchronize with the second step any code uses
 * spinlocks that disable DPC ints.  
 *
 * In FreeBSD, both steps are performed synchronously from the interrupt
 * handler - the first one is performed, then the second one. The processing
 * runs with our HW ints disabled, so, both steps are already synchronized.
 * The rest of the driver code needs to synchronize with the second step,
 * so, spinlocks used for this must disable our HW ints.
 *
 * Since I went along the way of implementing some NT functionality that
 * this driver missed in FreeBSD, I implemented the above mentioned
 * synchronization function that acquires interrupt spinlock. For this,
 * PROCBASE object includes interrupt spinlock, and the routine acquires this
 * spinlock. This spinlock is actually not needed for the current driver
 * because only the second step needs it, but the second step is already
 * synchronized. The driver would work without this spinlock. But implementing
 * this function this way provides a general mechanism for synchronization
 * with the first step that can be used from _any_ code that needs it. Also,
 * if FreeBSD kernel is ever to be executed by multiple CPUs, this spinlock
 * might become useful. So, being optimistic, I kept this redundancy.
 *
 * To decide to which of FreeBSD's interrupt groups to put our interrupt,
 * we need to consider that 7751 is not a time-critical device. Disabling
 * its ints will never cause data loss as it may be in the case of network
 * interfaces. This suggests to put it into tty_imask group. At the other hand,
 * the primary and typical purpose of 7751 is to perform IPSec operations on
 * network packets. In such processing, network packets will typically be
 * buffered until IPSec processing completes. If we assign 7751 to be in
 * tty_imask, network controllers will have priority over it, but this doesn't
 * seem to be a problem unless network HW interrupt processing consumes most of 
 * CPU time, at which case the system will have more serious problems to worry
 * about anyway :-)
 *
 * BUGBUG: Thus I feel that putting 7751 into tty_imask is reasonable, though
 * I haven't tried it and feel there might be some gotchas with registering
 * our interrupt handler in this imask, just because I haven't investigated
 * that deep enough. In the case of a problem, revert to use net_imask by
 * changing "#if 1" to "if 0" below, this is the only change needed.
 * net_imask always worked before for this driver.
 * 
 * THe Linux version of the code uses save_flags() and cli() to do its dirty
 * work. Unfortunately, I think cli() disables *all* interrupts, not just ones
 * in a certain mask, but it seems to work.
 */

#ifdef __FreeBSD__
#if 1
#define PWRCRYPT_IMASK  tty_imask
#define splpwrcrypt()   spltty()
#else
#define PWRCRYPT_IMASK  net_imask
#define splpwrcrypt()   splimp()
#endif

typedef struct _PCRPT_SPIN_LOCK
{
    int     old_pl;

} PCRPT_SPIN_LOCK, *PPCRPT_SPIN_LOCK;

#define PCRPT_DECLARE_SPINLOCK( VarName)  PCRPT_SPIN_LOCK VarName

#define PCrptSpinLockInit( pThis)
#define PCrptSpinLockAcquire( pThis)                (pThis)->old_pl = splpwrcrypt()
#define PCrptSpinLockAcquireAtDpcLevel( pThis)      PCrptSpinLockAcquire( pThis)
#define PCrptSpinLockRelease( pThis)                splx( (pThis)->old_pl)
#define PCrptSpinLockReleaseFromDpcLevel( pThis)    PCrptSpinLockRelease( pThis)

#elif defined(linux)

typedef struct _PCRPT_SPIN_LOCK
{
	int flags;

} PCRPT_SPIN_LOCK, *PPCRPT_SPIN_LOCK;  

#define PCRPT_DECLARE_SPINLOCK( VarName) PCRPT_SPIN_LOCK VarName
#define PCrptSpinLockInit( pThis) 
#define PCrptSpinLockAcquire( pThis) { save_flags((pThis)->flags); cli(); }
#define PCrptSpinLockAcquireAtDpcLevel( pThis) PCrptSpinLockAcquire( pThis)
#define PCrptSpinLockRelease( pThis) restore_flags((pThis)->flags)
#define PCrptSpinLockReleaseFromDpcLevel( pThis) PCrptSpinLockRelease( pThis)

#endif /* __FreeBSD__ */

/*---------------------------------------------------------------------------
 * Make a delay specified in ms. Used only for making delays during hardware
 * initialization, not at the normal operation time.
 */
#define delay( ms)     DELAY( 1000*(ms))

/*---------------------------------------------------------------------------
 * memory-mapped register accessors
 */
#define READ_MEM_REG32( RegAddr)         *( ( unsigned long*)RegAddr)
#define WRITE_MEM_REG32( RegAddr, Value) *( ( unsigned long*)RegAddr) = Value

/*
 * The list describing physical memory that backs up a locked virtual memory
 * and some corresponding routines
 */
typedef MDL MEM_DESC_LIST, *PMEM_DESC_LIST;

#define CNT_FROM_MDL( pMdl)             MmGetMdlByteCount( pMdl)
#define VIRT_ADDR_FROM_MDL( pMdl)       MmGetMdlVirtualAddress( pMdl)

#if 0

#define ALLOCATE_MDL( nPages)           AllocateMdl( NULL,                \
                                                       nPages * PAGE_SIZE,  \
                                                       FALSE,               \
                                                       FALSE,               \
                                                       NULL)

#define FREE_MDL( pMdl)                 FreeMdl( pMdl)

#define PCrptBuildPartialMdl( pSrcMdl, pTargetMdl, StartVA, Length)         \
                IoBuildPartialMdl( pSrcMdl, pTargetMdl, StartVA, Length)

#define PCrptPrepareMdlForReuse( pMdl)  MmPrepareMdlForReuse( pMdl)

#else

#ifdef linux
#define printf printk
#endif

#define ALLOCATE_MDL( nPages)   (void*)printf("BUGBUG")
#define FREE_MDL( pMdl)            printf("BUGBUG")
#define PCrptBuildPartialMdl( pSrcMdl, pTargetMdl, StartVA, Length)    printf("BUGBUG")
#define PCrptPrepareMdlForReuse( pMdl)    printf("BUGBUG")

#endif /* HW_BUG_SEARCH_BUILD */

#endif /* _BSD_DRIVER_BUILD_ */

#endif /* __OSINDEPENDENCE_H__ */
