#include "osindependence.h"
#include "debug.h"
#include "processor.h"
#include "proc_hw_dma.h"
#include "proc_hw_proc.h"

/****************************************************************************
DMA UNIT ROUTINES
****************************************************************************/

/****************************************************************************
DmaDiagnose
  DMA Unit diagnostics for proper mapping, address, data lines
****************************************************************************/
BOOLEAN
DmaDiagnose( IN PPROCESSOR   pThis)
{
    BOOLEAN bError;
    ULONG val, readVal;

    DebugPrint((1, "DmaDiagnose\n"));
    ASSERT( pThis->Type == PROCESSOR_SIGN);

    bError = FALSE;

    /* Check data lines by walking ones and zeroes*/
    for( val = 1; ( val & 0xffff ) && !bError; val <<= 1 )
    {
        WRITE_DMA32( COM_RING_ADR, val );
        readVal = READ_DMA32( COM_RING_ADR);
        if( ( readVal & 0x0000ffff ) != val )
        {
            bError = TRUE;
        }

        WRITE_DMA32( COM_RING_ADR, ~val & 0xffff );
        readVal = READ_DMA32( COM_RING_ADR);
        if( ( readVal | 0xffff0000 ) != ~val )
        {
            bError = TRUE;
        }
    }

    /* Check address lines */
    WRITE_DMA32( COM_RING_ADR,      0);

    /* A2: 40 vs 44 */
    WRITE_DMA32( INT_ENABLE,        0);
    WRITE_DMA32( STATUS_AND_CTL,    0x3e3e3c3d);

    readVal = READ_DMA32( INT_ENABLE);
    if( readVal != 0 )
    {
        bError = TRUE;
    }

    /* A3: 40 vs 48 */

    /* Note! Writing 0 to CONFIG clears the "master reset" bit, and it must
     * be set at some time later to return the proc.unit to normal life
     */
    DmaSetPollingFreq( pThis, 0);
    WRITE_DMA32( CONFIG,          0);
    WRITE_DMA32( STATUS_AND_CTL,  0x3e3e3c3d);

    readVal = READ_DMA32( CONFIG);
    if( readVal != 0 )
    {
        bError = TRUE;
    }

    /* A4: 0C vs 1C */

    WRITE_DMA32( SRC_RING_ADR,    0xffffffff );
    readVal = READ_DMA32( COM_RING_ADR);
    if( readVal != 0 )
    {
        bError = TRUE;
    }

    /* A5: 0C vs 2C */

    WRITE_DMA32( RLT_RING_ADR,    0xffffffff );
    readVal = READ_DMA32( COM_RING_ADR);
    if( readVal != 0 )
    {
        bError = TRUE;
    }

    readVal = READ_DMA32( CONFIG);
    DebugPrint((1, "    conf before = %08x\n", readVal));

    /* Note! Set the "master reset" bit to enable proc.unit
     */
    WRITE_DMA32( CONFIG, D_CONF_MASK_MASTER_RESET);

    readVal = READ_DMA32( CONFIG);
    DebugPrint((1, "    conf AFTER1 = %08x\n", readVal));
    delay(1);
    readVal = READ_DMA32( CONFIG);
    DebugPrint((1, "    conf after2 = %08x\n", readVal));

    return !bError;
}

/****************************************************************************
  initDMAUnit -- brings the DMA unit out of reset and disables polling until
  the ring address registers are loaded.
  Returns 0 on success
****************************************************************************/

#if 0
void
InitDmaDisablePoll( IN PPROCESSOR   pThis)
{
    ULONG nConfigRegSave;

    ASSERT( pThis->Type == PROCESSOR_SIGN);

   /*
    * Start with a clean configuration, setting poll.freq. to 0
    */
   WRITE_DMA32( CONFIG, D_CONF_LAST);

   /*
    * After setting poll.freq., wait for 1 ms
    */
   delay(1);

   /*
    * And reset the chip
    */
   nConfigRegSave = READ_DMA32( CONFIG);
   WRITE_DMA32( CONFIG, D_CONF_MASTER_RESET);

   /* Bring it back out of reset, specify polling off, save rest of
    * configuration
    */
   WRITE_DMA32( CONFIG,
                   (nConfigRegSave & ~D_CONF_POLLFREQ_MASK) |
                   D_CONF_MASTER_RESET | D_CONF_DMA_RESET);
}
#endif

/****************************************************************************

Resets DMA and preserves value in the DMA config reg.

p.32 - 9.7.2 DMAReset#

    - set poll.freq. to 0 (waits for 1 ms afterwards)
    - clear DMA RESET bit
    - set DMA RESET bit

****************************************************************************/
void
DmaReset( IN PPROCESSOR   pThis)
{
    ULONG ConfSave;
    ULONG PollFreqSave;

    DebugPrint((1, "DmaReset\n"));
    ASSERT( pThis->Type == PROCESSOR_SIGN);

    /* Save config reg. value and poll.freq.
     */
    PollFreqSave = DmaGetPollingFreq( pThis);
    ConfSave = READ_DMA32( CONFIG);

    DmaSetPollingFreq( pThis, 0);

    /* Zero config reg. (except master reset bit)
     */
    WRITE_DMA32( CONFIG, 0UL |
                         D_CONF_MASK_MASTER_RESET |
                         ~D_CONF_MASK_DMA_RESET);

    /* Write old config with DMA reset bit set and zero poll.freq.
     */
    WRITE_DMA32( CONFIG, (ConfSave & ~D_CONF_POLLFREQ_MASK) |
                         D_CONF_MASK_MASTER_RESET |
                         D_CONF_MASK_DMA_RESET);

    /* Restore old poll.freq.
     */
    DmaSetPollingFreq( pThis, PollFreqSave);

#ifdef notdef
/*
    WORKWORK This is from the doc, but _they_ use the write below */
    Do other conf.reg. fields make a difference ??? */
*/
    DmaSetPollingFreq( pThis, 0);

    WRITE_DMA32( CONFIG, D_CONF_MASK_MASTER_RESET);

   /* Set config with 0 poll.freq and issue DMA reset */

   WRITE_DMA32( CONFIG, (ConfSave & ~D_CONF_POLLFREQ_MASK) |
                      D_CONF_MASK_MASTER_RESET |
                      D_CONF_DMA_RESET);
   delay(1);
#endif
}

/****************************************************************************
****************************************************************************/
/*
 * Polling frequency - the frequency with which the DMA unit switches
 * rings to keep DMA going in both directions.
 */

/****************************************************************************
****************************************************************************/
ULONG
DmaGetPollingFreq( IN PPROCESSOR   pThis)
{
    ULONG u;

    ASSERT( pThis->Type == PROCESSOR_SIGN);

    u = READ_DMA32( CONFIG);

    return D_CONF_POLLFREQ_BITS_TO_VAL( u);
}

/****************************************************************************
DmaSetPollingFreq()

Writes new polling frequency into the configuration reg,
preserving other bits, then waits for 1 ms as the doc specifies.

****************************************************************************/
void
DmaSetPollingFreq( IN PPROCESSOR   pThis, ULONG freq)
{
    ULONG u;

    ASSERT( pThis->Type == PROCESSOR_SIGN);

    /*
     * Read the curr value of the conf.reg, clear the freq. bits,
     * put new bits, write it back and wait for 1 ms
     */
    u = READ_DMA32( CONFIG);
    u &= ~D_CONF_POLLFREQ_MASK;
    u |= D_CONF_VAL_TO_POLLFREQ_BITS( freq);
    WRITE_DMA32( CONFIG, u);
    delay(1);
}
/****************************************************************************
typedef struct _SG_CTL_AREA
{
    ULONG           PhysThis;
    ULONG           SmallRingDataSize;

    RING            SmallRing;
    SG_DESC         SR[ COM_AND_RES_RING_LEN - 1];

    RING            BigRing;
    SG_DESC         BR[ DATA_RING_LEN - 1];

    UCHAR           SmallRingData[ 1];

} SG_CTL_AREA, *PSG_CTL_AREA;
****************************************************************************/

#ifdef DBG

void
DmaDumpStatReg( IN PPROCESSOR   pThis, ULONG StatRegVal)
{
    ASSERT( pThis->Type == PROCESSOR_SIGN);

    DebugPrint((1, "        ComRing %s%s%s%s\n",
                StatRegVal & 0x00000010 ? "Done " : "     ",
                StatRegVal & 0x00000008 ? "Last " : "     ",
                StatRegVal & 0x00000004 ? "Wait " : "     ",
                "     "
                ));

    DebugPrint((1, "        SrcRing %s%s%s%s\n",
                StatRegVal & 0x00001000 ? "Done " : "     ",
                StatRegVal & 0x00000800 ? "Last " : "     ",
                StatRegVal & 0x00000400 ? "Wait " : "     ",
                "     "
                ));

    DebugPrint((1, "        DstRing %s%s%s%s\n",
                StatRegVal & 0x10000000 ? "Done " : "     ",
                StatRegVal & 0x08000000 ? "Last " : "     ",
                StatRegVal & 0x04000000 ? "Wait " : "     ",
                StatRegVal & 0x02000000 ? "Over " : "     "
                ));

    DebugPrint((1, "        RltRing %s%s%s%s\n",
                StatRegVal & 0x00100000 ? "Done " : "     ",
                StatRegVal & 0x00080000 ? "Last " : "     ",
                StatRegVal & 0x00040000 ? "Wait " : "     ",
                StatRegVal & 0x00020000 ? "Over " : "     "
                ));
}

void
DmaDumpRegs( IN PPROCESSOR   pThis)
{
    ULONG           regval;

    PSG_CTL_AREA    pArea;
    PSG_DESC        pFirstDesc;
    ULONG           FirstDescPA;
    int             DescIndex;

    PRING           pRing;

    DebugPrint((1, "DMA regs mapped at 0x%08x\n", pThis->pDMARegisters));
    ASSERT( pThis->Type == PROCESSOR_SIGN);

    regval = READ_DMA32( STATUS_AND_CTL);
    DebugPrint((1, "    Stat&Ctl    0x%08x\n", (ULONG)regval));
    DmaDumpStatReg( pThis, regval);
    regval = READ_DMA32( INT_ENABLE);
    DebugPrint((1, "    Int.Enable  0x%08x\n", (ULONG)regval));
    regval = READ_DMA32( CONFIG);
    DebugPrint((1, "    Config      0x%08x\n", (ULONG)regval));

    /***************************
     * Src
     */
    pArea = DmaCtxGetSgCtlAreaPtr( pThis->DmaCB[ SRC_XFER].pDmaCtx);

    /*--------------------------
     * Command
     */
    pFirstDesc = SgCtlAreaGetFirstSmallRingDescPtr( pArea);
    FirstDescPA = pArea->PhysThis + ( (ULONG) pFirstDesc - (ULONG) pArea);
    regval = READ_DMA32( COM_RING_ADR);
    DescIndex = (regval - FirstDescPA) / sizeof( SG_DESC);
    ASSERT( (regval - FirstDescPA) % sizeof( SG_DESC) == 0);

    DebugPrint((1, "    ComRingAdr  0x%08x  Desc# 0x%04x\n", (ULONG)regval, DescIndex));

    /*--------------------------
     * SrcData
     */
    pFirstDesc = SgCtlAreaGetFirstBigRingDescPtr( pArea);
    FirstDescPA = pArea->PhysThis + ( (ULONG) pFirstDesc - (ULONG) pArea);
    regval = READ_DMA32( SRC_RING_ADR);
    DescIndex = (regval - FirstDescPA) / sizeof( SG_DESC);
    ASSERT( (regval - FirstDescPA) % sizeof( SG_DESC) == 0);

    DebugPrint((1, "    SrcRingAdr  0x%08x  Desc# 0x%04x\n", (ULONG)regval, DescIndex));

    /***************************
     * Dst
     */
    pArea = DmaCtxGetSgCtlAreaPtr( pThis->DmaCB[ DST_XFER].pDmaCtx);

    /*--------------------------
     * DstData
     */
    pFirstDesc = SgCtlAreaGetFirstBigRingDescPtr( pArea);
    FirstDescPA = pArea->PhysThis + ( (ULONG) pFirstDesc - (ULONG) pArea);
    regval = READ_DMA32( DST_RING_ADR);
    DescIndex = (regval - FirstDescPA) / sizeof( SG_DESC);
    ASSERT( (regval - FirstDescPA) % sizeof( SG_DESC) == 0);

    DebugPrint((1, "    DstRingAdr  0x%08x  Desc# 0x%04x\n", (ULONG)regval, DescIndex));

    /*--------------------------
     * Result
     */
    pFirstDesc = SgCtlAreaGetFirstSmallRingDescPtr( pArea);
    FirstDescPA = pArea->PhysThis + ( (ULONG) pFirstDesc - (ULONG) pArea);
    regval = READ_DMA32( RLT_RING_ADR);
    DescIndex = (regval - FirstDescPA) / sizeof( SG_DESC);
    ASSERT( (regval - FirstDescPA) % sizeof( SG_DESC) == 0);

    DebugPrint((1, "    RltRingAdr  0x%08x  Desc# 0x%04x\n", (ULONG)regval, DescIndex));
}

#endif

/****************************************************************************
PROCESSING UNIT ROUTINES
****************************************************************************/

/****************************************************************************
Description:
    Wait for P_CTL_RESET bit in the CTL reg. to be clear

 Use the DMA reset mechanism to perform a hardware reset of the compression
 engine chip. Wait for a while afterward to allow the compression engine chip
 to become available again. Note that the busy wait loop does reads over the
 PCI bus to make it independent of processor speed. True, now it's dependent
 on PCI bus speed, but that's much more constant and predictable.  0x400
 iterations should be about 1ms on most 33MHz PCI systems.

Return
    TRUE if Ok, FALSE on timeout (the bit doesn't become clear)
****************************************************************************/
BOOLEAN
PuWaitForResetCompletion( IN PPROCESSOR   pThis)
{
    ULONG regval;
    int nTries;

    DebugPrint((1, "WaitForResetCompletion\n"));
    ASSERT( pThis->Type == PROCESSOR_SIGN);

    regval = P_CTL_RESET;
    nTries = 0x4000;
    while ( nTries > 0 && ( regval & P_CTL_RESET))
    {
        regval = READ_PROC32( CTL);
        nTries --;
    }

    DebugPrint((1, "    tries = 0x%x\n", 0x4000 - 1 - nTries));

    return nTries > 0;
}

/* ---------------------------------------------------------------------------
 * resetProcUnit --- Reset the Proc Unit.
 *
 * Sets the P_CTL_RESET bit in the CTL reg and waits for it to be cleared
 * by the 7751 - this indicates the reset completion.
 * ------------------------------------------------------------------------
 */
BOOLEAN
PuReset( IN PPROCESSOR   pThis)
{
    BOOLEAN bSuccess;

    DebugPrint((1, "PuReset\n"));
    ASSERT( pThis->Type == PROCESSOR_SIGN);

    bSuccess = FALSE;

    /* First see whether the reset (if initiated somewhere) is complete
     */
    if ( PuWaitForResetCompletion( pThis))
    {
        ULONG regval;

        regval = ( READ_PROC32( CTL) & P_CTL_ENABLE_DMA) | P_CTL_RESET;
        WRITE_PROC32( CTL, regval);
        bSuccess = PuWaitForResetCompletion( pThis);
    }
    return bSuccess;
}


/* ---------------------------------------------------------------------------
 * checkProcUnit --- Proc Unit diagnostic for mapping, address, data lines
 * ------------------------------------------------------------------------
 */
BOOLEAN
PuDiagnose( IN PPROCESSOR   pThis)
{
    ULONG       val;
    ULONG        ChipID;
    ULONG        readVal;
    SecurLevel  CurSecLevel;
    
    DebugPrint((1, "PuDiagnose\n"));
    ASSERT( pThis->Type == PROCESSOR_SIGN);

    /* First see whether the reset (if initiated somewhere) is complete
     */
    if ( !PuWaitForResetCompletion( pThis))
    {
        return FALSE;
    }

    /*
     * Check Chip ID (security level).
     *
     * (This should verify that we're talking to a Hi/fn chip not just memory
     * This also proves that we're hitting addrs 3 and 5 correctly) - hi/fn
     */
    if ( !PuGetSecurityLevel( pThis, &CurSecLevel))
    {
        return FALSE;
    }

    DebugPrint((1, "Current security level = %d\n", CurSecLevel));
    
    /*
     * Check data lines by walking ones and zeroes
     */
    for( val = 2; ( val & 0xffff ); val <<= 1 )
    {
        WRITE_PROC32( CONFIG, val);
        readVal = READ_PROC32( CONFIG);
        if( ( readVal & 0x0000fffe ) != val )
        {
            DebugPrint((1, "    data lines 1 failed\n"));
            return FALSE;
        }

        WRITE_PROC32( CONFIG, ~val & 0xfffe );
        readVal = READ_PROC32( CONFIG);
        if( ( readVal | 0xffff0001 ) != ~val )
        {
            DebugPrint((1, "    data lines 2 failed\n"));
            return FALSE;
        }
    }

    /* Check address lines
     * If we got a proper chip ID, then we know we can get addrs
     * 011b and 101b. That means no stuck@s on a2 or a1. Let's wiggle A0
     * by hitting address 010b (int status) and seeing if it changes 011b.
     */
    WRITE_PROC32( CONFIG, 0);
    WRITE_PROC32( INT_STATUS, 0xfbfc);
    readVal = READ_PROC32( CONFIG);
    if( ( readVal & 0xfffe ) != 0 )
    {
        DebugPrint((1, "    address lines failed\n"));
        return FALSE;
    }

    DebugPrint((1, "    Ok\n"));
    return TRUE;
}


#ifdef DBG

void
PuDumpRegs( IN PPROCESSOR   pThis)
{
    ULONG regval;
    USHORT Cnt;

    DebugPrint((1, "Proc regs mapped at 0x%08x\n", pThis->pProcRegisters));
    ASSERT( pThis->Type == PROCESSOR_SIGN);

    regval = READ_PROC32( CTL);
    DebugPrint((1, "    Control     0x%04x\n", (USHORT)regval));

    regval = READ_PROC32( STATUS);
    DebugPrint((1, "    Status      0x%04x\n", (USHORT)regval));
    DebugPrint((1, "        Invalid command : %s\n", ( regval & 0x8000) ? "Yes": "No"));
    DebugPrint((1, "        Data error      : %s\n", ( regval & 0x4000) ? "Yes": "No"));
    DebugPrint((1, "        Src FIFO ready  : %s\n", ( regval & 0x2000) ? "Yes": "No"));
    DebugPrint((1, "        Dst FIFO ready  : %s\n", ( regval & 0x1000) ? "Yes": "No"));
    DebugPrint((1, "        Dst overrun     : %s\n", ( regval & 0x0200) ? "Yes": "No"));
    DebugPrint((1, "        Src Command     : %s\n", ( regval & 0x0080) ? "Yes": "No"));
    DebugPrint((1, "        Src Context     : %s\n", ( regval & 0x0040) ? "Yes": "No"));
    DebugPrint((1, "        Src Data        : %s\n", ( regval & 0x0020) ? "Yes": "No"));
    DebugPrint((1, "        Dst Data        : %s\n", ( regval & 0x0010) ? "Yes": "No"));
    DebugPrint((1, "        Dst Result      : %s\n", ( regval & 0x0004) ? "Yes": "No"));

    regval = READ_PROC32( INT_STATUS);
    DebugPrint((1, "    IntStatus   0x%04x\n", (USHORT)regval));

    regval = READ_PROC32( INT_ENABLE);
    DebugPrint((1, "    IntEnable   0x%04x\n", (USHORT)regval));

    regval = READ_PROC32( CONFIG);
    DebugPrint((1, "    Config      0x%04x\n", (USHORT)regval));

    regval = READ_PROC32( FIFO_STATUS);
    DebugPrint((1, "    FIFO status 0x%04x\n", (USHORT)regval));
    Cnt = ((USHORT)regval >> 8) & 0x007f;
    DebugPrint((1, "        Src FIFO free bytes 0x%04x\n", Cnt));
    Cnt = (USHORT)regval & 0x007f;
    DebugPrint((1, "        Dst FIFO used bytes 0x%04x\n", Cnt));

    regval = READ_PROC32( FIFO_CONFIG);
    DebugPrint((1, "    FIFO config 0x%04x\n", (USHORT)regval));
}

#endif

/****************************************************************************
COMMON DMA & PROCESSING UNIT ROUTINES
    Group separated primitive DMA and Proc. Unit routines to form more
    complex actions required by the PROCESSOR logic.
****************************************************************************/

/****************************************************************************

DmaPuShutdown()

p.31-32 - 9.7.1 MSTReset#

    - set poll.freq. to 0
    - wait 1 ms
    - clear MASTER RESET bit
    - set MASTER RESET bit

****************************************************************************/
void
DmaPuShutdown( IN PPROCESSOR   pThis)
{
    ULONG u;

    DebugPrint((1, "TotalReset\n"));
    ASSERT( pThis->Type == PROCESSOR_SIGN);

    /*
     * Disable all ints
     */
    WRITE_DMA32( INT_ENABLE, 0);
    WRITE_PROC32( INT_ENABLE, 0);

    /*
     * Stop DMA engines
     */
    WRITE_DMA32( STATUS_AND_CTL, D_STATCTL_DISABLE_ALL_DMA);
    DmaSetPollingFreq( pThis, 0);

    /*
     * Clear the "master reset" bit
     */
    u = READ_DMA32( CONFIG);
    WRITE_DMA32( CONFIG, u | ~D_CONF_MASK_MASTER_RESET);

    /*
     * Set the "master reset" bit
     * (If the "master reset" bit is not set, the Proc.Unit is unaccessible).
     */
    WRITE_DMA32( CONFIG, D_CONF_MASK_MASTER_RESET);
}

/********************************************************************
 */
void
DmaPuLaunch(
    IN PPROCESSOR   pThis
    )
{
    DebugPrint((2, "DmaPuLaunch\n"));
    ASSERT( pThis->Type == PROCESSOR_SIGN);

    /********************************************************************
     * Configure processing unit and enable its ints, though they will
     * actually be enabled later, when we'll enable DMA unit ints.
     *
     * Note that ints will never come until this function returns, because
     * ISR is installed later.
     */
    WRITE_PROC32( FIFO_CONFIG, 0x00000400);

    WRITE_PROC32( CONFIG, P_CONF_CONST_CONF);
    WRITE_PROC32( CTL, P_CTL_ENABLE_DMA);

    pThis->PUEnabledInts = P_INT_CONST_INTS;
    WRITE_PROC32( INT_ENABLE, pThis->PUEnabledInts);


    /********************************************************************
     * Initialize DMA unit
     */

    /*
     * Invalidate SG descriptors in the rings and reset "current descriptor"
     * indexes to point to the first descriptors in the rings.
     * Load returned DMA addresses to the 7751 Ring Address Regs
     */
    {
        ULONG   RingPhysAddr;

        #if DBG
            ASSERT( pThis->pComRing != NULL);
            ASSERT( pThis->pRltRing != NULL);
            ASSERT( pThis->DmaCB[ SRC_XFER].pDataRing != NULL);
            ASSERT( pThis->DmaCB[ DST_XFER].pDataRing != NULL);
        #endif

        RingPhysAddr = RingResetAndGetPhysAddr( pThis->pComRing);
        WRITE_DMA32( COM_RING_ADR, RingPhysAddr);

        RingPhysAddr = RingResetAndGetPhysAddr( pThis->pRltRing);
        WRITE_DMA32( RLT_RING_ADR, RingPhysAddr);

        RingPhysAddr = RingResetAndGetPhysAddr( pThis->DmaCB[ SRC_XFER].pDataRing);
        WRITE_DMA32( SRC_RING_ADR, RingPhysAddr);

        RingPhysAddr = RingResetAndGetPhysAddr( pThis->DmaCB[ DST_XFER].pDataRing);
        WRITE_DMA32( DST_RING_ADR, RingPhysAddr);
    }

    /*
     * Set the permanent DMA configuration - take poll frequency and invalid poll
     * scalar bits from the configuration, "or" them to "fixed" configuration bits
     * and write to the 7751.
     */
    WRITE_DMA32( CONFIG,
                     D_CONF_CONST_CONFIG |
                     D_CONF_VAL_TO_POLLFREQ_BITS( ConfigGetPollFreq( pThis->pConfig)) |
                     D_CONF_VAL_TO_POLLSCALAR_BITS( ConfigGetPollScalar( pThis->pConfig)));

    /* WORKWORK. Clear DISAB_SCHED bit in STATUS_AND_CTL here, */
    /* as Hi/fn does??? */
    /* */

    /*
     * Tell the DMA engines that they can begin polling their rings
     */
    WRITE_DMA32( STATUS_AND_CTL, D_STATCTL_ENABLE_ALL_DMA);

    /*
     * Enable necessary DMA interrupts
     *
     * Note that ints will never come until this function returns, because
     * the ISR is installed later (BUGBUG - is this a good idea???)
     */
    pThis->DUEnabledInts = D_INT_CONST_INTS;
    WRITE_DMA32( INT_ENABLE, pThis->DUEnabledInts);

    #ifdef DBG
        PuDumpRegs( pThis);
        DmaDumpRegs( pThis);
    #endif
}

/********************************************************************
 */
BOOLEAN
DmaPuTest(
    IN PPROCESSOR   pThis
    )
{
    DebugPrint((2, "DmaPuTest\n"));
    ASSERT( pThis->Type == PROCESSOR_SIGN);

    if(!DmaDiagnose( pThis))
	return FALSE;
    PuReset( pThis);
    if(!PuDiagnose( pThis))
	return FALSE;

    PuReset( pThis);
    return TRUE;  

}
