/**********************************************************************

    Module: proc-proc.h

        Product: kernel mode driver for "PowerCrypt" encryption
        acceleration board based on Hi/fn 7751 chip
        by Ashley Laurent, Inc.

    ---------------------------------------------------------------

    Copyright:

        Ashley Laurent, Inc., 1999
        All Rights Reserved.

    ---------------------------------------------------------------

    Functional Description:

        Layout and bits of the 7751 processing unit control registers.

        The regs are mapped into virtual memory, so they are defined as array.
        Not all significant bits are defined, only those used by this driver.

    ---------------------------------------------------------------

    Environment:

        NT or BSD kernel mode

    ---------------------------------------------------------------

    Author:

        Stan Shkolnyy

    ---------------------------------------------------------------

    Revision history:

        *   15-Feb-99   initial revision

**********************************************************************/

#ifndef __PROC_PROC_H__
#define __PROC_PROC_H__

/****************************************************************************
 * 7751 procesing unit memory mapped registers layout.
 */
typedef struct _PROC_REGISTERS {

    ULONG  REG_00;         /* 00-03 - not defined by hardware */
    ULONG  CTL;            /* 04-07 */
    ULONG  INT_STATUS;     /* 08-0B */
    ULONG  CONFIG;         /* 0C-0F */
    ULONG  INT_ENABLE;     /* 10-13 */
    ULONG  STATUS;         /* 14-17 */
    ULONG  FIFO_STATUS;    /* 18-1B */
    ULONG  FIFO_CONFIG;    /* 1C-1F */

} PROC_REGISTERS, *PPROC_REGISTERS;

/*---------------------------------------------------------------------------
 * CTL - PU Control Register
 */
#define P_CTL_ENABLE_DMA                        0x00000002L
#define P_CTL_RESET                             0x00000001L

/*---------------------------------------------------------------------------
 * CONFIG - PU Configuration Register
 */

/*
 * Set when it's necessary to read chip ID
 */
#define P_CONF_GET_CHIP_ID                      0x00000020L

/*
 * The bits that are documented as "must be set"
 */
#define P_CONF_MUST_BE_SET_BITS                 0x00000340L

/*
 * Data processing
 */
#define P_CONF_BIG_ENDIAN                       0x00000080L
#define P_CONF_LITTLE_ENDIAN                    0x00000000L

/*
 * DRAM context memory columns
 */
#define P_CONF_CTX_MEM_DRAM_COLS_9              0x00000000L
#define P_CONF_CTX_MEM_DRAM_COLS_10             0x40000000L
#define P_CONF_CTX_MEM_DRAM_COLS_11             0x80000000L
#define P_CONF_CTX_MEM_DRAM_COLS_12             0xc0000000L

/*
 * DRAM refesh rate
 */
#define P_CONF_CTX_MEM_DRAM_REFRESH_FREQ_512    0x00000000L
#define P_CONF_CTX_MEM_DRAM_REFRESH_FREQ_256    0x00000800L
#define P_CONF_CTX_MEM_DRAM_REFRESH_FREQ_128    0x00001000L

/*
 * Context memory hardware type
 */
#define P_CONF_CTX_MEM_TYPE_DRAM                0x00000010L
#define P_CONF_CTX_MEM_TYPE_SRAM                0x00000000L

/*
 * Context memory configuration
 */
#define P_CONF_CTX_MEM_CONF_MULTICOMP           0x00000000L
#define P_CONF_CTX_MEM_CONF_UNICOMP_NO_RC4      0x00000006L
#define P_CONF_CTX_MEM_CONF_UNICOMP_RC4         0x00000004L

/*
 * Fixed CONFIG that is set once at init.time
 */
#define P_CONF_CONST_CONF       ( P_CONF_MUST_BE_SET_BITS | \
                                  P_CONF_CTX_MEM_DRAM_COLS_9 | \
                                  P_CONF_CTX_MEM_DRAM_REFRESH_FREQ_128 | \
                                  P_CONF_LITTLE_ENDIAN | \
                                  P_CONF_CTX_MEM_TYPE_DRAM | \
                                  P_CONF_CTX_MEM_CONF_UNICOMP_RC4)

/*
 * Fixed interrupt configuration that is set once at init.time -
 * interrupt on "abort" errors.
 */
#define P_INT_DATA_ERROR        0x00004000L
#define P_INT_INVAL_COMM        0x00008000L
#define P_INT_CONST_INTS        ( P_INT_DATA_ERROR | \
                                  P_INT_INVAL_COMM)

/****************************************************************************
 * hi/fn defines
 */
#define HIFN_REGISTER_NUMBER(x)      ((x) * 4)


/* -- Chip register definitions -- */

#define HIFN_DATA_REG                HIFN_REGISTER_NUMBER(0)

#define HIFN_CONFIG_REG              HIFN_REGISTER_NUMBER(3)
#define HIFN_CONFIG_DRAM_SIZE_BITS   3
#define HIFN_CONFIG_DRAM_SIZE_SHIFT  13
#define HIFN_CONFIG_DRAM_SIZE_MASK   MIDMASK(HIFN_CONFIG_DRAM_SIZE_SHIFT, HIFN_CONFIG_DRAM_SIZE_BITS)
#define HIFN_CONFIG_DRAM_SIZE_256K   (0L << HIFN_CONFIG_DRAM_SIZE_SHIFT)
#define HIFN_CONFIG_DRAM_SIZE_512K   (1L << HIFN_CONFIG_DRAM_SIZE_SHIFT)
#define HIFN_CONFIG_DRAM_SIZE_1M     (2L << HIFN_CONFIG_DRAM_SIZE_SHIFT)
#define HIFN_CONFIG_DRAM_SIZE_2M     (3L << HIFN_CONFIG_DRAM_SIZE_SHIFT)
#define HIFN_CONFIG_DRAM_SIZE_4M     (4L << HIFN_CONFIG_DRAM_SIZE_SHIFT)
#define HIFN_CONFIG_DRAM_SIZE_8M     (5L << HIFN_CONFIG_DRAM_SIZE_SHIFT)
#define HIFN_CONFIG_DRAM_SIZE_16M    (6L << HIFN_CONFIG_DRAM_SIZE_SHIFT)
/* refresh rate = input clock/this value */
#define HIFN_CONFIG_DRAM_REFR_BITS   2
#define HIFN_CONFIG_DRAM_REFR_SHIFT  11
#define HIFN_CONFIG_DRAM_REFR_MASK   MIDMASK(HIFN_CONFIG_DRAM_REFR_SHIFT, HIFN_CONFIG_DRAM_REFR_BITS)
#define HIFN_CONFIG_DRAM_REFR_512    (0L << HIFN_CONFIG_DRAM_REFR_SHIFT)
#define HIFN_CONFIG_DRAM_REFR_256    (1L << HIFN_CONFIG_DRAM_REFR_SHIFT)
#define HIFN_CONFIG_DRAM_REFR_128    (2L << HIFN_CONFIG_DRAM_REFR_SHIFT)
#define HIFN_CONFIG_TC_DRIVE_TOTEM   (1L << 8)
#define HIFN_CONFIG_BIG_ENDIAN       (1L << 7)
#define HIFN_CONFIG_LITTLE_ENDIAN    (0L << 7)
#define HIFN_CONFIG_BUS_WIDTH_32     (1L << 6)
#define HIFN_CONFIG_BUS_WIDTH_16     (0L << 6)
#define HIFN_CONFIG_CHIP_ID          (1L << 5)
#define HIFN_CONFIG_USE_DRAM         (1L << 4)
#define HIFN_CONFIG_USE_SRAM         (0L << 4)

#define HIFN_INTR_ENAB_REG           HIFN_REGISTER_NUMBER(4)

#define HIFN_STATUS_REG              HIFN_REGISTER_NUMBER(5)
#define HIFN_STATUS_SOURCE_FIFO_RDY  (1L << 13)
#define HIFN_STATUS_DEST_FIFO_RDY    (1L << 12)
#define HIFN_STATUS_DEST_OVERRUN     (1L << 9)

#define HIFN_FIFO_STAT_REG           HIFN_REGISTER_NUMBER(6)
#define HIFN_FIFO_SOURCE_BITS        7
#define HIFN_FIFO_SOURCE_SHIFT       8
#define HIFN_FIFO_SOURCE_MASK        MIDMASK(HIFN_FIFO_SOURCE_SHIFT, HIFN_FIFO_SOURCE_BITS)
#define HIFN_FIFO_SOURCE_VALUE(x)    (((x) << HIFN_FIFO_SOURCE_SHIFT) & HIFN_FIFO_SOURCE_MASK)
#define HIFN_FIFO_DEST_BITS          7
#define HIFN_FIFO_DEST_SHIFT         0
#define HIFN_FIFO_DEST_MASK          MIDMASK(HIFN_FIFO_DEST_SHIFT, HIFN_FIFO_DEST_BITS)
#define HIFN_FIFO_DEST_VALUE(x)      (((x) << HIFN_FIFO_DEST_SHIFT) & HIFN_FIFO_DEST_MASK)

#define HIFN_FIFO_CONFIG_REG         HIFN_REGISTER_NUMBER(7)
#define HIFN_FIFO_THRESH_SOURCE      0x04
#define HIFN_FIFO_THRESH_DEST        0x00
#define HIFN_FIFO_THRESHOLD          (HIFN_FIFO_SOURCE_VALUE(HIFN_FIFO_THRESH_SOURCE) | \
                                        HIFN_FIFO_DEST_VALUE(HIFN_FIFO_THRESH_DEST))

#define HIFN_REG_SPACE_SIZE          HIFN_REGISTER_NUMBER(8)

/* -- Chip register definitions -- */

#define PU_COMMAND0_ENCODE         (0L << 13)
#define PU_COMMAND0_DECODE         (1L << 13)
#define PU_COMMAND0_READ_RAM       (2L << 13)
#define PU_COMMAND0_WRITE_RAM      (3L << 13)

#define PU_DATA_REG                HIFN_DATA_REG

#define PU_INTR_STATUS_REG         HIFN_REGISTER_NUMBER(2)
#define PU_INTR_STAT_INVALID_CMD   (1L << 15)
#define PU_INTR_STAT_DATA_ERROR    (1L << 14)
#define PU_INTR_STAT_SRC_FIFO_RDY  (1L << 13)
#define PU_INTR_STAT_DST_FIFO_RDY  (1L << 12)
#define PU_INTR_STAT_SRC_NOT_ZERO  (1L << 11)
#define PU_INTR_STAT_DST_OVERRUN   (1L << 9)
#define PU_INTR_STAT_MAC_MISCOMP   (1L << 8)
#define PU_INTR_STAT_SRC_COMMAND   (1L << 7)
#define PU_INTR_STAT_SRC_CONTEXT   (1L << 6)
#define PU_INTR_STAT_SRC_DATA      (1L << 5)
#define PU_INTR_STAT_DST_DATA      (1L << 4)
#define PU_INTR_STAT_DST_CONTEXT   (1L << 3)
#define PU_INTR_STAT_DST_RESULT    (1L << 2)

#define PU_CONFIG_REG              HIFN_CONFIG_REG
#define PU_CONFIG_DRAM_SIZE_BITS   HIFN_CONFIG_DRAM_SIZE_BITS
#define PU_CONFIG_DRAM_SIZE_SHIFT  HIFN_CONFIG_DRAM_SIZE_SHIFT
#define PU_CONFIG_DRAM_SIZE_MASK   HIFN_CONFIG_DRAM_SIZE_MASK
#define PU_CONFIG_DRAM_SIZE_256K   HIFN_CONFIG_DRAM_SIZE_256K
#define PU_CONFIG_DRAM_SIZE_512K   HIFN_CONFIG_DRAM_SIZE_512K
#define PU_CONFIG_DRAM_SIZE_1M     HIFN_CONFIG_DRAM_SIZE_1M
#define PU_CONFIG_DRAM_SIZE_2M     HIFN_CONFIG_DRAM_SIZE_2M
#define PU_CONFIG_DRAM_SIZE_4M     HIFN_CONFIG_DRAM_SIZE_4M
#define PU_CONFIG_DRAM_SIZE_8M     HIFN_CONFIG_DRAM_SIZE_8M
#define PU_CONFIG_DRAM_SIZE_16M    HIFN_CONFIG_DRAM_SIZE_16M
/* refresh rate = input clock/this value */
#define PU_CONFIG_DRAM_REFR_BITS   HIFN_CONFIG_DRAM_REFR_BITS
#define PU_CONFIG_DRAM_REFR_SHIFT  HIFN_CONFIG_DRAM_REFR_SHIFT
#define PU_CONFIG_DRAM_REFR_MASK   HIFN_CONFIG_DRAM_REFR_MASK
#define PU_CONFIG_DRAM_REFR_1024   HIFN_CONFIG_DRAM_REFR_1024
#define PU_CONFIG_DRAM_REFR_512    HIFN_CONFIG_DRAM_REFR_512
#define PU_CONFIG_DRAM_REFR_256    HIFN_CONFIG_DRAM_REFR_256
#define PU_CONFIG_TC_ALL_PHASES    (1L << 9)
#define PU_CONFIG_TC_DRIVE_TOTEM   HIFN_CONFIG_TC_DRIVE_TOTEM
#define PU_CONFIG_BIG_ENDIAN       HIFN_CONFIG_BIG_ENDIAN
#define PU_CONFIG_LITTLE_ENDIAN    HIFN_CONFIG_LITTLE_ENDIAN
#define PU_CONFIG_BUS_WIDTH_32     HIFN_CONFIG_BUS_WIDTH_32
#define PU_CONFIG_BUS_WIDTH_16     HIFN_CONFIG_BUS_WIDTH_16
#define PU_CONFIG_CHIP_ID          HIFN_CONFIG_CHIP_ID
#define PU_CONFIG_USE_DRAM         HIFN_CONFIG_USE_DRAM
#define PU_CONFIG_USE_SRAM         HIFN_CONFIG_USE_SRAM
#define PU_CONFIG_CONTEXT_IN_HOST  (1L << 3)
#define PU_CONFIG_ONE_COMP_CONTEXT (1L << 2)
#define PU_CONFIG_64B_ENCR_CONTEXT (1L << 1)
#define PU_CONFIG_FREEZE_ON_OVRRUN (1L << 0)

#define PU_INTR_ENAB_REG           HIFN_INTR_ENAB_REG
#define PU_INTR_ENAB_INVALID_CMD   (1L << 15)
#define PU_INTR_ENAB_DATA_ERROR    (1L << 14)
#define PU_INTR_ENAB_SRC_FIFO_RDY  (1L << 13)
#define PU_INTR_ENAB_DST_FIFO_RDY  (1L << 12)
#define PU_INTR_ENAB_SRC_NOT_ZERO  (1L << 11)
#define PU_INTR_ENAB_DST_OVERRUN   (1L << 9)
#define PU_INTR_ENAB_MAC_MISCOMP   (1L << 8)
#define PU_INTR_ENAB_SRC_COMMAND   (1L << 7)
#define PU_INTR_ENAB_SRC_CONTEXT   (1L << 6)
#define PU_INTR_ENAB_SRC_DATA      (1L << 5)
#define PU_INTR_ENAB_DST_DATA      (1L << 4)
#define PU_INTR_ENAB_DST_CONTEXT   (1L << 3)
#define PU_INTR_ENAB_DST_RESULT    (1L << 2)

#define PU_STATUS_REG              HIFN_STATUS_REG
#define PU_STATUS_INVALID_CMD      (1L << 15)
#define PU_STATUS_DATA_ERROR       (1L << 14)
#define PU_STATUS_SOURCE_FIFO_RDY  HIFN_STATUS_SOURCE_FIFO_RDY
#define PU_STATUS_DEST_FIFO_RDY    HIFN_STATUS_DEST_FIFO_RDY
#define PU_STATUS_SOURCE_NOT_ZERO  (1L << 11)
#define PU_STATUS_DEST_OVERRUN     HIFN_STATUS_DEST_OVERRRUN
#define PU_STATUS_MAC_MISCOMPARE   (1L << 8)
#define PU_STATUS_SOURCE_COMMAND   (1L << 7)
#define PU_STATUS_SOURCE_CONTEXT   (1L << 6)
#define PU_STATUS_SOURCE_DATA      (1L << 5)
#define PU_STATUS_DEST_DATA        (1L << 4)
#define PU_STATUS_DEST_CONTEXT     (1L << 3)
#define PU_STATUS_DEST_RESULT      (1L << 2)

#define PU_FIFO_STAT_REG           HIFN_FIFO_STAT_REG
#define PU_FIFO_SOURCE_BITS        HIFN_FIFO_SOURCE_BITS
#define PU_FIFO_SOURCE_SHIFT       HIFN_FIFO_SOURCE_SHIFT
#define PU_FIFO_SOURCE_MASK        HIFN_FIFO_SOURCE_MASK
#define PU_FIFO_SOURCE_VALUE(x)    HIFN_FIFO_SOURCE_VALUE(x)
#define PU_FIFO_DEST_BITS          HIFN_FIFO_DEST_BITS
#define PU_FIFO_DEST_SHIFT         HIFN_FIFO_DEST_SHIFT
#define PU_FIFO_DEST_MASK          HIFN_FIFO_DEST_MASK
#define PU_FIFO_DEST_VALUE(x)      HIFN_FIFO_DEST_VALUE(x)

#define PU_FIFO_CONFIG_REG         HIFN_FIFO_CONFIG_REG
#define PU_FIFO_THRESH_SOURCE      HIFN_FIFO_THRESH_SOURCE
#define PU_FIFO_THRESH_DEST        HIFN_FIFO_THRESH_DEST
#define PU_FIFO_THRESHOLD          HIFN_FIFO_THRESHOLD


/* ---------------------------------------------------------------------------
 * The actual sizes of structures used in the chip */

#define PU_SIZEOFBASE         8
#define PU_SIZEOFCOMPRESSION  8
#define PU_SIZEOFPADDING      4
#define PU_SIZEOFMAC          8
#define PU_SIZEOFENCRYPTION   8
#define PU_SIZEOFRBASE        8
#define PU_SIZEOFRCOMPRESSION 4
#define PU_SIZEOFRMAC         4
#define PU_SIZEOFRENCRYPTION  4

#define PU_MACLENGTH_TRUNC   12
#define PU_MACLENGTH_MD5     16
#define PU_MACLENGTH_SHA     20

#define PU_MAC_KEY_SIZE      64
#define PU_MAC_CONTEXT_SIZE  40

#define PU_DES_KEY_SIZE       8
#define PU_3DES_KEY_SIZE     24
#define PU_DES_IV_SIZE        8
#define PU_3DES_IV_SIZE       8
#define PU_RC4_KEY_SIZE     260
#define PU_RC4_CONTEXT_SIZE 260

#define PU_CRAM_CONTEXT_BITS 14

/* Number of bytes of padding required after a write/read RAM command struc */
#define PU_SIZEOFCRAMCOMMANDFLUFF 8

/* Number of bytes of padding required after a enc/dec command that uses
   padding but not any of the other three ops */
#define PU_SIZEOFPADDINGONLYFLUFF 4

/* Number of bytes of padding required after a encode/decode command struct
   that uses none of the four processing units */
#define PU_SIZEOFPASSTHROUGHFLUFF 8

#endif

