#include "osindependence.h"
#include "debug.h"
#include "processor.h"
#include "proc_hw_proc.h"
#include "proc_hw_dma.h"

#define SIGNATURE_MASK 0x80080125

/****************************************************************************
 * XOR all bits together to calculate odd parity.
 * Returns 1 if an odd number of bits are set, 0 if even.
 */
static
unsigned
Parity(
    ULONG val
    )
{
    val ^= val >> 16;
    val ^= val >> 8;
    val ^= val >> 4;
    val ^= val >> 2;
    val ^= val >> 1;
    val &= 1;
    return val;
}

/****************************************************************************
 * Calculates next signature by the current one
 */
static
void
ProcNextSignature
    (
        IN  PPROCESSOR   pThis,
        IN  ULONG        iterations
    )
{
    unsigned i;

    ASSERT( pThis->Type == PROCESSOR_SIGN);

    for( i = 0; i < iterations; i++ )
    {
        pThis->curSignature = ( pThis->curSignature << 1 ) |
                              Parity( pThis->curSignature & SIGNATURE_MASK );
    }
}

/****************************************************************************
typedef enum
{
    INITIAL_LOCKED = 0,
    PARTIALLY_UNLOCKED,
    FULLY_UNLOCKED,
    LOCKED

} SecurLevel;
****************************************************************************/
BOOLEAN
PuGetSecurityLevel
    (
        IN  PPROCESSOR  pThis,
        OUT SecurLevel  *pSecLevel
    )
{
    ULONG       config;
    ULONG       ChipID;
    BOOLEAN     bFailure;

    ASSERT( pThis->Type == PROCESSOR_SIGN);

    config = READ_PROC32( CONFIG );
    WRITE_PROC32(CONFIG, config | P_CONF_GET_CHIP_ID);
    ChipID = READ_PROC32( STATUS) & 0x0000FF00L;

    DebugPrint((3, "    chip ID = 0x%04x\n", ChipID));

    bFailure = FALSE;
    switch( ChipID)
    {
        case 0x3000:
            *pSecLevel = INITIAL_LOCKED;
            break;
        case 0x1000:
            *pSecLevel = PARTIALLY_UNLOCKED;
            break;
        case 0x1100:
            *pSecLevel = FULLY_UNLOCKED;
            break;
        default:
            DebugPrint((3, "    invalid chip ID!!!\n"));
            bFailure = TRUE;
            break;
    }

    return !bFailure;
}

/****************************************************************************
****************************************************************************/

static int  desiredIndex[] = { -1, 8, 12, 13 };

BOOLEAN
PuUnlock
    (
        IN  PPROCESSOR  pThis,
        IN  SecurLevel  NewLevel,
        IN  PUNLOCK_KEY UnlockKey
    )
{
    SecurLevel  DetectedLevel;
    SecurLevel  ResultingLevel;

    int     i;
    int     initialIndex;
    int     finalIndex;

    BOOLEAN bFullUnlock;

    DebugPrint(( 3, "PuUnlock( newlevel=%d )\n", NewLevel));
    ASSERT( pThis->Type == PROCESSOR_SIGN);

    /*
     * Start with a clean configuration
     */
    WRITE_DMA32( CONFIG,    D_CONF_UNLOCK |
                            D_CONF_MASK_MASTER_RESET |
                            D_CONF_MASK_DMA_RESET    |
                            D_CONF_CPU_CONTROLS_LAST);

    /*
     * Get chip state and select initial index by this state
     */
    if ( !PuGetSecurityLevel( pThis, &DetectedLevel))
    {
        return FALSE;
    }

    /*
     * Select starting loop index by the detected security level
     *
     * Detected Level
     *
     * ini_unl      0
     * part_unl     9
     * fully_unl    13
     * locked       14
     */
    initialIndex = desiredIndex[ DetectedLevel] + 1;

    /*
     * Select final loop index by the new security level
     *
     * New Level
     *
     * ini_unl      -1
     * part_unl     8
     * fully_unl    12
     * locked       13
     */
    finalIndex = desiredIndex[ NewLevel];

    /*
     * If we need to move from the initial locked level, read the starting
     * signature, otherwise, the current signature will be used
     */
    if( ( DetectedLevel == INITIAL_LOCKED ) && ( NewLevel != INITIAL_LOCKED ) )
    {
        /*
         * Read in initial signature
         */
        pThis->curSignature = READ_DMA32( START_SIGNATURE );

        /*
         * Do a dummy write to wake up Signature logic
         */
        WRITE_DMA32( UNLOCK_SIGNATURE, 0);

        /*
         * Allow time for signature to settle
         */
        delay(1);
    }

    /*
     * If the signature is still 0 (locked up), there's no good we can do.
     * We only get one chance to read that signature register.
     */
    if( pThis->curSignature == 0)
    {
        DebugPrint(( 3, "Signature returned 0! Hi/fn may already be unlocked, unlock cancelled\n"));
        return FALSE;
    }

    /*
     * Check indexes
     */
    if ( initialIndex == 0 &&
         finalIndex == 8)
    {
        bFullUnlock = FALSE;
    }
    else if ( initialIndex == 0 &&
              finalIndex == 12)
    {
        bFullUnlock = TRUE;
    }
    else
    {
        DebugPrint(( 3, "Strange unlock key array range! Unlock cancelled\n"));
        return FALSE;
    }

    /*
     * Use the first part of the unlock key and perform the partial unlock
     */
    for( i = 0; i < arraysize( UnlockKey->Level1Value); i++ )
    {
        /*
         * Calculate the next signature by the current one and unlock key,
         * write it to the device and
         * make a time delay for the signature to settle
         */
        ProcNextSignature( pThis, 257 + UnlockKey->Level1Value[ i] );
        WRITE_DMA32( UNLOCK_SIGNATURE, pThis->curSignature);
        delay(1);
    }

    /*
     * If the full unlock is required, continue the procedure with the appendix
     * to the unlock key to perform the full unlock
     */
    if ( bFullUnlock)
    {
        for( i = 0; i < arraysize( UnlockKey->Level2Appendix); i++ )
        {
            /*
             * Calculate the next signature by the current one and unlock key,
             * write it to the device and
             * make a time delay for the signature to settle
             */
            ProcNextSignature( pThis, 257 + UnlockKey->Level2Appendix[ i] );
            WRITE_DMA32( UNLOCK_SIGNATURE, pThis->curSignature);
            delay(1);
        }
    }

    return TRUE;
}
