/**********************************************************************

    Module: proc_xfer.c

        Product: kernel mode driver for "PowerCrypt" encryption
        acceleration board based on Hi/fn 7751 chip
        by Ashley Laurent, Inc.

    ---------------------------------------------------------------

    Copyright:

        Ashley Laurent, Inc., 1999
        All Rights Reserved.

    ---------------------------------------------------------------

    Functional Description:

        Implementation of the PROCESSOR object that performs all the
        work of the 7751 programming.

    ---------------------------------------------------------------

    Environment:

        NT or BSD kernel mode

    ---------------------------------------------------------------

    Author:

        Stan Shkolnyy

    ---------------------------------------------------------------

    Revision history:

        *   15-Feb-99   initial revision

**********************************************************************/

#include "osindependence.h"
#include "debug.h"
#include "processor.h"
#include "procbase.h"

/****************************************************************************
****************************************************************************/
void
ProcTryToStartDMATransfer(
    IN      PPROCESSOR      pThis,
    IN      PROC_XFER_DIR   XferDir
    )
{
    PPROC_DMA_CTL_BLOCK    pDmaCB;
    PDMA_CTX               pDmaCtx;
    PPROC_DMA_CHANNEL      pDmaChan;

    PPROC_QUEUE_ENTRY       pQEntry;
    ULONG                   nMaxPages;

    DebugPrint((2, "ProcTryToStartTransfer[%d]\n", XferDir));
    ASSERT( pThis->Type == PROCESSOR_SIGN);

    ASSERT( XferDir < 2 && XferDir >= 0);

    pDmaCB = &pThis->DmaCB[ XferDir];
    pDmaCtx = pDmaCB->pDmaCtx;

    /*
     * Acquire the spinlock, because we are going to look into the request
     * queue that it protects and also need to serialize the actions being
     * performed on the DMA contexts. 
     */
    PCrptSpinLockAcquire( &pThis->SpinLock);

    /*
     * If the DMA ctx is busy with the waiting for mapping, exit now.
     */
    if ( pDmaCB->bDmaCtxBusy)
    {
        DebugPrint((2, "    DMA Ctx is busy with waiting for mapping - exiting\n"));
        PCrptSpinLockRelease( &pThis->SpinLock);
        return;
    }

    /*
     * If the DMA ctx is busy with other active parallel mappings (there are
     * no free DMA channels) - exit now
     */
    if ( pDmaCB->nMapsInProgress >= DMA_CTX_MAX_PARALLEL_MAPS)
    {
        ASSERT( pDmaCB->nMapsInProgress == DMA_CTX_MAX_PARALLEL_MAPS);
        DebugPrint((2, "    DMA Ctx already has %d active mappings - exiting\n", pDmaCB->nMapsInProgress));
        PCrptSpinLockRelease( &pThis->SpinLock);
        return;
    }

    /* */
    /* BUGBUG - can this func. execute on the empty queue? when? */
    /* */

    /*
     * Ok, we CAN do the mapping, the DMA Ctx state allows it. Check to see
     * whether we NEED TO. We need, if there is a request in the proc.queue
     * which has the non-zero DataRemainCnt for our given transfer direction
     *
     * If there are no requests in the PROCESSOR queue - there is nothing to map.
     */
    if ( pThis->BusyEntryCnt == 0)
    {
        DebugPrint((2, "    No requests found in the proc.queue - exiting\n"));
        PCrptSpinLockRelease( &pThis->SpinLock);
        return;
    }

    /*
     * There are requests in the PROCESSOR queue - try to find one that
     * needs to be transferred in the given direction (DMA ctx). Walk the
     * request queue from the oldest entry toward the newest and check the
     * Transfer Progress Blocks for non-zero DataRemainCnt to detect which
     * request needs a data transfer. 
     */ 
    pQEntry = NULL;
    {
        int i;
        int QEntriesToCheck;

        QEntriesToCheck = pThis->BusyEntryCnt;

        /*
         * Begin to look from the oldest entry, because no newer entry's data
         * should be mapped until the data of older ones are all mapped.
         */
        i = pThis->iOldestQueueEntry;

        while ( QEntriesToCheck != 0 && pQEntry == NULL)
        {
            DebugPrint((2, "    checking QEntry %d\n", i));

            if ( pThis->Queue[ i].XferProgress[ XferDir].DataRemainCnt != 0)
            {
                DebugPrint((2, "        starting transfer for QEntry %d\n", i));

                /*
                 * This is what we've been looking for
                 */
                pQEntry = &pThis->Queue[ i];

                /*
                 * Since the DataRemainCnt != 0, the queue entry must have
                 * been marked as "started" by the ProcessorStartQueuedRequest,
                 * since that func puts non-zero value before marking it "started"
                 */
                ASSERT( pQEntry->bStarted);
            }
            else
            {
                IncrementQueueIndex( i);
            }
            QEntriesToCheck --;
        }
    }

    /*
     * If no request needs a transfer, exit now
     */
    if ( pQEntry == NULL)
    {
        DebugPrint((2, "    no requests found that need transfer\n"));
        PCrptSpinLockRelease( &pThis->SpinLock);
        return;
    }

    /*
     * We have just determined that the following conditions are met:
     *
     *  1. DMA Ctx is available for mapping
     *  2. DMA Ctx has a free channel
     *  3. There is a request in the proc.queue that needs mapping
     *
     * We can and need to start a new transfer - set the pDmaCB->bDmaCtxBusy
     * to grab the DMA context so that no other invocation of this func could
     * execute until this flag will be cleared by the map control routine.
     * (Note that this func checks the flag in its very beginning).
     *
     * Now it's perfectly safe to manipulate that state, change request's
     * mapping-related counters, build the partial MDL and make the new
     * mapping in the MapControl routine.
     */

    /*
     * Invalidate condition 1, to prevent others from using this DMA Ctx.
     * (It will be freed back in the ProcMapControl func)
     */
    pDmaCB->bDmaCtxBusy = TRUE;

    /*
     * Possibly invalidate condition 2 by acquiring a DMA channel
     * (It will be freed back as soon as the data transfer ends, in
     * ProcTryToCompleteDMATransfers func) - increment MapsCnt and
     * mark a channel as "busy". Note, that a free channel is guaranteed
     * to be found, since we've detected condition 2.
     */
    pDmaCB->nMapsInProgress++;
    pDmaChan = NULL;
    {
        int i;
        for ( i = 0; pDmaChan == NULL && i < DMA_CTX_MAX_PARALLEL_MAPS; i++)
        {
            if ( !pDmaCB->Chan[ i].bInUse)
            {
                pDmaChan = &pDmaCB->Chan[ i];
            }
        }
    }
    ASSERT( pDmaChan != NULL);

    pDmaChan->bInUse = TRUE;
 
    ASSERT( pDmaChan->pLastSGDesc == NULL);

    /*
     * Setting last desc. ptr to NULL protects this channel from being checked
     * and freed by ProcTryToCompleteDMATransfers. This desc.ptr will be set
     * by forthcoming ProcMapControl func, when the actual mapping occurs.
     */
    pDmaChan->pLastSGDesc = NULL;

    /*
     * We've locked the DMA ctx and reserved a DMA Channel, so drop the spinlock.
     */
    PCrptSpinLockRelease( &pThis->SpinLock);

    /*
     * (!) The following code never executes concurrently for the same DMA Ctx
     * and assumes that the QEntry is not changed in any way. Though the
     * DataRemainCnt in the current request is still non-zero, another invocation
     * of this func can't get here, because we've locked the DMA Ctx. When we
     * unlock it in the ProcMapControl, we've already decremented the
     * DataRemainCnt, so the new invocation will work properly - either map
     * the next chunk or ignore this request as been already fulfilled in part
     * of tranfer.
     */

    /*
     * Remember which DMA channel is to be used by the forthcoming invocation
     * of the ProcMapControl in the current DMA ctx. Since the DMA Ctx can't
     * be touched until we free it, we are free to use the "cargo" var inside of
     * it to pass the channel ptr to the ProcMapControl. We have to do this
     * because we have more than one channel and which of them was reserved
     * for the forthcoming transfer is an arbitrary choice - just the first
     * free one.
     */
    pDmaCtx->UserUlong4 = (ULONG) pDmaChan;

    /*
     * Attention! This function may not be able to map _all_ of the data that are
     * to be mapped for the current request. DmaCtxIsPartialTransfers( pDmaCtx)
     * gives the status whether this is ever possible. This status is
     * constant and is determined at the time when the PROCBASE initializes
     * the DMA Ctx.
     *
     * So, this function may be invoked not only to start transfer for a fresh
     * request, but also to start a next chunk of a request data.
     *
     * The "if" below checks the mentioned limiting condition and knows not to
     * bother with creating the partial MDL, if it sees that there is no
     * limitation and transfers can always be mapped wholly ("else" branch).
     * Otherwise, each QEntry has two XferProgress structs (for two directions)
     * that maintain the transfer progress state in terms of how much data
     * have been already mapped (this is where DataRemainCnt lives).
     */

    /*
     * This number must match the num.of pages in the partial MDL below
     * WORKWORK = - currently we always ask AllocAdapChan for this number
     * of map regs - it is fixed as the half of available map regs cnt.
     * Maybe, we can ask less, if possible
     */
    nMaxPages = DmaCtxGetMaxMapRegs( pDmaCtx) / 2;
    ASSERT( nMaxPages == 0x41);

    /*
     * Note the code below assigning pDmaChan->bLastMap. This field tells the
     * ProcMapControl() whether it should set the LAST bit in the last SG
     * descriptor it will load in the course of mappping of the specified MDL.
     */
    if ( DmaCtxIsPartialTransfers( pDmaCtx))
    {
        ULONG                   Length;

        ASSERT(0);

        /*
         * We operate with a limited number of map regs - build the partial MDL
         * and map only part of the data.
         */
        Length = nMaxPages * PAGE_SIZE - BYTE_OFFSET( pQEntry->XferProgress[ XferDir].DataCurrVa);
        if ( pQEntry->XferProgress[ XferDir].DataRemainCnt <= Length)
        {
            pDmaChan->bLastMap = TRUE;
            Length = pQEntry->XferProgress[ XferDir].DataRemainCnt;
        }
        else
        {
            pDmaChan->bLastMap = FALSE;
        }

        DebugPrint((2, "    making partial MDL. Va=0x%08x, Len=0x%08x\n", pQEntry->XferProgress[ XferDir].DataCurrVa, Length));

        PCrptPrepareMdlForReuse( pDmaChan->pMdlToMap);

        PCrptBuildPartialMdl(   pQEntry->XferProgress[ XferDir].pFullMdl,
                                pDmaChan->pMdlToMap,
                                pQEntry->XferProgress[ XferDir].DataCurrVa,
                                Length);

        pQEntry->XferProgress[ XferDir].DataCurrVa += Length;
        pQEntry->XferProgress[ XferDir].DataRemainCnt -= Length;
    }
    else
    {
        /*
         * We operate with enough map regs to always map transfers wholly -
         * map the source full MDL. The DataRemainCnt goes to 0 immediately to
         * indicate that the data no more needs to be transferred in this DMA Ctx 
         */
        DebugPrint((2, "    using full MDL. Va=0x%08x\n", pQEntry->XferProgress[ XferDir].DataCurrVa));

        pDmaChan->bLastMap = TRUE;

        pDmaChan->pMdlToMap = pQEntry->XferProgress[ XferDir].pFullMdl;
        pQEntry->XferProgress[ XferDir].DataRemainCnt   = 0;
    }

    /*
     * This num is stored for later call to the DmaCtxCompleteTransfer
     */
    pDmaChan->NumOfMapRegs = nMaxPages;
    
    DebugPrint((2, "    nMaxPages= 0x%x\n", nMaxPages));

    /*
     * Order the ProcMapControl routine to be called when the map regs
     * become available.
     *
     * Remember that this call may return without calling ProcMapControl()
     * in NT, but, in FreeBSD, it will call it unconditionally.
     */
    DmaCtxAllocateAdapterChannel(   pDmaCtx,
                                    pDmaChan->NumOfMapRegs,
                                    pDmaChan->pMdlToMap,
                                    ProcMapControl,
                                    (PVOID) pDmaCtx);

    DebugPrint((2, "    finishing ProcTryToStartDMATransfer[%d]\n", XferDir));
}

/****************************************************************************
 * DMA_CTX user-defined routine that is called as a consequence of a call to
 * DmaCtxAllocateAdapterChannel. In this routine, the user can call
 * DmaCtxMapTransfer to map its phys.RAM range to a number of logical DMA
 * ranges.
 *
 * The callback must store away the MapRegisterBase value passed to it.
 * It will need this value to pass it to DmaCtxFreeMapRegisters later, when
 * the transfer is over.
 *
 * We chose not to store this and other values internally, because there may be
 * two simultaneous transfers in progress in a single DMA_CTX and we would have
 * to store 2 sets of such values and introduce "handles" to give the user, etc.
 *
 * They are valid only until the user returns from the callback routine.
 * After the routine returns, a new transfer may be initiated in the DMA
 * context, since the DMA_CTX frees "adapter channel" after calling this
 * routine.
 */

void
ProcMapControl
    (
        IN PVOID        Context,
        IN PVOID        MapRegBase
    )
{
    PPROCESSOR              pThis;
    PDMA_CTX                pDmaCtx;
    PROC_XFER_DIR           XferDir;

    PPROC_DMA_CTL_BLOCK     pDmaCB;
    PPROC_DMA_CHANNEL       pDmaChan;

    PMEM_DESC_LIST          pMdl;
    ULONG                   RemainingDataLength;
    PUCHAR                  CurVa;

    /*
     * Withdraw the "coordinates" of the work from the DMA context "cargo" vars.
     * They are were stored there when this MapControl routine was queued.
     * Since the access to MapControl queueing is serialized, it's safe to
     * pass the "coordinates" this way
     */
    pDmaCtx     = (PDMA_CTX)            Context;
    pThis       = (PPROCESSOR)          pDmaCtx->UserUlong1;
    XferDir     = (PROC_XFER_DIR)       pDmaCtx->UserUlong2;

    ASSERT( pDmaCtx->Type == DMA_CTX_SIGN);
    ASSERT( pThis->Type == PROCESSOR_SIGN);
    ASSERT( pThis->DmaCB[ XferDir].pDmaCtx == pDmaCtx);


    pDmaChan    = (PPROC_DMA_CHANNEL)   pDmaCtx->UserUlong4;
    pDmaCB      = &pThis->DmaCB[ XferDir];

    DebugPrint((2, "ProcMapControl[%d]\n", XferDir));

    ASSERT( XferDir < 2 && XferDir >= 0);
    ASSERT( pDmaChan->bInUse);
    ASSERT( pDmaCB->nMapsInProgress > 0);

    /*
     * This ptr is stored for later call to the DmaCtxCompleteTransfer
     */
    pDmaChan->pMapRegBase = MapRegBase;

    /*
     * This loop always maps the whole MDL. The num of map regs is always
     * enough, because we split data to smaller chunks in
     * ProcTryToStartDMATransfer, if necessary
     */
    pMdl                = pDmaChan->pMdlToMap;
    RemainingDataLength = CNT_FROM_MDL( pMdl);
    CurVa               = VIRT_ADDR_FROM_MDL( pMdl);

    while ( RemainingDataLength != 0)
    {
        ULONG   Length;
        ULONG   LogAddr;

#ifdef DBG
        PUCHAR  _dbg_VA;
        _dbg_VA = CurVa;
#endif
        /*
         * Map the next chunk of the buffer - determine its "Length" and
         * "LogAddr". We always request to map the whole remaining length,
         * but the function may decide to map only part of it.
         */
        Length = RemainingDataLength;
        LogAddr = DmaCtxMapTransfer( pDmaCtx,
                                     MapRegBase,
                                     pMdl,
                                     CurVa,
                                     &Length);
        ASSERT( Length != 0);

/*        DebugPrint((2, "    LA=0x%08x VA=0x%08x Len=0x%08x DescFlags=0x%04x\n", LogAddr, CurVa, Length, DescFlags)); */

        CurVa += Length;
        RemainingDataLength -= Length;

        /*
         * This loop splits the mapped range to pieces suitable for one
         * S/G descriptor and loads as many descriptors as necessary to
         * transfer the logical address range that has just been mapped.
         * The 7751 SG descriptor can address only SG_DESC_MAX_CNT bytes.
         *
         * There is always enough free SG descriptors, since they are allocated
         * for the worst case when each logical range <= PAGE_SIZE.
         */
        while ( Length != 0)
        {
            PSG_DESC    pCurSGDesc;
            USHORT      DescFlags;

#ifdef DBG
            USHORT      _dbg_OldFlags;
            ULONG       _dbg_i1stFreeDesc;
            USHORT      _dbg_DataSize;
            ULONG       _dbg_LA;
#endif
            /*
             * Get ptr to the current descriptor in the ring
             */
            pCurSGDesc = RingGetFirstFreeDescPtr( pDmaCB->pDataRing);

#ifdef DBG

            _dbg_i1stFreeDesc = RingGetFirstFreeDescIndex( pDmaCB->pDataRing);

            /*
             * We never load already loaded descriptors :-)
             */
            _dbg_OldFlags = SgDescGetFlags( pCurSGDesc);

    #ifndef _SKIP_MODE
/*    #ifndef _NO_DST */

            ASSERT( !( _dbg_OldFlags & SG_DESC_FLAG_VALID));
/*    #endif */
    #endif

            _dbg_OldFlags &= ~( SG_DESC_FLAG_LAST | SG_DESC_FLAG_MASKDONE);

/*    #ifndef _NO_DST */
            ASSERT( _dbg_OldFlags == 0);
/*    #endif */

    #ifndef _SKIP_MODE
/*            ASSERT( SgDescGetDataSize( pCurSGDesc) == 0); */
    #endif

#endif
            /*
             * Load the "logical address" into it
             */
            SgDescSetDataPhysAddr( pCurSGDesc, LogAddr);

#ifdef DBG
            _dbg_LA = LogAddr;
#endif
            /*
             * Calculate and Load the data chunk size
             */
            if ( Length <= SG_DESC_MAX_CNT)
            {
                SgDescSetDataSize( pCurSGDesc, LOWORD( Length));
                Length = 0;
            }
            else
            {
                SgDescSetDataSize( pCurSGDesc, SG_DESC_MAX_CNT);
                Length -= SG_DESC_MAX_CNT;
                LogAddr += SG_DESC_MAX_CNT;
            }
#ifdef DBG
            _dbg_DataSize = SgDescGetDataSize( pCurSGDesc);
#endif
            /*
             * Calculate which control bits (apart from the essential VALID)
             * are to be set.
             *
             * If the S/G descriptor is the last in this mapping, not set the
             * MASK_DONE bit and that will trigger a DONE int.
             *
             * In addition to the above action, if the current mapping maps the
             * last MDL of the user request, set the LAST bit that will stop
             * the corresponding 7751 data DMA engine.
             */
#ifndef _SKIP_MODE
            DescFlags = SG_DESC_FLAG_VALID;
#else
            DescFlags = 0;
#endif

            if ( Length == 0 && RemainingDataLength == 0)
            {
                /*
                 * Last desc in the current mapping - don't disable DONE int
                 * (as it's done in the "else"), to trigger a DONE interrupt
                 * which is used to call ProcTryToCompleteDMATransfers which
                 * will free this mapping and this DMA channel.
                 */
                if ( pDmaChan->bLastMap)
                {
                    /*
                     * This is also the last mapping in the user req. - set the
                     * LAST bit to cause the 7751 to stop walking this desc.ring
                     * until the next request.
                     */
                    DescFlags |= SG_DESC_FLAG_LAST;
                }

                /*
                 * Acquire the spinlock, because we're going to change the state of
                 * DMA channel by setting a value in pDmaChan->pLastSGDesc.
                 * We can do it here, because the loop will never be executed again -
                 * we've reached the finishing conditions
                 */
                PCrptSpinLockAcquire( &pThis->SpinLock);

                /*
                 * Put the pointer to the last SG desc into the current DMA
                 * channel - the ProcTryToCompleteTransfer will then consider
                 * it as a candidate for freeing and will check the VALID bit
                 * in this desc. If there will be no VALID bit
                 * in this descriptor, it will complete the related transfer,
                 * and free this mapping by calling DmaCtxCompleteTransfer and
                 * mark this DMA channel as free.
                 *
                 * Note also! It's important to assign this ptr _before_
                 * setting bits in the descriptor, to prevent the occurence of
                 * the corresponding DONE int before the ptr is stored in the
                 * DMA channel
                 */
                ASSERT( pDmaChan->pLastSGDesc == NULL);
                pDmaChan->pLastSGDesc = pCurSGDesc;
            }
            else
            {
                /*
                 * The current mapping is not finished yet - disable DONE int
                 * at this descriptor
                 */
                DescFlags |= SG_DESC_FLAG_MASKDONE;
            }

            DebugPrint((2, "    VA=%08x Desc#%04x LA=%08x Size=%04x flags=0x%04x\n",
                        _dbg_VA, _dbg_i1stFreeDesc, _dbg_LA, _dbg_DataSize, DescFlags));

            /*
             * Finally, put the bits into the descriptor to cause it to be
             * processed by the 7751.
             */
            SgDescSetFlags( pCurSGDesc, DescFlags);

            /*
             * We've loaded the descriptor, switch the ring to the next one
             */
            RingAdvanceFirstFreeDescIndex( pDmaCB->pDataRing);

        } /* while ( Length != 0) */

    } /* while ( RemainingDataLength != 0) */

    /*
     * Release this DMA context acquired earlier in ProcTryToStartTransfer.
     * After that, other invocations of ProcTryToStartTransfer may compete for
     * this DMA ctx and acquire it again (if there is a free DMA channel in it).
     */
    pDmaCB->bDmaCtxBusy = FALSE;

    /*
     * Release the spinlock since we've finished chanding the DMA Ctl Block state
     */
    PCrptSpinLockRelease( &pThis->SpinLock);

    #ifdef _NT_DRIVER_BUILD_
        /*
         * Request "software" mapping interrupt (implemented with DPC),
         * so that the handler could perform another mapping, if necessary.
         *
         * It will cause our ProcServiceRegisteredInts() routine to run as
         * so the true int has happened, though it hasn't. We need it to run to
         * continue possibly pending DMA mappings through calling
         * ProcTryToStartDMATransfer() which in turn will again call this
         * ProcMapControl().
         */
        #ifdef _NO_DST
            DebugPrint((2, "    finishing ProcMapControl\n"));
        #else
            DebugPrint((2, "    finishing ProcMapControl and request \"mapping\" int\n"));
            PCrptInterlockedIncrement( &pThis->IntsState.MapIntCnt[ XferDir]);
            ProcBaseRequestInt( pThis->pProcBase);
        #endif
    #else
        /*
         * We are in FreeBSD and don't need to request a "software interrupt"
         * to organize a sequence of DMA mappings. This is needed in NT where
         * the ProcTryToStartDMATransfer() may not perform mapping even if
         * there is a need (due to map registers unavailability), so we may
         * need to try later. FreeBSD doesn't operate on computers with map
         * regs, so there can be no such limitations and such a need.
         */
    #endif
}

/****************************************************************************
****************************************************************************/
void
ProcTryToCompleteDMATransfers(
    IN      PPROCESSOR      pThis,
    IN      PROC_XFER_DIR   XferDir
    )
{
    PPROC_DMA_CTL_BLOCK    pDmaCB;
    PPROC_DMA_CHANNEL       pDmaChan;

    DebugPrint((2, "ProcTryToCompleteDMATransfers[%d]\n", XferDir));
    ASSERT( pThis->Type == PROCESSOR_SIGN);

    ASSERT( XferDir < 2 && XferDir >= 0);

    pDmaCB = &pThis->DmaCB[ XferDir];

    do
    {
        /*
         * Acquire the spinlock, because we are going to walk DMA channels in
         * the DMA Control Block and complete a mapping, so we want nobody to
         * change the CtlBlock while we are browsing it (it could be changed
         * by ProcTryToStartTransfer())
         */
        PCrptSpinLockAcquire( &pThis->SpinLock);

        /*
         * Find a DMA channel that is marked "in-use", then check to see
         * whether the "last" SG descriptor that it points to is processed
         * (has the VALID bit cleared). Then the DMA transfer in this channel
         * has been finished and we free the DMA mapping and mark the channel
         * "not-in-use".
         *
         * Note, that after we've copied the ptr to the "last" SG desc. to
         * a local var and assigned NULL to its old place in the DMA channel,
         * another execution of this func won't try to free this channel,
         * so we can drop the spinlock. However, the channel will remain busy
         * and not available for a new acquisition until we call the
         * DmaCtxCompleteTransfer and finally free it.
         */
        pDmaChan = NULL;
        {
            int i;
            for ( i = 0; pDmaChan == NULL && i < DMA_CTX_MAX_PARALLEL_MAPS; i++)
            {
                if ( pDmaCB->Chan[ i].bInUse)
                {
                    PSG_DESC    pCurSGDesc;

                    pCurSGDesc = pDmaCB->Chan[ i].pLastSGDesc;


#ifndef _NO_DST
                    if ( pCurSGDesc != NULL &&
                         !( SgDescGetFlags( pCurSGDesc) & SG_DESC_FLAG_VALID))
#else
                    if ( pCurSGDesc != NULL &&
                         ( !( SgDescGetFlags( pCurSGDesc) & SG_DESC_FLAG_VALID) || XferDir == DST_XFER))
#endif
                    {
                        /*
                         * No VALID flag - the descriptor has been processed
                         */
                        pDmaChan = &pDmaCB->Chan[ i];

                        /*
                         * Some weird stuff here. Hi/fn says in its FreeBSD
                         * driver that when DMA engine walks the SG descriptor
                         * array and encounters a "non-VALID" descriptor, it
                         * can "pre-fetch" it and then continue polling only
                         * its VALID bit. If this "non-VALID" descriptor has
                         * the LAST bit set from previous operations, it
                         * may be "pre-fetched" and the engine will deem this
                         * descriptor "LAST" even if the flag will be cleared
                         * later, at the time when the VALID bit is eventually
                         * being set. It's difficult to believe this, but it is
                         * not difficult to clear the descriptor flags now.
                         * (for reference, see their InterruptPending() func.
                         * in operation.c). However, this, if not corrected,
                         * would not lead to malfunction of this driver, only
                         * to an unnecessary interrupt and a void run of this
                         * routine.
                         */
                        SgDescSetFlags( pCurSGDesc, 0);
                    }
                }
            }
        }

        if( pDmaChan != NULL)
        {
            /*
             * Found DMA channel with finished DMA transfer.
             *
             * This will protect this DMA channel from being touched by
             * another invocation of this proc
             */
            pDmaChan->pLastSGDesc = NULL;

            /*
             * Now its safe to drop the spinlock and free the DMA mapping of
             * the data in the OS
             */
            PCrptSpinLockRelease( &pThis->SpinLock);

            DmaCtxCompleteTransfer( pDmaCB->pDmaCtx,
                                    pDmaChan->pMapRegBase,
                                    pDmaChan->NumOfMapRegs,
                                    pDmaChan->pMdlToMap);

            /*
             * Finally, this DMA channel can be freed for new DMA mappings
             */
            PCrptSpinLockAcquire( &pThis->SpinLock);

            pDmaChan->bInUse = FALSE;
            pDmaCB->nMapsInProgress--;
        }

        PCrptSpinLockRelease( &pThis->SpinLock);
    }
    while( pDmaChan != NULL);
}

#ifdef DBG

void
ProcDumpXferStatus
    (
        IN      PPROCESSOR   pThis,
        IN      PROC_XFER_DIR   XferDir
    )
{
    int i;

    DebugPrint((2, "ProcDumpXferStatus[%d]\n", XferDir));
    ASSERT( pThis->Type == PROCESSOR_SIGN);

    for ( i = 0; i < DMA_CTX_MAX_PARALLEL_MAPS; i++)
    {
        PPROC_DMA_CHANNEL pChan;

        pChan = &pThis->DmaCB[ XferDir].Chan[ i];

        DebugPrint((2, "    chan %d - %s %s\n",
                    i,
                    pChan->bInUse ? "in use " : "free   ",
                    pChan->pLastSGDesc == NULL ? "" :
                    SgDescGetFlags( pChan->pLastSGDesc) & SG_DESC_FLAG_VALID ? "working" : "completed"
                    ));
    }
}

#endif
