/**********************************************************************

    Module: processor.c

        Product: kernel mode driver for "PowerCrypt" encryption
        acceleration board based on Hi/fn 7751 chip
        by Ashley Laurent, Inc.

    ---------------------------------------------------------------

    Copyright:

        Ashley Laurent, Inc., 1999
        All Rights Reserved.

    ---------------------------------------------------------------

    Functional Description:

        Implementation of the PROCESSOR object that performs all the
        work of the 7751 programming.

    ---------------------------------------------------------------

    Environment:

        NT or BSD kernel mode

    ---------------------------------------------------------------

    Author:

        Stan Shkolnyy

    ---------------------------------------------------------------

    Revision history:

        *   15-Feb-99   initial revision

**********************************************************************/

#include "osindependence.h"
#include "debug.h"
#include "processor.h"
#include "procbase.h"
#include "kmi_private.h"    /* for KmInterfaceCompleteCalcRequest() */

#ifdef ALLOC_PRAGMA
#pragma alloc_text(INIT, ProcessorConstructor)
#pragma alloc_text(INIT, ProcessorPlug)
#pragma alloc_text(INIT, QueueEntryBuild)
#endif

/****************************************************************************
****************************************************************************/
void
ProcessorConstructor(
    IN OUT  PPROCESSOR          pThis,
    IN      struct _PROCBASE    *pProcBase,
    IN      PCONFIG             pConfig
    )
{
    int i;

    DebugPrint((2, "ProcessorConstructor\n"));

    pThis->Type = PROCESSOR_SIGN;

    pThis->pConfig = pConfig;
    pThis->pProcBase = pProcBase;
    PCrptSpinLockInit( &pThis->SpinLock);

    /*
     * For the dtor to work correctly.
     */
    pThis->DmaCB[ SRC_XFER].pDmaCtx = NULL;
    pThis->DmaCB[ DST_XFER].pDmaCtx = NULL;
    for ( i = 0; i < DMA_CTX_MAX_PARALLEL_MAPS; i++)
    {
        pThis->DmaCB[ SRC_XFER].Chan[ i].pMdlToMap = NULL;
        pThis->DmaCB[ DST_XFER].Chan[ i].pMdlToMap = NULL;
    }

    /*
     * Unlock procedure related value, changed by Proc7751Unlock()
     */
    pThis->curSignature = 0;

    /*
     * Init.statistics
     */
    pThis->nReqsTotalReceived = 0;
    pThis->nReqsFailed = 0;

    #if DBG
        pThis->pComRing = NULL;
        pThis->pRltRing = NULL;
        pThis->DmaCB[ SRC_XFER].pDataRing = NULL;
        pThis->DmaCB[ DST_XFER].pDataRing = NULL;
	
        for( i = 0; i < PROCESSOR_QUEUE_LENGTH; i++)
	{
	   pThis->Queue[ i].pCalcReq = NULL;
	}
    #endif
}

/****************************************************************************
****************************************************************************/
void
ProcessorDestructor(
    IN PPROCESSOR pThis
    )
{
    int i;

    DebugPrint((2, "ProcessorDestructor\n"));
    ASSERT( pThis->Type == PROCESSOR_SIGN);

    /*
     * Free partial MDLs, if allocated
     */
    if ( pThis->DmaCB[ SRC_XFER].pDmaCtx != NULL)
    {
        if ( DmaCtxIsPartialTransfers( pThis->DmaCB[ SRC_XFER].pDmaCtx))
        {
            int i;

            for ( i = 0; i < DMA_CTX_MAX_PARALLEL_MAPS; i++)
            {
                if ( pThis->DmaCB[ SRC_XFER].Chan[ i].pMdlToMap != NULL)
                {
                    FREE_MDL( pThis->DmaCB[ SRC_XFER].Chan[ i].pMdlToMap);
                }
            }
        }
    }

    if ( pThis->DmaCB[ DST_XFER].pDmaCtx != NULL)
    {
        if ( DmaCtxIsPartialTransfers( pThis->DmaCB[ DST_XFER].pDmaCtx))
        {
            int i;

            for ( i = 0; i < DMA_CTX_MAX_PARALLEL_MAPS; i++)
            {
                if ( pThis->DmaCB[ DST_XFER].Chan[ i].pMdlToMap != NULL)
                {
                    FREE_MDL( pThis->DmaCB[ DST_XFER].Chan[ i].pMdlToMap);
                }
            }
        }
    }
}

/****************************************************************************

Description:

    Reports to the PROCESSOR the virtual addresses of the 7751 mem-mapped regs,
    the virtual and DMA addresses of the scatter/gather control
    area and a ptr to the DISPATCHER. Activates the PROCESSOR and makes it
    completely ready to work.

    No other PROCESSOR initializations are needed - as soon as this procedure
    completes and then the client interface is eventually activated, user
    requests can reach the PROCESSOR.

    This is the first time when PROCESSOR comes in touch with its hardware.

****************************************************************************/

PWRCRYPT_STATUS
ProcessorPlug(
    IN PPROCESSOR   pThis,
    IN PDMA_CTX     pSrcDmaCtx,
    IN PDMA_CTX     pDstDmaCtx,
    IN PDISPATCHER  pDispatcher,
    IN PVOID        pDMARange,
    IN PVOID        pProcRange
    )
{
    PWRCRYPT_STATUS status;

    DebugPrint((2, "ProcessorPlug\n"));
    ASSERT( pThis->Type == PROCESSOR_SIGN);

    /********************************************************************
     * Initialize PROCESSOR's constant members, those which never change
     * during its life.
     * (This includes making necessary memory allocations).
     */

    /*
     * Save the pointers to key objects the caller passes us.
     * They'll be needed throughout the PROCESSOR's life.
     */
    pThis->pDispatcher              = pDispatcher;
    pThis->DmaCB[ SRC_XFER].pDmaCtx = pSrcDmaCtx;
    pThis->DmaCB[ DST_XFER].pDmaCtx = pDstDmaCtx;
    pThis->pProcRegisters           = (PPROC_REGISTERS) pProcRange;
    pThis->pDMARegisters            = (PDMA_REGISTERS) pDMARange;

    /*
     * Put back pointers to this PROCESSOR and identificators of kind of
     * context into each of DMA contexts. They are used in the MapControl
     * routine, since that routine is only given the pointer to the DMA Ctx
     * object
     */
    pSrcDmaCtx->UserUlong1 = (ULONG) pThis;
    pSrcDmaCtx->UserUlong2 = SRC_XFER;

    pDstDmaCtx->UserUlong1 = (ULONG) pThis;
    pDstDmaCtx->UserUlong2 = DST_XFER;

    /*
     * Initialize both DMA Contexts
     */
    status = DmaCtxInit( pSrcDmaCtx,
                         DATA_RING_LEN - 1,         /* mapregs needed */
                         PROCESSOR_MAX_XFER_SIZE);  /* max DMA transfer size */

    if ( !PWRCRYPT_SUCCESS( status))
    {
        DPF("DmaCtxInit for src failed. error 0x%08x\n", status);
        return status;
    }

    status = DmaCtxInit( pDstDmaCtx,
                         DATA_RING_LEN - 1,         /* mapregs needed */
                         PROCESSOR_MAX_XFER_SIZE);  /* max DMA transfer size */

    if ( !PWRCRYPT_SUCCESS( status))
    {
        DPF("DmaCtxInit for dest failed. error 0x%08x\n", status);
        return status;
    }

    /*
     * Ask the DMA Contexts to allocate the memory for 2 SG memory areas -
     * one area in each of them, "source" and "destination".
     * Construct the areas (the area ctor also calls the ctors of SG desc.
     * rings inside of the area).
     */
    {
        ULONG   SgAreaDmaAddr;

        pThis->DmaCB[ SRC_XFER].pSgCtlArea =
            (PSG_CTL_AREA) DmaCtxAllocateDmaMemory(
                                pThis->DmaCB[ SRC_XFER].pDmaCtx,
                                SRC_SG_CTL_AREA_SIZE,
                                &SgAreaDmaAddr);

        if ( pThis->DmaCB[ SRC_XFER].pSgCtlArea == NULL)
        {
	    DPF("Couldn't allocate SgCtlArea for src\n");
            return PWRCRYPT_STATUS_INSUFFICIENT_RESOURCES;
        }

        SgCtlAreaConstructor( pThis->DmaCB[ SRC_XFER].pSgCtlArea,
                              SgAreaDmaAddr,
                              sizeof( BIGGEST_COMMAND)
                              );

        pThis->DmaCB[ DST_XFER].pSgCtlArea =
            (PSG_CTL_AREA) DmaCtxAllocateDmaMemory(
                                pThis->DmaCB[ DST_XFER].pDmaCtx,
                                DST_SG_CTL_AREA_SIZE,
                                &SgAreaDmaAddr);

        if ( pThis->DmaCB[ DST_XFER].pSgCtlArea == NULL)
        {
	    DPF("Couldn't allocate SgCtlArea for dest\n");
	    
            return PWRCRYPT_STATUS_INSUFFICIENT_RESOURCES;
        }

        SgCtlAreaConstructor( pThis->DmaCB[ DST_XFER].pSgCtlArea,
                              SgAreaDmaAddr,
                              sizeof( BIGGEST_RESULT)
                              );
    }

    /*
     * After the construction, the SG control areas know where their SG rings
     * live inside of them and we ask them, since we need direct access to
     * the rings.
     * The PROCESSOR deals with the Command and Result rings, whereas the
     * data rings are logically associated with DMA Control Blocks because
     * the single routine governs both DMA CBs - "src" and "dst".
     */
    pThis->pComRing = &pThis->DmaCB[ SRC_XFER].pSgCtlArea->SmallRing;
    pThis->pRltRing = &pThis->DmaCB[ DST_XFER].pSgCtlArea->SmallRing;

    pThis->DmaCB[ SRC_XFER].pDataRing = &pThis->DmaCB[ SRC_XFER].pSgCtlArea->BigRing;
    pThis->DmaCB[ DST_XFER].pDataRing = &pThis->DmaCB[ DST_XFER].pSgCtlArea->BigRing;

    /*
     * Allocate associated partial MDLs for channels of those DMA contexts
     * that don't have enough map regs. On any allocation failure - return,
     * the dtor will deallocate any MDLs that were allocated.
     *
     * Each map register in the DMA Ctx corresponds to a page of memory,
     * and we use map regs counter as an argument to MDL allocation where
     * the page cnt is needed.
     */
    {
        int         i;
        PPROC_DMA_CTL_BLOCK pDmaCB;

        pDmaCB = &pThis->DmaCB[ SRC_XFER];
        for ( i = 0; i < DMA_CTX_MAX_PARALLEL_MAPS; i++)
        {
            if ( DmaCtxIsPartialTransfers( pDmaCB->pDmaCtx))
            {
                pDmaCB->Chan[ i].pMdlToMap =
                    ALLOCATE_MDL( DmaCtxGetMaxMapRegs( pDmaCB->pDmaCtx) / 2);

                if ( pDmaCB->Chan[ i].pMdlToMap == NULL)
                {
                    DPF("Can't allocate partial MDL for src\n");
                    return PWRCRYPT_STATUS_INSUFFICIENT_RESOURCES;
                }
            }
        }

        pDmaCB = &pThis->DmaCB[ DST_XFER];
        for ( i = 0; i < DMA_CTX_MAX_PARALLEL_MAPS; i++)
        {
            if ( DmaCtxIsPartialTransfers( pDmaCB->pDmaCtx))
            {
                pDmaCB->Chan[ i].pMdlToMap =
                    ALLOCATE_MDL( DmaCtxGetMaxMapRegs( pDmaCB->pDmaCtx) / 2);

                if ( pDmaCB->Chan[ i].pMdlToMap == NULL)
                {
                    DPF ("Can't allocate partial MDL for dest\n");
                    return PWRCRYPT_STATUS_INSUFFICIENT_RESOURCES;
                }
            }
        }
    }

    /*
     * Build each entry in the PROCESSOR's request queue.
     * Queue entry build includes setting ptrs to S/G area data areas
     */
    {
        int i;
        for ( i = 0; i < PROCESSOR_QUEUE_LENGTH; i++)
        {
            QueueEntryBuild( pThis, i);
        }
    }

    /********************************************************************
     * Clear and reset PROCESSOR's queue and other mechanics.
     */
    ProcReset( pThis);

    /********************************************************************
     * Initialize the PROCESSOR's hardware - reset the 7751, perform some
     * rudimentary diagnostics, unlock the strong encryption, load valid
     * values into the control regs.
     */
    DmaPuShutdown( pThis);
    if ( !DmaPuTest( pThis))
    {
        DPF1("ProcessorPlug - diags failed\n");
        return PWRCRYPT_STATUS_DEVICE_NOT_CONNECTED;
    }

    if ( !ProcUnlock( pThis))
    {
        DebugPrint((2, "ProcessorPlug - encryption unlock failed\n"));
        #ifdef DBG
            PuDumpRegs( pThis);
        #endif
        return PWRCRYPT_STATUS_NOT_IMPLEMENTED;
    }

    DmaPuLaunch( pThis);

    return PWRCRYPT_STATUS_SUCCESS;
}

/********************************************************************
 * ProcReset()
 * Initializes PROCESSOR's variable members which change in the course
 * of its work.
 * (No memory allocations, only vars are reset to initial values).
 */
void
ProcReset(
    IN PPROCESSOR   pThis
    )
{
    DebugPrint((2, "ProcReset\n"));
    ASSERT( pThis->Type == PROCESSOR_SIGN);

    #if DBG
        ASSERT( pThis->pComRing != NULL);
        ASSERT( pThis->pRltRing != NULL);
        ASSERT( pThis->DmaCB[ SRC_XFER].pDataRing != NULL);
        ASSERT( pThis->DmaCB[ DST_XFER].pDataRing != NULL);
    #endif

    /*
     * Queue control:
     * Set newest/oldest entry indexes in such a way that adding a first
     * request, preceded by moving the "iNewest", will make
     * iNewest == iOldest.
     */
    pThis->iOldestQueueEntry = 0;
    pThis->iNewestQueueEntry = PROCESSOR_QUEUE_LENGTH - 1;
    pThis->BusyEntryCnt = 0;

    /*
     * Queue entries:
     */
    {
        int i;
        for ( i = 0; i < PROCESSOR_QUEUE_LENGTH; i++)
        {
            PPROC_QUEUE_ENTRY   pQEntry;

            pQEntry = &pThis->Queue[ i];

            /*
             * Sentinel value preventing DMA mapping of a request in the queue entry
             * when the entry is not in use or the request has no data to transfer.
             * Analyzed and used in ProcTryToStartTransfer.
             */
            pQEntry->XferProgress[ SRC_XFER].DataRemainCnt = 0;
            pQEntry->XferProgress[ DST_XFER].DataRemainCnt = 0;

            /*
             * It was never started before
             */
/*            pQEntry->bStarted = FALSE; */

            #ifdef DBG
                PCrptFillMemory( pQEntry->pCommand, sizeof(BIGGEST_COMMAND), 0x33);
                PCrptFillMemory( pQEntry->pResult, sizeof(BIGGEST_RESULT), 0x77);
            #endif
        }
    }

    /*
     * DMA control blocks and DMA channels inside of the blocks.
     */
    pThis->DmaCB[ SRC_XFER].bDmaCtxBusy = FALSE;
    pThis->DmaCB[ SRC_XFER].nMapsInProgress = 0;
    {
        int i;

        for ( i = 0; i < DMA_CTX_MAX_PARALLEL_MAPS; i++)
        {
            pThis->DmaCB[ SRC_XFER].Chan[ i].bInUse = FALSE;
            pThis->DmaCB[ SRC_XFER].Chan[ i].pLastSGDesc = NULL;
        }
    }

    pThis->DmaCB[ DST_XFER].bDmaCtxBusy = FALSE;
    pThis->DmaCB[ DST_XFER].nMapsInProgress = 0;
    {
        int i;

        for ( i = 0; i < DMA_CTX_MAX_PARALLEL_MAPS; i++)
        {
            pThis->DmaCB[ DST_XFER].Chan[ i].bInUse = FALSE;
            pThis->DmaCB[ DST_XFER].Chan[ i].pLastSGDesc = NULL;
        }
    }

    /*
     * Clear interrupt registry - no ints has occurred yet.
     */
    IntsStateReset( pThis);
}

/****************************************************************************

Description:

****************************************************************************/
void
ProcessorUnplug(
    IN PPROCESSOR   pThis
    )
{
    DebugPrint((2, "ProcessorUnplug\n"));
    ASSERT( pThis->Type == PROCESSOR_SIGN);

    DmaPuShutdown( pThis);

    SgCtlAreaDestructor( (PSG_CTL_AREA) DmaCtxGetSgCtlAreaPtr( pThis->DmaCB[ SRC_XFER].pDmaCtx));
    SgCtlAreaDestructor( (PSG_CTL_AREA) DmaCtxGetSgCtlAreaPtr( pThis->DmaCB[ DST_XFER].pDmaCtx));
}


/****************************************************************************
QueueEntryBuild()

sets queue entry's ptrs to
 - the Command and the Command SG descriptor located in the "source" SG area,
 - the Result and the Result SG descriptor located in the "destination" SG area

This "linkage" has no other meaning except having the pointers at hand,
in the queue entry, since they will be used often.

****************************************************************************/
void
QueueEntryBuild
    (
        IN OUT  PPROCESSOR  pThis,
        IN      int         EntryIndex
    )
{
    PPROC_QUEUE_ENTRY   pQEntry;
    PSG_CTL_AREA        pSrcSgCtlArea, pDstSgCtlArea;

    DebugPrint((2, "QueueEntryBuild[%d]\n", EntryIndex));
    ASSERT( pThis->Type == PROCESSOR_SIGN);
    ASSERT( EntryIndex < PROCESSOR_QUEUE_LENGTH);

    pQEntry       = &pThis->Queue[ EntryIndex];
    pSrcSgCtlArea = (PSG_CTL_AREA) DmaCtxGetSgCtlAreaPtr( pThis->DmaCB[ SRC_XFER].pDmaCtx);
    pDstSgCtlArea = (PSG_CTL_AREA) DmaCtxGetSgCtlAreaPtr( pThis->DmaCB[ DST_XFER].pDmaCtx);

    /*
     * "Small ring data" in SrcSgCtlArea are BIGGEST_COMMANDs and
     * "Small ring data" in DstSgCtlArea are BIGGEST_RESULTs.
     * Pick one of them by queue entry index.
     */
    ASSERT( SgCtlAreaGetSmallRingDataSize( pSrcSgCtlArea) ==
            PROCESSOR_QUEUE_LENGTH * sizeof( BIGGEST_COMMAND));
    ASSERT( SgCtlAreaGetSmallRingDataSize( pDstSgCtlArea) ==
            PROCESSOR_QUEUE_LENGTH * sizeof( BIGGEST_RESULT));

    pQEntry->pCommand = (PUCHAR)
                        &( (BIGGEST_COMMAND*) SgCtlAreaGetSmallRingDataPtr( pSrcSgCtlArea)
                         )[ EntryIndex];
    pQEntry->pCommandDesc =
        &( SgCtlAreaGetFirstSmallRingDescPtr( pSrcSgCtlArea))[ EntryIndex];

    pQEntry->pResult =  (PUCHAR)
                        &( (BIGGEST_RESULT*) SgCtlAreaGetSmallRingDataPtr( pDstSgCtlArea)
                         )[ EntryIndex];
    pQEntry->pResultDesc  =
        &( SgCtlAreaGetFirstSmallRingDescPtr( pDstSgCtlArea))[ EntryIndex];

    DebugPrint((2, "    sizeof(sg_ctl_area)=0x%04x\n", sizeof( SG_CTL_AREA)));

    DebugPrint((2, "    src sg area=0x%08x\n", pSrcSgCtlArea));
    DebugPrint((2, "    sizeof(biggest_comm)=0x%04x\n", sizeof( BIGGEST_COMMAND)));
    DebugPrint((2, "        command=0x%08x\n", pQEntry->pCommand));
    DebugPrint((2, "        command desc=0x%08x\n", pQEntry->pCommandDesc));
    DebugPrint((2, "        data jump desc=0x%08x\n", RingGetJumpDescPtr( &pSrcSgCtlArea->BigRing)));

    DebugPrint((2, "    dst sg area=0x%08x\n", pDstSgCtlArea));
    DebugPrint((2, "    sizeof(biggest_rlt) =0x%04x\n", sizeof( BIGGEST_RESULT)));
    DebugPrint((2, "        result=0x%08x\n", pQEntry->pResult));
    DebugPrint((2, "        result desc=0x%08x\n", pQEntry->pResultDesc));
    DebugPrint((2, "        data jump desc=0x%08x\n", RingGetJumpDescPtr( &pDstSgCtlArea->BigRing)));
}

/********************************************************************
 * ProcUnlock()
 * Tries to unlocks strong encryption.
 * Returns TRUE on successful switching to the sec.level specified
 * in the config or if that level is already set. Remember - once it is
 * set, it can't be returned to a lower level without the 7751 power cycle.
 */
BOOLEAN
ProcUnlock(
    IN PPROCESSOR   pThis
    )
{
    UNLOCK_KEY  UnlockKey;          /* the key for strong encryption unlocking */
    SecurLevel  NewLevel;           /* the target level of 7751 unlock */
    SecurLevel  ResultingLevel;     /* the result of 7751 unlock */

    DebugPrint((2, "ProcUnlock\n"));
    ASSERT( pThis->Type == PROCESSOR_SIGN);

    ConfigGetUnlockKey( pThis->pConfig, &UnlockKey);
    NewLevel = ConfigGetUnlockSecurityLevel( pThis->pConfig);

    PuUnlock( pThis, NewLevel, &UnlockKey);
    if ( PuGetSecurityLevel( pThis, &ResultingLevel))
    {
        if ( ResultingLevel == NewLevel)
        {
            return TRUE;
        }
    }
    return FALSE;
}
