/**********************************************************************

    Module: processor.h

        Product: kernel mode driver for "PowerCrypt" encryption
        acceleration board based on Hi/fn 7751 chip
        by Ashley Laurent, Inc.

    ---------------------------------------------------------------

    Copyright:

        Ashley Laurent, Inc., 1999
        All Rights Reserved.

    ---------------------------------------------------------------

    Functional Description:

        Declaration of the PROCESSOR object that performs all the
        work of the 7751 programming.

    ---------------------------------------------------------------

    Environment:

        NT or BSD kernel mode

    ---------------------------------------------------------------

    Author:

        Stan Shkolnyy

    ---------------------------------------------------------------

    Revision history:

        *   15-Feb-99   initial revision

**********************************************************************/

#ifndef _PROCESSOR_H__
#define _PROCESSOR_H__

#include "xfer.h"
#include "proc_hw_dma.h"            /* for DMA_REGISTERS */
#include "proc_hw_proc.h"           /* for PROC_REGISTERS */
#include "sg_ctlarea.h"             /* for SG_CTL_AREA */
#include "dma_ctx.h"                /* for DMA_CTX */
#include "dispatcher.h"             /* for PDISPATCHER */
#include "pwrcryptioctl.h"          /* for IOCTL_PWRCRYPT_CALCULATE_INPUT_OUTPUT */
#include "km_interface.h"           /* for PWRCRYPT_CALCULATE_KM_WORKAREA */
#include "driverdef.h"              /* for UNLOCK_KEY */
#include "config.h"                 /* for PCONFIG */

/****************************************************************************
 */
typedef struct _PROCESSOR_INTS_STATE
{
    /*
     * "software" ints "generated" when a DMA mapping completes
     */
    LONG        MapIntCnt[2];

    /*
     * The bits in this vars exactly correspond to bits in DMA and
     * proc. units that caused interrupts
     */
    ULONG       DUInts;
    ULONG       PUInts;

} PROCESSOR_INTS_STATE, *PPROCESSOR_INTS_STATE;

/****************************************************************************
 */
typedef enum
{
    SRC_XFER = 0,
    DST_XFER = 1

} PROC_XFER_DIR;

/****************************************************************************
 */
typedef struct _PROC_QENTRY_XFER_PROGRESS
{
    /*
     * Full MDLs (taken directly from CalcReq) describing source and
     * destination data buffers.
     * (We have no VA to access these data and we don't need to. Only the
     * hardware is interested in these data).
     */

    PMEM_DESC_LIST      pFullMdl;

    /*
     * This many bytes remain to be transferred for each data stream
     */
    ULONG               DataRemainCnt;

    /*
     * The current "VA for MDL" where the next DMA mapping should occur.
     * Actually, used only when the corresponding DMA context doesn't have
     * enough map regs and partial MDLs are used.
     */
    PCHAR               DataCurrVa;

} PROC_QENTRY_XFER_PROGRESS, *PPROC_QENTRY_XFER_PROGRESS;

/****************************************************************************
 */
typedef struct _PROC_QUEUE_ENTRY
{
    /*-----------------------------------------------------------------------
     * The following members are initialized only once and never change
     */

    /*
     * Two ptrs to the DMA-accessible memory areas where the Command
     * part of the request is to be copied to and the Result part (when the
     * 7751 finishes) is to be taken from.
     */
    PUCHAR              pCommand;
    PUCHAR              pResult;

    /*
     * Two ptrs to the DMA SG descriptors corresponding to the Command
     * and Result areas above. These ptrs are needed, because the data sizes
     * and "valid" bits need to be set in those descriptors for each request.
     * Note, that data ptrs in the descriptors always point to the same
     * Command and Result areas, but data sizes change from request to request.
     */
    PSG_DESC            pCommandDesc;
    PSG_DESC            pResultDesc;

    /*-----------------------------------------------------------------------
     * The following members change as the entry is used for a new request
     */

    /*
     * Indicates whether the entry was started (by ProcessorStartQueuedRequest()).
     */
    BOOLEAN             bStarted;

    /*
     * Points to the calculating request being processed
     */
    volatile PIOCTL_PWRCRYPT_CALCULATE_INPUT_OUTPUT  pCalcReq;

    /*
     * Two transfer control data - for SrcData and DstData. Each one contains
     * the _full_ MDL for the corresponding buffer and the current position
     * in the buffer where the next DMA mapping will occur.
     *
     * The full MDL is assigned once at request start. The VA and length
     * are advanced when a chunk of the buffer is mapped for DMA.
     *
     * When there is no shortage of DMA map regs, the transfer always occurs
     * in single DMA mapping and the XferProgress[x].DataRemainCnt goes to 0
     * at the very first call to ProcTryStartDMATransfer(). If not, the
     * ProcTryStartDMATransfer() will be invoked several times, each time
     * creating a partial MDL for the part of the buffer and will
     * decrement the DataRemainCnt each time, until it gets to 0.
     */
    PROC_QENTRY_XFER_PROGRESS XferProgress[ 2];

} PROC_QUEUE_ENTRY, *PPROC_QUEUE_ENTRY;

/****************************************************************************
 * DMA Channel
 *
 * pMdlToMap may either point constantly to the same partial MDL or
 * alternatively to different "full" MDLs if the computer is capable
 * to always map the whole user buffer for the DMA operation.
 *
 * Partial MDLs describing chunks of source and destination data buffers
 * are needed if the whole data can't be transferred in a single
 * DMA transfer. These MDLs describe the chunks which are currently being
 * transfered. As the transfer completes, the partial MDL is re-used for
 * the next chunk of the data. Each partial MDL is capable to describe as
 * many RAM pages as the number of map registers available in corresponding
 * (Src or Dst) DMA context (DMA_CTX members of the PROCESSOR). In fact,
 * the number of map regs in the corresponding DMA_CTX is directly used to
 * create each of these partial MDLs.
 *
 * Indication of which of two MDL schemes is used is returned by
 * DmaCtxIsPartialTransfers()
 */
typedef struct _PROC_DMA_CHANNEL
{
    BOOLEAN             bInUse;
    PMEM_DESC_LIST      pMdlToMap;
    BOOLEAN             bLastMap;
    PSG_DESC            pLastSGDesc;
    PVOID               pMapRegBase;
    ULONG               NumOfMapRegs;

} PROC_DMA_CHANNEL, *PPROC_DMA_CHANNEL;

/****************************************************************************
 * DMA Control Block wraps the DMA Context, the abstraction of the O/S-specific
 * DMA architecture designed for this drivers needs. The PROCESSOR has two of
 * this Control Blocks - for transferring the data to or from the 7751.
 * "Source" one is used for Commands and Src Data, "destination" -
 * for Dst Data and Results.
 *
 * Each Control Block has two "DMA channels", representing the fact that
 * two MDLs can be simultaneously mapped for DMA transfer.
 */

/*
 * Scatter/gather control areas contain descriptor rings and some other data
 *
 * SG_CTL_AREA structure includes everything except SmallRingData
 * array which is defined as 1-byte array because its size is different
 * for src and dst DMA contexts. These macros define the actual sizes of
 * S/G areas for these two cases.
 */
#define SRC_SG_CTL_AREA_SIZE \
            ( sizeof( SG_CTL_AREA) - 1 + \
              PROCESSOR_QUEUE_LENGTH * sizeof( BIGGEST_COMMAND) \
            )

#define DST_SG_CTL_AREA_SIZE \
            ( sizeof( SG_CTL_AREA) - 1 + \
              PROCESSOR_QUEUE_LENGTH * sizeof( BIGGEST_RESULT) \
            )


typedef struct _PROC_DMA_CTL_BLOCK
{
    PDMA_CTX            pDmaCtx;
    PSG_CTL_AREA        pSgCtlArea;
    PRING               pDataRing;

    /*-----------------------------------------------------------------------
     * Flag that indicates that the DMA context is waiting for the
     * AdapterControl to be called. The flag protects form an attempt to
     * start another mapping in this DMA context.
     */
    BOOLEAN             bDmaCtxBusy;

    /*-----------------------------------------------------------------------
     * How many transfers are in progress, limited to DMA_CTX_MAX_PARALLEL_MAPS
     */
    ULONG               nMapsInProgress;

    PROC_DMA_CHANNEL    Chan[ DMA_CTX_MAX_PARALLEL_MAPS];

} PROC_DMA_CTL_BLOCK, *PPROC_DMA_CTL_BLOCK;

/****************************************************************************
 * Processor class
 *
 *  This O/S-independent class defines the data processing functionality.
 *  An O/S-dependent code creates the PROCESSOR and provides it with
 *  memory addresses where the 7751 registers can be accessed, and calls
 *  PROCESSOR methods as appropriate, to deliver O/S events, like ints, to it.
 *
 *  Remark:
 *  PROCESSOR _does_ contain O/S dependent code. This is *_MEM_REGISTER
 *  macros defined at the bottom of this file. The macros access 7751
 *  registers through O/S-dependent functions. The macros resolve to
 *  different functions during compilation for different OSes as defined
 *  by osindependence.h.
 */

typedef struct _PROCESSOR
{
    ULONG           Type;
    struct _PROCBASE    *pProcBase;

    /*
     * How many requests are currently in progress (0..PROCESSOR_QUEUE_LENGTH).
     *
     * This member is for the DISPATCHER's exclusive using. The DISPATCHER
     * increments the counter and then calls ProcessorQueueRequest() and
     * ProcessorStartQueuedRequest().
     *
     * After the request is completed, the processor dequeues it and calls
     * DISPATCHER.StartNewRequest() which decrements the counter and possibly
     * starts a new request by repeating increment/queue/start sequence.
     */
    ULONG               nReqsInQueue;

    /*
     * This member is for the DISPATCHER's exclusive using.
     * Indicates whether the PROCESSOR shouldn't receive requests because
     * it is restarting after a fatal error.
     */
    BOOLEAN             bInactive;    

    /*
     * Statistics
     */
    ULONG               nReqsTotalReceived;
    ULONG               nReqsFailed;

    /*
     * The queue (one entry per calc.request) and the spinlock that protects
     * it during modifications (adding or removing requests)
     */
    PROC_QUEUE_ENTRY    Queue[ PROCESSOR_QUEUE_LENGTH];
    PCRPT_DECLARE_SPINLOCK( SpinLock);

    /*
     * Queue indexes that point to the oldest and newest entries.
     * Moved by ProcessorQueueRequest and ProcessorServiceRegisteredInterrupts
     */
    UCHAR               iOldestQueueEntry;
    UCHAR               iNewestQueueEntry;

    /*
     * Processor's own queued requests counter
     */
    ULONG               BusyEntryCnt;

    /*
     * These are needed to call "validate jump" method of these rings
     */
    PRING               pComRing;
    PRING               pRltRing;

    /*
     * Interrupt counters incremented by ProcessorConfirmAndRegisterInterrupt()
     * and used and zeroed out by ProcessorServiceRegisteredInterrupts()
     */
    PROCESSOR_INTS_STATE    IntsState;

    /* Ptr to the DISPATCHER that gives this processor work and waits
     * for completion notifications.
     */
    PDISPATCHER         pDispatcher;

    /* Ptr to the configuration storage that holds some initialization
     * parameters for the PROCESSOR
     */
    PCONFIG             pConfig;

    /* Ptrs to two memory ranges mapped to 7751 I/O ranges - supplied by
     * O/S dependent code. They are used by READ_* and WRITE_* macros to
     * read/write 7751 registers -
     */
    PPROC_REGISTERS     pProcRegisters;
    PDMA_REGISTERS      pDMARegisters;

    /* Indicates which ints are enabled, so that processor's ISR
     * (ProcessorConfirmAndRegisterInterrupt) would not waste time
     * on reading int.enabling regs at interrupt time
     */
    ULONG               DUEnabledInts;
    ULONG               PUEnabledInts;

    /* Two, source and destination, transfer control blocks used for
     * transferring data to the 7751. Source one is used for Commands
     * and Src Data, destination - for Dst Data and Results.
     */
    PROC_DMA_CTL_BLOCK DmaCB[ 2];

    /*
     * Unlock procedure related value
     */
    ULONG curSignature;

} PROCESSOR, *PPROCESSOR;

/**********************************************************************
 * PUBLIC methods
 */
void
ProcessorConstructor(
    IN OUT  PPROCESSOR       pThis,
    IN      struct _PROCBASE *pProcBase,
    IN      PCONFIG          pConfig
    );

void
ProcessorDestructor(
    IN PPROCESSOR pThis
    );

PWRCRYPT_STATUS
ProcessorPlug(
    IN PPROCESSOR   pThis,
    IN PDMA_CTX     pSrcDmaCtx,
    IN PDMA_CTX     pDstDmaCtx,
    IN PDISPATCHER  pDispatcher,
    IN PVOID        pDMARange,
    IN PVOID        pProcRange
    );

void
ProcessorUnplug(
    IN PPROCESSOR   pThis
    );

void
ProcessorEnqueueRequest(
    IN PPROCESSOR   pThis,
    IN PIOCTL_PWRCRYPT_CALCULATE_INPUT_OUTPUT   pCalcReq
    );

void
ProcessorStartQueuedRequest(
    IN PPROCESSOR   pThis
    );

BOOLEAN
ProcessorConfirmAndRegisterInterrupt(
    IN PPROCESSOR   pThis
    );

void
ProcessorServiceRegisteredInterrupts(
    IN PPROCESSOR   pThis
    );

/****************************************************************************
 * PRIVATE methods
 */
void
ProcReset(
    IN PPROCESSOR   pThis
    );

BOOLEAN
ProcUnlock(
    IN PPROCESSOR   pThis
    );

void
ProcTryToCompleteRequests(
    IN PPROCESSOR   pThis
    );

/*
 * Data DMA transfer stuff
 */
void
ProcTryToStartDMATransfer(
    IN      PPROCESSOR      pThis,
    IN      PROC_XFER_DIR   XferDir
    );

void
ProcTryToCompleteDMATransfers(
    IN      PPROCESSOR      pThis,
    IN      PROC_XFER_DIR   XferDir
    );

void
ProcMapControl
    (
        IN PVOID        Context,
        IN PVOID        MapRegBase
    );

/*
 * request queue stuff
 */
void
QueueEntryBuild
    (
        IN OUT  PPROCESSOR  pThis,
        IN      int         EntryIndex
    );

#define IncrementQueueIndex( i)             \
{                                           \
    ASSERT( PROCESSOR_QUEUE_LENGTH == 2);   \
    i ^= 1;                                 \
}

#define CheckQueueIndexes( pThis)           \
{                                           \
    ASSERT( ((pThis)->BusyEntryCnt == 1 && (pThis)->iOldestQueueEntry == (pThis)->iNewestQueueEntry) ||   \
            ((pThis)->BusyEntryCnt != 1 && (pThis)->iOldestQueueEntry != (pThis)->iNewestQueueEntry));  \
}

/*
 * Interrupt stuff
 */
void
IntsStateReset(
    IN OUT PPROCESSOR   pThis
    );

/*
 * Hardware stuff
 */
void
DmaPuShutdown(
    IN PPROCESSOR   pThis
    );

BOOLEAN
DmaPuTest(
    IN PPROCESSOR   pThis
    );

void
DmaPuLaunch(
    IN PPROCESSOR   pThis
    );

void
DmaReset(
    IN PPROCESSOR   pThis
    );

BOOLEAN
DmaDiagnose(
    IN PPROCESSOR   pThis
    );

ULONG
DmaGetPollingFreq(
    IN PPROCESSOR   pThis
    );

void
DmaSetPollingFreq(
    IN PPROCESSOR   pThis,
    ULONG freq
    );

BOOLEAN
PuReset(
    IN PPROCESSOR   pThis
    );

BOOLEAN
PuDiagnose(
    IN PPROCESSOR   pThis
    );

BOOLEAN
PuWaitForResetCompletion(
    IN PPROCESSOR   pThis
    );

BOOLEAN
PuGetSecurityLevel
    (
        IN  PPROCESSOR  pThis,
        OUT SecurLevel  *pSecLevel
    );

BOOLEAN
PuUnlock
    (
        IN  PPROCESSOR  pThis,
        IN  SecurLevel  NewLevel,
        IN  PUNLOCK_KEY UnlockKey
    );

#ifdef DBG

void
PuDumpRegs
    (
        IN PPROCESSOR   pThis
    );

void
DmaDumpRegs
    (
        IN PPROCESSOR   pThis
    );

void
DmaDumpStatReg
    (
        IN PPROCESSOR   pThis,
        ULONG StatRegVal
    );

void
ProcDumpXferStatus
    (
        IN PPROCESSOR   pThis,
        IN      PROC_XFER_DIR   XferDir
    );

#endif

/*---------------------------------------------------------------------------
 * Macros used by PROCESSOR to access its memory-mapped registers.
 *
 * May be used only inside of PROCESSOR methods since they contain
 * pThis->pXxxRegisters
 */

#define READ_DMA32( RegXX)               \
                (READ_MEM_REG32(&( pThis->pDMARegisters)->RegXX))

#define WRITE_DMA32( RegXX, ValueXX)     \
                (WRITE_MEM_REG32(&( pThis->pDMARegisters)->RegXX, ValueXX))

#define READ_PROC32( RegXX)               \
                (READ_MEM_REG32(&( pThis->pProcRegisters)->RegXX))

#define WRITE_PROC32( RegXX, ValueXX)     \
                (WRITE_MEM_REG32(&( pThis->pProcRegisters)->RegXX, ValueXX))

#endif
