/**********************************************************************

    Module: sg_ctlarea.h

        Product: kernel mode driver for "PowerCrypt" encryption
        acceleration board based on Hi/fn 7751 chip
        by Ashley Laurent, Inc.

    ---------------------------------------------------------------

    Copyright:

        Ashley Laurent, Inc., 1999
        All Rights Reserved.

    ---------------------------------------------------------------

    Functional Description:

        Layout and accessors of a half-duplex scatter/gather memory area
        used by PROCESSOR. PROCESSOR has two such areas - for incoming data
        and for outcoming data.

        This area is special in that it is located in the non-cached memory
        and its "logical" DMA address is known in advance and ready to be used
        by a DMA device.

    ---------------------------------------------------------------

    Environment:

        NT or BSD kernel mode

    ---------------------------------------------------------------

    Author:

        Stan Shkolnyy

    ---------------------------------------------------------------

    Revision history:

        *   15-Feb-99   initial revision

**********************************************************************/

#ifndef __SG_CTLAREA_H__
#define __SG_CTLAREA_H__

#include "xfer.h"
#include "sg_ring.h"

#ifdef _WIN32
    #pragma pack(push, _SGCTLAREA_STRUCT_DEF_, 1)
#endif

/****************************************************************************
 * Half-duplex DMA control area
 *
 *  This area, usually less than a page in size, contains 2 DMA descriptor
 *  rings - a small one for command or result and a big one for source data or
 *  destination data.
 *
 *  This area should be made non-cached, since it is accessed often
 *  by both the 7751 and CPU(s).
 *
 *  The instance of this object is created by a DMA_CTX constructor, which
 *  also defines its SmallRingData array size.
 */

typedef struct _SG_CTL_AREA
{
    ULONG           Type;

    ULONG           PhysThis;
    ULONG           SmallRingDataSize;

    /*
     * -1 is here because RING struct definition includes 1 descriptor.
     */
    RING            SmallRing;
    SG_DESC         SR[ COM_AND_RES_RING_LEN - 1];

    RING            BigRing;
    SG_DESC         BR[ DATA_RING_LEN - 1];

    /* This array actually contains SmallRingDataSize bytes.
     * It is declared this way because it may be of two different sizes -
     * either contain two BIGGEST_COMMANDs or two BIGGEST_RESULTs, depending on
     * which data transmission direction is this S/G Control Area for.
     */
    UCHAR           SmallRingData[ 1];

} SG_CTL_AREA, *PSG_CTL_AREA;

/**********************************************************************
 * PUBLIC methods
 */
void
SgCtlAreaConstructor(
    PSG_CTL_AREA    pThis,
    ULONG           PhysThis,
    ULONG           SmallRingDataFragmentSize
    );

#define SgCtlAreaDestructor( pThis)

#define SgCtlAreaGetFirstSmallRingDescPtr( pThis) \
            &((pThis)->SmallRing.Descs[0])

#define SgCtlAreaGetFirstBigRingDescPtr( pThis) \
            &((pThis)->BigRing.Descs[0])

#define SgCtlAreaGetPhysAddrOfFirstSmallRingDesc( pThis) \
            (pThis)->PhysThis + FIELD_OFFSET( SG_CTL_AREA, SmallRing.Descs[0])

#define SgCtlAreaGetPhysAddrOfFirstBigRingDesc( pThis) \
            (pThis)->PhysThis + FIELD_OFFSET( SG_CTL_AREA, BigRing.Descs[0])

#define SgCtlAreaGetSmallRingDataPtr( pThis)    &(pThis)->SmallRingData[0]

#define SgCtlAreaGetSmallRingDataSize( pThis)   (pThis)->SmallRingDataSize

/**********************************************************************
 * PRIVATE methods
 */

#ifdef _WIN32
    #pragma pack(pop, _SGCTLAREA_STRUCT_DEF_)
#endif

#endif
