/**********************************************************************

    Module: sg_ring.h

        Product: kernel mode driver for "PowerCrypt" encryption
        acceleration board based on Hi/fn 7751 chip
        by Ashley Laurent, Inc.

    ---------------------------------------------------------------

    Copyright:

        Ashley Laurent, Inc., 1999
        All Rights Reserved.

    ---------------------------------------------------------------

    Functional Description:

        Layout and accessors of scatter/gather descriptor ring used by
        PROCESSOR and the 7751 DMA unit. Encapsulates the actual descriptor
        ring and some control data helping to track the state of it.

    ---------------------------------------------------------------

    Environment:

        NT or BSD kernel mode

    ---------------------------------------------------------------

    Author:

        Stan Shkolnyy

    ---------------------------------------------------------------

    Revision history:

        *   15-Feb-99   initial revision

**********************************************************************/

#ifndef __SG_RING_H__
#define __SG_RING_H__

#include "sg_desc.h"

#ifdef _WIN32
    #pragma pack(push, _SGRING_STRUCT_DEF_, 1)
#endif

/****************************************************************************
 * RING - descriptor "ring" descriptor (used in DMA_CONTROL_AREA)
 *
 * FirstFreeDescIndex -
 *  when a 7751 walks a ring, it has some descriptors ahead
 *  of it to process. These descriptors point to data and have "valid" bit set.
 *  This index points to the first descriptor that must be used for further data
 *  sibmission - that one that will be polled after the last "valid" is processed.
 *  As the mentioned descriptor is "made valid", the index is advanced. If it is
 *  found pointing to the last (jump) descriptor, that "jump" is marked valid
 *  and the index wraps to 0.
 *
 * DescCnt -
 *  indicates the number of descriptors in the ring, including the last (jump)
 *  descriptor, which always point to the first and never used to point to data.
 *  Used for "end-of-ring" detection when the FirstFreeDescIndex is to be
 *  advanced.
 *
 * Descs -
 *  array of the actual descriptors, used for data submission to the 7751.
 *  The number of elements = DescriptorCnt.
 *  The last descriptor in the array is always "jump" to the first one.
 *  Only one array element is defined here, by other ones follow it.
 */

typedef struct _RING
{
    ULONG           Type;

    ULONG   FirstFreeDescIndex;
    ULONG   DescCnt;
    SG_DESC Descs[ 1];          /* array of DescCnt elements */
} RING, *PRING;

/**********************************************************************
 * PUBLIC methods
 */
void
RingConstructor(
    PRING   pThis,
    ULONG   DescCnt,
    ULONG   ThisPhysAddr
    );

ULONG
RingResetAndGetPhysAddr(
    PRING   pThis
    );

/*
 * This void destructor would be called by SgCtlAreaDestructor would that one
 * be itself not void.
 */
#define RingDestructor( pThis)

/*
 * Returns the current value of FirstFreeDescIndex
 */
#define RingGetFirstFreeDescIndex( pThis)    \
            ( (pThis)->FirstFreeDescIndex)

/*
 * Returns the ptr to the descriptor pointed to by FirstFreeDescIndex
 */
#define RingGetFirstFreeDescPtr( pThis)    \
            ( &(pThis)->Descs[ pThis->FirstFreeDescIndex])

/*
 * Advances FirstFreeDescIndex to the next descriptor with "end-of-array"
 * checking and proper index "wrapping" to the first descriptor in the array
 */
void
RingAdvanceFirstFreeDescIndex(
    PRING   pThis
    );

/*
 * Sets the VALID bit in the jump descriptor at the end of the ring
 */
void
RingValidateJump(
    PRING   pThis
    );

/*
 * Returns the ptr to the descriptor pointed to by FirstFreeDescIndex
 */
#define RingGetJumpDescPtr( pThis)    \
            ( &(pThis)->Descs[  (pThis)->DescCnt - 1])

/* 
 *  RingXxx() funcs example usage:
 * 
 *  Get ptrs to the first free desc. and its command
 * 
 *  pDesc = RingGetFirstFreeDescPtr( &ComRing);
 *  pCommand = &Commands[ RingGetFirstFreeDescIndex( &ComRing);
 *
 *  Assemble( pCommand, <assembly data>);       fill in the command structure
 *  SgDescSetDataSize( pDesc, <new com.size>);  put its size into the desc.
 *  RingAdvanceFirstFreeDescIndex( &ComRing);   advance ring ptr
 *  SgDescAddFlags( pDesc, SG_DESC_FLAG_VALID); mark the desc.as valid - DMA starts
 */

/**********************************************************************
 * PRIVATE methods
 */

#ifdef _WIN32
    #pragma pack(pop, _SGRING_STRUCT_DEF_)
#endif

#endif
