#include "osindependence.h"
#include "debug.h"

#include "slotman.h"

#ifdef ALLOC_PRAGMA
#pragma alloc_text(INIT, SlotmanConstructor)
#pragma alloc_text(INIT, SlotmanInit)
#endif

/****************************************************************************
****************************************************************************/
void
SlotmanConstructor(
    IN OUT  PSLOTMAN pThis
    )
{
    DebugPrint((2, "SlotmanConstructor\n"));

    pThis->Type = SLOTMAN_SIGN;

    pThis->pSlots = NULL;
    PCrptSpinLockInit( &pThis->SpinLock);
}

/****************************************************************************
****************************************************************************/
void
SlotmanInit(
    IN OUT  PSLOTMAN pThis,
    IN      PCONFIG  pConfig
    )
{
    ULONG   nSlotIndex;

    DebugPrint((2, "SlotmanInit\n"));
    ASSERT( pThis->Type == SLOTMAN_SIGN);

    pThis->TotalSlotCnt = ConfigGetNumOfCtxMemSlots( pConfig);
/*    pThis->NumOfDESSlots = ConfigGetNumOfCtxMemSlots( pConfig); */
    ASSERT( pThis->TotalSlotCnt != 0);

    /* Allocate memory for array of slot descriptors
     */
    pThis->pSlots = PCrptNonPagedMAlloc( pThis->TotalSlotCnt * sizeof( K_SLOT_DESC));
    if ( pThis->pSlots == NULL)
    {
        return;
    }
    else
    {
        DebugPrint((2, "    allocated 0x%08x bytes for 0x%04x slots\n", pThis->TotalSlotCnt * sizeof( K_SLOT_DESC), pThis->TotalSlotCnt));
    }

    /* Initialize all slot descriptors to "free" state
     */
    for ( nSlotIndex = 0; nSlotIndex < pThis->TotalSlotCnt; nSlotIndex ++)
    {
        ASSERT( (PUCHAR)&pThis->pSlots[ nSlotIndex] <= (PUCHAR)pThis->pSlots + (pThis->TotalSlotCnt-1) * sizeof( K_SLOT_DESC));
        pThis->pSlots[ nSlotIndex].State = FreeState;
    }
}

/****************************************************************************
****************************************************************************/
void
SlotmanDestructor(
    IN PSLOTMAN pThis
    )
{
    DebugPrint((2, "SlotmanDestructor\n"));
    ASSERT( pThis->Type == SLOTMAN_SIGN);

    if ( pThis->pSlots != NULL)
    {
        PCrptNonPagedMFree( pThis->pSlots);
    }
}

/****************************************************************************
****************************************************************************/
BOOLEAN
SlotmanReserveSlot(
    IN PSLOTMAN pThis,
    OUT PULONG  pCoreSlotNumber
    )
{
    ULONG       nSlotIndex;
    BOOLEAN     bSuccess;

    DebugPrint((2, "SlotmanReserveSlot\n"));
    ASSERT( pThis->Type == SLOTMAN_SIGN);

    PCrptSpinLockAcquire( &pThis->SpinLock);

    /*
     * Walk the slot descriptor array looking for a free slot
     */
    bSuccess = FALSE;
    for ( nSlotIndex = 0;
          nSlotIndex < pThis->TotalSlotCnt && !bSuccess;
          nSlotIndex ++)
    {
        if ( pThis->pSlots[ nSlotIndex].State == FreeState)
        {
            /*
             * Mark the found free slot as "reserved" and clear its affinities,
             * because it was not used yet
             */
            pThis->pSlots[ nSlotIndex].State = ReservedState;
            pThis->pSlots[ nSlotIndex].PreferredAffinity[ MAC] = 0;
            pThis->pSlots[ nSlotIndex].PreferredAffinity[ Cryp] = 0;
            pThis->pSlots[ nSlotIndex].PreferredAffinity[ Comp] = 0;

            /*
             * Copy the slot number into the caller's storage
             */
            *pCoreSlotNumber = nSlotIndex;

            /*
             * Terminate the search with success
             */
            bSuccess = TRUE;
        }
    }
    PCrptSpinLockRelease( &pThis->SpinLock);
    return bSuccess;
}

/****************************************************************************
****************************************************************************/
void
SlotmanFreeSlot(
    IN PSLOTMAN pThis,
    IN ULONG    CoreSlotNumber
    )
{
    DebugPrint((2, "SlotmanFreeSlot: CoreSlotNumber=%ld\n", CoreSlotNumber));
    ASSERT( pThis->Type == SLOTMAN_SIGN);

    ASSERT( CoreSlotNumber < pThis->TotalSlotCnt);
    ASSERT( pThis->pSlots[ CoreSlotNumber].State == ReservedState);

    PCrptSpinLockAcquire( &pThis->SpinLock);

    /* Mark the specified reserved slot as "free"
     */
    pThis->pSlots[ CoreSlotNumber].State = FreeState;
    PCrptSpinLockRelease( &pThis->SpinLock);
}

/****************************************************************************

    The method is called when it becomes known on which processor the command
    will be executed. It reflect that knowing by updating both the Brief Command
    and the target Slot affinities.

    First, it may clear some slot affinities, those for which the Brief Command
    indicates "new context" requirement, which means that a new context is to
    be written to the context memory slot during the command execution. This
    also means, that that context is being discarded in all of the processors,
    since we maintain a view of a single processor. For instance, if an
    encryption command has come that has the "new key" flag, the encryption
    affinity is cleared for all of the processors, so that the old enc.key
    that may reside in the context memory of some processors would never be
    used for any of the next commands. 

    Second, it decides which of the context information supplied with the
    command needs to be sent to the selected processor and which already
    resides in its context memory and sholdn't be sent. This decision is made
    based on the current slot affinities. For instance, if the slot's
    encryption affinity indicates that the selected processor already has an
    encryption key in its context memory, then the enc.key supplied with the
    command should be ignored. Note, that the clearing of the affinities,
    described earlier, must occur before that. The final decision on which
    context info will be sent and which won't is recorded in the "newflags"
    field and used then by the PROCESSOR when it assembles the command/context
    block to be sent to the hardware by combining the command and parts of the
    context info sent by the client. Also, the bit for the selected processor
    is set in the corresponding affinity, if the command will store the context.
    The value of the "newflags" is formed as follows. The value of "flags" is
    taken as a starting value and new flags are added in the places when the
    command doesn't specify the "new context" because it is not the first
    command sent with such a context and another processor already has it in
    its context memory, but the current one doesn't and we will need to take
    it from the client-supplied context, which (remember!) always contains
    _all_ of the context info. In fact, the "flags" is what the client think
    we should do, and the "newflags" is what we are actually going to do.
    This is not the same since the client thinks it deals with a single
    processor but we actually deal with several ones, so a given context info
    may need to be sent several times into different processors.

    Called by the DISPATCHER with its spinlock held.

    Note, that we don't acquire the SLOTMAN's spinlock that is acquired by
    its Reserve/FreeSlot methods, because, first, the slot is already reserved
    and nobody, except the owner, touches it and, second, the access to it
    is serialized by acquisition of the DISPATCHER's spinlock.

****************************************************************************/
void
SlotmanUpdateSlotAffinitiesByCommand(
    IN      PSLOTMAN         pThis,
    IN OUT  PBRIEF_COMMAND   pBriefComm,
    IN      int              iProcessor
    )
{
    PK_SLOT_DESC    pSlotDesc;
    SLOT_AFFINITY   ProcAffinityBit;

    DebugPrint((2, "SlotmanUpdateSlotAffinities\n"));
    ASSERT( pThis->Type == SLOTMAN_SIGN);

    ASSERT( pBriefComm->SlotNumber < pThis->TotalSlotCnt);
     
    DebugPrint((3,"     Slotnumber = %ld, TotalSlotCnt = %ld\n",
		pBriefComm->SlotNumber, pThis->TotalSlotCnt));

    /*
     * Point to the target slot descriptor
     */
    pSlotDesc = &pThis->pSlots[ pBriefComm->SlotNumber];

    /************************************************************************
     * Part 1 - clear some slot affinities
     *
     * If the "flags" field indicates that an engine is engaged and its old
     * context must be discarded, the corresponding slot affinity affinity
     * is cleared. The only special case can happen with the encryption.
     */
    if ( ( pBriefComm->flags & MAC_ENGAGED) &&
         ( pBriefComm->flags & MAC_NEW_CTX))
    {
        pSlotDesc->PreferredAffinity[ MAC] = 0;
    }

    /*
     * Note! The enc.command which doesn't store enc.context is essentially
     * a phantom that doesn't affect the crypto-state of the slot and we don't
     * touch the crypto-affinity in this case
     */
    if ( ( pBriefComm->flags & CRYP_ENGAGED) &&
         ( pBriefComm->flags & CRYP_NEW_CTX) &&
         !( pBriefComm->flags & CRYP_DONT_STORE_CTX))
    {
        pSlotDesc->PreferredAffinity[ Cryp] = 0;
    }

    if ( ( pBriefComm->flags & COMP_ENGAGED) &&
         ( pBriefComm->flags & COMP_NEW_CTX))
    {
        pSlotDesc->PreferredAffinity[ Comp] = 0;
    }

    /************************************************************************
     * Part 2 - decide which context info needs to be sent, record this in
     * the "newflags" field and set the bit in the corresponding affinity.
     * The encryption may again present a special case.
     */

    /*
     * Make affinity mask for the selected processor
     */
    ProcAffinityBit = ( 1 << iProcessor);

    /*
     * The "newflags" are formed by taking the "flags" and adding some "new
     * context" bits.
     */
    pBriefComm->newflags = pBriefComm->flags;

    if ( pBriefComm->flags & MAC_ENGAGED)
    {
        if ( !( pSlotDesc->PreferredAffinity[ MAC] & ProcAffinityBit))
        {
            /*
             * The MAC key is not there - need to send the key and set the
             * affinity bit
             */
            pBriefComm->newflags |= MAC_NEW_CTX;
            pSlotDesc->PreferredAffinity[ MAC] |= ProcAffinityBit;
        }
    }

    if ( pBriefComm->flags & CRYP_ENGAGED)
    {
        if ( !( pSlotDesc->PreferredAffinity[ Cryp] & ProcAffinityBit))
        {
            /*
             * The Enc.key is not there - need to send the key and set the
             * affinity bit. However, if the enc.context is not to be stored,
             * the affinity is not set!
             */
            pBriefComm->newflags |= CRYP_NEW_CTX;
            if ( !( pBriefComm->flags & CRYP_DONT_STORE_CTX))
            {
                pSlotDesc->PreferredAffinity[ Cryp] |= ProcAffinityBit;
            }
        }
    }

    if ( pBriefComm->flags & COMP_ENGAGED)
    {
        if ( !( pSlotDesc->PreferredAffinity[ Comp] & ProcAffinityBit))
        {
            /*
             * The Enc.key is not there - need to send the key and set the
             * affinity bit
             */
            pBriefComm->newflags |= COMP_NEW_CTX;
            pSlotDesc->PreferredAffinity[ Comp] |= ProcAffinityBit;
        }
    }
}
