/**********************************************************************

    Module: slotman.h

        Product: kernel mode driver for "PowerCrypt" encryption
        acceleration board based on Hi/fn 7751 chip
        by Ashley Laurent, Inc.

    ---------------------------------------------------------------

    Copyright:

        Ashley Laurent, Inc., 1999
        All Rights Reserved.

    ---------------------------------------------------------------

    Functional Description:

        Declaration of the SLOTMAN (Slot Manager) object that performs
        the control over the 7751 internal context memory slot usage to
        allow multiple clients to use the slots without conflicts.

    ---------------------------------------------------------------

    Environment:

        NT or BSD kernel mode

    ---------------------------------------------------------------

    Author:

        Stan Shkolnyy

    ---------------------------------------------------------------

    Revision history:

        *   15-Feb-99   initial revision

**********************************************************************/

#ifndef __SLOTMAN_H__
#define __SLOTMAN_H__

#include "driverdef.h"      /* for MAXIMUM_DEVICES_SUPPORTED */
#include "config.h"         /* for CONFIG */

#include "pwrcryptioctl.h"
#include "km_interface.h"   /* for PBRIEF_COMMAND */

/**********************************************************************
 * Slot Manager maintains a descriptor for each context memory slot.
 * Though several 7751 processors can be installed, the slots are reserved and
 * freed in parallel in all the processors. Thus, the number of slots doesn't
 * depend on the number of the processors, but on the minimum amount of the
 * context memory amongst the processors and on the memory configuration which
 * is also identical for all the processors.
 *
 * SLOT_DESC.State indicates whether the slot is reserved (in use) by a client.
 *
 * The decision to reserve and use slots in parallel in several processors
 * results in the need of bookkeeping which processors already contain context
 * information in their slots and which do not.
 *
 * SLOT_DESC.PreferredAffinity members carry that bookkeeping information.
 * Each of members corresponds to a particular kind of context info, such as
 * Encryption or MAC. Each bit of PreferredXxxAffinity corresponds to a
 * processor. If it is set, the context info is already stored in the
 * processor's context memory, hence the name PreferredAffinity - means that
 * the corresponding operation can be carried out on this processor without
 * initial writing that context info into its context memory.
 *
 */
#if MAXIMUM_DEVICES_SUPPORTED <= 8
    typedef UCHAR   SLOT_AFFINITY;
#else
    #error Sorry guys, I did not count on MAXIMUM_DEVICES_SUPPORTED > 8
#endif

/*
 * Define the types of operations that the 7751 can perform with respect to
 * the context memory usage
 */
typedef enum _SLOT_AFFINITY_TYPE
{
    FirstAffinityType   = 0,
    MAC                 = 0,    /* MAC key */
    Cryp,                       /* Enc.key */
    Comp,                       /* Compression history */
    LastAffinityType    = Comp

} SLOT_AFFINITY_TYPE, *PSLOT_AFFINITY_TYPE;

typedef struct _K_SLOT_DESC
{
    UCHAR           State;
    SLOT_AFFINITY   PreferredAffinity[ LastAffinityType - FirstAffinityType + 1];

} K_SLOT_DESC, *PK_SLOT_DESC;

/* Two possible states of the descriptor
 */
#define FreeState           0
#define ReservedState       1

/**********************************************************************
 * SLOTMAN is an abstraction of the
 */

typedef struct _SLOTMAN
{
    ULONG           Type;

/*    ULONG           NumOfDESSlots; */
/*    ULONG           NumOfRC4Slots; */
/*    ULONG           NumOfCompHistSlots; */

/*    { */
        ULONG           TotalSlotCnt;
        ULONG           FreeSlotCnt;
        PK_SLOT_DESC    pSlots;         /* array of TotalSlotCnt length */
/*    } */

/*MAXIMUM_DEVICES_SUPPORTED */

    /* Spinlock acquired by slot reservation/freeing methods
     */
    PCRPT_DECLARE_SPINLOCK( SpinLock);

} SLOTMAN, *PSLOTMAN;

/**********************************************************************
 * PUBLIC methods
 */
void
SlotmanConstructor(
    IN OUT  PSLOTMAN pThis
    );

void
SlotmanInit(
    IN OUT  PSLOTMAN pThis,
    IN      PCONFIG  pConfig
    );

void
SlotmanDestructor(
    IN PSLOTMAN pThis
    );

BOOLEAN
SlotmanReserveSlot(
    IN PSLOTMAN pThis,
    OUT PULONG  pCoreSlotNumber
    );

void
SlotmanFreeSlot(
    IN PSLOTMAN pThis,
    IN ULONG    CoreSlotNumber
    );

#define SlotmanIsValid( pThis)   ( (pThis)->pSlots != NULL)

#define SlotmanGetNumOfSlots( pThis)         ( (pThis)->TotalSlotCnt)

/*#define SlotmanGetNumOfDESSlots( pThis)         ( (pThis)->NumOfDESSlots) */
/*#define SlotmanGetNumOfRC4Slots( pThis)         ( (pThis)->NumOfRC4Slots) */
/*#define SlotmanGetNumOfCompHistSlots( pThis)    ( (pThis)->NumOfCompHistSlots) */

void
SlotmanUpdateSlotAffinitiesByCommand(
    IN      PSLOTMAN         pThis,
    IN OUT  PBRIEF_COMMAND   pBriefComm,
    IN      int              iProcessor    
    );

#endif
