/**********************************************************************

    Module: slotrack.c

        Product: kernel mode driver providing user mode interface
        to the PwrCrypt driver for encryption acceleration board
        based on Hi/fn 7751 chip by Ashley Laurent, Inc.

    ---------------------------------------------------------------

    Copyright:

        Ashley Laurent, Inc., 1999
        All Rights Reserved.

    ---------------------------------------------------------------

    Functional Description:

        Implements the PwrCrypt slot tracker object that provides
        conflict free slot reservation.

    ---------------------------------------------------------------

    Environment:

        NT or BSD kernel and user mode

    ---------------------------------------------------------------

    Author:

        Stan Shkolnyy

    ---------------------------------------------------------------

    Revision history:

        *   16-Mar-99   initial revision

**********************************************************************/

#include "osindependence.h"

#include "debug.h"
#include "driverdef.h"
#include "slotrack.h"

#ifdef ALLOC_PRAGMA
#pragma alloc_text(INIT, SlotTrackerConstructor)
#pragma alloc_text(INIT, SlotTrackerInit)
#endif

/****************************************************************************
****************************************************************************/
void
SlotTrackerConstructor(
    IN OUT  PSLOTTRACKER pThis
    )
{
    DebugPrint((2, "SlotTrackerConstructor\n"));
    pThis->Type = SLOTRACK_SIGN;
    pThis->pSlots = NULL;
    PCrptSpinLockInit( &pThis->SpinLock);
}

/****************************************************************************
****************************************************************************/
void
SlotTrackerInit(
    IN OUT  PSLOTTRACKER pThis,
    IN      ULONG        NumOfSlots
    )
{
    ULONG   nSlotIndex;

    DebugPrint((2, "SlotTrackerInit 0x%04x slots\n", NumOfSlots));
    ASSERT( pThis->Type == SLOTRACK_SIGN);

    pThis->NumOfSlots = NumOfSlots;
    ASSERT( pThis->NumOfSlots != 0);

    /* Allocate memory for array of slot descriptors
     */
    pThis->pSlots = PCrptNonPagedMAlloc( pThis->NumOfSlots * sizeof( SLOT_DESC));
    if ( pThis->pSlots == NULL)
    {
        DebugPrint((2, "    Can't allocate memory\n", NumOfSlots));
        return;
    }

    /* Initialize all slot descriptors to "free" state
     */
    for ( nSlotIndex = 0; nSlotIndex < pThis->NumOfSlots; nSlotIndex ++)
    {
        pThis->pSlots[ nSlotIndex].nKey = 0;
        pThis->pSlots[ nSlotIndex].RefCnt = 0;
    }
}

/****************************************************************************
****************************************************************************/
void
SlotTrackerDestructor(
    IN PSLOTTRACKER pThis
    )
{
    DebugPrint((2, "SlotTrackerDestructor\n"));

    if ( pThis->pSlots != NULL)
    {
        PCrptNonPagedMFree( pThis->pSlots);
    }
}

/****************************************************************************
SlotTrackerRegisterSlot()
    Finds an free place in the Slot Tracker registry,
    places the "core slot number"/"key" pair into it
    and returns an index of the found place.

    Not finding a free entry is a bug indicator.
    It should never happen, since Slot Tracker must have the registry spacious
    enough to register all of the PwrCrypt core driver slots.
****************************************************************************/
BOOLEAN
SlotTrackerRegisterSlot(
    IN PSLOTTRACKER pThis,
    IN ULONG        CoreSlotNumber,
    IN ULONG        nKey,
    OUT ULONG       *pSlotNumber
    )
{
    ULONG i;
    BOOLEAN bSuccess;

    DebugPrint((2, "SlotTrackerRegisterSlot %ld, pThis=0x%x\n", 
		CoreSlotNumber, (unsigned int) pThis));
    ASSERT( pThis->Type == SLOTRACK_SIGN);
    ASSERT( nKey != 0);         /* nKey = 0 is not an allowed value */

    bSuccess = FALSE;
    PCrptSpinLockAcquire( &pThis->SpinLock);
    for ( i = 0; i < pThis->NumOfSlots && !bSuccess; i ++)
    {
        if ( pThis->pSlots[ i].nKey == 0)
        {
	    DebugPrint((3, "    found empty slot %ld\n", i));
            ASSERT( pThis->pSlots[ i].RefCnt == 0);
            pThis->pSlots[ i].nKey = nKey;
            pThis->pSlots[ i].CoreSlotNumber = CoreSlotNumber;
	    ASSERT(pThis->pSlots[ i].CoreSlotNumber == CoreSlotNumber);
            *pSlotNumber = i;
            bSuccess = TRUE;
        }
    }
    PCrptSpinLockRelease( &pThis->SpinLock);
    return bSuccess;
}

/****************************************************************************
SlotTrackerUnregisterSlot()
    By the passed index of the entry in the Slot Tracker's registry and the
    key, veryfies that the index is valid and the entry contains the key.
    If it is so, makes the entry free and returns the PwrCrypt's "core slot
    number" stored in the entry.
****************************************************************************/
BOOLEAN
SlotTrackerUnregisterSlot(
    IN PSLOTTRACKER pThis,
    IN ULONG        SlotNumber,
    IN ULONG        nKey,
    OUT ULONG       *pCoreSlotNumber
    )
{
    BOOLEAN bSuccess;

    DebugPrint((2, "SlotTrackerUnregisterSlot 0x%08x, pThis=0x%x\n", 
		SlotNumber, (unsigned int) pThis));
    ASSERT( pThis->Type == SLOTRACK_SIGN);
    ASSERT( nKey != 0);         /* nKey = 0 is not an allowed value */

    /*
     * This slot number is given by user, it can be invalid
     */
    if ( SlotNumber >= pThis->NumOfSlots)
    {
        DebugPrint((1, "App.error - impossible slot number 0x%08x passed\n", SlotNumber));
        return FALSE;
    }

    PCrptSpinLockAcquire( &pThis->SpinLock);
    bSuccess = FALSE;
    if ( pThis->pSlots[ SlotNumber].nKey != 0)
    {
        if ( pThis->pSlots[ SlotNumber].nKey == nKey)
        {
            if ( pThis->pSlots[ SlotNumber].RefCnt == 0)
            {
		DebugPrint((3, "    found entry for slot %d, core=%d\n",
			    SlotNumber, pThis->pSlots[SlotNumber].CoreSlotNumber));
                pThis->pSlots[ SlotNumber].nKey = 0;
                *pCoreSlotNumber = pThis->pSlots[ SlotNumber].CoreSlotNumber;
                bSuccess = TRUE;
            }
            else
            {
                DebugPrint((1, "App.error - attempt to free a busy slot 0x%08x\n", SlotNumber));
            }
        }
        else
        {
            DebugPrint((1, "App.error - attempt to free other people's slot 0x%08x\n", SlotNumber));
        }
    }
    else
    {
        DebugPrint((1, "App.error - attempt to free a free slot 0x%08x\n", SlotNumber));
    }
    PCrptSpinLockRelease( &pThis->SpinLock);
    return bSuccess;
}

/****************************************************************************
SlotTrackerReferenceSlot()
     nKey is needed in this method merely to prevent the caller
     from touching wrong slots (those which it didn't reserved)
****************************************************************************/
BOOLEAN
SlotTrackerReferenceSlot(
    IN PSLOTTRACKER pThis,
    IN ULONG        SlotNumber,
    IN ULONG        nKey
    )
{
    BOOLEAN bSuccess;

    DebugPrint((2, "SlotTrackerReferenceSlot 0x%08x\n", SlotNumber));
    ASSERT( pThis->Type == SLOTRACK_SIGN);
    ASSERT( nKey != 0);         /* nKey = 0 is not an allowed value */

    /*
     * This slot number is given by user, it can be invalid
     */
    if ( SlotNumber >= pThis->NumOfSlots)
    {
        DebugPrint((1, "    App.error - impossible slot number 0x%08x passed\n", SlotNumber));
        return FALSE;
    }

    bSuccess = FALSE;
    PCrptSpinLockAcquire( &pThis->SpinLock);
    if ( pThis->pSlots[ SlotNumber].nKey != 0)
    {
        if ( pThis->pSlots[ SlotNumber].nKey == nKey)
        {
            pThis->pSlots[ SlotNumber].RefCnt++;
            bSuccess = TRUE;
        }
        else
        {
            DebugPrint((1, "    App.error - attempt to use other people's slot 0x%08x\n", SlotNumber));
        }
    }
    else
    {
        DebugPrint((1, "    App.error - attempt to use free slot 0x%08x\n", SlotNumber));
    }
    PCrptSpinLockRelease( &pThis->SpinLock);
    return bSuccess;
}

/****************************************************************************
SlotTrackerDereferenceSlot()
     nKey is needed in this method merely to prevent the caller
     from touching wrong slots (those which it didn't reserved)
****************************************************************************/
BOOLEAN
SlotTrackerDereferenceSlot(
    IN PSLOTTRACKER pThis,
    IN ULONG        SlotNumber,
    IN ULONG        nKey
    )
{
    BOOLEAN bSuccess;

    DebugPrint((2, "SlotTrackerDereferenceSlot 0x%08x\n", SlotNumber));
    ASSERT( pThis->Type == SLOTRACK_SIGN);
    ASSERT( nKey != 0);         /* nKey = 0 is not an allowed value */

    /*
     * This slot number is given by user, it can be invalid
     */
    if ( SlotNumber >= pThis->NumOfSlots)
    {
        DebugPrint((1, "    App.error - impossible slot number 0x%08x passed\n", SlotNumber));
        return FALSE;
    }

    bSuccess = FALSE;
    PCrptSpinLockAcquire( &pThis->SpinLock);
    if ( pThis->pSlots[ SlotNumber].nKey != 0)
    {
        if ( pThis->pSlots[ SlotNumber].nKey == nKey)
        {
            pThis->pSlots[ SlotNumber].RefCnt--;
            bSuccess = TRUE;
        }
        else
        {
            DebugPrint((1, "    App.error - attempt to use other people's slot 0x%08x\n", SlotNumber));
        }
    }
    else
    {
        DebugPrint((1, "    App.error - attempt to use free slot 0x%08x\n", SlotNumber));
    }
    PCrptSpinLockRelease( &pThis->SpinLock);
    return bSuccess;
}

/****************************************************************************
SlotTrackerEnumSlotsByKey()
     Walks the slot tracking array and calls the callback if the array entry
     contains the nKey, passing the callback its "Context" and the Slot
     Tracker's index of the found registry entry.
****************************************************************************/
void
SlotTrackerEnumSlotsByKey(
    IN PSLOTTRACKER         pThis,
    IN PSLOT_ENUM_CALLBACK  pCallBack,
    IN PVOID                Context,
    IN ULONG                nKey
    )
{
    ULONG   i;

    DebugPrint((2, "SlotTrackerEnumSlotsByKey 0x%08x\n", nKey));
    ASSERT( pThis->Type == SLOTRACK_SIGN);
    ASSERT( nKey != 0);         /* nKey = 0 is not an allowed value */

    for ( i = 0; i < pThis->NumOfSlots; i ++)
    {
        if ( pThis->pSlots[ i].nKey == nKey)
        {
            (pCallBack) ( Context, i);
        }
    }
}
