/**********************************************************************

    Module: slotrack.c

        Product: kernel mode driver providing user mode interface
        to the PwrCrypt driver for encryption acceleration board
        based on Hi/fn 7751 chip by Ashley Laurent, Inc.

    ---------------------------------------------------------------

    Copyright:

        Ashley Laurent, Inc., 1999
        All Rights Reserved.

    ---------------------------------------------------------------

    Functional Description:

        Implements the PwrCrypt slot tracker object that provides
        conflict free slot reservation.

    ---------------------------------------------------------------

    Environment:

        NT or BSD kernel and user mode

    ---------------------------------------------------------------

    Author:

        Stan Shkolnyy

    ---------------------------------------------------------------

    Revision history:

        *   16-Mar-99   initial revision

**********************************************************************/

#include "osindependence.h"

#include "debug.h"
#include "driverdef.h"
#include "slotrack.h"

#ifdef ALLOC_PRAGMA
#pragma alloc_text(INIT, SlotTrackerConstructor)
#pragma alloc_text(INIT, SlotTrackerInit)
#endif

/****************************************************************************
****************************************************************************/
void
SlotTrackerConstructor(
    IN OUT  PSLOTTRACKER pThis
    )
{
    DebugPrint((2, "SlotTrackerConstructor\n"));
    pThis->Type = SLOTRACK_SIGN;
    pThis->pSlots == NULL;
    PCrptSpinLockInit( &pThis->SpinLock);
}

/****************************************************************************
****************************************************************************/
void
SlotTrackerInit(
    IN OUT  PSLOTTRACKER pThis,
    IN      ULONG        NumOfSlots
    )
{
    ULONG   i;

    DebugPrint((2, "SlotTrackerInit 0x%04x slots\n", NumOfSlots));
    ASSERT( pThis->Type == SLOTRACK_SIGN);

    pThis->NumOfSlots = NumOfSlots;
    ASSERT( pThis->NumOfSlots != 0);

    /* Allocate memory for array of slot descriptors
     */
    pThis->pSlots = PCrptNonPagedMAlloc( pThis->NumOfSlots * sizeof( SLOT_DESC));
    if ( pThis->pSlots == NULL)
    {
        DebugPrint((2, "    Can't allocate memory\n", NumOfSlots));
        return;
    }

    /* Initialize all slot descriptors to "free" state
     */
    for ( i = 0; i < pThis->NumOfSlots; i ++)
    {
        pThis->pSlots[ i].nKey = 0;
        pThis->pSlots[ i].RefCnt = 0;
    }
}

/****************************************************************************
****************************************************************************/
void
SlotTrackerDestructor(
    IN PSLOTTRACKER pThis
    )
{
    DebugPrint((2, "SlotTrackerDestructor\n"));

    if ( pThis->pSlots != NULL)
    {
        PCrptNonPagedMFree( pThis->pSlots);
    }
}

/****************************************************************************
    Finds a free slot descriptor, puts the key and slot number into it and
    returns a ptr to it or NULL which is an exceptional case, since that
    would mean that PwrCrypt has disbursed more slots than it told this
    user-mode interface driver it had or we somehow didn't free slots that
    must have to be freed.

****************************************************************************/
PVOID
SlotTrackerRegisterSlot(
    IN PSLOTTRACKER pThis,
    IN ULONG        SlotNumber,
    IN ULONG        nKey
    )
{
    PVOID   pSlotDesc;

    DebugPrint((2, "SlotTrackerReserveSlot 0x%08x\n", SlotNumber));
    ASSERT( pThis->Type == SLOTRACK_SIGN);
    ASSERT( nKey != 0);         /* nKey = 0 is not an allowed value */

    pSlotDesc = NULL;

    PCrptSpinLockAcquire( &pThis->SpinLock);

    for ( i = 0; i < pThis->NumOfSlots; i ++)
    {
        if ( pThis->pSlots[ i].nKey == 0)
        {
            ASSERT( pThis->pSlots[ i].RefCnt == 0);
            pThis->pSlots[ i].nKey = nKey;
            pThis->pSlots[ i].SlotNumber = SlotNumber;
            pSlotDesc = &pThis->pSlots[ i];
        }
    }

    PCrptSpinLockRelease( &pThis->SpinLock);

    return pSlotDesc;
}

/****************************************************************************
    Checks whether the pSlotDesc points to a slot desc. in this Slot Tracker's
    descriptor array.
****************************************************************************/
static
BOOLEAN
StTkIsSlotDescPtr(
    IN PSLOTTRACKER pThis,
    IN PVOID        pSlotDesc
    )
{
    ULONG   nOffset;

    ASSERT( pThis->Type == SLOTRACK_SIGN);

    /*
     * Check offset of slot desc.ptr from the array beginning
     */
    nOffset = (ULONG) pSlotDesc - (ULONG) &pThis->pSlots[ 0];
    if ( nOffset % sizeof( SLOT_DESC) != 0)
    {
        DebugPrint((1, "App.error(align) - impossible slot desc 0x%08x passed\n", pSlotDesc));
        return FALSE;
    }

    /*
     * Alignment is Ok, check the offset to be in the array bounds
     */
    if ( nOffset < (ULONG) &pThis->pSlots[ 0] &&
         nOffset > (ULONG) &pThis->pSlots[ pThis->NumOfSlots - 1])
    {
        DebugPrint((1, "App.error(bounds) - impossible slot desc 0x%08x passed\n", pSlotDesc));
        return FALSE;
    }

    /*
     * Ok, it has hitted a slot descriptor
     */
    return TRUE;
}

/****************************************************************************
****************************************************************************/
BOOLEAN
SlotTrackerUnregisterSlot(
    IN PSLOTTRACKER pThis,
    IN PVOID        pSlotDesc,
    IN ULONG        nKey
    )
{
    BOOLEAN bSuccess;

    DebugPrint((2, "SlotTrackerFreeSlot 0x%08x\n", SlotNumber));
    ASSERT( pThis->Type == SLOTRACK_SIGN);
    ASSERT( nKey != 0);         /* nKey = 0 is not an allowed value */

    /*
     * This slot number is given by user, it can be invalid
     */
    if ( !StTkIsSlotDescPtr( pThis, pSlotDesc))
    {
        return FALSE;
    }

    PCrptSpinLockAcquire( &pThis->SpinLock);
    bSuccess = FALSE;

    /*
     * nKey is needed in this method merely to prevent the caller
     * from freeing wrong slots (belonging to others).
     */
    if ( ((PSLOT_DESC) pSlotDesc)->nKey != 0)
    {
        if ( ((PSLOT_DESC) pSlotDesc)->nKey == nKey)
        {
            if ( ((PSLOT_DESC) pSlotDesc)->RefCnt == 0)
            {
                ((PSLOT_DESC) pSlotDesc)->nKey = 0;
                bSuccess = TRUE;
            }
            else
            {
                DebugPrint((1, "App.error - attempt to free a busy slot 0x%08x\n", SlotNumber));
            }
        }
        else
        {
            DebugPrint((1, "App.error - attempt to free other people's slot 0x%08x\n", SlotNumber));
        }
    }
    else
    {
        DebugPrint((1, "App.error - attempt to free a free slot 0x%08x\n", SlotNumber));
    }

    PCrptSpinLockRelease( &pThis->SpinLock);
    return bSuccess;
}

/****************************************************************************
****************************************************************************/
BOOLEAN
SlotTrackerReferenceSlot(
    IN PSLOTTRACKER pThis,
    IN PVOID        pSlotDesc,
    IN ULONG        nKey
    )
{
    BOOLEAN bSuccess;

    DebugPrint((2, "SlotTrackerReferenceSlot 0x%08x\n", SlotNumber));
    ASSERT( pThis->Type == SLOTRACK_SIGN);
    ASSERT( nKey != 0);         /* nKey = 0 is not an allowed value */

    /*
     * This slot number is given by user, it can be invalid
     */
    if ( !StTkIsSlotDescPtr( pThis, pSlotDesc))
    {
        return FALSE;
    }

    bSuccess = FALSE;

    PCrptSpinLockAcquire( &pThis->SpinLock);

    /*
     * nKey is needed in this method merely to prevent the caller
     * from touching wrong slots (those which it didn't reserved)
     */
    if ( ((PSLOT_DESC) pSlotDesc)->nKey != 0)
    {
        if ( ((PSLOT_DESC) pSlotDesc)->nKey == nKey)
        {
            ((PSLOT_DESC) pSlotDesc)->RefCnt++;
            bSuccess = TRUE;
        }
        else
        {
            DebugPrint((1, "    App.error - attempt to use other people's slot 0x%08x\n", SlotNumber));
        }
    }
    else
    {
        DebugPrint((1, "    App.error - attempt to use free slot 0x%08x\n", SlotNumber));
    }

    PCrptSpinLockRelease( &pThis->SpinLock);
    return bSuccess;
}

/****************************************************************************
****************************************************************************/
BOOLEAN
SlotTrackerDereferenceSlot(
    IN PSLOTTRACKER pThis,
    IN PVOID        pSlotDesc,
    IN ULONG        nKey
    )
{
    BOOLEAN bSuccess;

    DebugPrint((2, "SlotTrackerDereferenceSlot 0x%08x\n", SlotNumber));
    ASSERT( pThis->Type == SLOTRACK_SIGN);
    ASSERT( nKey != 0);         /* nKey = 0 is not an allowed value */

    /*
     * This slot number is given by user, it can be invalid
     */
    if ( !StTkIsSlotDescPtr( pThis, pSlotDesc))
    {
        return FALSE;
    }

    bSuccess = FALSE;

    PCrptSpinLockAcquire( &pThis->SpinLock);

    /*
     * nKey is needed in this method merely to prevent the caller
     * from touching wrong slots (those which it didn't reserved)
     */
    if ( ((PSLOT_DESC) pSlotDesc)->nKey != 0)
    {
        if ( ((PSLOT_DESC) pSlotDesc)->nKey == nKey)
        {
            ((PSLOT_DESC) pSlotDesc)->RefCnt--;
            bSuccess = TRUE;
        }
        else
        {
            DebugPrint((1, "    App.error - attempt to use other people's slot 0x%08x\n", SlotNumber));
        }
    }
    else
    {
        DebugPrint((1, "    App.error - attempt to use free slot 0x%08x\n", SlotNumber));
    }

    PCrptSpinLockRelease( &pThis->SpinLock);
    return bSuccess;
}

/****************************************************************************
****************************************************************************/
void
SlotTrackerEnumSlotsByKey(
    IN PSLOTTRACKER         pThis,
    IN PSLOT_ENUM_CALLBACK  pCallBack,
    IN PVOID                Context,
    IN ULONG                nKey
    )
{
    ULONG   i;

    DebugPrint((2, "SlotTrackerEnumSlotsByKey 0x%08x\n", nKey));
    ASSERT( pThis->Type == SLOTRACK_SIGN);
    ASSERT( nKey != 0);         /* nKey = 0 is not an allowed value */

    /* Walk the slot tracking array and call the callback if the array entry
     * contains the nKey
     */
    for ( i = 0; i < pThis->NumOfSlots; i ++)
    {
        if ( pThis->pSlots[ i].nKey == nKey)
        {
            (pCallBack) ( Context, &pThis->pSlots[ i]);
        }
    }
}
